package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.awt.*;  
import javax.swing.*;
import java.awt.image.*;
import java.io.*;
import java.net.*;
import java.lang.*;
import java.util.*;

class BestSegCanvas extends Canvas   implements java.awt.event.MouseListener, java.awt.event.MouseMotionListener {	
	
	BestSegJFrame parent=null;
	int	mouseStartX, mouseStartY,stampXorg=-1,stampYorg=-1; 		// Used when dragging center of postage stamp region around
	SegPic bestPic = null;
	int width=0,height=0;
	int stampWidth=0, stampHeight=0;
	int stampXorig=0, stampYorig=0;
	int oldStampXorig=-1,  oldStampYorig=-1;
	Image bestImage=null;
	int [] grayRawDat=null;
	
	public void sepuku() {
		bestPic=null;
		grayRawDat=null;
		bestImage=null;
		parent=null;
		//    talk("        BestSegCanvas.sepuku()");
		setVisible(false);
		}

	public BestSegCanvas(int stampWidth,int stampHeight, int stampXorig, int stampYorig, 
				SegPic bestPic, BestSegJFrame parent) {
		this.stampWidth=stampWidth;
		this.stampHeight=stampHeight;
		this.stampXorig=stampXorig;
		this.stampYorig=stampYorig;
		this.height=bestPic.wholeRows;
		this.width=bestPic.wholeCols;
		this.parent=parent;
		relocateStampBox((stampXorig+stampWidth)/2, (stampYorig+stampHeight)/2);
		setBackground(Color.white);
		setVisible(true);
	   	addMouseListener(this);
	   	addMouseMotionListener(this);
		newBestPic(bestPic);
		}
		
	public void changeStampSizes(int newStampWidth,int newStampHeight) {
		int stampX = stampXorig + stampWidth/2;
		int stampY = stampYorig + stampHeight/2;
		this.stampHeight=newStampHeight;
		this.stampWidth=newStampWidth;
		relocateStampBox(stampX,stampY); // this repaints
		}
		
	public void relocateStampBox(int x, int y) {
		x -= stampWidth/2;
		y -= stampHeight/2;
		if((stampXorig=x) < 0)
			stampXorig = 0;
		if(stampXorig >= width-stampWidth)
			stampXorig = width-stampWidth -1;
		if((stampYorig=y) < 0)
			stampYorig = 0;
		if(stampYorig >= height-stampHeight)
			stampYorig =  height-stampHeight -1;
		repaint();
		}
		
	public void newBestPic(SegPic theBestPic) {
		bestPic=theBestPic;
		int i;
		if(bestPic==null) return;
		int [] curIm = parent.parent.parent.curStack.curIm;

		if( height != bestPic.wholeRows || width != bestPic.wholeCols || grayRawDat==null) {
			width = bestPic.wholeCols;
			height = bestPic.wholeRows;
			setSize(width,height);
			}
				
	    bestImage = createImage(new MemoryImageSource(width, height,
			 	ColorModel.getRGBdefault(),bestPic.wholePic, 0, width));
		setVisible(true);
		repaint();
		}
			
	public void paint(Graphics g) {
		// talk("This should never happen. BestSegCanvas.paint() called");
		paint2D((Graphics2D)g);
		}
		
	public void update(Graphics g) {
		// talk("BestSegCanvas.update() called. Should never happen");
		paint2D((Graphics2D)g);
		}

	public void update2D(Graphics2D g) {
		paint2D(g);
		}

	public void paint2D(Graphics2D g) {
		if(bestPic==null) return;
	    if(bestImage==null) {
	    	bestImage = createImage(new MemoryImageSource(width, height,
			 	ColorModel.getRGBdefault(), bestPic.wholePic, 0, width));
			}
		try {
			 g.drawImage(bestImage, 0, 0, width,height, this);
			 }
		  catch (Exception e) {
			System.out.println("Error painting bestPic panel: " + e.toString());
			}
		drawTheStampBox(g);
		}

	static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}

	public void drawTheStampBox(Graphics g) {
		g.setColor(Color.cyan);
		g.drawRect(stampXorig,stampYorig,stampWidth,stampHeight);
		}
		
	public void mouseMoved(java.awt.event.MouseEvent event){
		int x = event.getX();
		int y = event.getY();
		int bri = bestPic.getGlow(x,y);
		int avg = bestPic.getNbhdGlow(x,y);
		parent.pixelReport(x,y,bri,avg);
		}
			
	public void mouseDragged(java.awt.event.MouseEvent event){
		int x=event.getX(),y=event.getY();
		//talk("Mouse Dragged to "+mouseStartX+", "+mouseStartX);		    
		//canvas.relocateStampBox(mouseStartX,mouseStartY);
		}
	
// catch all other mouse events 
	
	public void mousePressed(java.awt.event.MouseEvent event) {
		}
				
	public void mouseReleased(java.awt.event.MouseEvent event) {
		int x= event.getX(), y=event.getY();
		//talk("BestSegJFrame.Mouse Pressed at "+x+", "+y);		    
		//canvas.relocateStampBox(x,y);
		}
		
		
	public void mouseClicked(java.awt.event.MouseEvent event) {
		int x=event.getX(),y=event.getY();
		relocateStampBox(x,y);
		}
		
	public void mouseEntered(java.awt.event.MouseEvent event) {
		// talk("Mouse entered at ("+event.getX()+", "+event.getY()+")");
		}
		
	public void mouseExited(java.awt.event.MouseEvent event) {
		// talk("Mouse exited at ("+event.getX()+", "+event.getY()+")");
		if(stampXorig != oldStampXorig || stampYorig != oldStampYorig) {
			 oldStampXorig=stampXorig;
			 oldStampYorig=stampYorig;
			parent.parent.newStampLocation(stampXorig,stampYorig);
			}
		}

}