package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
import java.awt.*;  import javax.swing.*;
import java.awt.image.*;
import java.util.*;
import java.lang.*;
import java.io.*;

class BioradCanvas extends Canvas {
	String sectionTypeName = "XY-";
	int sectionType =0;
	boolean ifdiags=false; // set to true to print diagnostic messages...
	ShowPicJPanel parent=null;
	SectionNavigator navigator=null;
	double windowScaleFac = 1.0;
	Image		rawImage=null;
	Image		offScrImage;
	BioradFile	redfile=null,greenfile=null,bluefile=null;
	int curWidth=384,curHeight=256;
	static int curJFrame=0;
	int NJFrames=0,nwide=10,nhigh=10;
	int [] curIm = null;
	int curRows=256;
	int curCols=384;
	
	int lastJFrameShown = -1;  // used by showAJFrame();
	boolean ifForceNewDrawing = true;

	String capString="sections from biorad z-series files";
	String myRedFileName="",myGreenFileName="",myBlueFileName="";
	boolean ifred=false,ifgreen=false,ifblue=false;
	int [] Xshifts = {0, 0, 0}; // the 0 position for the red channel,...
	int [] Yshifts = {0, 0, 0}; // the 2 position for the blue channel
	int [] Zshifts = {0, 0, 0};
	
	/* these are lookup tables to be indexed by (byte_pix + 256) whose values range from
		256 to 383 for non-negative values of byte_pix and whose values range from
		1 to 128 for negative values of bte_pix. The idea is to avoid the arithmetic of
		LUT[( byte[i]<0? byte[i]+256 : byte[i]) ]
		for each conversion of a byte value to a  a color table value and just use the faster
		LUT[byte[i]+256]
		  */
	int [] redLUT = new int[384];
	int [] redShiftLUT = new int[384];
	int [] greenLUT = new int[384];
	int [] greenShiftLUT = new int[384];
	int [] blueLUT = new int[384];
	
	//  these are lookup tables to be indexed by ints ranging from 0 to 255.
	int [] redLUT256 = new int[256];
	int [] redShiftLUT256  = new int[256];
	int [] greenLUT256  = new int[256];
	int [] greenShiftLUT256  = new int[256];
	int [] blueLUT256 = new int[256];
	int [] redRedness = new int[256];
	int [] greenRedness = new int[256];
	int [] blueRedness = new int[256];
	int [] redGreenness = new int[256];
	int [] greenGreenness = new int[256];
	int [] blueGreenness = new int[256];
	int [] redBlueness = new int[256];
	int [] greenBlueness = new int[256];
	int [] blueBlueness = new int[256];
	

	double redness=1.0, greenness=1.0, blueness=1.0;
	boolean ifUseGrayScale=true;
	boolean ifUsingArtColors=false;
	int singleChan=-1;
	int tooZfar2Draw = 16; // the z-nearness criterion determining whether markers get drawn
	long  timeMark=0, time2Mark=-1;;
	private static short [] add256 = null;
	byte [][] redByteChanStack = null;
	byte [][] greenByteChanStack = null;
	byte [][] blueByteChanStack = null;
	boolean [] byteChanStackLoaded = null;

	public void initializeByteChanStack() {
		if(NJFrames>0) {
			redByteChanStack = new byte[NJFrames][];	
			greenByteChanStack = new byte[NJFrames][];	
			blueByteChanStack = new byte[NJFrames][];
			byteChanStackLoaded = new boolean[NJFrames];
			for(int i=0;i<NJFrames;i++)
				byteChanStackLoaded[i]=false;
			}
		}
	
	public void disposeOfByteChanStack() {
		int i, instack;
		if(redByteChanStack != null && (instack=redByteChanStack.length)>0)
			for(i=0;i<instack;i++)
				if(redByteChanStack[i] != null)
					redByteChanStack[i] =null;
		
		if(greenByteChanStack != null && (instack=greenByteChanStack.length)>0)
			for(i=0;i<instack;i++)
				if(greenByteChanStack[i] != null)
					greenByteChanStack[i] =null;
		
		if(blueByteChanStack != null && (instack=blueByteChanStack.length)>0)
			for(i=0;i<instack;i++)
				if(blueByteChanStack[i] != null)
					blueByteChanStack[i] =null;
		redByteChanStack=null;
		greenByteChanStack=null;
		blueByteChanStack=null;
		byteChanStackLoaded=null;
		}

	private static final void makeAdd256() {
		add256 = new short[256+128];
		byte bytebright;
		int i,j;
		for(i=0;i<256;i++) {
			bytebright = (byte)i;
			j = bytebright+256;
			add256[j] = (short)i;
			}
		}
			

	public void sepuku() {
	parent=null;
	rawImage=null;
	navigator=null;
	offScrImage=null;
	redfile=null;
	greenfile=null;
	bluefile=null;
	curIm = null;
	Xshifts = null;
	Yshifts = null;
	Zshifts = null;
	redLUT = null;
	redShiftLUT = null;
	greenLUT = null;
	greenShiftLUT = null;
	blueLUT = null;
	redLUT256 = null;
	redShiftLUT256  = null;
	greenLUT256  = null;
	greenShiftLUT256  = null;
	blueLUT256 = null;
	redRedness = null;
	greenRedness = null;
	blueRedness = null;
	redGreenness = null;
	greenGreenness =null;
	blueGreenness = null;
	redBlueness = null;
	greenBlueness = null;
	blueBlueness = null;
	}
				
	BioradCanvas(BioradFile redfile,BioradFile greenfile,BioradFile bluefile, 
			ShowPicJPanel parent, int sectionType,SectionNavigator navigator) {
		int		i;
		if(ifdiags) talk("Constructor for BioradCanvas");
		this.navigator = navigator;
		this.sectionType = sectionType;
		switch (sectionType) {
			default :
			case 0:
				sectionTypeName= "XY-";
				break;
			case 1:
				sectionTypeName= "XZ-";
				break;
			case 2:
				sectionTypeName= "ZY-";
				break;
			}

		if(redfile != null) {
			this.redfile = redfile;
			myRedFileName=redfile.filesName;
			ifred=true;
			}
		else {
			this.redfile = null;
			myRedFileName=null;
			ifred=false;
			}
			
		if(greenfile != null) {
			this.greenfile = greenfile;
			myGreenFileName=greenfile.filesName;
			ifgreen=true;
			}
		else {
			this.greenfile = null;
			myGreenFileName=null;
			ifgreen=false;
			}
			
		if(bluefile != null) {
			this.bluefile = bluefile;
			myBlueFileName=bluefile.filesName;
			ifblue=true;
			}
		else {
			this.bluefile = null;
			myBlueFileName=null;
			ifblue=false;
			}
      	this.parent = parent;
 		NJFrames=howManySections();
		if(ifdiags) talk("BioradCanvas constructor, NJFrames="+NJFrames);
		setImageSize();
      	curJFrame=NJFrames/2; // this will be overridden when new files are loaded.
      	tellStatsJPanelWhichJFrame();
		requestFocus();
	    setLocation(400,20);
	}

   private void tellStatsJPanelWhichJFrame() {
   	if(parent==null || parent.parent==null || parent.parent.histogramJPanel==null) return;
   	parent.parent.histogramJPanel.sectionChangedAlert();
   	}
   	
   private void markTime() {
    	timeMark=(new Date()).getTime();
		}
		
    private long timeSinceMark() {
		return((new Date()).getTime() - timeMark);
		}
		
   private void mark2Time() {
    	time2Mark=(new Date()).getTime();
		}
		
    private long timeSince2Mark() {
		return((new Date()).getTime() - time2Mark);
		}

	public void unScaleCoords(double [] XYZ) {
		double x = XYZ[0]-parent.parent.xDrawOrg;
		double y = XYZ[1]-parent.parent.yDrawOrg;
		if(x<parent.XMIN) x=parent.XMIN;
		if(x>=parent.fullWidth) x=parent.fullWidth-1;
		if(y<parent.YMIN)y=parent.YMIN;
		if(y>=parent.fullHeight) y=parent.fullHeight-1;
		XYZ[0] = x;
		XYZ[1] = y;
		}
	public void scaleCoords(double [] XYZ) {
		//obsolete;
		}
		
	public void newLUTS(int nsection) {
		int i,bright,itab;
		// byte 0->0(256), byte1->1(257), ..., byte127->127(383),, 
		// pixel 128-> byte-127 -> 128,   pixel 255 -> byte -1 -> 255,  ...
		int [] redColors = parent.getRedsColors();		
		int [] greenColors = parent.getGreensColors();		
		int [] blueColors = parent.getBluesColors();
		int [] priRedLUT = parent.getRedsCustLUT();
		int [] priGreenLUT = parent.getGreensCustLUT();
		int [] priBlueLUT = parent.getBluesCustLUT();
		double [] boosts = parent.getBoosts(nsection);
		// talk("newLUTS() gets boosts = ("+boosts[0]+","+boosts[1]+","+boosts[2]+")");
	    // redColors[3],greenColors[3],blueColors[3] determine up-concavity...		    
		for(i=0;i<256;i++) {
			if(i<128) itab = 256+i;
			else itab = i; // this weird packing of the LUT is because java byte variables
			// are signed. The negative values correspond with brightness values 128 to 255
			if(redness<0) bright=0;
			else bright = (int)(.49 + boosts[0]*priRedLUT[i]); // chnged july 2001 to implement stackGradient
			if(bright>255) bright=255;
			redLUT[itab]=bright;
			redShiftLUT[itab]=(bright << 16);
			redLUT256[i] = bright;
			redShiftLUT256[i] = (bright <<16);
			redRedness[i] = (bright*redColors[0])/255;
			redGreenness[i] = (bright*redColors[1])/255;
			redBlueness[i] = (bright*redColors[2])/255;
			
			if(greenness<0) bright=0;
			else bright = (int)(.49 + boosts[1]*priGreenLUT[i]);// chnged july 2001 to implement stackGradient
			if(bright>255) bright=255;
			greenLUT[itab]=bright;
			greenShiftLUT[itab]= (bright << 8);
			greenLUT256[i] = bright;
			greenShiftLUT256[i] = (bright <<8);
			greenRedness[i] = (bright*greenColors[0])/255;
			greenGreenness[i] = (bright*greenColors[1])/255;
			greenBlueness[i] = (bright*greenColors[2])/255;
			
			
			if(blueness<0) bright=0;
			else bright = (int)(.49 + boosts[2]*priBlueLUT[i]);// chnged july 2001 to implement stackGradient
			if(bright>255) bright=255;
			blueLUT256[i] = bright;
			blueLUT[itab]=bright;
			blueRedness[i] = (bright*blueColors[0])/255;
			blueGreenness[i] = (bright*blueColors[1])/255;
			blueBlueness[i] = (bright*blueColors[2])/255;
			}
		}
	

	public void setBrightness(double redness, double greenness, double blueness,
				boolean ifUseGrayScale, int singleChan) {
		this.redness=redness;
		this.greenness=greenness;
		this.blueness=blueness;
		this.ifUseGrayScale=ifUseGrayScale;
		this.singleChan = singleChan;
		newLUTS(curJFrame);
		}			

	public int howManySections() {
		int nf=0;
		nhigh=10;	nwide=10;
		if(ifred && redfile.numJFrames>nf) {
			 nf = redfile.numJFrames;
			 nwide=redfile.width;
			 nhigh=redfile.height;
			 }
		if(ifgreen && greenfile.numJFrames>nf) {
				nf = greenfile.numJFrames;
			 nwide=greenfile.width;
			 nhigh=greenfile.height;
			 }
		if(ifblue && bluefile.numJFrames>nf) {
			nf = bluefile.numJFrames;
			 nwide=bluefile.width;
			 nhigh=bluefile.height;
			 }
		//talk("The stack has "+nf+" sections, each with width "+nwide+", and height "+nhigh);
		return nf;
		}
		
	static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
	
	public int adjustCacheSize(int cs) {
		if(cs<1) cs=1;
		int cacheSize=0;
		int ncac=0;
		if(redfile != null) {
			if((ncac=cs)>redfile.numJFrames) ncac = redfile.numJFrames;
			cacheSize=redfile.adjustCacheSize(ncac);
			}
		if(greenfile != null) {
			if((ncac=cs)>greenfile.numJFrames) ncac = greenfile.numJFrames;
			cacheSize=greenfile.adjustCacheSize(ncac);
			}
		if(bluefile != null) {
			if((ncac=cs)>bluefile.numJFrames) ncac = bluefile.numJFrames;
			cacheSize=bluefile.adjustCacheSize(ncac);
			}
		double megs = (3.0*parent.fullWidth*parent.fullHeight*cs)/1000000.0;
		talk("raster caches use "+megs+" megabytes, and holds "+cs+" biorad pics");
		return(cacheSize);
		}

	public void setImageWindow() {
		showAJFrame(curJFrame);
		}
	
	public String getRedStackName() {	
		// This attempts to return the filename, without the directory path
		String str = redfile.filename;
		int i = str.lastIndexOf(File.separator);
		if(i == -1) return(str);
		return(str.substring(i + 1));
	}
	public String getGreenStackName() {	
		// This attempts to return the filename, without the directory path
		String str = greenfile.filename;
		int i = str.lastIndexOf(File.separator);
		if(i == -1) return(str);
		return(str.substring(i + 1));
	}
	public String getBlueStackName() {	
		// This attempts to return the filename, without the directory path
		String str = bluefile.filename;
		int i = str.lastIndexOf(File.separator);
		if(i == -1) return(str);
		return(str.substring(i + 1));
	}
	
	public void showNextJFrame() {
		if(curJFrame < NJFrames-1)
			showAJFrame(curJFrame+1);
		}
		
	public void showPreviousJFrame() {
		if(curJFrame>0)
			showAJFrame(curJFrame-1);
		}

	public void showNextJFrameUrgently(int maxDelay) {
		if(curJFrame < NJFrames-1)
			showAJFrameUrgently(curJFrame+1,maxDelay);
		}
		
	public void showPreviousJFrameUrgently(int maxDelay) {
		if(curJFrame>0)
			showAJFrameUrgently(curJFrame-1,maxDelay);
		}

	public void showMiddleJFrame() {
		curJFrame = (NJFrames-1)/2;
		showAJFrame(curJFrame);
		}
		
	public void showFirstJFrame() {
		curJFrame = 0;
		showAJFrame(curJFrame);
		}
		
	public void showLastJFrame() {
		curJFrame = NJFrames-1;
		showAJFrame(curJFrame);
		}
		
	public void showJumpJFrame(int njump) {
		if(njump>=0) { // jump forward
			curJFrame += njump;
			if(curJFrame >= NJFrames-1)
				curJFrame = NJFrames-1;
			}
		if(njump<0) { // jump forward
			curJFrame += njump;
			if(curJFrame < 0)
				curJFrame = 0;
			}
		showAJFrame(curJFrame);
		}
		
	public void setImageSize() {
      	curWidth = nwide;  
      	curHeight = nhigh;
    	setSize(curWidth, curHeight);
		}
	
	private int [] drawArtColors(int nframe) {
		// get the colors...
		int [] redColors = parent.getRedsColors();		
		int [] greenColors = parent.getGreensColors();		
		int [] blueColors = parent.getBluesColors();
	    int npix=curWidth*curHeight,artred, artgreen, artblue, 
		red, green,  blue, i,alpha=(255 << 24);
	    int [] rasRGB = null;
	    byte [] redRaster=getOneByteChan(nframe, 0);
	    byte [] greenRaster=getOneByteChan(nframe, 1);
	    byte [] blueRaster=getOneByteChan(nframe, 2);
	    // talk("drawArtColors got the red, green, blue rasters");
	    try { // allocate the RGB raster 
		rasRGB = new int[npix];
		}
	    catch(RuntimeException e) {		// Probably a memory error
	    talk("drawArtColors with nframe="+nframe+", got exception "+e);
		Runtime.getRuntime().gc();	// Clean up what we just got rid of
		return null;
		}
	    red=0;  green=0;	blue=0;
	    for(i=0;i<npix;i++) {
		if(redRaster != null && (red = redRaster[i])<0) red += 256;
		if(greenRaster != null && (green = greenRaster[i])<0) green += 256;
		if(blueRaster != null && (blue = blueRaster[i])<0) blue += 256;
		if((artred = redRedness[red]+ greenRedness[green] + blueRedness[blue])>255)
		    artred=255;
		if((artgreen = redGreenness[red]+ greenGreenness[green] + blueGreenness[blue])>255)
		    artgreen=255;
		if((artblue = redBlueness[red]+ greenBlueness[green] + blueBlueness[blue])>255)
		    artblue=255;
		rasRGB[i] = (alpha | (artred << 16) | (artgreen << 8) | (artblue));
		}
	    return (rasRGB);
	}
	
  private byte [] getOneByteChan(int nframe, int theChan) {
	int npix=curWidth*curHeight,alpha=(255 << 24);
	int pix=0, rowf,row,col,column;
	int [] LUT=null;
	int rowshf,shiftx,shifty, row1, row2, col1, col2, rowtab;
	boolean ifGotPic=false;
	byte [] rasRGB = null;
	byte [] raster=null;
		
	try { // allocate the RGB raster 
	    rasRGB = new byte[npix];
	    }
	catch(RuntimeException e) {		// Probably a memory error
	    Runtime.getRuntime().gc();	// Clean up what we just got rid of
	    talk("Probable out of memory exception... rasRGB failed to allocate");
	    return null;
	    }

	try {
	    // get the image raster...
	    switch (theChan) {
		    case 0: default:
			if(redfile == null) return null; 
			raster = redfile.readBioradRaster(nframe-Zshifts[0]);
			if(raster==null) return null;
			break;
		    case 1:
			if(greenfile == null) return null; 
			raster = greenfile.readBioradRaster(nframe-Zshifts[1]);
			if(raster==null) return null;
			break;
		    case 2:
			if(bluefile == null) return null; 
			raster = bluefile.readBioradRaster(nframe-Zshifts[2]);
			if(raster==null) return null;
			break;
		    }
	    }
	 catch(RuntimeException e) {		// Probably a memory error, so guess that that's what it is
		Runtime.getRuntime().gc();	// Clean up what we just got rid of
		talk("getOneByteChan failed for frame number "+nframe);
		return null;
		}
	   pix=0;  shiftx=Xshifts[theChan];	shifty=Yshifts[theChan];
	   //  get row limits
	   //  rowshf = row+shifty ==0 <==> row== - shifty;
	   //  rowshf = row+shifty == nhigh-1 <==> row== (nhigh-1- shifty);
	    if((row1 = - shifty)<0) row1=0;
	    if((row2 = (nhigh-1- shifty))>= curHeight) row2=curHeight-1;
	    // get column limits...
	    //  col=(col+shiftx)==0 <==> col == -shiftx
	    //  col=(col+shiftx)==nwide-1 <==> col == (nwide-1-shiftx)
	    if((col1 =  -shiftx)<0) col1=0;
	    if((col2 =  nwide-1-shiftx)>=curWidth) 
	    	col2 = curWidth-1;
	    rowf = curWidth*(row1 - 1);
	    rowtab = nwide;
	    if((rowshf = shifty + row1-1)+1 <0)
	    	rowshf = -1;
	    rowshf *= nwide;
	    for(row=row1;row<=row2;row++) { 
		    rowf += curWidth; // compute everything dependent on row index
				      // outside of col++ loop
		    // should have:  rowshf = nwide*(row+shifty)  = 
		    //			    nwide*shifty +(nwide)*row;
		    rowshf += rowtab;
		    column = (col1-1 + shiftx);
		    for(col=col1;col<=col2;col++) {
			    //column should equal col;
			    column += 1;
			// get the pixel at (col+shiftx, row+shifty)
			    rasRGB[rowf+col] = raster[rowshf+column];
			    }
		   }
	    
	    // black out the edges...
	    for(row=0;row<curHeight;row++) {
		rowf = curWidth*row;
		for(col=0;col<col1;col++) {
		    rasRGB[rowf+col]=0;
		    }
		for(col=col2;col<curWidth;col++) {
		    rasRGB[rowf+col]=0;
		    }
		}
		
	    for(row=row2;row<curHeight;row++) {
		rowf = curWidth*row;
		for(col=0;col<curWidth;col++) {
		    rasRGB[rowf+col]=0;
		    }
		}
		
	    for(row=0;row<row1;row++) {
		rowf = curWidth*row;
		for(col=0;col<curWidth;col++) {
		    rasRGB[rowf+col]=0;
		    }
		}
	    return rasRGB;
	}

 private int [] drawJustOneArtChan(int nframe, int theChan) {
 	int npix=curWidth*curHeight,alpha=(255 << 24);
	int pix=0, rowf,row,col,column, artred, artgreen, artblue, bright;
	int [] redNessLUT=null,  greenNessLUT=null,  blueNessLUT=null;
	int rowshf,shiftx,shifty, row1, row2, col1, col2, rowtab;
	boolean ifGotPic=false;
	int [] rasRGB = null;
	byte [] raster=null;
		
	if(ifdiags) talk("drawJustOneArtChan("+nframe+", "+theChan+")");
	try { // allocate the RGB raster 
	    rasRGB = new int[npix];
	    }
	catch(RuntimeException e) {		// Probably a memory error
	    Runtime.getRuntime().gc();	// Clean up what we just got rid of
	    talk("Probable out of memory exception... rasRGB failed to allocate");
	    return null;
	    }

	try {
	    // get the image raster...
	    switch (theChan) {
		    case 0: default:
				if(redfile == null) return null; 
				raster = redfile.readBioradRaster(nframe-Zshifts[0]);
				if(raster==null) return null;
				break;
		    case 1:
				if(greenfile == null) return null; 
				raster = greenfile.readBioradRaster(nframe-Zshifts[1]);
				if(raster==null) return null;
				break;
		    case 2:
				if(bluefile == null) return null; 
				raster = bluefile.readBioradRaster(nframe-Zshifts[2]);
				if(raster==null) return null;
				break;
		    }
	    }
	 catch(RuntimeException e) {		// Probably a memory error, so guess that that's what it is
		Runtime.getRuntime().gc();	// Clean up what we just got rid of
		talk("drawJustOneArtChan failed for frame number "+nframe);
		return null;
		}
		if(ifdiags) talk("drawJustOneArtChan() got the raster....");
	    switch(theChan) {
		    case 0 :
			redNessLUT = redRedness;
			greenNessLUT = redGreenness; 
			blueNessLUT = redBlueness; 
			break;
			
		    case 1 :
			redNessLUT = greenRedness;
			greenNessLUT = greenGreenness; 
			blueNessLUT = greenBlueness; 
			break;
			
		    case 2 :
			redNessLUT = blueRedness;
			greenNessLUT = blueGreenness; 
			blueNessLUT = blueBlueness; 
			break;
		    }
	   pix=0;  shiftx=Xshifts[theChan];	shifty=Yshifts[theChan];
	   if(ifdiags) talk("drawJustOneArtChan("+theChan+"), shiftx="+shiftx+", shifty="+shifty);
	   //  get row limits
	   //  rowshf = row+shifty ==0 <==> row== - shifty;
	   //  rowshf = row+shifty == nhigh-1 <==> row== (nhigh-1- shifty);
	    if((row1 = - shifty)<0) row1=0;
	    if((row2 = (nhigh-1- shifty))>= curHeight) row2=curHeight-1;
	    // get column limits...
	    //  col=(col+shiftx)==0 <==> col == -shiftx
	    //  col=(col+shiftx)==nwide-1 <==> col == (nwide-1-shiftx)
	    if((col1 =  -shiftx)<0) col1=0;
	    if((col2 =  (nwide-1-shiftx))>=curWidth) col2 = curWidth-1;
	    rowf = curWidth*(row1 - 1);
	    rowtab = nwide;
	    if((rowshf = shifty + row1-1) <0)
	    	rowshf = -1;
	    rowshf *= nwide;
	    for(row=row1;row<=row2;row++) { 
		    rowf += curWidth; // compute everything dependent on row index
				      // outside of col++ loop
		    // should have:  rowshf = nwide*(row+shifty)  = 
		    //			    nwide*shifty +(nwide)*row;
		    rowshf += rowtab;
		    column = ((col1-1) + (shiftx));
		    for(col=col1;col<=col2;col++) {
			    //column should equal col;
			column += 1;
			// get the pixel at (col+shiftx, row+shifty)
			if((bright = raster[rowshf+column])<0) bright += 256;
			artred = redNessLUT[bright];
			artgreen = greenNessLUT[bright];
			artblue = blueNessLUT[bright];
			rasRGB[rowf+col] = (alpha |artred<<16 | artgreen<<8 | artblue);
			}}
	    // black out the edges...
	    for(row=0;row<curHeight;row++) {
		rowf = curWidth*row;
		for(col=0;col<col1;col++) {
		    rasRGB[rowf+col]=alpha;
		    }
		for(col=col2;col<curWidth;col++) {
		    rasRGB[rowf+col]=alpha;
		    }
		}
		
	    for(row=row2;row<curHeight;row++) {
		rowf = curWidth*row;
		for(col=0;col<curWidth;col++) {
		    rasRGB[rowf+col]=alpha;
		    }
		}
		
	    for(row=0;row<row1;row++) {
		rowf = curWidth*row;
		for(col=0;col<curWidth;col++) {
		    rasRGB[rowf+col]=alpha;
		    }
		}
		
		return rasRGB;

    }
   
 private int [] drawJustOneChan(int nframe, int theChan, boolean ifgray) {
	int npix=curWidth*curHeight,alpha=(255 << 24);
	int pix=0, rowf,row,col,column;
	int [] LUT=null;
	int rowshf,shiftx,shifty, row1, row2, col1, col2, rowtab;
	boolean ifGotPic=false;
	int [] rasRGB = null;
	byte [] raster=null;
		
	if(ifdiags) talk("drawJustOneChan("+nframe+", "+theChan+")");
	try { // allocate the RGB raster 
	    rasRGB = new int[npix];
	    }
	catch(RuntimeException e) {		// Probably a memory error
	    Runtime.getRuntime().gc();	// Clean up what we just got rid of
	    talk("Probable out of memory exception... rasRGB failed to allocate");
	    return null;
	    }

	try {
	    // get the image raster...
	    switch (theChan) {
		    case 0: default:
				if(redfile == null) return null; 
				raster = redfile.readBioradRaster(nframe-Zshifts[0]);
				if(raster==null) return null;
				break;
		    case 1:
				if(greenfile == null) return null; 
				raster = greenfile.readBioradRaster(nframe-Zshifts[1]);
				if(raster==null) return null;
				break;
		    case 2:
				if(bluefile == null) return null; 
				raster = bluefile.readBioradRaster(nframe-Zshifts[2]);
				if(raster==null) return null;
				break;
		    }
	    }
	 catch(RuntimeException e) {		// Probably a memory error, so guess that that's what it is
		Runtime.getRuntime().gc();	// Clean up what we just got rid of
		talk("drawJustOneChan failed for frame number "+nframe);
		return null;
		}
		if(ifdiags) talk("drawJustOneChan() got the raster....");
	    switch(theChan) {
		    case 0 :LUT = ifgray? redLUT : redShiftLUT; break;
		    case 1 :LUT = ifgray? greenLUT : greenShiftLUT; break;
		    case 2 :LUT = blueLUT; break;
		    }
	   pix=0;  shiftx=Xshifts[theChan];	shifty=Yshifts[theChan];
	   if(ifdiags) talk("drawJustOneChan("+theChan+"), shiftx="+shiftx+", shifty="+shifty);
	   //  get row limits
	   //  rowshf = row+shifty ==0 <==> row== - shifty;
	   //  rowshf = row+shifty == nhigh-1 <==> row== (nhigh-1- shifty);
	    if((row1 = - shifty)<0) row1=0;
	    if((row2 = (nhigh-1- shifty))>= curHeight) row2=curHeight-1;
	    // get column limits...
	    //  col=(col+shiftx)==0 <==> col == -shiftx
	    //  col=(col+shiftx)==nwide-1 <==> col == (nwide-1-shiftx)
	    if((col1 =  -shiftx)<0) col1=0;
	    if((col2 =  (nwide-1-shiftx))>=curWidth) col2 = curWidth-1;
	    rowf = curWidth*(row1 - 1);
	    rowtab = nwide;
	    if((rowshf = shifty + (row1-1)) <0)
	    	rowshf = -1;
	    rowshf *= nwide;
	    for(row=row1;row<=row2;row++) { 
		    rowf += curWidth; // compute everything dependent on row index
				      // outside of col++ loop
		    // should have:  rowshf = nwide*(row+shifty)  = 
		    //			    nwide*shifty +(nwide)*row;
		    rowshf += rowtab;
		    column = ((col1-1) + (shiftx));
		    if(ifgray) {
			  for(col=col1;col<=col2;col++) {
			    //column should equal col;
			    column += 1; 
			// get the pixel at (col+shiftx, row+shifty)
			    pix = LUT[(256+raster[rowshf+column])];
			    rasRGB[rowf+col] = (alpha | (pix << 16) | (pix << 8) | pix);
			    }}
		    else {
			for(col=col1;col<=col2;col++) {
			    //column should equal col;
			    column += 1;
			// get the pixel at (col+shiftx, row+shifty)
			    pix = LUT[256+raster[rowshf+column]];
			    rasRGB[rowf+col] = (alpha | pix);
			    }}
		   }
	    // black out the edges...
	    for(row=0;row<curHeight;row++) {
		rowf = curWidth*row;
		for(col=0;col<col1;col++) {
		    rasRGB[rowf+col]=alpha;
		    }
		for(col=col2;col<curWidth;col++) {
		    rasRGB[rowf+col]=alpha;
		    }
		}
		
	    for(row=row2;row<curHeight;row++) {
		rowf = curWidth*row;
		for(col=0;col<curWidth;col++) {
		    rasRGB[rowf+col]=alpha;
		    }
		}
		
	    for(row=0;row<row1;row++) {
		rowf = curWidth*row;
		for(col=0;col<curWidth;col++) {
		    rasRGB[rowf+col]=alpha;
		    }
		}
		
		return rasRGB;
	}

     private int [] draw3Pri(int nframe) {
	int [] rasRGB = drawJustOneChan(nframe, 0, false); // start with the red channel
	
	if(rasRGB == null) rasRGB=drawJustOneChan(nframe, 1, false);
	else rasRGB = addOneChan(nframe, 1,  rasRGB); // add the green channel
	
	if(rasRGB == null) rasRGB=drawJustOneChan(nframe, 2, false);
	else rasRGB = addOneChan(nframe, 2,  rasRGB); // add the green channel
	
	return rasRGB;
	}
	
 private int [] addOneChan(int nframe, int theChan,  int [] rasRGB) {
	int npix=curWidth*curHeight,alpha=(255 << 24);
	int pix=0, rowf,row,col,column;
	int [] LUT=null;
	int rowshf,shiftx,shifty, row1, row2, col1, col2, rowtab;
	boolean ifGotPic=false;
	byte [] raster=null;
		

	if(ifdiags) talk("addOneChan("+nframe+", "+theChan+")");
	try {
	    // get the image raster...
	    switch (theChan) {
		    case 0: default:
			if(redfile == null) return rasRGB; 
			raster = redfile.readBioradRaster(nframe-Zshifts[0]);
			if(raster==null) return rasRGB;
			break;
		    case 1:
			if(greenfile == null) return rasRGB; 
			raster = greenfile.readBioradRaster(nframe-Zshifts[1]);
			if(raster==null) return rasRGB;
			break;
		    case 2:
			if(bluefile == null) return rasRGB; 
			raster = bluefile.readBioradRaster(nframe-Zshifts[2]);
			if(raster==null) return rasRGB;
			break;
		    }
	    }
	 catch(RuntimeException e) {		// Probably a memory error, so guess that that's what it is
		Runtime.getRuntime().gc();	// Clean up what we just got rid of
		talk("addOneChan failed for frame number "+nframe);
		return rasRGB;
		}
	   switch(theChan) {
		    case 0 :LUT = redShiftLUT; break;
		    case 1 :LUT = greenShiftLUT; break;
		    case 2 :LUT = blueLUT; break;
		    }
	   pix=0;  shiftx=Xshifts[theChan];	shifty=Yshifts[theChan];
	   if(ifdiags) talk("addOneChan("+theChan+"), shiftx="+shiftx+", shifty="+shifty);
	   //  get row limits
	   //  rowshf = row+shifty ==0 <==> row== - shifty;
	   //  rowshf = row+shifty == nhigh-1 <==> row== (nhigh-1- shifty);
	    if((row1 = - shifty)<0) row1=0;
	    if((row2 = (nhigh-1- shifty))>= curHeight) row2=curHeight-1;
	    // get column limits...
	    //  col=(col+shiftx)==0 <==> col == -shiftx
	    //  col=(col+shiftx)==nwide-1 <==> col == (nwide-1-shiftx)
	    if((col1 =  -shiftx)<0) col1=0;
	    if((col2 =  (nwide-1-shiftx))>=curWidth) col2 = curWidth-1;
	    rowf = curWidth*(row1 - 1);
	    rowtab = nwide;
	    if((rowshf = shifty + (row1-1)) <0)
	    	rowshf = -1;
	    rowshf *= nwide;
	    for(row=row1;row<=row2;row++) { 
		    rowf += curWidth; // compute everything dependent on row index
				      // outside of col++ loop
		    // should have:  rowshf = nwide*(row+shifty)  = 
		    //			    nwide*shifty +(nwide)*row = 
		    //			   nwide*shifty * rowtab*row
		    rowshf += rowtab;
		    column = ((col1-1) + (shiftx));
		    for(col=col1;col<=col2;col++) {
			//column should equal col;
			column += 1;
			// get the pixel at (col+shiftx, row+shifty)
			pix = LUT[256+raster[rowshf+column]];
			rasRGB[rowf+col] = (rasRGB[rowf+col] | pix);
			}
		   }
	    return rasRGB;
	}
	
	private int minOf3(int [] aray) {
	    int min = aray[0];
	    if(aray[1]<min) min=aray[1];
	    if(aray[2]<min) min=aray[2];
	    return min;
	    }
	private int maxOf3(int [] aray) {
	    int max = aray[0];
	    if(aray[1]>max) max=aray[1];
	    if(aray[2]>max) max=aray[2];
	    return max;
	    }
	
	public int [] edgeTrim(int [] ras) { // paint to black the edge strips where 
			// the three color plaes don't overlap
	    int maxx = -minOf3(Xshifts);
	    int minx= -maxOf3(Xshifts);  
	    int maxy = -minOf3(Yshifts);
	    int miny= -maxOf3(Yshifts);
	    int row, col;
	    if(maxy>0) 
		for(row=0; row<maxy; row++) 
		    for(col=0;col<parent.fullWidth;col++)
			ras[row*parent.fullWidth+col]=0;  
	    
	    if(miny<0)
		for(row=parent.fullHeight-1+miny; row<parent.fullHeight; row++) 
		    for(col=0;col<parent.fullWidth;col++)
			ras[row*parent.fullWidth+col]=0;  
	    
	    if(maxx>0)
		for(col=0; col<maxx; col++)
		    for(row=0; row<parent.fullHeight; row++)
			ras[row*parent.fullWidth+col]=0;  
		    	    
	    if(minx<0)
		for(col=parent.fullWidth-1+minx; col<parent.fullWidth; col++)
		    for(row=0; row<parent.fullHeight; row++)
			ras[row*parent.fullWidth+col]=0;  
	    return ras;
	    }    	    
	
	/*  private static final int paintblend(int RGBvox, int Tvox, int filmRGB, int filmCapacityRemaining) {
		given the current brightness of the film in filmRGB, and the brightness of the voxel just hit by the ray-tracing
		ray in RGBvox, and given the transparency of "the voxel" in Tvox, this method updates values of 
		filmRGB and filmCapacityRemaining. No arguments are passed to save time.
	
	Recipr used in e6.c is:
	(Rpix,Gpix,Bpix) is the red-green-blue content of the voxel, and Tpix is the transparency of the voxel
	(red,green,blue) is the current red-green-blue content of the fil pixel
	tran is the current holding capacity of the flm pixel (starts at 255
	depthfader() is the depth cue factor for the depth of the current voxel
	cube255 = 255*255*255
	To incorporate the current voxel,
	trandepth=tran*depthfader()*(255-Tpix);
	red -> red + (trandepth*Rpix)/cube255;
	green -> green + (trandepth*Gpix)/cube255;
	blue -> blue + (trandepth*Bpix)/cube255;
	tran -> (tran*Tpix)/255
				
		blender = filmCapacityRemaining*depthFadeVox*(255-Tvox)*RGBvox;
		filmRGB += (int)(blender/cube255);
		filmCapacityRemaining -> (filmCapacityRemaining*Tvox)/255;
		}
		
	private static final int filmCapacity(int filmCapacityRemaining, int Tvox) {
		return (filmCapacityRemaining*Tvox)/255;
		} */

	private static final void accumulateOneSectIntoStereoPair(int stereoWidth, int currentWidth, int currentHeight, int leftMargin, int rightMargin, 
		int from, int to, int ifr, byte [] byteChan, short [] rgbmax, short [] rgbtrace, short [] filmCapacityRemaining, int [] transpLUT, 
			boolean ifdomaxproj, boolean ifdoraytrace, boolean ifStereo,
		    	double colShiftPerSection,double swingOffset, int depthFadeVox) {
		int col,row,rightcol,leftcol,rowcol,offset,tces,Tvox;
		int leftViewRightStop = (3*stereoWidth)/4, rightViewLeftStop = stereoWidth/4;
		long blender, cube255 = 255*255*255;
		short brightshort;

		if(add256==null)
			makeAdd256();
			
	    if(colShiftPerSection >= 0) 
	    	tces = from;
	    else {
	    	tces = to;
	    	colShiftPerSection = -colShiftPerSection;
	    	}
				
		for(col=0;col<currentWidth;col++) {
			leftcol = leftMargin + col + (int)(.49 + (swingOffset + colShiftPerSection)*(ifr - tces));
			if(leftcol<0 || leftcol > leftViewRightStop)
				leftcol=-1; // negative value flags illegal column
			if(ifStereo) {
				rightcol = rightMargin + col + (int)(.49 +  (swingOffset - colShiftPerSection)*(ifr - tces));
				if(rightcol < rightViewLeftStop || rightcol >= stereoWidth)
					rightcol = -1;
				}
			else rightcol=-1; // negative value flags illegal column = way to supress sterep pairs
			for(row=0;row<currentHeight;row++){
				rowcol = row*currentWidth + col;
		    	brightshort = add256[256+byteChan[rowcol]];
		    	if(leftcol >= 0) {
		    		offset = row*stereoWidth + leftcol;		    				
		    		if(ifdomaxproj && brightshort > rgbmax[offset]) 
		    				rgbmax[offset] = brightshort;
		    		if(ifdoraytrace) {
/*George, the Next tfour lines of code are the ones that accumulate brightness into the film pixel along a traced ray. The same identical lines
appear immediately below. It is faster to just duplicate these lines instead of using coding tricks to use one copy of the code twice. One copy is for the
left eye view and the other copy is for the right eye view. */
		    			Tvox = transpLUT[brightshort];     // this is the transparency of the voxel that the ray hit. That transparency is determined
		    												// from the brightness of the pixel.
		    			blender = filmCapacityRemaining[offset]*depthFadeVox*(255-Tvox)*brightshort; /* (255-Tvox) is the opacity of the voxel. brightshort
		    										is the brightness of the voxel. depthFadeVox is an attenuation factor for the current optical section
		    										filmCapacityRemaining[offset] is how much brightness can be added to the current film pixel from which
		    										the tracd ray emanated. */ 
						rgbtrace[offset] += (int)(blender/cube255); // all factors in the equation for blender are between 0 and 255. Dividing by 255^3 normalizes.
						filmCapacityRemaining[offset] = (short)((filmCapacityRemaining[offset]*Tvox)/255); // decrement the capacity of the pixel
																			// according to the transparency of the voxel just processed
		    			}
		    		}	
		    	if(rightcol >= 0) {
		    		offset = row*stereoWidth + rightcol;
		    		if(ifdomaxproj && brightshort > rgbmax[offset]) 
		    			rgbmax[offset] = brightshort;
		    		if(ifdoraytrace) {
		    			Tvox = transpLUT[brightshort];
		    			blender = filmCapacityRemaining[offset]*depthFadeVox*(255-Tvox)*brightshort;
						rgbtrace[offset] += (int)(blender/cube255);
						filmCapacityRemaining[offset] = (short)((filmCapacityRemaining[offset]*Tvox)/255);
		    			}
		    		}	
		        } 
		    }
		}
		    		  
	private static final int inlimits(int low,int hi,int i) {
		if(i<low)
			i=low;
		if(i>hi)
			i=hi;
		return i;
		}
			   		  		
	public void makeStereoPairOverlay(int [] rayTrace, boolean ifdoraytrace, int [] maxProj, boolean ifdomaxproj, 
		int stereoWidth, int leftMargin, int rightMargin, int from, int to, int depthCueBrightnessOfFromSection, int depthCueBrightnessOfToSection, 
			boolean ifdored, boolean ifdogreen, boolean ifdoblue, 
				int [] redtransp, int [] greentransp, int [] bluetransp, double redNonMaskFrac, double greenNonMaskFrac,
					double blueNonMaskFrac, double colShiftPerSection,double swingOffset, 
						boolean ifStereo, int movieFrameNumber, JTextField progressReportField) {
		/* returns two arrays. One is a maximum projection (in the maxProj array) and the other is a ray-traced projection that uses
		different transparencies for each channel.
			If ifStereo==true, they these arrays contain stereo pairs, and the algebraic sign of colShiftPerSection determines whether
		these are for cross-eyed viewing (when colShiftPerSection>0) or diverged-eye viewing (when colShiftPerSection<0).
			The transparencies[] array tells transparencies of the channels. transpae[rencies0] is for the red channel, etc.
		transparencies[1] = 0 means the green channel voxels are opaque. transparencies[1] means the green channel voxels are 
		completely transparent.
			For ray tracing, make from < to in order to send the ray from the first-scanned sections to the last-scanned sections.
		Make from > to in order to send the ray from the last-scanned sections to the first-scanned sections.
				*/
			
		if(!ifdoraytrace && !ifdomaxproj) {
			talk("Do at least one projection, maximum or raytracing");
			progressReportField.setText("Do at least one projection, maximum or raytracing");
			return;
			}		
			
		if(!ifdored && !ifdogreen && !ifdoblue) {
			talk("Include at least one channel, red, green, or blue");
			progressReportField.setText("Include at least one channel, red, green, or blue");
			return;
			}	
		
		depthCueBrightnessOfFromSection = inlimits(0,255,depthCueBrightnessOfFromSection);		
		depthCueBrightnessOfToSection = inlimits(0,255,depthCueBrightnessOfToSection);		
				
		int depthFadeVox;
	    int alpha=(255 << 24);
	    byte [] byteChan;
	    int npix = stereoWidth*curHeight; // both the maxProj and stereoRas arrays must have this length
	    short [] redwork=null, greenwork=null, bluework=null, redmax=null, greenmax=null, bluemax=null, 
	    	redFilmCapacityRemaining=null,greenFilmCapacityRemaining=null,blueFilmCapacityRemaining=null;
	    int ifr, i, ifrinc;
	    boolean ifdidred=false, ifdidgreen=false, ifdidblue=false;

	    ifrinc = (to >= from)? 1 : -1; 
	    	
	    if(redfile != null && ifdored) { // set up working arrays for red channel	    
			ifdidred=true;
			if(ifdoraytrace) {
				redwork =  new short[npix];
				redFilmCapacityRemaining=new short[npix];
				for(i=0;i<npix;i++) {
					redFilmCapacityRemaining[i]=255;
					redwork[i]=0;
					}
				}
			if(ifdomaxproj) {
				redmax =  new short[npix];
				for(i=0;i<npix;i++) 
					redmax[i]=0;
				}
	    	}
	    
	     	
	    if(greenfile != null && ifdogreen) { // set up working arrays for green channel	     	
			if(ifdoraytrace) {
				greenwork =  new short[npix];
				greenFilmCapacityRemaining = new short[npix];
				for(i=0;i<npix;i++) {
					greenFilmCapacityRemaining[i] = 255;
					greenwork[i] = 0;
					}
				}
			if(ifdomaxproj) {
				greenmax =  new short[npix];
				for(i=0;i<npix;i++) 
					greenmax[i]=0;
				}	
			ifdidgreen=true;
	    	}

	     	
	    if(bluefile != null && ifdoblue) {// set up working arrays for blue channel	     	
			if(ifdoraytrace) {
				bluework =  new short[npix];
				blueFilmCapacityRemaining=new short[npix];
				for(i=0;i<npix;i++) {
					blueFilmCapacityRemaining[i] = 255;
					bluework[i] = 0;
					}
				}
			if(ifdomaxproj) {
				bluemax =  new short[npix];
				for(i=0;i<npix;i++) 
					bluemax[i]=0;
				}	
		ifdidblue = true;
	    }

	    ifr=from;
	    byte [] rgbByteChanFromFilterChain;
	    int wasAtSection = curJFrame;
	    String progressString;
	    while(ifr != (to+ifrinc)) {
 			depthFadeVox = (int)(depthCueBrightnessOfFromSection + 
		     		(double)(ifr-from)*(depthCueBrightnessOfToSection-depthCueBrightnessOfFromSection)/((double)(to-from)));
	     	
	 		 if(!byteChanStackLoaded[ifr] && Stax.hub.segmentationJPanel.ifPutDataThroughFilterChain())
	 			 	Stax.hub.segmentationJPanel.getReadyToSegmentNewSection(ifr);
	     	
			if(ifdidred) { // accumulate red section into 3D reconstruction
				progressString = "red chan sect "+ifr;
				if(movieFrameNumber>=0)
					progressString += (" of movie frame "+movieFrameNumber);
				progressReport(progressString,progressReportField);
				if(!byteChanStackLoaded[ifr])
					redByteChanStack[ifr] = Stax.hub.segmentationJPanel.ifPutDataThroughFilterChain() ?
							Stax.hub.segmentationJPanel.processPrimaryChannelData(0,redNonMaskFrac) : 
								getOneByteChan(ifr, 0);
				byteChan = redByteChanStack[ifr]; 

				if(byteChan != null) {
					accumulateOneSectIntoStereoPair(stereoWidth,curWidth,curHeight,leftMargin,rightMargin, 
							from, to, ifr,byteChan,redmax, redwork, redFilmCapacityRemaining, redtransp, 
								ifdomaxproj, ifdoraytrace,ifStereo,colShiftPerSection,swingOffset, depthFadeVox); 
					}
				}
	     	     		     	
			if(ifdidgreen) {// accumulate green section into 3D reconstruction
				progressString = "green chan sect "+ifr;
				if(movieFrameNumber>=0)
					progressString += (" of movie frame "+movieFrameNumber);
				progressReport(progressString,progressReportField);

				if(!byteChanStackLoaded[ifr])
					greenByteChanStack[ifr] = Stax.hub.segmentationJPanel.ifPutDataThroughFilterChain()? 
						Stax.hub.segmentationJPanel.processPrimaryChannelData(1,greenNonMaskFrac) : 
									getOneByteChan(ifr, 1);
					byteChan = greenByteChanStack[ifr]; 
					if(byteChan != null) {
						accumulateOneSectIntoStereoPair(stereoWidth,curWidth,curHeight,leftMargin,rightMargin, 
								from, to, ifr,byteChan,greenmax, greenwork, greenFilmCapacityRemaining, greentransp, 
									ifdomaxproj, ifdoraytrace,ifStereo,colShiftPerSection,swingOffset, depthFadeVox); 
					}
			      }
		     	
			if(ifdidblue) {// accumulate blue section into 3D reconstruction
				progressString = "blue chan sect "+ifr;
				if(movieFrameNumber>=0)
					progressString += (" of movie frame "+movieFrameNumber);
				progressReport(progressString,progressReportField);

				if(!byteChanStackLoaded[ifr])
					blueByteChanStack[ifr] = Stax.hub.segmentationJPanel.ifPutDataThroughFilterChain()? 
						Stax.hub.segmentationJPanel.processPrimaryChannelData(2,blueNonMaskFrac) : 
									getOneByteChan(ifr, 2);
				byteChan = greenByteChanStack[ifr]; 
				if(byteChan != null)  {
					accumulateOneSectIntoStereoPair(stereoWidth,curWidth,curHeight,leftMargin,rightMargin, 
						from, to, ifr,byteChan, bluemax, bluework, blueFilmCapacityRemaining, bluetransp, 
								ifdomaxproj, ifdoraytrace,ifStereo,colShiftPerSection,swingOffset, depthFadeVox); 
					}
				}
			byteChanStackLoaded[ifr]=true;
			ifr += ifrinc;
    	}
	     	
	     /* now blend the arrays that were made. There are 7 possibilities:
	    alpha | red | green | blue :  case 1
	    alpha  | red | green       : case 2
	    alpha | red | blue         : case 3
	    alpha  | red               :case 4
	    
	    alpha | green | blue       : case 5
	    alpha  | green             : case 6
	    alpha | blue               : case 7
	     */
	    if(ifdoraytrace) { // normalize the ray-trace work arrays
	    	if(ifdidred)
	    		normalize(redwork);
	    	if(ifdidgreen)
	    		normalize(greenwork);
	    	if(ifdidblue)
	    		normalize(bluework);
	    	}
	    	
	     if(ifdidred) {
	     	if(ifdidgreen) {
	     		if(ifdidblue) { // case 1
 	    			if(ifdomaxproj) 
 	     				for(i=0; i<npix;i++) 
							maxProj[i] = alpha | (redmax[i] << 16) | (greenmax[i] << 8) | bluemax[i];
					if(ifdoraytrace)
 	     				for(i=0; i<npix;i++) 
							rayTrace[i] = alpha | (redwork[i] << 16) | (greenwork[i] << 8) | bluework[i];			
	     			}
	     		else {  // case 2
 	    			if(ifdomaxproj) 
 	     				for(i=0; i<npix;i++) 
							maxProj[i] = alpha | (redmax[i] << 16) | (greenmax[i] << 8);
					if(ifdoraytrace)
 	     				for(i=0; i<npix;i++) 
							rayTrace[i] = alpha | (redwork[i] << 16) | (greenwork[i] << 8);			
	     			}
	     		}
	     	else {  // dored but not green
	     		if(ifdidblue) { // case 3
 	    			if(ifdomaxproj) 
 	     				for(i=0; i<npix;i++) 
							maxProj[i] = alpha | (redmax[i] << 16) | bluemax[i];
					if(ifdoraytrace)
 	     				for(i=0; i<npix;i++) 
							rayTrace[i] = alpha | (redwork[i] << 16) | bluework[i];			
	     			}
	     		else {  // case 4
 	    			if(ifdomaxproj) 
 	     				for(i=0; i<npix;i++) 
							maxProj[i] = alpha | (redmax[i] << 16) ;
					if(ifdoraytrace)
 	     				for(i=0; i<npix;i++) 
							rayTrace[i] = alpha | (redwork[i] << 16);			
	     			}
	     		}
	     	}
	     		
	     else { // do not do red
	     	if(ifdidgreen) {
	     		if(ifdidblue) { // case 5
 	    			if(ifdomaxproj) 
 	     				for(i=0; i<npix;i++) 
							maxProj[i] = alpha |(greenmax[i] << 8) | bluemax[i];
					if(ifdoraytrace)
 	     				for(i=0; i<npix;i++) 
							rayTrace[i] = alpha |(greenwork[i] << 8) | bluework[i];			
	     			}
	     		else {  // case 6
 	    			if(ifdomaxproj) 
 	     				for(i=0; i<npix;i++) 
							maxProj[i] = alpha | (greenmax[i] << 8) ;
					if(ifdoraytrace)
 	     				for(i=0; i<npix;i++) 
							rayTrace[i] = alpha | (greenwork[i] << 8);			
	     			}
	     		}
	     	else if(ifdidblue) { // case 7
 	    			if(ifdomaxproj) 
 	     				for(i=0; i<npix;i++) 
							maxProj[i] = alpha |  bluemax[i];
					if(ifdoraytrace)
 	     				for(i=0; i<npix;i++) 
							rayTrace[i] = alpha |  bluework[i];			
	     		}
	     	}
	     		     	
	    redwork=null;
	    greenwork=null;
	    bluework=null;
	    redmax=null;
	    greenmax=null;
	    bluemax=null;
	    byteChan=null; 
	    redFilmCapacityRemaining=null;
	    greenFilmCapacityRemaining=null;
	    blueFilmCapacityRemaining=null;
	    showAJFrame(wasAtSection);    
	    }

	private static void normalize(short [] rgbwork) {
		short themax=1;
		int i,npix = rgbwork.length;
		for(i=0; i<npix;i++)
			if(rgbwork[i]>themax)
				themax = rgbwork[i];
		for(i=0; i<npix;i++)
			rgbwork[i] = (short)((255*rgbwork[i])/themax);
		}
			
	private void progressReport(String stg, JTextField progressReportField) {
		progressReportField.setText(stg);
		try {
			Thread.sleep(15);
			}
		catch(InterruptedException e) {
			;
			}
		}				

	public void overlaySections(int [] raster, int from,  int to,  int whichBlend, double redNonMaskFrac , double greenNonMaskFrac , 
			double blueNonMaskFrac, JTextField progressReportField) { 
	    int npix=curWidth*curHeight,alpha=(255 << 24);
	    byte [] byteChan;
	    int [] workR = new int[npix], workG = new int[npix], workB = new int[npix];
	    int ifr, i,  quot, bright, navgR=0,navgG=0,navgB=0;
	    for(i=0; i<npix;i++) { 
		     raster[i] = alpha; // load in transparency value to raster.
		     workR[i]=0;
		     workG[i]=0;
		     workB[i]=0;
	    	}
	    
	   if(whichBlend==0) {  // averaging
	    // get the red channel plus alpha into raster array
	 	 for(ifr=from; ifr <= to;ifr++) { 
	 		 if(Stax.hub.segmentationJPanel.ifPutDataThroughFilterChain())
	 			Stax.hub.segmentationJPanel.getReadyToSegmentNewSection(ifr);
						
	 		 if(redfile != null) {
			     byteChan = Stax.hub.segmentationJPanel.ifPutDataThroughFilterChain()? 
			    		 Stax.hub.segmentationJPanel.processPrimaryChannelData(0,redNonMaskFrac) : 
										getOneByteChan(ifr, 0);
	 			 if(byteChan != null) {
	 				 ++navgR;
	 				 progressReport(("overlay red chan sect "+ifr), progressReportField);
				
	 				 for(i=0;i<npix;i++)
	 					 workR[i] += (byteChan[i]<0? 256+byteChan[i] : byteChan[i]);
	 			 	}
	 		 }
	    
	 		 if(greenfile != null) {
	 			 // add the green channel into the raster array
				 byteChan = Stax.hub.segmentationJPanel.ifPutDataThroughFilterChain()?
						 Stax.hub.segmentationJPanel.processPrimaryChannelData(1,greenNonMaskFrac) : 
										getOneByteChan(ifr, 1);
	 			 if(byteChan != null) {
	 				 ++navgG;
	 				 progressReport(("overlay green chan sect "+ifr), progressReportField);
	 				 for(i=0;i<npix;i++)
	 					 workG[i] += (byteChan[i]<0? 256+byteChan[i] : byteChan[i]);
	 			 	}
	 		 	}

	 		if(bluefile != null) {
	 			// add the blue channel into the raster array
				byteChan = Stax.hub.segmentationJPanel.ifPutDataThroughFilterChain()? 
						Stax.hub.segmentationJPanel.processPrimaryChannelData(2,blueNonMaskFrac) : 
									getOneByteChan(ifr, 2);
	 			if(byteChan != null) {
	 				for(i=0;i<npix;i++)
	 					workB[i] += (byteChan[i]<0? 256+byteChan[i] : byteChan[i]);
	 				++navgB;
	 				progressReport(("overlay blue chan sect "+ifr),progressReportField);
	      			}
	 			}
	 		} // end of loop: for(ifr=from+1; ifr <= to;ifr++)
	 	 
	    if(navgR==0) navgR=1;
	    if(navgG==0) navgG=1;
	    if(navgB==0) navgB=1;
	    for(i=0; i<npix;i++) 
	    	  raster[i] = raster[i] | ( (workR[i]/navgR) << 16 |  (workG[i]/navgG) << 8 | (workB[i]/navgB));
	    } // end of if(whichBlend==1) {
	    
	    
	    if(whichBlend==1) {  // finding max
		 for(ifr=from; ifr <= to;ifr++) { 
	 		 if(Stax.hub.segmentationJPanel.ifPutDataThroughFilterChain())
	 			 parent.parent.segmentationJPanel.getReadyToSegmentNewSection(ifr);
						
			 if(redfile != null) {
			     byteChan = Stax.hub.segmentationJPanel.ifPutDataThroughFilterChain()? 
			    		 Stax.hub.segmentationJPanel.processPrimaryChannelData(0,redNonMaskFrac) : 
										getOneByteChan(ifr, 0);
		 		if(byteChan != null) 
		 			for(i=0;i<npix;i++) {
		 				bright= (byteChan[i]<0? 256+byteChan[i] : byteChan[i]);
		 				if(bright>workR[i]) workR[i]=bright;
		 				} 
		 		}
	    
			 if(greenfile != null) {
			     byteChan = Stax.hub.segmentationJPanel.ifPutDataThroughFilterChain()? 
			    		 Stax.hub.segmentationJPanel.processPrimaryChannelData(1,greenNonMaskFrac) : 
										getOneByteChan(ifr, 1);
		 		if(byteChan != null) 
		 			for(i=0;i<npix;i++) {
		 				bright= (byteChan[i]<0? 256+byteChan[i] : byteChan[i]);
		 				if(bright>workG[i]) workG[i]=bright;
		 				} 
		 		}
			 
			 if(bluefile != null) {
			     byteChan = Stax.hub.segmentationJPanel.ifPutDataThroughFilterChain()? 
			    		 Stax.hub.segmentationJPanel.processPrimaryChannelData(2,blueNonMaskFrac) : 
										getOneByteChan(ifr, 2);
		 		if(byteChan != null) 
		 			for(i=0;i<npix;i++) {
		 				bright= (byteChan[i]<0? 256+byteChan[i] : byteChan[i]);
		 				if(bright>workB[i]) workB[i]=bright;
		 				} 
		 		}
		 
		 	} // end of loop: for(ifr=from; ifr <= to;ifr++)
		    for(i=0; i<npix;i++) 
		    	  raster[i] = raster[i] | (workR[i] << 16 |  workG[i] << 8 | workB[i]);
	    } // end of if(whichBlend==1) {
		workR=null;
		workG=null;
		workB=null;
	    byteChan=null;    
	    }

	public int [] showThisRaster(int [] raster,  int from,  int to) {
	    if(parent == null) return null;
		return showThisRaster(raster,("overlay of sections "+from+" to "+to));
		}
	
	public int [][][] getTriColorLUTS() {
		/* returns a int[3][3][255] array called A, where
		redByte   -> (A[0][0][redByte],   A[0][1][redByte],   A[0][2][redByte]) 
		greenByte -> (A[1][0][greenByte], A[1][1][greenByte], A[1][2][greenByte]) 
		blueByte  -> (A[2][0][blueByte],  A[2][1][blueByte],  A[2][2][blueByte])
		and this is regardless of whether we are using artCOlors, RGB channel colors, or single channels */
 		int colorSwitch = getColorSwitch(); // == -1 for all 3 priomary colors, 0 for red only, 
					    // 1 for green only, 2 for blue only, 4 for artColors
		int i,j,k;
		int [][][] A = new int[3][3][256];
		for(i=0;i<3;i++)
			for(j=0;j<3;j++)
				for(k=0;k<256;k++)
					A[i][j][k]=0;
					
		int [] redshades, greenshades, blueshades;
		int [] grayshades = {255,255,255};
		int pix; 			
		switch (colorSwitch) {
			case 4 : // art colors
			case -1 :
				redshades = parent.getRedsColors();
				greenshades = parent.getGreensColors();
				blueshades = parent.getBluesColors();
				for(j=0;j<3;j++) {
					A[0][j] = anLUT(redLUT256,redshades[j]);
					A[1][j] = anLUT(greenLUT256,greenshades[j]);
					A[2][j] = anLUT(blueLUT256, blueshades[j]);
					}						
				break;
			
			case 0 : // redonly
				if(ifUseGrayScale) 
					redshades = grayshades;
				else
					redshades = parent.getRedsColors();
				
				for(j=0;j<3;j++) 
					A[0][j] = anLUT(redLUT256,redshades[j]);
				break;
				
			case 1 : // greenonly
				if(ifUseGrayScale) 
					greenshades = grayshades;
				else
					greenshades = parent.getGreensColors();
				
				for(j=0;j<3;j++) 
					A[1][j] = anLUT(greenLUT256,greenshades[j]);
				break;
				
			case 2 : // blueonly
				if(ifUseGrayScale) 
					blueshades = grayshades;
				else
					blueshades = parent.getBluesColors();
				
				for(j=0;j<3;j++) 
					A[2][j] = anLUT(blueLUT256,blueshades[j]);
				break;
			}
		return A;		
		}
	
	private int [] anLUT(int [] priLUT, int shade) {
		int [] retn = new int[256];
		for(int i=0;i<256;i++)
			retn[i] = (shade*priLUT[i])/255;
		return retn;
		}
			
	public int [] showThisRaster(int [] raster,  String CanvasTitle) {
	    if(parent == null) return null;
		
	    talk("BioradCanvas.showThisRaster()");
	    
		int colorSwitch = getColorSwitch(); // == -1 for all colors, 0 for red only, 
					    // 1 for green only, 2 for blue only, 4 for artColors
		int i, npix=curWidth*curHeight,red,green,blue;
		int redmask=(255 << 16), greenmask=(255<<8), bluemask=255;
		int alpha=(255<<24);
		if(curIm == null)
			curIm = new int[npix];
	    switch (colorSwitch) {
	    	case -1 : // use the 3 primary colors for all three channels
	    		for(i=0;i<npix;i++) 
	    /*			red = redShiftLUT256[((raster[i]&redmask)>>16)];
	    			green = greenShiftLUT256[((raster[i]&greenmask)>>8)];
	    			blue = blueLUT256[(raster[i]&bluemask)];
	    			curIm[i] = alpha |  red  |  green  | blue;  */
	    		curIm[i] = alpha | (redShiftLUT256[((raster[i] & redmask)>>16)]) | 
	    			( greenShiftLUT256[((raster[i]&greenmask)>>8)]) | 
	    					 blueLUT256[(raster[i]&bluemask)];	 
	    		break;
	    	case 0: // red channel only
	    		if(ifUseGrayScale) 
	    			for(i=0;i<npix;i++)  {
	    				red = redLUT256[((raster[i] & redmask)>>16)];
	    				curIm[i] =alpha | (red <<16) | (red <<8) | red;
	 					}   			
	    		else {
	    			if(ifUsingArtColors) {
			  			int [] redColors = parent.getRedsColors();		
			  			int artred,artgreen,artblue;
	    	  			for(i=0;i<npix;i++) {
	    					red = (raster[i]&redmask)>>16;
	curIm[i] = (alpha | (redRedness[red] << 16) | (redGreenness[red] << 8) | (redBlueness[red]));
	    					}
	    				}
	    			else for(i=0;i<npix;i++) 
	    				curIm[i] =alpha | redShiftLUT256[((raster[i]&redmask)>>16)];
	    			}
	    		break;
	    	case 1: //green channel only
	    		if(ifUseGrayScale) 
	    			for(i=0;i<npix;i++)  {
	    				green = greenLUT256[((raster[i]&greenmask)>>8)];
	    				curIm[i] =alpha | (green <<16) | (green <<8) | green;
	 					}   			
	    		else {
	    			if(ifUsingArtColors) {
			  			int [] greenColors = parent.getGreensColors();		
			  			int artred,artgreen,artblue;
	    	  			for(i=0;i<npix;i++) {
	    					green = (raster[i]&greenmask)>>8;
curIm[i] = (alpha | (greenRedness[green] << 16) | (greenGreenness[green] << 8) | (greenBlueness[green]));
	    					}
	    				}
	    			else for(i=0;i<npix;i++) 
	    				curIm[i] =alpha | greenShiftLUT256[((raster[i]&greenmask)>>8)];
	    			}
	    		break;
	    	case 2: //blue channel only
	    		if(ifUseGrayScale) 
	    			for(i=0;i<npix;i++)  {
	    				blue = blueLUT256[(raster[i]&bluemask)];
	    				curIm[i] =alpha | (blue <<16) | (blue <<8) | blue;
	 					}   			
	    		else {
	    			if(ifUsingArtColors) {
			  			int [] blueColors = parent.getBluesColors();
			  			int artred,artgreen,artblue;
	    	  			for(i=0;i<npix;i++) {
	    					blue = raster[i]&bluemask;
curIm[i] = (alpha | (blueRedness[blue] << 16) | (blueGreenness[blue] << 8) | (blueBlueness[blue]));
	    					}
	    				}
	    			else for(i=0;i<npix;i++) 
	    				curIm[i] =alpha | blueLUT256[(raster[i]&bluemask)];
	    			}
	    		break;
	    		
	    	case 4: // use art colors
			  int [] redColors = parent.getRedsColors();		
			  int [] greenColors = parent.getGreensColors();		
			  int [] blueColors = parent.getBluesColors();
			  int artred,artgreen,artblue;
	    	  for(i=0;i<npix;i++) {
	    		red = (raster[i]&redmask)>>16;
	    		green = (raster[i]&greenmask)>>8;
	    		blue = (raster[i]&bluemask);
			if((artred = redRedness[red]+ greenRedness[green] + blueRedness[blue])>255)
				artred=255;
			if((artgreen = redGreenness[red]+ greenGreenness[green] + blueGreenness[blue])>255)
				artgreen=255;
			if((artblue = redBlueness[red]+ greenBlueness[green] + blueBlueness[blue])>255)
				artblue=255;
			curIm[i] = (alpha | (artred << 16) | (artgreen << 8) | (artblue));
				}
	    		break;
	    	}
	    rawImage = createImage(new MemoryImageSource(curWidth, curHeight,
			 ColorModel.getRGBdefault(), curIm, 0, curWidth));

		parent.setJPanelTitle(CanvasTitle);
		parent.picInvalid();
		parent.repaint();
	    return curIm;
	    }

	private int getColorSwitch() {
		int colorSwitch = singleChan; // == -1 for all colors, 0 for red only, 
					    // 1 for green only, 2 for blue only
		// get the colors...
		int [] redColors = parent.getRedsColors();		
		int [] greenColors = parent.getGreensColors();		
		int [] blueColors = parent.getBluesColors();
		newLUTS(curJFrame); 
		// always use threshold numbers in redColors[3],greenColors[3],blueColors[3] 		    
		ifUsingArtColors=false;
		if(redColors[0]!=255 || redColors[1]!=0 || redColors[2]!=0 ||
		    greenColors[0]!=0 || greenColors[1]!=255 || greenColors[2]!=0 ||
			blueColors[0]!=0 || blueColors[1]!=0 || blueColors[2]!=255) {
			if(singleChan<0) colorSwitch=4;
			ifUsingArtColors=true;
			// talk("using art colors instead of primary RGBs");
			}
		 return colorSwitch;
		 }
	
	 		
  	public void showAJFrame(int nframe) {  // calls showAnotherJFrame  
  	         // if a new frame really needs to be drawn. 
  			if(lastJFrameShown != nframe || ifForceNewDrawing) {
  				showAnotherJFrame(nframe);
  				}
  			lastJFrameShown=nframe;
  			ifForceNewDrawing=false;
  			talk("bioradCanvas shows section "+nframe);
			if(parent != null) 
				parent.repaint(); 
  			}
  			
  	public void showAJFrameUrgently(int nframe, int maxDelay) {  // calls showAnotherJFrame  
  	         // if a new frame really needs to be drawn.
  	        mark2Time(); 
  			if(lastJFrameShown != nframe || ifForceNewDrawing) {
  				showAnotherJFrameNoRepaint(nframe);
  				if(parent.ifTellTiming) 
  					talk("BioradCanvas.showAJFrameUrgently() took "+timeSince2Mark()+" ms to create image");
  				}
  			lastJFrameShown=nframe;
  			ifForceNewDrawing=false;
			if(parent != null) 
				parent.repaintUrgently(maxDelay);
  			}
  		
  	public void forceNewDrawing() {
   		ifForceNewDrawing=true;
   		}
 		

   	public void showAnotherJFrame(int nframe) {
		int colorSwitch = getColorSwitch(); // == -1 for all colors, 0 for red only, 
					    // 1 for green only, 2 for blue only, 4 for artColors
		
		// decide if we need to redraw the micrograph. This determined by whether
		rawImage=null; 
		// System.gc();
		newLUTS(nframe); 
		parent.picInvalid();
		markTime();
		switch (colorSwitch)  {
		    case 0 : // red channel only
		    case 1 : // green channel only
		    case 2 : // blue channel only
			if(!ifUsingArtColors || ifUseGrayScale)
			    curIm=drawJustOneChan(nframe, colorSwitch,  ifUseGrayScale);
			else
			    curIm=drawJustOneArtChan(nframe, colorSwitch);
				break;
		    case (-1):
		    default : // this is the case when more than one color is present
				curIm = draw3Pri(nframe);
			break;
				
		    case (4):   // this is the case when more than one color is present
				    // and we are using fancy art colors from the arrays
				    // redColors, greenColors, blueColors
				curIm = drawArtColors(nframe);
			break;
		    }
			
		//talk("that took "+timeSinceMark());
		if(curIm != null)  {
		    rawImage = createImage(new MemoryImageSource(curWidth, curHeight,
			 	ColorModel.getRGBdefault(), curIm, 0, curWidth));
			curRows=parent.fullHeight;
			curCols=parent.fullWidth;
		    }
		else {
		    talk("failed to create rawImage in showAJFrame()");
		    rawImage=null;
		    }
		curJFrame=nframe;
		tellStatsJPanelWhichJFrame();
		
		if(rawImage != null) {
			capString = sectionTypeName+"section "+(new Integer(curJFrame)).toString();
			capString += "/";
			capString += (new Integer(NJFrames-1)).toString();
			if(singleChan>=0) 
				switch (singleChan) {
					case 0:
						capString += " red";
					break;
					case 1:
						capString += " green";
					break;
					case 2:
						capString += " blue";
					break;
					}
			if(parent != null) {
				parent.setJPanelTitle(capString);
				parent.picInvalid();
				parent.repaint();
				}
			}
	else talk("OOps, showAJFrame got no valid picture to show");
	parent.repaintExhibitRGBChanLUTS();
	} 

   	public void showAnotherJFrameNoRepaint(int nframe) {
		int colorSwitch = getColorSwitch(); // == -1 for all colors, 0 for red only, 
					    // 1 for green only, 2 for blue only, 4 for artColors
		
		// decide if we need to redraw the micrograph. This determined by whether
		rawImage=null; 
		// System.gc();
		newLUTS(nframe); 
		parent.picInvalid();
		markTime();
		switch (colorSwitch)  {
		    case 0 : // red channel only
		    case 1 : // green channel only
		    case 2 : // blue channel only
			if(!ifUsingArtColors || ifUseGrayScale)
			    curIm=drawJustOneChan(nframe, colorSwitch,  ifUseGrayScale);
			else
			    curIm=drawJustOneArtChan(nframe, colorSwitch);
				break;
		    case (-1):
		    default : // this is the case when more than one color is present
				curIm = draw3Pri(nframe);
			break;
				
		    case (4):   // this is the case when more than one color is present
				    // and we are using fancy art colors from the arrays
				    // redColors, greenColors, blueColors
				curIm = drawArtColors(nframe);
			break;
		    }
			
		if(parent.ifTellTiming) talk("                   curIm creation took "+timeSinceMark());
		markTime();
		if(curIm != null)  {
		    rawImage = createImage(new MemoryImageSource(curWidth, curHeight,
			 	ColorModel.getRGBdefault(), curIm, 0, curWidth));
			curRows=parent.fullHeight;
			curCols=parent.fullWidth;
			if(parent.ifTellTiming) 
				talk("                   rawImage = createImage(curIm) took "+timeSinceMark());
		    }
		else {
		    talk("failed to create rawImage in showAJFrame()");
		    rawImage=null;
		    }
		curJFrame=nframe;
		// tellStatsJPanelWhichJFrame();
		
		if(rawImage != null) {
			capString = sectionTypeName+"section "+(new Integer(curJFrame)).toString();
			capString += "/";
			capString += (new Integer(NJFrames-1)).toString();
			if(singleChan>=0) 
				switch (singleChan) {
					case 0:
						capString += " red";
					break;
					case 1:
						capString += " green";
					break;
					case 2:
						capString += " blue";
					break;
					}
			if(parent != null) {
				parent.setJPanelTitle(capString);
				parent.picInvalid();
				}
			}
	else talk("OOps, showAJFrame got no valid picture to show");
	} 

	public void writeTiffFile() {
		        if(curHeight >= -1) 
			    	talk("writing tiff images temporarily disabled");
		    	else {
			    	try{(new TiffFuncs()).writeColor768Tiff(edgeTrim(curIm));}
			    	catch(IOException e) {
						talk("Oops, writeColor768Tiff(...) failed ");
						}
			    	}
			}


	public void writeGmoRasFile() {     
			    try{(new GMORasFuncs()).writeGmoRgbRas(curRows,curCols,edgeTrim(curIm));}
			    catch(IOException e) {
				talk("Oops, writeGmoRgbRas(...) failed ");
				}
			}

	public void writeJpeg() {     
		try{(new GMORasFuncs()).writeJpeg(curRows,curCols,edgeTrim(curIm),1);}
			    catch(IOException e) {
				talk("Oops, writeJpeg(...) failed ");
				}
 		}

	public  void writeJpeg(int [] anIm, int itswidth, int itsheight) {     
			    try{(new GMORasFuncs()).writeJpeg(itsheight,itswidth,edgeTrim(anIm),1);}
			    catch(IOException e) {
				talk("Oops, writeJpeg(...) failed ");
				}
			}


	public void update(Graphics g) {
		// talk("BioradCanvas.update() called. Should never happen");
		paint2D((Graphics2D)g);
		}

	public void paint(Graphics g) {
		paint2D((Graphics2D)g);
		}
		
	public void update2D(Graphics2D g) {
		paint2D(g);
		}

	public void paint2D(Graphics2D g) {
		// talk("BioradCanvas.paint2D() called");
		int	x, y;	
		GraphicsPs gg;	
		parent.parent.setDrawOrg(curWidth,curHeight);


// talk("xDrawOrg="+parent.parent.xDrawOrg+", yDrawOrg="+parent.parent.yDrawOrg+", winwidth="+parent.parent.winwidth);
		 	// that sets what location of the drawn image goes at the upper left corner
		 	// it's controlled by sliders...
		// gg = new GraphicsPs(parent.parent.winwidth,parent.parent.winheight,g,false);
		//talk("BioradCanvas.paint() has curWidth="+curWidth+" and curHeight="+curHeight);	
		try {
		    if(rawImage != null) {
			    g.drawImage(rawImage, parent.parent.xDrawOrg, parent.parent.yDrawOrg, curWidth,curHeight, parent);
			    parent.picValid(); 
			    }
		    //else talk("paint(): rawImage==null ????");
		    }
		catch (Exception e) {
			System.out.println("Error painting biorad file: " + e.toString());
			}
		// gg.finish();
		setVisible(true);
 		}
		
	public int [] getColors(int x, int y) {
		int [] colors = new int[3];
		int row,col;
		boolean ifrd = ifred, ifgr=ifgreen, ifbl=ifblue; /* local versions to be modified to
						account for using only one channel when singleChan>0
		         repeat code from showAJFrame(). all rasters should be in the cache... */
		byte [] redRaster=null;
		byte [] greenRaster=null;
		byte [] blueRaster=null;
		
		switch (singleChan)  {
			case 0 :
				ifgr=false;
				ifbl=false;
				break;
			case 1 :
				ifrd=false;
				ifbl=false;
				break;
			case 2 :
				ifrd=false;
				ifgr=false;
				break;
			default :
				break;
			}
		try {
			// get the image rasters...
			if(ifrd &&  (redRaster = redfile.readBioradRaster(curJFrame-Zshifts[0])) == null)
				ifrd=false;
			if(ifgr && (greenRaster = greenfile.readBioradRaster(curJFrame-Zshifts[1])) == null)
				 ifgr=false;
			if(ifbl &&(blueRaster = bluefile.readBioradRaster(curJFrame-Zshifts[2])) == null)
				 ifbl=false;

		colors[0]=-1;	colors[1]=-1;	colors[2]=-1;
		if(ifrd) {
			row = y + Yshifts[0];
			col = x + Xshifts[0];
			if(row>=0 && row < nhigh && col >=0 && col < nwide) {
				colors[0] = redRaster[row*curWidth+col];
				if(colors[0]<0) colors[0] += 256;
				}
			}

		if(ifgr) {
			row = y + Yshifts[1];
			col = x + Xshifts[1];
			if(row>=0 && row < nhigh && col >=0 && col < nwide) {
				colors[1] = greenRaster[row*curWidth+col];
				if(colors[1]<0) colors[1] += 256;
				}
			}

		if(ifbl) {
			row = y + Yshifts[2];
			col = x + Xshifts[2];
			if(row>=0 && row < nhigh && col >=0 && col < nwide) {
				colors[2] = blueRaster[row*curWidth+col];
				if(colors[2]<0) colors[2] += 256;
				}
			}
		}
		catch(RuntimeException e) {		// Probably a memory error, so guess that that's what it is
			Runtime.getRuntime().gc();	// Clean up what we just got rid of
			talk("getColors() failed for frame number "+curJFrame);
			}
		redRaster=null;
		greenRaster=null;
		blueRaster=null;
		return colors;
		}
	    			
}
