package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.io.*;
import java.lang.*;
import java.util.*;
import java.awt.*;  
import javax.swing.*;


// this will eventually be read from the header in the first section in the stack

public class BioradStackConfig extends JFrame{
double [] depth,controlBrightness,boostFacToUnBleach;
int [] mapsto;
int [] dimThresh;
double [] gainCurveFitCoeffs = {-0.0, 0.0, 0., 0.0}; 
double concentrationMappedTo255 = 0.1;
double lambda1=0., lambda2=0., fastrat=0.1;
int firstRealSection=28;
int nsects=164;
boolean ifGainCurveCorrected_bleachCorrected_stack=false;
Stringer ringer;
int stdlen = 120;
int ifRGB=0; // 0->red, 1->green, 3-> blue
String [] titles = {"Raw data used as is","UnGainCurve transformation","UnGain/UnBleach correction"};
String magicNumber = "g odell modified this biorad picture     <-- the magic number";

ShowPicJPanel picJPanel=null;
BioradFile myBioradFile=null;
ShowLUT showMyLUT=null;

byte [] theLines=null;
int startOfNextLine=0;
int lineNumber=0;

public void blotOutHoles(byte [] ras, int ncols, int nrows, int nsect ) {
		picJPanel.blotOutHoles(this, ras, ncols, nrows, nsect);
		}
		
public double getZHeightFromSectNum(int sectnum) {
	return depth[sectnum];
	}		
	
public void sepuku() {
	depth=null;
	controlBrightness=null;
	boostFacToUnBleach=null;
	mapsto=null;
	dimThresh=null;
	ringer=null;	
	picJPanel=null;
	myBioradFile=null;
	}
	
public void lowPassBoostFacs(int npass) {
	// low pass filter (npass passes) the actually measured values in the
	//  boostFacToUnBleach[] array
	int i,jpass,j;
	double [] smth = new double[firstRealSection];
	double [] smooth = new double[firstRealSection];
	double [] temp;
	double prevFac;
	for(i=0;i<firstRealSection;i++)
		smooth[i] = boostFacToUnBleach[mapsto[i]]; // these are the points to filter

	for(jpass=0;jpass<npass;jpass++) {
		for(i=1;i<firstRealSection-1;i++)
			smth[i] = (smooth[i-1] + smooth[i] + smooth[i+1])/3.0;
		smth[0]=smooth[0];
		smth[firstRealSection-1]=smooth[firstRealSection-1];
		temp=smth;
		smth=smooth;
		smooth=temp;
		} // when you exit, the smoothed result is in the smooth array

	// now rebuild the interpolated values.
		for(i=firstRealSection;i<nsects;i++) 
		  	boostFacToUnBleach[i]=-1;
		for(i=0;i<firstRealSection;i++) // restore the actually measured values
			boostFacToUnBleach[mapsto[i]] = smooth[i]; 
		for(i=firstRealSection;i<nsects;i++) 
		  	if(boostFacToUnBleach[i]<0.) {
		  		// get the value of myStackConfig.boostFacToUnBleach[i] by interpolation 
		  		prevFac=boostFacToUnBleach[i-1];
		  		j=i+1;
		  		while(j<nsects && boostFacToUnBleach[j]<0.) ++j;
		  		if(j==nsects) boostFacToUnBleach[i]=prevFac;
		  		else boostFacToUnBleach[i] = 
		  			((j-i)*prevFac + boostFacToUnBleach[j])/(double)(j-i+1.0);
		  		}
	}


public BioradStackConfig(int ifRGB, ShowPicJPanel picJPanel,BioradFile myBioradFile) { // constructor
	int i,PriorScans;
	this.ifRGB = ifRGB;
	this.myBioradFile = myBioradFile;
	this.picJPanel=picJPanel;
	nsects=myBioradFile.numJFrames;
	ifGainCurveCorrected_bleachCorrected_stack=false;
	depth = new double[nsects];
	controlBrightness = new double[nsects];	
	boostFacToUnBleach = new double[nsects];
	dimThresh = new int[nsects];
	mapsto = new int[nsects];
	ringer = new Stringer();
	theLines = myBioradFile.readGmoHeader();
	if(theLines == null || !parseHeader() )
		nullSetup();
	switch (ifRGB) {
		case 0:
		showMyLUT = picJPanel.parent.exhibitRedChanLUT;
		break;
			
		case 1:
		showMyLUT = picJPanel.parent.exhibitGreenChanLUT;
		break;
			
		case 2:
		showMyLUT = picJPanel.parent.exhibitBlueChanLUT;
		break;
		}
	}

	public void init() {
		setLambdas();
		showMyLUT.setNewGainCurveFit(concentrationMappedTo255, gainCurveFitCoeffs,false);
		}

	private void setLambdas() {
		if(showMyLUT != null)
			showMyLUT.setLambdas(lambda1,lambda2, fastrat,concentrationMappedTo255);
		}
			
	private void getLambdas() {
		if(showMyLUT != null)
			showMyLUT.getLambdas();
		}
			
	public void finalize() {
		//talk("                                                                           BioradStackConfig with ifRGB="+ifRGB+" snuffs");
		}
	
	private String nextLine() {
	startOfNextLine += stdlen;
	++lineNumber;
	String stg = new String(theLines,0,startOfNextLine,stdlen);
	// talk(stg);
	return stg;
	}

	private boolean parseHeader() {
		int i,nlines = 15+nsects;
		double [] tmp;
		String stg;
		startOfNextLine=0;  lineNumber=0;
 		stg=new String(theLines,0,startOfNextLine,stdlen);
		if(stg.indexOf(magicNumber) < 0) {
		talk("There is no autobleach correction header in this BioradFile");
		return false;
		}
		// OK got past the magic number header	
		
 		nextLine();
//L1   | -1             gainCurveCorrected_bleachCorrected_stack <--- tells this is ready for those corrections 
		ifGainCurveCorrected_bleachCorrected_stack = (parseDoubles(1, startOfNextLine, 16))[0] >0;

 		nextLine();
//L2   | -0.000258006    0.000281793     4.14219e-07     0.0                    <--- inverse gain curve coefficients      
		gainCurveFitCoeffs = parseDoubles(4, startOfNextLine, 20);
		/* talk("The gain curve coefficients are: "+gainCurveFitCoeffs[0]+", "+gainCurveFitCoeffs[0]+
					", "+gainCurveFitCoeffs[0]); */

 		nextLine();
//L3   | 0.0991993             <-- concentration mapped to pixelBrightness 255                                            
		concentrationMappedTo255 = (parseDoubles(1, startOfNextLine, 16))[0];
		//talk("concentrationMappedTo255 = "+concentrationMappedTo255);

		
 		nextLine();
//L4   | 0.00237         e);0.02552         0.11684                <--- lambda1, lambda2, fastrat        
		tmp = parseDoubles(3, startOfNextLine, 16);
		lambda1=tmp[0];	lambda2=tmp[1];   fastrat=tmp[2];
		talk("Just read in lambda1="+lambda1+", lambda2="+lambda2+", fastrat="+fastrat);
		setLambdas();
                    
 		nextLine();
//L5   | 28               <--- first of the real data sections.                                                           
		firstRealSection = (int)((parseDoubles(1, startOfNextLine, 16))[0]);


 		nextLine();
//L6   | 164              <--- total number of data sections.                                                             
		nsects  = (int)((parseDoubles(1, startOfNextLine, 16))[0]);

 		nextLine(); // place holders for later thoughts...
//L7   | 0 0 0 0 0 0         <--- something I forgot to include up front                                                  
 		nextLine();
//L8   | 1 1 1 1 1 1         <--- something I forgot to include up front                                                  
 		nextLine();
//L9   | 2 2 2 2 2 2         <--- something I forgot to include up front                                                  
 		nextLine();
//L10  | 3 3 3 3 3 3         <--- something I forgot to include up front                                                  
 		nextLine();
//L11  | 4 4 4 4 4 4         <--- something I forgot to include up front                                                  
 		nextLine();
//L12  | raw section     depth in        map to          boost           Gain-corrected/avg in                            
 		nextLine();
//L13  | number          microns         section         factor to       avg in control/pre-scan                          
 		nextLine();
//L14  |                                 number          un-bleach               region/region   dimThresh 
		for(i=0;i<nsects;i++) {               
 			nextLine();
 			tmp = parseDoubles(7, startOfNextLine, 16);
 			if(i != (int)tmp[0]) talk("when i="+i+", the section number I read was "+(int)tmp[0]);
 			depth[i]=tmp[1];
 			mapsto[i] = (int)tmp[2];
 			boostFacToUnBleach[i]=tmp[3];
 			controlBrightness[i]=tmp[4];
 			dimThresh[i] = (int)tmp[6]; 
 			}			
	return true;
	}

  private double [] parseDoubles(int howMany, int offset, int fieldLength) {
  	String oneNum;
  	int i;
  	double[] retn = new double[howMany];
  	offset += 7; // to get past the L123 | line number header
  	for(i=0;i<howMany;i++) {
   		oneNum = new String(theLines,0,offset,fieldLength);
   		offset += fieldLength;
		retn[i] = myDoubleParser(oneNum);
		}
	return retn;
	}

	private double  myDoubleParser(String stg) {
		int at=0;
		int iwh=-1;
		boolean ifNeg=false;
		boolean ifPastPoint=false;
		int lenstg = stg.length();
		double val=0.;
		double base=.10;
		while(at<lenstg && (iwh=isPartOfNumber(stg.charAt(at))) <0 ) ++at;
		if(at >= lenstg-1) {
			talk("OOPs this string <"+stg+"> gave a bad number conversion");
			return -1.0;
			}
		// first meaningful character:
		switch (iwh) {
			case 12 : ifNeg=true;	++at;	break; // minus sign
			case 11 :  	++at;               break;  // plus sign
			case 10 :  ifPastPoint=true;   ++at; break;    // decimal point;
			default: break;
			}
		while(at<lenstg) {
			iwh=isPartOfNumber(stg.charAt(at));
			switch (iwh) {
				case 12 : // minus sign
				case 11 : // plus sign
				case (-1) : // white space of text characters;
				default: 
					break;
					
				case 10 :    // decimal point;
					ifPastPoint=true;  
					break;  
				
				case (0):
				case (1):
				case (2):
				case (3):
				case (4):
				case (5):
				case (6):
				case (7):
				case (8):
				case (9):
					if(ifPastPoint) {
						val += base*iwh;
						base = base/10;
						}
					else {
						val = 10*val + iwh;
						}
					break;
				}
			++at;
			}
		
		if(ifNeg) {
			//   talk(stg+" parsed to "+(-val));
			return (-val);
			}
		else {
			//     talk(stg+" parsed to "+val);
			return val;	
			}
		}
			              
   private int isPartOfNumber(char achar) {
   		if(achar == '-') return 12;
   		if(achar == '+') return 11;
   		if(achar == '.') return 10;
   		if(achar == '0') return 0;
   		if(achar == '1') return 1;
   		if(achar == '2') return 2;
   		if(achar == '3') return 3;
   		if(achar == '4') return 4;
   		if(achar == '5') return 5;
   		if(achar == '6') return 6;
   		if(achar == '7') return 7;
   		if(achar == '8') return 8;
   		if(achar == '9') return 9;
		return -1;
		}   		            
   private void nullSetup() {
		firstRealSection=0;
		for(int i=0;i<nsects;i++) {
			mapsto[i]= i;
			boostFacToUnBleach[i] = 1.;
			depth[i] = .1*Math.round(10.*i*0.2); // micrometers
			controlBrightness[i] =-1;
			dimThresh[i]=0;
			}
		ifGainCurveCorrected_bleachCorrected_stack=false;
		}
		
	public void bleachCorrectPreScanSections() {
		getLambdas();
		for(int i=0;i<firstRealSection;i++)
			boostFacToUnBleach[i] = 1./preBleach(i);		
		}
		
			
	public boolean initBleachCorrection() { 
		int i;
		if(ifGainCurveCorrected_bleachCorrected_stack) {
			talk("This feature was already initiated. ");
			return true;
			}
		// initialize to a no-bleach-correction state.
		talk("Initiated auto bleach correction... nsects="+nsects);
		switch (nsects) {
			case 164 : // 27 prescan sections 1 uM apart. then 27*5 sections spaced 0.2 uM apart.
				firstRealSection=28;
				for(i=0;i<firstRealSection;i++) {
					mapsto[i]= 28 + i*5;
					boostFacToUnBleach[i] = 1;
					depth[i]=i;
					controlBrightness[i] =-1;
					}
				for(i=firstRealSection;i<nsects;i++) {
					depth[i] = .1*Math.round(10.*(i-28)*0.2); // micrometers
					controlBrightness[i] =0;
					mapsto[i]= i; // means no remapping
					boostFacToUnBleach[i] = 1; // temporary estimate
					}
				ifGainCurveCorrected_bleachCorrected_stack=true;
				getLambdas();
				break;
			default :
				nullSetup();
				break;
			}
		if(ifGainCurveCorrected_bleachCorrected_stack)  {
			setLambdas();
			printHeaderData();
			}
		GainBleachCorrectionLUT_ForSect(0);	// this tells the showLUT the ID of this BioradStackConfig
		return ifGainCurveCorrected_bleachCorrected_stack;			
	}

public int takeWanderingCoreBoundaryFromSection(int isect) {
	if(!ifGainCurveCorrected_bleachCorrected_stack) 
		return isect;
	if(isect<0) isect=0;
	if(isect>=nsects) isect=nsects-1;
	return mapsto[isect];
	}

public int closestSectMappedToMe(int isect) {
	if(isect<firstRealSection || !ifGainCurveCorrected_bleachCorrected_stack) return isect;
	int j,i,imt=0,dist=10000;
	for(i=0;i<firstRealSection;i++) {
		j = isect - takeWanderingCoreBoundaryFromSection(i);
		if(j<0) j=-j;
		if(j<dist) {
			dist=j;
			imt=i;
			}
		}
	return imt;
	}

public byte [] GainBleachCorrectionLUT_ForSect(int isect) {
	// returns the LUT that corrects raw data in section isect for both bleaching
	// and GainCurveNon-linearities 
	int i,k,pix,mode;
	int [] temp = new int[256];;
	byte [] LUT = new byte[384];
	
	mode = showMyLUT.modeLUT;
	
	if(mode==2 && !ifGainCurveCorrected_bleachCorrected_stack)
		mode = 1;
	getLambdas();	
	
	switch (mode) {
		case 0:
			for(i=0;i<256;i++)  {
				if(i<128)
					k = 256+i;
				else 
					k = i;
				temp[i] =i;
				LUT[k] = (byte)(i);
				}
			break;
		case 1:
			for(i=0;i<256;i++)  {
				if(i<128)
					k = 256+i;
				else 
					k = i;
				LUT[k] = showMyLUT.gainCurveLUT[k];
				if((pix=LUT[k]) <0) pix += 256;
				temp[i] =pix;
				}
			break;
		case 2:
			LUT = new byte[384];
			for(i=0;i<256;i++)  {
				if(i<128)
					k = 256+i;
				else 
					k = i;
				pix=(int)(.499+  boostFacToUnBleach[isect]*unGain(i));
				if(pix<0) pix=0;
				if(pix>255) pix=255;
				temp[i] =pix;
				LUT[k] = (byte)(pix);
				}
			break;
		}
	
		switch (ifRGB) {
			case 0:
				showMyLUT.acceptNextLUT(temp,"Red channel's LUT", titles[showMyLUT.modeLUT]);
				break;
			case 1:
				showMyLUT.acceptNextLUT(temp,"Green channel's LUT", titles[showMyLUT.modeLUT]);
				break;
			case 2:
				showMyLUT.acceptNextLUT(temp,"Blue channel's LUT", titles[showMyLUT.modeLUT]);
				break;
			}
	return LUT;
	}

double unGain(int pix) { // maps a pixel's brightness, pix, to the brightness
	// which the corresponding dye standard concentration would produce given that
	// the concentration concentrationMappedTo255 has a brightness of 255.
	return 255.0*(gainCurveFitCoeffs[0] + pix*(gainCurveFitCoeffs[1] +
			pix*gainCurveFitCoeffs[2]))/concentrationMappedTo255;
	}
		

public double preBleach(int scanNumber) {
	getLambdas();
	return ((1-fastrat)*Math.pow(1.0-lambda1,scanNumber) + 
						fastrat*Math.pow(1.0 - lambda2,scanNumber));
	}

private static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}

	
private void lineAdvance(boolean iftalk, String stg) {
		stg.getBytes(0,stdlen,theLines,startOfNextLine);
		startOfNextLine += stdlen;
		theLines[startOfNextLine-1] = (byte)10;  // line end for use in reading/editing the file.
		// if(iftalk) 
			talk(stg);
		}

private String lineHead() {
	String stg = "L"+lineNumber;
	stg = ringer.padToLength(5,stg);	
	++lineNumber;
	return (stg+"| ");
	}

public void writeHeaderDataToDiskFile() {
	writeHeaderData(false);
	}
	
public void printHeaderData() {
	writeHeaderData(true);
	}
	
public void writeHeaderData(boolean iftalk) {
	String stg;
	int kfrom,kto;
	double denominator;
	
	int maxBytes = (20+nsects)*stdlen;
	theLines = new byte[maxBytes];
	
	getLambdas();
	startOfNextLine=0; // initialize the array that will hold all the lines this method writes.
	lineAdvance(iftalk,ringer.padToLength(stdlen,magicNumber)); 								// line 0
	lineNumber=1;
	kto = ifGainCurveCorrected_bleachCorrected_stack? 1:-1;	
	stg = ringer.appendNumber(lineHead(),kto,16);	
	lineAdvance(iftalk,ringer.padToLength(stdlen,
		(stg+" gainCurveCorrected_bleachCorrected_stack <--- +1 says this is ready for those corrections"))); 	// line 1

	stg=ringer.appendNumber(lineHead(), gainCurveFitCoeffs[0],16, 20, true);
	stg=ringer.appendNumber(stg, gainCurveFitCoeffs[1], 16, 20, true);
	stg=ringer.appendNumber(stg, gainCurveFitCoeffs[2], 16, 20, true);
	stg=ringer.appendNumber(stg, gainCurveFitCoeffs[3], 16, 20, true) + " =invGainCurveCoeffs";
    lineAdvance(iftalk,ringer.padToLength(stdlen,stg )); 		// line 2

	stg = ringer.appendNumber(lineHead(), concentrationMappedTo255,6, 16) + "      <-- concentration mapped to pixelBrightness 255";
	lineAdvance(iftalk,ringer.padToLength(stdlen,stg)); 																// line 3	
	
	talk("What's going on? lambda1="+lambda1+", lambda2="+lambda2+", fastrat="+fastrat);
	stg=ringer.appendNumber(lineHead(), lambda1, 8, 16);
	stg=ringer.appendNumber(stg, lambda2, 8, 16);
	stg=ringer.appendNumber(stg, fastrat, 8, 16)+"       <--- lambda1, lambda2, fastrat";
	//talk("write this line to file: <"+stg);
	lineAdvance(iftalk,ringer.padToLength(stdlen,stg)); 																// line 4

	stg=ringer.appendNumber(lineHead(), firstRealSection, 16)+"     <--- first of the real data sections.";
	lineAdvance(iftalk,ringer.padToLength(stdlen,stg)); 																// line 5

	stg=ringer.appendNumber(lineHead(), nsects, 16)+"     <--- total number of data sections.";
	lineAdvance(iftalk,ringer.padToLength(stdlen,stg)); 																// line 6

	lineAdvance(iftalk,ringer.padToLength(stdlen,lineHead()+"0 0 0 0 0 0         <--- something I forgot to include up front")); // line 7
	lineAdvance(iftalk,ringer.padToLength(stdlen,lineHead()+"1 1 1 1 1 1         <--- something I forgot to include up front")); // line 8
	lineAdvance(iftalk,ringer.padToLength(stdlen,lineHead()+"2 2 2 2 2 2         <--- something I forgot to include up front")); // line 9
	lineAdvance(iftalk,ringer.padToLength(stdlen,lineHead()+"3 3 3 3 3 3         <--- something I forgot to include up front")); // line 10
	lineAdvance(iftalk,ringer.padToLength(stdlen,lineHead()+"4 4 4 4 4 4         <--- something I forgot to include up front")); // line 11
	lineAdvance(iftalk,ringer.padToLength(stdlen,lineHead()+"raw section     depth in        map to          boost           Gain-corrected/avg in            ")); // line 12
	lineAdvance(iftalk,ringer.padToLength(stdlen,lineHead()+"number          microns         section         factor to       avg in control/pre-scan          ")); // line 13
	lineAdvance(iftalk,ringer.padToLength(stdlen,lineHead()+"                                number          un-bleach               region/region   dimThresh")); // line 14

	for(int i=0;i<nsects;i++) {
		stg=ringer.appendNumber(lineHead(), i, 16);
		stg = ringer.appendNumber(stg, depth[i],1, 16);
		stg = ringer.appendNumber(stg, mapsto[i], 16);
		stg = ringer.appendNumber(stg, boostFacToUnBleach[i], 6, 16);
		stg = ringer.appendNumber(stg, controlBrightness[i], 6,16);

		kfrom = closestSectMappedToMe(i);
		kto=takeWanderingCoreBoundaryFromSection(i);
		if(kto==i) denominator = controlBrightness[kfrom];
		else denominator=-1;
		stg = ringer.appendNumber(stg, denominator, 6, 16);
		stg = ringer.appendNumber(stg, dimThresh[i], 16);		
		lineAdvance(iftalk,ringer.padToLength(stdlen,stg));
		} 
		
	RandomAccessFile outfile;
	if(!iftalk) {
		outfile = openWritingFile(this,"specify header file name");
		if(outfile != null) {
			try{ 
				outfile.write(theLines,0,startOfNextLine); 
				outfile.close();
				}
			catch (IOException exc){
		   		talk("Trouble writing or closing the file:" + exc);
				}
			}
		else talk("Something went wrong");
		}
	outfile=null;	
	theLines=null;
	startOfNextLine=0;
	}

	static RandomAccessFile openWritingFile(JFrame theJFrame,String prompt) {
		RandomAccessFile outfile;
		FileDialog getFileToWrite = 
		   new FileDialog(theJFrame,prompt, FileDialog.SAVE);
		getFileToWrite.show();
		String filename = getFileToWrite.getDirectory()+"/"+getFileToWrite.getFile();
		if(filename == null) return null;
		try{
			outfile = new RandomAccessFile(filename,"rw");
			}
		catch (IOException e) {
			talk("openWritingFile("+prompt+") failed");
			return null;
			}
		try{outfile.seek((long)0);}
		catch(IOException e) {};
		return outfile;
		}
		
	
}
