package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.lang.*;
import java.util.*;
import java.awt.*;  
import java.io.*;
import javax.swing.*;

class BitMask {
int NCOLS=-1,NROWS=-1,NSECTS=-1;
byte [] bytes = null;
int nbytes;
byte [] bitMask;
boolean [] toDoXYsects;
boolean [] toDoXZsects;
boolean [] toDoZYsects;
ArchiveMarks maskWriter=null;

public BitMask(int NCOLS, int NROWS,  int NSECTS, boolean ifModifyExistingFile) {
	bitMask = new byte [] {1,2,4,8,16,32,64,(byte)128};
	maskWriter = new ArchiveMarks();
	int i;
	if(ifModifyExistingFile) { 	// create BitMask from a disk file.
		// see if size of the BitMask read from disk is compatible with present stack
		if(!maskWriter.readBitMask(this) || this.NCOLS != NCOLS || this.NROWS != NROWS || this.NSECTS != NSECTS) {
			talk("Something went wrong in reading the BitMask file");
			talk("The biorad stack has "+NCOLS+" columns, "+NROWS+" rows, and "+NSECTS+" sections");
			talk("But the mask read from disk has "+this.NCOLS+" columns, "+this.NROWS+" rows, and "+this.NSECTS+" sections");
			talk("The mismatch means you must create a new BitMask.");
			maskWriter.closeFile();
			ifModifyExistingFile = false;
			}
		else talk("BitMask created new bit mask from a disk file");
		}
		
	if(!ifModifyExistingFile) { 	// create a new BitMask.
		this.NCOLS = NCOLS;
		this.NROWS = NROWS;
		this.NSECTS = NSECTS; // create a new BitMask
		nbytes = 1 + (NCOLS*NROWS*NSECTS)/8;
		talk("New BitMask with "+NCOLS+" cols, "+NROWS+" rows, and"+NSECTS+
			" sections occupies "+nbytes+" bytes."); 
		bytes = new byte[nbytes];
		for(i=0;i<nbytes;i++)
			bytes[i]=0;
			
		toDoXYsects = new boolean[NSECTS];			
		toDoXZsects = new boolean[NROWS];			
		toDoZYsects = new boolean[NCOLS];
		markAllSectsUnDone();
			
		maskWriter.writeBitMask(this);
		talk("BitMask created a brand new bit mask");
		}
	}
	
	public void markAllSectsUnDone() {
		int i;
		for(i=0;i<NSECTS;i++)
			toDoXYsects[i]=true;
		for(i=0;i<NROWS;i++)
			toDoXZsects[i]=true;
		for(i=0;i<NCOLS;i++)
			toDoZYsects[i]=true;
		}
		
public void sepuku() {
	closeMaskFile();
	bytes=null;
	bitMask=null;
	maskWriter=null;
	talk("BitMask with "+NCOLS+" columns, "+NROWS+" rows, and "+NSECTS+" sections completed.");
	}

public void writeBitMaskToDisk() {
	if(!maskWriter.writeBitMask(this))
		talk("BitMask.writeBitMaskToDisk() failed??");
	}
	
public int countOnBits() {
	int non=0,i;
	for(i=0;i<nbytes;i++)
		non += countOnBits(bytes[i]);
	return non;
	}

public int tellHowManyRemainToDo(int iwh) {
	boolean [] bol;
	int i=0,j,k;
	String stg;
	switch (iwh) {
		default :
		case 1 :
			bol = toDoXYsects;
			stg = " XY";
			break;
		case 2 :
			bol = toDoXZsects;
			stg = " XZ";
			break;
		case 3 :
			bol = toDoZYsects;
			stg = " ZY";
			break;
		}
	k = bol.length;
	for(j=0;j<k;j++)
		if(bol[j])
			++i;
	talk(i+stg+" sections remain undone");
	return i;
	}
	
private int countOnBits(byte abyte) {
	int i,on = (abyte & bitMask[0]);
	for(i=1;i<8;i++)
		if((abyte & bitMask[i]) != 0)
			++on;
	return on;
	}
			
public void turnBitOn(int x, int y, int z) {
	long whichBit = NCOLS*(z*NROWS + y) + x;
	int whichByte = (int)((whichBit)/8);
	int bit = (int)(whichBit - 8*whichByte);
	bytes[whichByte] = (byte)(bytes[whichByte] | bitMask[bit]);
	}

public boolean isBitOn(int x, int y, int z) {
	long whichBit = NCOLS*(z*NROWS + y) + x;
	int whichByte = (int)((whichBit)/8);
	int bit = (int)(whichBit - 8*whichByte);
	return ((bytes[whichByte] & bitMask[bit]) != 0);
	}

		
public void closeMaskFile() {
		if(maskWriter != null)
			maskWriter.closeFile();
		}	
		
public boolean orShortTwoDimMask(short [] ras, int xyORxzORzy, int position) {
	int x,y,z,offset;
	if(ras==null) {
		talk("BitMask.orShortTwoDimMask() got null short [] ras");
		return false;
		}
		
	if(bytes==null) {
		talk("BitMask.orShortTwoDimMask() had null byte [] bytes");
		return false;
		}
		
	switch (xyORxzORzy) {
		default:
			talk("Nonsense case = "+xyORxzORzy+" sent to orShortTwoDimMask");
			return false;
			
		case 0 : // it's an xy section and position==z
			if(ras.length != NCOLS*NROWS) {
				talk("wrong size xy section sent to orShortTwoDimMask");
				return false;
				}
			for(y=0;y<NROWS;y++) {
				offset = y*NCOLS;
				for(x=0;x<NCOLS;x++)
					if(ras[offset+x]>128)
						turnBitOn(x,y,position);
				}
			break;
			
		case 1 : // it's an xz section and position==y
			if(ras.length != NCOLS*NSECTS) {
				talk("wrong size xz section sent to orShortTwoDimMask");
				return false;
				}
			for(z=0;z<NSECTS;z++) {
				offset = z*NCOLS;
				for(x=0;x<NCOLS;x++)
					if(ras[offset+x]>128)
						turnBitOn(x,position,z);
				}
			break;
			
		case 2 : // it's a zy section and position==x
			if(ras.length != NSECTS*NROWS) {
				talk("wrong size zy ssection sent to orShortTwoDimMask");
				return false;
				}
			for(z=0;z<NSECTS;z++) {
				offset = z*NROWS;
				for(y=0;y<NROWS;y++) 
					if(ras[offset + y]>128)
						turnBitOn(position,y,z);
				}
			break;
		}
	
	return true;
	}

public byte [] bytePicBlendedWithMask(short [] ras, int xyORxzORzy, int position, 
			int blendRat) {
	int i,npix;
	if(xyORxzORzy<0 || xyORxzORzy>2) {
		talk("Bad request to bytePicBlendedWithMask");
		return null;
		}
		
	short [] shortPic = shortPicBlendedWithMask(ras, xyORxzORzy, position, blendRat);
	switch (xyORxzORzy) {
		default:
		case 0 : // it's an xy section and position==z
			npix = NCOLS*NROWS;
			break;

		case 1 : // it's an xz section and position==y
			npix = NCOLS*NSECTS;
			break;
			
		case 2 : // it's a zy section and position==x
			npix = NSECTS*NROWS;
			break;
		}	
	byte [] bytePic = new byte[npix];
	for(i=0;i<npix;i++)
		bytePic[i] = (byte)(shortPic[i]);
		
	return bytePic;
	}
	
	
public short [] shortPicBlendedWithMask(short [] ras, int xyORxzORzy, int position, 
			int blendRat) {
	int x,y,z,offset,location,ratBlend = 255*(255-blendRat);
	short [] pic=null;
	switch (xyORxzORzy) {
		default:
			talk("Nonsense case = "+xyORxzORzy+" senet to blendShortXYPicWithMask");
			return null;
			
		case 0 : // it's an xy section and position==z
			if(ras.length != NCOLS*NROWS) {
				talk("wrong size xy section sent to blendShortXYPicWithMask");
				return null;
				}
			pic = new short[NCOLS*NROWS];
			for(y=0;y<NROWS;y++) {
				offset = y*NCOLS;
				for(x=0;x<NCOLS;x++) {
					location = offset+x;
					if(isBitOn(x, y, position))
						pic[location] = (short)((blendRat*ras[location] + ratBlend)/255);
					else
						pic[location] = 0;
					}						
				}
			break;
			
		case 1 : // it's an xz section and position==y
			if(ras.length != NCOLS*NSECTS) {
				talk("wrong size xz section sent to blendShortXYPicWithMask");
				return null;
				}
			pic = new short[NCOLS*NSECTS];
			for(z=0;z<NSECTS;z++) {
				offset = z*NCOLS;
				for(x=0;x<NCOLS;x++) {
					location = offset+x;
					if(isBitOn(x, position, z))
						pic[location] =  (short)((blendRat*ras[location] + ratBlend)/255);
					else
						pic[location] = 0;
					}						
				}
			break;
			
		case 2 : // it's a zy section and position==x
			if(ras.length != NSECTS*NROWS) {
				talk("wrong size zy ssection sent to blendShortXYPicWithMask");
				return null;
				}
			pic = new short[NSECTS*NROWS];
			for(z=0;z<NSECTS;z++) {
				offset = z*NROWS;
				for(y=0;y<NROWS;y++) {
					location = offset + y;
					if(isBitOn(position, y, z))
						pic[location] =  (short)((blendRat*ras[location] + ratBlend)/255);
					else
						pic[location] = 0;
					}						
				}
			break;
		}	
	return pic;
	}
		
static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
}
