package staxx;
/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.lang.*;
import java.util.*;
import java.awt.*;  
import javax.swing.*;
import java.io.*;

class CallibrationStats extends JFrame{ //This class computes and holds callibration statistics
static double  TINY = 1.e-20; 
	int tube=0,slit=0,gain=70,gainIndex;
	int npics=0;
	int width=120,blur=1;
	int thresh  = 0; // all intensities less than this are ignored.
	int [] prof = null;
	int [] normalProf = null;
	int profmax=1;
	int captionX = 10;
	boolean ifCallibrationStatsValid=false;
	static int maxNumCurveSets = 10;
	static String [] prefix = {"!","@","#","$","%","^","&","*",")","+"};
	static String [] curveSetDirName={"","","","","","","","","",""};
	static Color [] curveSetColor = new Color[maxNumCurveSets];
	/* Use a triply indexed array with first index ~ gain (actually 10*(gain-6.0), so the gain index runs from 0->60, 1->61,... 40->100),
	and with the second index = slit width (from 0 to 15)
	Each gain curve holds all the measurements made on all dye tubes at that particular gain and slit width. */
	GainCurve [][][] gncv = new GainCurve[maxNumCurveSets][16][41];
	// For example, gncv[0][2][25] holds all the measurements made in curveSet 0 at slit width=2 
	// gain = .1*(60+25) = 8.5	
	int curveSet =0;
	int leftTube=25;  // the current left boundary of the plot
	int rightTube=1;  // the current right boundary of the plot.
	GainCurve gainCurveToFit = null;
	double [] gainCurveFitCoeffs = null;
	double concentrationMappedTo255 = .07;	
	
	double [] tubeNum;
	double [] tubeIntens;
	boolean hasProfileChanged=true;
	Color mycolor = Color.red;
	HistogramBox myBox=null;
	OneProfile myProf = null;
	Histogram parent=null;
	RandomAccessFile theDataFile=null;
	TransectLine myTransect;
	boolean ifRed=false;
	
	int npts=256;
	static int grafMax = 511;
	
	int drawWhat = 0; // tells draw() what to draw.
	int [] menuy; // y-heights of clickable menu items...
	boolean [] menuItemIsOn;
	String [] clickJMenuItem;
	int numClickJMenuItems=0; // the number of clickable menu items.
	boolean ifLogScale=true;
	int [] clilckTx; // x coordinates of where to click on a T-limit
	int clickLowTy=0,clickHiTy=0;
	static int howWide=1000;
	static int [] menuCol = {howWide+35,howWide+85,howWide+135,howWide+185,howWide+235};
	static double minConc = .00024;
	static double maxConc = 1.0;
	static double logScale = (double)howWide/Math.log(maxConc/minConc);
	static double linScale = (double)howWide/(maxConc-minConc);
	static int menuGap=12;
	static int maxNumJMenuRows=44;
	static int maxNumClickJMenuItems=4*maxNumJMenuRows;
	int firstClickForSetOnOff=0;
	
	public void sepuku() {
		menuCol=null;
		tubeNum = null;
		normalProf = null;
		prof = null;
		curveSetColor = null;
		gainCurveFitCoeffs = null;
		gainCurveToFit = null;
		tubeIntens = null;
		gncv = null;
		myBox=null;
		myProf = null;
		parent=null;
		theDataFile=null;
		menuy=null;
		menuItemIsOn=null;
		clickJMenuItem=null;
		clilckTx=null;
		if(myTransect != null) myTransect.sepuku();
		myTransect=null;
		}
		
	static double [] redDyeConcentration = {0, // relative dye concentrations in the red (Bodipy) tubes..
			1.000, 		// T01
			0.71428,	// T02
			0.5,		// T03
			0.3529,		// T04
			0.250,		// T05
			0.17647,	// T06
			0.125,		// T07
			0.090909,	// T08
			0.0625,		// T09
			0.043478, 	// T10
			0.03125, 	// T11	
			0.022222,	// T12 
			0.01554,	// T13
			0.010989,	// T14
			.0079365,	// T15
			.00552486,	// T16
			.00390625,	// T17
			0.0027701,	// T18 
			.0019569,	// T19
			.00138696,	// T20 
			.0009794,	// T21
			.0006892,	// T22 
			.0004876,	// T23
			.000344708,	// T24
			.000243842}; // T25

	static double [] greenDyeConcentration = {0, // relative dye concentrations in the green (fluorescein) tubes..
			1.000, 		// T01
			0.71428,	// T02
			0.5,		// T03
			0.3529,		// T04
			0.250,		// T05
			0.17647,	// T06
			0.125,		// T07
			0.090909,	// T08
			0.0625,		// T09
			0.043478, 	// T10
			0.03125, 	// T11	
			0.022222,	// T12 
			0.01554,	// T13
			0.010989,	// T14
			.0079365,	// T15
			.00552486,	// T16
			.00390625,	// T17
			0.0027701,	// T18 
			.0019569,	// T19
			.00138696,	// T20 
			.0009794,	// T21
			.0006892,	// T22 
			.0004876,	// T23
			.000344708,	// T24
			.000243842}; // T25

	double [] dyeConcentration=redDyeConcentration;
	
	public CallibrationStats(Histogram parent, HistogramBox myBox, OneProfile myProf) {
		this.myProf = myProf;
		this.myBox=myBox;
		this.parent = parent;
		int i;
		myTransect = myBox.myTransect;
		npts = 525;
		myTransect.width=width;
		myTransect.blur=blur;
		prof = new int[npts+1];
		normalProf = new int[npts+1];
		for(i=0;i<npts;i++) 
			prof[npts]=1;
		
		hasProfileChanged=true;
		drawWhat = 0; // tells draw() what to draw.
		menuy = new int[maxNumClickJMenuItems]; // y-heights of clickable menu items...
		menuItemIsOn = new boolean[maxNumClickJMenuItems]; // whether this menu item is on or off
		clickJMenuItem = new String[maxNumClickJMenuItems];
		clilckTx = new int[26];
		for(i=0;i<maxNumClickJMenuItems;i++) 
			menuItemIsOn[i]=true;
		
	    numClickJMenuItems=0; // the number of clickable menu items.
		for(int cs=0;cs<maxNumCurveSets;cs++)
			for(slit=0;slit<16;slit++)		
		  		for(gainIndex=0;gainIndex<41;gainIndex++)
					gncv[cs][slit][gainIndex] = null;
		}

	private int recordIntensity(int g,int s,int t,double c,String bioradFileName) {
		int row,col,intens=0,light;
		BioradFile thepic=null;
		double dialGain = 0.1*g;
		byte [] picras=null;
		int gainIndex = g-60;
		if(gainIndex>40) gainIndex=40;
		if(gainIndex<0) gainIndex=0;	

		if(c <= 0.0) {
			talk("Bad file name =<"+bioradFileName);
			return 0;
			}
		
		// extract the brightest line-average intensty from the PIC file
		try {thepic = new BioradFile(bioradFileName);}
		catch (IOException e) {
			talk("Creation of single-frame BioradFile <"+bioradFileName+"> failed... IOException= "+e);
			// return 0;
			}

		try {picras = thepic.readBioradRaster(0);}
		catch(Exception e) {		// Probably a memory error, so guess that that's what it is
			talk("read <"+bioradFileName+"> failed...Runtime exception = "+e);
			return 0;
			}
		if(!thepic.isValid()) {
			talk("read <"+bioradFileName+"> failed... picras = "+picras+" thepic.isValid() = "+thepic.isValid());
			return (0);
			}
			
		intens = findMaxBarAvg(picras);			
		// tube  slit  gain  concentration intensity
		String stg = ""+t+"  "+s+"  "+g+"  "+intens+"  "+c+" from "+bioradFileName;
		if(theDataFile != null) {
			//talk("Write to RAWDAT file this datum: "+stg);
			writeAString(stg,theDataFile);
			}
		else 
			//talk("found but did not write to disk this datum: "+stg);
		if(gncv[curveSet][s][gainIndex]==null) {
			//talk(" made a new gncv["+curveSet+"]["+s+"]["+gainIndex+"]");
			gncv[curveSet][s][gainIndex]=new GainCurve(s,gainIndex,curveSet,prefix[curveSet],ifRed);
			//talk("make new gncv[][][] succeeded");
			}
		//talk("addMeasurement");
		gncv[curveSet][s][gainIndex].addMeasurement(t, c, intens, bioradFileName);
		curveSetColor[curveSet] = gncv[curveSet][s][gainIndex].myBrightestColor();
		++npics;
		//talk("recordIntensity returns with npics = "+npics);
		return intens;
		}
		
	private int findMaxBarAvg(byte [] ras) {
		int halflength=8;
		int xmid=parent.statsJPanel.picJPanel.fullWidth/2;
		int xcen=xmid,delxb=0;
		int ycen=1;
		int xlo,xhi,ylo,yhi;
		int avg;
		int xbest=xcen,ybest=0;
		int bestavg=-1;
		double delangb=0, delxcen,delang,angbest=0.,anglo,anghi,ang=0.;
		
		// first coarse pass to get in the ballpark
		int fullHeight=parent.statsJPanel.picJPanel.fullHeight;
		for(ycen=1;ycen<fullHeight;ycen += 10) 
			if((avg=barAvg(xcen,ycen,halflength,ang,ras)) >= bestavg) {
				if(avg>bestavg) { 
					bestavg=avg; xbest=xcen; ybest=ycen;
					if((delxb=xbest-xmid)<0) delxb=-delxb;
					}
				else {
					if((delxcen=xcen-xmid)<0) delxcen=-delxcen;
					if(delxcen<delxb) {  
						bestavg=avg; xbest=xcen; ybest=ycen;
						if((delxb=xbest-xmid)<0) delxb=-delxb;
						}
					}
				}
		if((delxb=xbest-xmid)<0) delxb=-delxb;
		talk("  ");	
		talk("max avg = "+bestavg+" after first pass along midline at ycen="+ybest+", xcen="+xbest+", anbest="+angbest);
		// second refinement
		int ymin=ybest-30,ymax=ybest+30;
		if(ymin<1) ymin=1;
		if(ymax>511) ymax=511;
		xlo = xbest-250;
		xhi=xbest+250;
		halflength=16; // after changing halflength, must recompute bestavg.
		bestavg=barAvg(xbest,ybest,halflength,angbest,ras);
		for(ycen=ymin;ycen<ymax;ycen += 2) 
		  for(xcen=xlo; xcen<=xhi; xcen += 20)
			if((avg=barAvg(xcen,ycen,halflength,ang,ras)) >= bestavg) {
				if(avg>bestavg) { 
					bestavg=avg; xbest=xcen; ybest=ycen;
					if((delxb=xbest-xmid)<0) delxb=-delxb;
					}
				else {
					if((delxcen=xcen-xmid)<0) delxcen=-delxcen;
					if(delxcen<delxb) {  
						bestavg=avg; xbest=xcen; ybest=ycen;
						if((delxb=xbest-xmid)<0) delxb=-delxb;
						}
					}
				}
		if((delxb=xbest-xmid)<0) delxb=-delxb;
		talk("max avg = "+bestavg+" after second  refinement at ycen="+ybest+", xcen="+xbest+", anbest="+angbest);
				
		// search a mesh of (ang,xcen,ycen) near (xbest,ybest,ang=0)
		if((ylo=ybest-6)<0) ylo=0;
		if((yhi=ybest+6)>511) yhi=511;
		xlo = xbest-20;
		xhi = xbest+20;
		halflength=30; // after changing halflength, must recompute bestavg.
		bestavg=barAvg(xbest,ybest,halflength,angbest,ras);
		for(xcen = xlo; xcen<=xhi;xcen += 2) 	// 20 passes
			for(ycen=ylo;ycen<=yhi;ycen += 2) 	// 6 passes
				for(ang=-10;ang<10;ang += 2)    // 10 passes 
					if((avg=barAvg(xcen,ycen,halflength,ang,ras)) >= bestavg) {
				if(avg>bestavg) { 
					bestavg=avg; xbest=xcen; ybest=ycen; angbest=ang;
					if((delxb=xbest-xmid)<0) delxb=-delxb;
					}
				else {
					if((delxcen=xcen-xmid)<0) delxcen=-delxcen;
					if(delxcen<delxb) {  
						bestavg=avg; xbest=xcen; ybest=ycen; angbest=ang;
						if((delxb=xbest-xmid)<0) delxb=-delxb;
						}
					}
				}
		if((delxb=xbest-xmid)<0) delxb=-delxb;
		talk("max avg = "+bestavg+" after third  refinement at ycen="+ybest+", xcen="+xbest+", anbest="+angbest);

		// Use a long bar and search a mesh of (ang,xcen,ycen) near (xbest,ybest,ang=0)
		if((ylo=ybest-4)<0) ylo=0;
		if((yhi=ybest+4)>511) yhi=511;
		xlo = xbest-8;
		xhi = xbest+8;
		anglo=angbest-6;
		anghi=angbest+6;
		halflength=40; // after changing halflength, must recompute bestavg.
		bestavg=barAvg(xbest,ybest,halflength,angbest,ras);
		for(xcen = xlo; xcen<=xhi;xcen += 1) 		// 16 passes
			for(ycen=ylo;ycen<=yhi;ycen += 1) 		// 8 passes
				for(ang=anglo;ang<anghi;ang += 2)    // 10 passes 
					if((avg=barAvg(xcen,ycen,halflength,ang,ras)) >= bestavg) {
				if(avg>bestavg) { 
					bestavg=avg; xbest=xcen; ybest=ycen; angbest=ang;
					if((delxb=xbest-xmid)<0) delxb=-delxb;
					}
				else {
					if((delxcen=xcen-xmid)<0) delxcen=-delxcen;
					if(delxcen<delxb) {  
						bestavg=avg; xbest=xcen; ybest=ycen; angbest=ang;
						if((delxb=xbest-xmid)<0) delxb=-delxb;
						}
					}
				}
		talk("max avg = "+bestavg+" after long-bar refinement at ycen="+ybest+", xcen="+xbest+", anbest="+angbest);

		anglo=angbest-12;
		anghi=angbest+12;
		halflength=50; // after changing halflength, must recompute bestavg.
		bestavg=barAvg(xbest,ybest,halflength,angbest,ras);
		for(ang=anglo;ang<anghi;ang += 1)    // 10 passes 
			if((avg=barAvg(xbest,ybest,halflength,ang,ras)) >= bestavg) {
				if(avg>bestavg) { 
					bestavg=avg; angbest=ang;
					if((delangb=angbest)<0) delangb=-delangb;
					}
				else {
					if((delang=ang)<0) delang=-delang;
					if(delang<delangb) {  
						bestavg=avg; angbest=ang;
						if((delangb=angbest)<0) delangb=-delangb;
						}
					}
				}
		talk("max avg = "+bestavg+" after after angular refinement at ycen="+ybest+", xcen="+xbest+", anbest="+angbest);
		return bestavg;
		}
		
	private int barAvg(int xcen, int ycen, int halflength, double ang, byte [] ras){
		int ix,iy;
		double cosang=Math.cos(ang*Math.PI/180.);
		double sinang=Math.sin(ang*Math.PI/180.);
		double s=-halflength;
		double ds = .5;
		double x = xcen + cosang*s;
		double y = ycen + sinang*s;
		int ixold = (int)x;
		int iyold = (int)y;
		int avg=pixat(ixold,iyold,ras);
		int npix=1;
		while((s += ds) <= halflength) {
			x = xcen + cosang*s;
			y = ycen + sinang*s;
			ix = (int)x;
			iy = (int)y;
			if(ix != ixold || iy != iyold) {
				avg += pixat(ix,iy,ras);
				++npix;
				ixold=ix;
				iyold=iy;
				}
			}
		return (avg/npix);	
		}
			
	private int pixat( int col, int row, byte [] ras) {
		int pix = ras[row*parent.statsJPanel.picJPanel.fullWidth +col];
		if(pix < 0) pix += 256;
		return pix;
		}
	
	public void purgeEmptyDirectories() { // delete any empty directory in the RED or GREEN dirs
		File chanDir,tubeDir,slitDir;
		String chanDirName,tubeDirName,slitDirName;
		double conc;
		int ndeleted=0,npics=0,nfiles,rorg,id;
		boolean ifMadeDiscardDir=false;
		String [] files;
		
		talk("Pick a file in the RED or GREEN channel sub directory whose empty subdirectories");
		talk("you want to purge");
		if((chanDirName = getDir("Pick a file in the directory you want to purge of empty directories"))==null)
				return; 
		talk("The channel directory is "+chanDirName);
		
		// see which tube directories exist...
		for(rorg=0;rorg<2;rorg++) {
		 if(rorg==1) {  // swap RED<--->GREEN
		 	tubeDirName=chanDirName;
		 	if(chanDirName.endsWith("/RED/")) {
		 		tubeDirName= (chanDirName.substring(0, chanDirName.indexOf("/RED/")))+"/GREEN/";
		 		}
		 	if(chanDirName.endsWith("/GREEN/")) {
		 		tubeDirName= (chanDirName.substring(0, chanDirName.indexOf("/GREEN/")))+"/RED/";
		 		}
		 	if(chanDirName.compareTo(tubeDirName)==0) 
		 		return;
		 	chanDirName = tubeDirName;
		 	talk("Now purge "+chanDirName);
		 	}
		 for(slit=0;slit<17;slit++){
			if(slit==16) slitDirName = chanDirName+"SRENAME/";
			else {
				if(slit<10)
					slitDirName = chanDirName+"S0"+slit+"/";
				else 
					slitDirName = chanDirName+"S"+slit+"/";
				}
			slitDir = new File(slitDirName);
			if(slitDir != null && slitDir.exists() && slitDir.isDirectory()) { // see which tube directories are in this slitDir...
				for(tube=1;tube<26;tube++) {
					if(tube<10)
						tubeDirName = slitDirName+"T0"+tube+"/";
					else 
						tubeDirName = slitDirName+"T"+tube+"/";
					tubeDir = new File(tubeDirName);
					if(tubeDir.exists() && tubeDir.isDirectory()) { 
						if((nfiles=tubeDir.list().length) ==0) { // this tube directory is empty; delete it
							tubeDir.delete();
							++ndeleted;
							}
						else { // count the PIC files
							files=tubeDir.list();
							for(id=0;id<nfiles;id++)
								if(files[id].endsWith(".PIC"))
									++npics;
							}
						}
					}
				if(slitDir != null && (slitDir.list()).length ==0) // the whole slit directory is empty; delete it.
					slitDir.delete();
				}
			}
			talk("Deleted "+ndeleted+" empty subdirectyories from "+chanDirName);
			talk(chanDirName+" contains "+npics+" PIC Files");
			ndeleted=0;
			npics=0;
		  }
		talk("Purge complete...");
		talk("");
		return;
		}				
		
	public boolean overlayAnotherCurveSet() {
		int id,nfiles,gainIndex,I,Imin,Imax;
		File chanDir,aFile,tubeDir,slitDir;
		String chanDirName,tubeDirName,slitDirName,entireFileName,instring;
		byte [] theBytes=null;
		String [] files;
		OnePoint aPoint=null;
		double conc;
		boolean ifDrawOldGainCurvFile=false;
		
		if(curveSet>=maxNumCurveSets) {
			talk("Cannot overlay any more curve sets. 10 is the limit...");
			return false;
			}
			
		talk("Pick a file in the channel directory you want to scan");
		talk("If this directory contains a file called 'RAWDAT', we draw the gain curves it contains.");
		talk("Else, we compute the gain curves from all the PIC files in this directory."); 
		talk("That takes a long time..."); 
		if((chanDirName = getDir("Pick a file in the directory you want to scan"))==null)
				return false; 
		talk("The channel directory is "+chanDirName);
		chanDir = new File(chanDirName);
		files = chanDir.list();
		nfiles = files.length;
		talk("there are "+nfiles+" entries in this directory, including these files:");
		for(id=0;id<nfiles;id++) {
			entireFileName=chanDirName+files[id];
			aFile = new File(entireFileName);
			if(aFile != null && aFile.isFile()) {
				talk("<"+entireFileName+">");
				if(files[id].compareTo("RAWDAT")==0) {
					talk("the 'RAWDAT' file exists, so take gain curves from it."); 
					ifDrawOldGainCurvFile=true;
					}
				}
			}
			
		boolean ifInValidGreenOrRedDirectory = false;
		if(chanDirName.indexOf("/GREEN/") >0 || chanDirName.indexOf("/green/") >0) { 
			ifRed=false;
			ifInValidGreenOrRedDirectory=true;
			talk("Using the green channel dye tubes");
			dyeConcentration=greenDyeConcentration;
			}

		if(chanDirName.indexOf("/RED/") >0 || chanDirName.indexOf("/red/") >0) { 
			ifRed=true;
			ifInValidGreenOrRedDirectory=true;
			talk("Using the red channel dye tubes");
			dyeConcentration=redDyeConcentration;
			}
		
		if(!ifInValidGreenOrRedDirectory) {
			talk("You are not in a subdirectory called red or RED or green or GREEN.");
			ifCallibrationStatsValid=false;
			return false;
			}

		if(ifDrawOldGainCurvFile)  { // just draw the previous GAINCURV file
			talk("drawing gain curves taken from file RAWDAT");
			entireFileName = chanDirName+"RAWDAT";
			theDataFile=openReadWriteFile(entireFileName); // get the file open to read the old data.
			instring=readOneLine(theDataFile);

			if(instring.compareTo("T  Sl gain brightness  dyeConc  <FILENAME.PIC>") != 0) {
				talk("The RAWDAT file exists but has the wrong first line. No action.");
				talk("<"+instring+"> = actual first line");
				talk("first line should have been: T  Sl gain brightness  dyeConc  <FILENAME.PIC>");
				ifCallibrationStatsValid=false;
				closeOutputFile(theDataFile);
				return false;
				}

			int [] gsti=new int[4];
			int bright;
			String fileName="not empty";
			while(fileName!=null && (instring=readOneLine(theDataFile)) != null && 
					instring.compareTo("overflow line") != 0) {
				if((fileName=parseRawDatLine(instring,gsti)) != null) {
					gain=gsti[0];	slit=gsti[1];	tube=gsti[2];	bright=gsti[3];
					conc=dyeConcentration[tube];
					talk("Got: g="+gain+", s="+slit+", t="+tube+", intens="+bright+" from "+fileName);
					gainIndex=gain-60;
					if(gncv[curveSet][slit][gainIndex] == null)
						gncv[curveSet][slit][gainIndex]= 
								new GainCurve(slit,gainIndex,curveSet,prefix[curveSet],ifRed);
					gncv[curveSet][slit][gainIndex].addMeasurement(tube, conc, bright, fileName);
					curveSetColor[curveSet] = gncv[curveSet][slit][gainIndex].myBrightestColor();
					++npics;
				/* typical instring is
9  0   7   56  0.0625  from /Macintosh%20HD/Desktop%20Folder/local/gmo/AUG0698/RED/T09/S00/G70.PIC
					*/
					}
				}
			ifCallibrationStatsValid=true;
			talk("There were "+npics+" .PIC files in the directory <"+chanDirName+">");
			closeOutputFile(theDataFile);
			++curveSet;
			}

		if(!ifDrawOldGainCurvFile && curveSet>0)  {
			talk("You cannot overlay another date's gain curves before you extract the intensity information");
			talk("from the PIC files..");
			}
		if(!ifDrawOldGainCurvFile && curveSet==0) {
		entireFileName = chanDirName+"RAWDAT";
		theDataFile=openReadWriteFile(entireFileName); // get the file open and ready to accept successive lines of data.
		writeAString("T  Sl gain brightness  dyeConc  <FILENAME.PIC>",theDataFile);
		
		if(theDataFile == null) {
			talk("cannot open <"+entireFileName+"> to write the data in. Bye");
			ifCallibrationStatsValid=false;
			return false;
			}
		// see which tube directories exist...
		for(slit=0;slit<16;slit++){
			if(slit<10)
				slitDirName = chanDirName+"S0"+slit+"/";
			else 
				slitDirName = chanDirName+"S"+slit+"/";
			slitDir = new File(slitDirName);
			if(slitDir.exists() && slitDir.isDirectory()) { // see which tube directories are in this slitDir...
				for(tube=1;tube<26;tube++) {
					if(tube<10)
						tubeDirName = slitDirName+"T0"+tube+"/";
					else 
						tubeDirName = slitDirName+"T"+tube+"/";
					tubeDir = new File(tubeDirName);
					if(tubeDir.exists() && tubeDir.isDirectory()) { // see if this slit directory has any G...PIC files in it
						conc=dyeConcentration[tube];
						files = tubeDir.list();
						nfiles = files.length;
						for(id=0;id<nfiles;id++) {
							entireFileName=tubeDirName+files[id];
							aFile = new File(entireFileName);
							if(aFile != null && aFile.isFile()) { // see if it is a validly named gain file
								if((gain=parseGainFileName(files[id])) >0) 
									talk("Intensity = "+recordIntensity(gain,slit,tube,conc,entireFileName)+
									" for tube="+tube+", slit="+slit+", and gain="+gain+", conc="+conc);
								else
									talk("stray file <"+entireFileName+"> ignored");
							}
						}
					}
				}
			}
		}
		// put some blank lines at the end of the file
		writeAString("777                                        ",theDataFile);
		writeAString("777                                        ",theDataFile);
		writeAString("777                                        ",theDataFile);
		writeAString("777                                        ",theDataFile);
		writeAString("777                                        ",theDataFile);
		writeAString("777                                        ",theDataFile);
		closeOutputFile(theDataFile);			

		talk("Create the GAINCURV file");
		entireFileName = chanDirName+"GAINCURV";
		theDataFile=openReadWriteFile(entireFileName); // get the file open and ready to accept successive lines of data.
		writeAString(("Gain Curve File for <"+chanDirName+">"),theDataFile);

		for(slit=0;slit<16;slit++)
		  for(gainIndex=0;gainIndex<41;gainIndex++) {
				if(gncv[curveSet][slit][gainIndex] != null &&
				   gncv[curveSet][slit][gainIndex].setupFirstConcentration()) {
					talk("Got valid gain curve for gain=="+(gainIndex+60)+", slit="+slit);
					while((aPoint=gncv[curveSet][slit][gainIndex].nextConcentration()) != null) {
						writeAString(aPoint.contents(),theDataFile);
						}}}						
		writeAString("                                           ",theDataFile);
		writeAString("                                           ",theDataFile);
		writeAString("                                           ",theDataFile);
		writeAString("                                           ",theDataFile);
		writeAString("                                           ",theDataFile);
		writeAString("                                           ",theDataFile);
		closeOutputFile(theDataFile);
		talk("There were "+npics+" .PIC files in the directory <"+chanDirName+">");
		++curveSet;
		ifCallibrationStatsValid=true;
		}
		setPlotScales();
		ifLogScale=true;
		talk("There are now "+curveSet+" curve sets");
		curveSetDirName[curveSet-1] = chanDirName;
		return true;
	}

	private String parseRawDatLine(String instring,int [] gsti){
				/* typical instring is
9  0   7    56   0.0625  from /Macintosh%20HD/Desktop%20Folder/local/gmo/AUG0698/RED/T09/S00/G70.PIC
					*/
talk("parseRawDatLine instring = <"+instring+">");
		char c;
		int number,nextDigit=0;
		int at=0;
		if(instring.length() < 20) return null;
		while(at <128 && (nextDigit=digitQ(instring.charAt(at))) == -32) // skip white space
			++at;
		if(at>10) return null; // stop parsing at first blank line;
		number=nextDigit; ++at;
		while(at <instring.length() && (nextDigit=digitQ(instring.charAt(at))) >= 0) {
			number = 10*number+nextDigit;
			++at;
			}
		if(number==777) return null; // flags last line of file
		// number is the tube number
		gsti[2]=number;
		
		while(at <instring.length() && (nextDigit=digitQ(instring.charAt(at))) == -32) // skip white space
			++at;
		number=nextDigit; ++at;
		while(at <instring.length() && (nextDigit=digitQ(instring.charAt(at))) >= 0) {
			number = 10*number+nextDigit;
			++at;
			}
		// number is the slit number
		gsti[1]=number;
		
		while(at <instring.length() && (nextDigit=digitQ(instring.charAt(at))) == -32) // skip white space
			++at;
		number=nextDigit; ++at;
		while(at <instring.length() && (nextDigit=digitQ(instring.charAt(at))) >= 0) {
			number = 10*number+nextDigit;
			++at;
			}
		// number is the gain
		gsti[0]=number;
		
		while(at <instring.length() && (nextDigit=digitQ(instring.charAt(at))) == -32) // skip white space
			++at;
		number=nextDigit; ++at;
		while(at <instring.length() && (nextDigit=digitQ(instring.charAt(at))) >= 0) {
			number = 10*number+nextDigit;
			++at;
			}
		// number is the intensity reading
		gsti[3]=number;
		if((at = instring.indexOf(" from "))>0) {
			at += 6;
			return(instring.substring(at));
			}
		return null;
		}				

	private int digitQ(char c) {
		if(c == '0') return 0;
		if(c == '1') return 1;
		if(c == '2') return 2;
		if(c == '3') return 3;
		if(c == '4') return 4;
		if(c == '5') return 5;
		if(c == '6') return 6;
		if(c == '7') return 7;
		if(c == '8') return 8;
		if(c == '9') return 9;
		if(c == ' ') return -32;
		if(c == ',') return -32;
		if(c == '\t') return -32;
		return -1;
		}
	
	private int parseGainFileName(String fname) { 
		// if File name starts with 'G' of 'g' followed by a 1 or two digit number, then return the appropriate gain number
		// else return -1;
		if(fname.charAt(0) != 'G' && fname.charAt(0) != 'g')
			return (-1);		
		int first = digitQ(fname.charAt(1));
		if(first<0) return (-1);
		int second = digitQ(fname.charAt(2));
		if(second<0) return 10*first;
		second += 10*first;
		if(second==10) return 100;
		else return second;
		}

	public double [] getGainCurveFit() { /* return null if anything goes wrong. Else returns the
		three coefficients of the best quadratic polynomial fit to the inverse of the 
		data points in the Gain curve specified by gainCurveToFit. */;
		if(gainCurveToFit==null || !(gainCurveToFit.setupFirstConcentration())) 
			return null;
		double [] tbcon = new double[16]; // there cannot possibly be that many tubes measured in any
		double [] tbglo = new double[16]; // particular gain curve.
		int i,j,k;
		for(i=0;i<16;i++) {tbcon[i]=-1;   tbglo[i]=-1;}
		int nglows=0;
		double conc;
		OnePoint aPoint=null;
		talk("getGainCurveFit: leftTube="+leftTube+"  rightTube="+rightTube);
		 tbcon[nglows]=0; // Make sure (0, 0) is one of the data points. 
		 tbglo[nglows]=0;
		 ++nglows;
		while((aPoint=gainCurveToFit.nextConcentration()) != null) {
		    conc=aPoint.getConcentration();
		    if(conc>= 0 && conc <= concentrationMappedTo255) {
		        tbcon[nglows] = conc;
				tbglo[nglows] = aPoint.getReading();
				++nglows;
				}
		    }
		// Ok now we have all the data recorded for gainCurveToFit in the tbcon,tbglo arrays.
		// gainCurveFitCoeffs[] holds the quadratic coefficients to be returned. 
		/* the map tbcon->tbglo assigns a glow to each of nglows concentrations. 
		 * the map tbglo->tbcon assigns a concentration to each of the nglows raw glow values. 
		 * it is the latter map we want to fit with a quadratic.
		 */
		 double [][] a = new double[3][3];
		 double [] b = new double[3];
		 double sum;
		 for(k=0;k<3;k++) {
		   	b[k]=0;
		   	for(j=0;j<nglows;j++)
		   		b[k] += tbcon[j]*phi(k,tbglo[j]);
		   	for(i=0;i<=k;i++) {
		   		sum=0;
		   		for(j=0;j<nglows;j++)
		   			sum += phi(k,tbglo[j])*phi(i,tbglo[j]);
		   		a[k][i] = sum;
		   		a[i][k] = sum;
		   		}
		   	}
		 // solve the system of linear equations...   		
		 gainCurveFitCoeffs = gauss(3,a,b);
		 if(gainCurveFitCoeffs==null) 
		 		talk("Singular matrix. Cannot fit a quadratic to that gain curve ");
		 else  		
		 	talk("The quadratic fit to the inverse gain curve has coefficients: "+gainCurveFitCoeffs[0]+
		 		", "+gainCurveFitCoeffs[1]+", "+gainCurveFitCoeffs[2]);
		 talk("There are nglows="+nglows+" data points to fit,  and they are");
		 talk("npt\t\t glow\t\t conc\t\tcurveFitValue");
		 for(i=0;i<nglows;i++) talk(i+"\t\t "+tbglo[i]+"\t\t"+tbcon[i]+"\t\t"+
		 		quadFitConc(tbglo[i]));  

		 return gainCurveFitCoeffs;
 		}
 		
public double  quadFitConc(double glo) {
	if(gainCurveFitCoeffs==null) return 0;
	return (gainCurveFitCoeffs[0] + glo*(gainCurveFitCoeffs[1] + glo*gainCurveFitCoeffs[2]));
	}
			
private double phi(int n, double x) {
	if(n==0) return 1.0;
	return Math.pow(x,n);
	}
		
private double unifnorm(double [] vec, int n) { // returns the uniform or max norm of an n-vector
	int i=0;	
	double dum = Math.abs(vec[i]);
	for(i=1;i<n;i++) 
		if(Math.abs(vec[i]) > dum)
			dum = Math.abs(vec[i]);
	return dum;
	}

public double [] gauss(int n, double [][]a, double [] b) {
	/*  function to solve the matrix equation a x=b where amat is an nxn matrix and
    b is an n-vector. Returns the solution, x, and destroys both the a matrix 
    and the right hand side, b.  
    Uses scaled partial pivoting.
    
    april 1998, GMO. 

    This function returns the answer in the x[] array. But if the matrix is singular
    (the constant TINY is used as "zero" for determining singularity), then his function
    returns null.

    Algorithm 6.3, page 293 in Burden, Faires, and Reynolds...   */

int [] nrow = new int[n];   /* an integer a pointer array to simulate row swaps... */
int i,j, k,ipiv;
double [] rownorms = new double[n];
double pivrat,piv;
boolean iftalk= false;


	for(i=0;i<n;i++) {
		nrow[i]=i;
		if((rownorms[i] = unifnorm(a[i],n)) <= TINY) {
			talk("       row "+i+" is zero. Bye");
			return null;
			}
		if(iftalk){ talk("       row norm for row "+i+" = "+rownorms[i]);}
		}
		
	for(i=0;i<n-1;i++) {
		ipiv=i;
		pivrat = Math.abs(a[nrow[i]][i]/rownorms[nrow[i]]);
		for(j=i+1;j<n;j++) // find the best pivot row among rows greater than or equal i
			if((piv = Math.abs(a[nrow[j]][i])/rownorms[nrow[j]]) >= pivrat) {
				pivrat = piv;
				ipiv=j;
				}
		if(Math.abs(a[nrow[ipiv]][i]) <= TINY) {
			talk("       gauss: singular matrix: pivot element "+i+" is too small");
			return null;
			}		
		if(nrow[ipiv] != (j=nrow[i])) { // simulate swapping rows
			nrow[i] = nrow[ipiv];
			nrow[ipiv] = j;
			}
			
		if(iftalk){ talk("       "+i+"-th pivot row is row "+nrow[ipiv]);}
		
		for(j=i+1;j<n;j++) {
			piv = a[nrow[j]][i]/a[nrow[i]][i];
			a[nrow[j]][i] = piv;
			b[nrow[j]] -= piv*b[nrow[i]];
			for(k=i+1;k<n;k++)
				a[nrow[j]][k] -= piv*a[nrow[i]][k];
			}
		if(Math.abs(a[nrow[n-1]][n-1]) < TINY) {
			talk("       gauss: singular matrix: pivot element "+(n-1)+" is too small");
			return null;
			}
		}
		// back substitution
		double [] x = new double[n];
		i=n-1;
		x[i] = b[nrow[i]]/a[nrow[i]][i];
		while(i>=0) {
			piv = b[nrow[i]];
			for(j=i+1;j<n;j++)
				piv -= a[nrow[i]][j]*x[j];
			x[i] = piv/a[nrow[i]][i];
			--i;
			}
	return x;	
}
	
	public void actOnMouseClick(int x,int y){
		int i,j,dely,mcol,csonoff,cs;
		boolean onOrOff=false, newFit;
		boolean needsCompleteRepainting=false;
		if(!parent.ifMonochrome && x < menuCol[4] && x>= menuCol[0] && y<clickHiTy-5) {
		  for(i=0;i<numClickJMenuItems;i++) {
		  	mcol = i/maxNumJMenuRows;
			if((dely=menuy[i]-y) < 0) dely=-dely;
			if(dely < 5 && x >= menuCol[mcol] && x <= menuCol[mcol+1]) {
				drawWhat = i;
				if(drawWhat>=firstClickForSetOnOff) { 
					// turn on (if drawWhat is even) or off (if drawWhat is odd) some curveSet
					drawWhat -= firstClickForSetOnOff;
					csonoff=drawWhat/2;
					if((drawWhat - 2*csonoff) == 0) 
						onOrOff=true;
					for(j=4;j<firstClickForSetOnOff;j++)
						if(clickJMenuItem[j].indexOf(prefix[csonoff])==0) {
							menuItemIsOn[j]=onOrOff;
							needsCompleteRepainting=true;
							}
					}				
				else switch (drawWhat) {
					case 0: // all gains on
						for(j=4;j<numClickJMenuItems;j++)
							menuItemIsOn[j]=true;
						break;
					case 1: // all gains off
						for(j=4;j<numClickJMenuItems;j++)
							menuItemIsOn[j]=false;
						needsCompleteRepainting=true;
						break;
					case 2: // log/Linear toggle
						if((menuItemIsOn[2] = !menuItemIsOn[2]))
							ifLogScale=true;
						else 
							ifLogScale=false;
						needsCompleteRepainting=true;
						setPlotScales();
						break;
					case 3: // full horizontal scale
						leftTube=25;
						rightTube=1;
						setPlotScales();
						needsCompleteRepainting=true;
						break;
					case 4: // set up for next click to select the gainCurveToFit
						gainCurveToFit=null;
						gainCurveFitCoeffs=null;
						talk("Next time you click on the name of any gain curve, that");
						talk("selects the single gain curve to fit... ");
						break;
					case 5: // boost conc ~ 255
						concentrationMappedTo255 *= 1.02;
						talk("concentrationMappedTo255 changed to "+concentrationMappedTo255);
						if(concentrationMappedTo255 > 1.0)
							concentrationMappedTo255=1.0;
						talk("concentrationMappedTo255 changed to "+concentrationMappedTo255);
						needsCompleteRepainting=true;
						break;
					case 6: // greatly boost conc ~ 255
						concentrationMappedTo255 *= 1.15;
						talk("concentrationMappedTo255 changed to "+concentrationMappedTo255);
						if(concentrationMappedTo255 > 1.0)
							concentrationMappedTo255=1.0;
						talk("concentrationMappedTo255 changed to "+concentrationMappedTo255);
						needsCompleteRepainting=true;
						break;
					case 7: // cut conc ~ 255
					    concentrationMappedTo255 *= .99;
						talk("concentrationMappedTo255 changed to "+concentrationMappedTo255);
						needsCompleteRepainting=true;
						break;
					case 8: // greatly cut conc ~ 255
					    concentrationMappedTo255 *= .89;
						talk("concentrationMappedTo255 changed to "+concentrationMappedTo255);
						needsCompleteRepainting=true;
						break;
					default:
						menuItemIsOn[drawWhat]=!menuItemIsOn[drawWhat];
						newFit=false;
						if(gainCurveToFit==null)  {
							talk("gainCurveToFit --> "+clickJMenuItem[drawWhat]);
						   for(cs=0;cs<curveSet;cs++)
			 				 for(slit=0;slit<16;slit++)
			  					for(gainIndex=0;gainIndex<41;gainIndex++)
									if(gncv[cs][slit][gainIndex] != null
									       && drawWhat==(gncv[cs][slit][gainIndex]).clickJMenuId) {
										gainCurveToFit = gncv[cs][slit][gainIndex];
										getGainCurveFit();
										newFit=true;
										}
							}
						if(!menuItemIsOn[drawWhat] || newFit)
							needsCompleteRepainting=true;
						break;
						}
				if(needsCompleteRepainting)
					parent.canvas.forceRepaint=true;
				parent.canvas.repaint();
				return;
				}
			}}

		if(!parent.ifMonochrome && y >= clickHiTy-5) {
			for(i=rightTube;i<=leftTube;i++) 
				if(x>=clilckTx[i] && x<clilckTx[i]+20) {
					if(y<=clickHiTy+3 && y>=clickHiTy-3) {
						if(i<leftTube) {
							if(i==rightTube && rightTube>1)
								--rightTube;
							else
								rightTube=i;
							setPlotScales();
							needsCompleteRepainting=true;
							}
						}
					if(y<=clickLowTy+3 && y>=clickLowTy-3) {
						if(i>rightTube) {
							if(i==leftTube && leftTube<25)
								++leftTube;
							else
								leftTube=i;
							setPlotScales();
							needsCompleteRepainting=true;
							}
						}
					}
			if(needsCompleteRepainting)
				parent.canvas.forceRepaint=true;
			parent.canvas.repaint();
			return;
			}
		return;
		}
		
	private void setPlotScales() {
		minConc = dyeConcentration[leftTube];
		maxConc = dyeConcentration[rightTube];
		logScale = howWide/Math.log(maxConc/minConc);
		linScale = (double)howWide/(maxConc-minConc);
		menuCol[0]=howWide+35; 
		menuCol[1]=howWide+85;
		menuCol[2]=howWide+135;
		menuCol[3]=howWide+185;	
		menuCol[4]=howWide+235;	
		}
		
	public void draw(Graphics g,int howWide, int howTall) {
		this.howWide=howWide;
		int ix1,ix2,iy1,iy2,i,gainIndex,mcol,cs;
		OnePoint aPoint;
		double cdec,c;
		boolean didDraw,alternate;
		Color curveOffColor = new Color(120,120,120);
		if(parent.ifMonochrome) {
			g.setColor(Color.black);
			numClickJMenuItems=0;
			}
		else {
			g.setColor(Color.cyan);
			iy1=menuGap;
			numClickJMenuItems=0;
			mcol=0;
			ix1=menuCol[mcol];
			clickJMenuItem[numClickJMenuItems]="All on";  
			menuy[numClickJMenuItems]=iy1-4;
			g.drawString(clickJMenuItem[numClickJMenuItems],ix1,iy1);
			++numClickJMenuItems;
			iy1 += menuGap;

			clickJMenuItem[numClickJMenuItems]="All off";  
			menuy[numClickJMenuItems]=iy1-4;
			g.drawString(clickJMenuItem[numClickJMenuItems],ix1,iy1);
			++numClickJMenuItems;
			iy1 += menuGap;

			if(menuItemIsOn[2])
				clickJMenuItem[numClickJMenuItems]="log";  
			else 
				clickJMenuItem[numClickJMenuItems]="linear";  
				
			menuy[numClickJMenuItems]=iy1-4;
			g.drawString(clickJMenuItem[numClickJMenuItems],ix1,iy1);
			++numClickJMenuItems;
			iy1 += menuGap;

			clickJMenuItem[numClickJMenuItems]="T1 - T25";  
			menuy[numClickJMenuItems]=iy1-4;
			g.drawString(clickJMenuItem[numClickJMenuItems],ix1,iy1);
			++numClickJMenuItems;
			iy1 += menuGap;

			clickJMenuItem[numClickJMenuItems]="click2fit";  
			menuy[numClickJMenuItems]=iy1-4;
			g.drawString(clickJMenuItem[numClickJMenuItems],ix1,iy1);
			++numClickJMenuItems;
			iy1 += menuGap;

			clickJMenuItem[numClickJMenuItems]="^ conc ~ 255";  
			menuy[numClickJMenuItems]=iy1-4;
			g.drawString(clickJMenuItem[numClickJMenuItems],ix1,iy1);
			++numClickJMenuItems;
			iy1 += menuGap;

			clickJMenuItem[numClickJMenuItems]="^^conc ~ 255";  
			menuy[numClickJMenuItems]=iy1-4;
			g.drawString(clickJMenuItem[numClickJMenuItems],ix1,iy1);
			++numClickJMenuItems;
			iy1 += menuGap;

			clickJMenuItem[numClickJMenuItems]="v conc ~ 255";  
			menuy[numClickJMenuItems]=iy1-4;
			g.drawString(clickJMenuItem[numClickJMenuItems],ix1,iy1);
			++numClickJMenuItems;
			iy1 += menuGap;

			clickJMenuItem[numClickJMenuItems]="VVconc ~ 255";  
			menuy[numClickJMenuItems]=iy1-4;
			g.drawString(clickJMenuItem[numClickJMenuItems],ix1,iy1);
			++numClickJMenuItems;
			iy1 += menuGap;

			for(cs=0;cs < curveSet;cs++)
			 for(slit=0;slit<16;slit++)
			  for(gainIndex=0;gainIndex<41;gainIndex++)
					if(gncv[cs][slit][gainIndex] != null && gncv[cs][slit][gainIndex].isValid()) {
						gncv[cs][slit][gainIndex].clickJMenuId=numClickJMenuItems;
						clickJMenuItem[numClickJMenuItems]=gncv[cs][slit][gainIndex].myName;
						menuy[numClickJMenuItems]=iy1-4;
						if(menuItemIsOn[numClickJMenuItems])
							g.setColor(gncv[cs][slit][gainIndex].myColor);
						else g.setColor(curveOffColor);	
						ix1=menuCol[mcol];
						g.drawString(clickJMenuItem[numClickJMenuItems],ix1,iy1);
						if(numClickJMenuItems<maxNumClickJMenuItems)
							++numClickJMenuItems;
						if(numClickJMenuItems>=(mcol+1)*maxNumJMenuRows) {
							if(mcol<3)
								++mcol;
							iy1 = menuGap;
							}
						else 
							iy1 += menuGap;
						}
			firstClickForSetOnOff = numClickJMenuItems;
			cs=0;
			g.setColor(Color.cyan);
			alternate=true;
			while(cs<curveSet && numClickJMenuItems<maxNumClickJMenuItems) {
				if(alternate) {
					clickJMenuItem[numClickJMenuItems]="all "+prefix[cs]+" on";
					alternate=false;
					}
				else {
					clickJMenuItem[numClickJMenuItems]="all "+prefix[cs]+" off";
					alternate=true;
					++cs;
					}
				g.drawString(clickJMenuItem[numClickJMenuItems],ix1,iy1);
				menuy[numClickJMenuItems]=iy1-4;
				++numClickJMenuItems;
				if(numClickJMenuItems>=(mcol+1)*maxNumJMenuRows) {
					if(mcol<3)
						++mcol;
					ix1=menuCol[mcol];
					iy1 = menuGap;
					}
				else 
					iy1 += menuGap;
				}
				
			g.setColor(Color.gray);// draw the linearly spaced horizontal grid lines in gray
				}
		
		for(iy1=50;iy1<howTall;iy1+=50) {		
			g.drawLine(0,grafMax-iy1,howWide,grafMax-iy1);
			g.drawString((""+iy1/2),howWide+5,grafMax-iy1+5);
			}
		if(!parent.ifMonochrome) g.setColor(Color.yellow);
		g.drawLine(0,grafMax-2*255,howWide,grafMax-2*255);
		
		
		// draw the log scale tic markers along horizontal axis.
		// the concentration values will range from minConc up to maxConc
		for(cdec=1.0; cdec>.0002; cdec = cdec/10.0) {
			if(!parent.ifMonochrome) g.setColor(Color.yellow);
			iy1=grafMax;
			iy2=iy1+10;
			if(cdec>=minConc && cdec <= maxConc)  
				g.drawString((""+cdec), scaledConcentration(cdec, howWide)-8, iy1-2);
			for(i=10;i>1;i--)  {
				c = 0.1*i*cdec;
				if(c>=minConc && c <= maxConc) {
					ix1=scaledConcentration(c, howWide);
					g.drawLine(ix1,iy1,ix1,iy2);
					}
				}
			}
			
	// draw tics at each of the tube standard concentrations.
		for(i=0;i<26;i++) clilckTx[i]=-777;
		iy2 = grafMax+20;
		if(dyeConcentration != null) {
		  iy1=grafMax-15;
		  int hilolim=2;
		  if(parent.ifMonochrome) hilolim=1;
		  for(int hilo=0;hilo<hilolim;hilo++) {
			if(!parent.ifMonochrome && hilo==0) {
				if(ifRed) g.setColor(Color.red);
				else   g.setColor(Color.green);
				}
			if(hilo==1) g.setColor(Color.gray);
			iy1 += 25;
			iy2=iy1+10;
			if(hilo==0) {
				clickHiTy = iy2+5;
				if(!parent.ifMonochrome) g.drawString("click tube to reset high concentration",howWide+20,iy2+10);
				}
			else {
				clickLowTy = iy2+5;
				g.drawString("click tube to reset low concentration",howWide+20,iy2+10);
				}
			for(i=rightTube;i<=leftTube;i++) {
				ix1 = scaledConcentration(dyeConcentration[i], howWide);
				g.drawLine(ix1,iy1,ix1,iy2);
				if(i<10)
					g.drawString("T0"+i, ix1-7, iy2+10);
				else 
					g.drawString("T"+i, ix1-7, iy2+10);
				clilckTx[i]=ix1-7;
				}
			 }
			 
			if(!parent.ifMonochrome) {
				g.setColor(Color.white);
				ix1 = scaledConcentration(concentrationMappedTo255, howWide);
				g.drawLine(ix1,0,ix1,iy2);
				} 
			}
				
		if(!ifCallibrationStatsValid) return;
		// print the names of the various curve sets.
		
		iy2 += 10;
		ix1=20;
		i=iy2;
		for(cs=0;cs < curveSet;cs++) {
			iy2 += 12;

			if(!parent.ifMonochrome) 
				g.setColor(curveSetColor[cs]);		
			g.drawString(prefix[cs]+"---> "+curveSetDirName[cs],ix1,iy2);
			if(cs==4) {
				ix1=30+howWide/2;
				iy2=i; // back up
				}
			}
		// now draw each gain curve...
		for(cs=0;cs < curveSet;cs++)
		  for(slit=0;slit<16;slit++)
			for(gainIndex=0;gainIndex<41;gainIndex++)
				if(gncv[cs][slit][gainIndex]!=null && gncv[cs][slit][gainIndex].setupFirstConcentration()) {
					ix2=-1;	iy2=0; didDraw=false;
					while((aPoint=gncv[cs][slit][gainIndex].nextConcentration()) != null && 
								menuItemIsOn[gncv[cs][slit][gainIndex].clickJMenuId])  {
						if(ix2<0 && aPoint.getConcentration() >= minConc) { 
							// this is the first point of the gain curve
							ix2 = scaledConcentration(aPoint.getConcentration(), howWide);
							iy2 = aPoint.getReading();
							if(!parent.ifMonochrome) g.setColor(gncv[cs][slit][gainIndex].myColor);
							if(aPoint.getMax() > aPoint.getMin())
								g.drawLine(ix2,grafMax-2*aPoint.getMin(),ix2,grafMax-2*aPoint.getMax()); 
							}
						else if(aPoint.getConcentration() <= maxConc){
							ix1=ix2;		iy1=iy2;
							didDraw=true;
							ix2 = scaledConcentration(aPoint.getConcentration(), howWide);
							iy2 = aPoint.getReading();
							g.drawLine(ix1,grafMax-2*iy1,ix2,grafMax-2*iy2);
							if(aPoint.getMax() > aPoint.getMin())
								g.drawLine(ix2,grafMax-2*aPoint.getMin(),ix2,grafMax-2*aPoint.getMax()); 
							}
						}
					if(didDraw && gncv[cs][slit][gainIndex]!=null &&
					  menuItemIsOn[gncv[cs][slit][gainIndex].clickJMenuId])
						g.drawString(gncv[cs][slit][gainIndex].myName,ix2,grafMax-2*iy2);
					}
		if(gainCurveFitCoeffs != null) { // there's a valid quadratic fit curve to exhibit
			double glo,conc;
				ix2=-1;
				if(!parent.ifMonochrome) g.setColor(Color.white);
				for(glo=0.;glo<256.;glo += 5.) {
					 conc = quadFitConc((double)glo);
				     if(ix2<0 && conc >= minConc) { 
							// this is the first point of the gain curve
							ix2 = scaledConcentration(conc, howWide);
							iy2 = (int)glo;
							}
						else if(conc <= maxConc){
							ix1=ix2;		iy1=iy2;
							didDraw=true;
							ix2 = scaledConcentration(conc, howWide);
							iy2 = (int)glo;
							g.drawLine(ix1,grafMax-2*iy1,ix2,grafMax-2*iy2);
							}
						}	
				}					
		}
		
	private int scaledConcentration(double conc, int howWide) {
		/* when conc == minConc, return 0
		when conc == maxConc, return howWide; */
		if(ifLogScale) 
			return ( (int)(0.499 + logScale*Math.log(conc/minConc)) );
			
		else 
			return( (int)(0.499 + linScale*(conc - minConc)) );			
		}
			
	private String getDir(String prompt) {
		FileDialog getAFile = 
		   new FileDialog(this,prompt, FileDialog.LOAD);
		getAFile.show();
		return getAFile.getDirectory();
		}
		
	private RandomAccessFile openReadWriteFile(String filename) {
		RandomAccessFile inoutfile;
		if(filename == null) {
			talk("bad file name");
			return null;
			}
		try{
			inoutfile = new RandomAccessFile(filename,"rw");
			}
		catch (FileNotFoundException e) {
			talk("openReadWriteFile("+filename+") failed -- file not found");
			return null;
			}
		catch (IOException e) {
			talk("openReadWriteFile("+filename+") failed IOexception" + e);
			return null;
			}
		try{inoutfile.seek((long)0);}
		catch(IOException e) {};
		return inoutfile;
		}

	private RandomAccessFile openReadFile(String filename) {
		RandomAccessFile inoutfile;
		if(filename == null) {
			talk("bad file name");
			return null;
			}
		try{
			inoutfile = new RandomAccessFile(filename,"r");
			}
		catch (FileNotFoundException e) {
			talk("openReadWriteFile("+filename+") failed -- file not found");
			return null;
			}
		catch (IOException e) {
			talk("openReadWriteFile("+filename+") failed IOexception" + e);
			return null;
			}
		try{inoutfile.seek((long)0);}
		catch(IOException e) {};
		return inoutfile;
		}

	private byte [] readTheFile(RandomAccessFile inoutfile) {
		int ngot;
		int  nbytes=0;
		
		try{nbytes = (int)inoutfile.length();}
		catch(IOException e) { 
			talk("Got bad length() IOException: "+e);
			return  null;
			}  
	
		if(nbytes==0) return null;
		byte [] theBytes = new byte[nbytes];

		try {ngot=inoutfile.read(theBytes);} // get the 16 byte header
		catch(IOException e) { 
			talk("Got RandomAccessFile IOException: "+e);
			return  null;
			} 
		try{inoutfile.seek((long)0);} // return file pointer to original start location 
		catch(IOException e) {
			talk("bad seek to beginning of e in readTheFile()");
			};
		
		return(theBytes);
		}
				
	private void writeAString(String stg,RandomAccessFile inout) {
		byte [] buf = new byte[stg.length()+1];
		stg.getBytes(0,stg.length(),buf,0);
		buf[stg.length()] = 10;
			try {
			inout.write(buf);
			}
		catch (IOException ioex) {
			talk("write to file failed: "+ioex);
			}
		}
		
	private String readOneLine(RandomAccessFile inout) {
		byte [] buf = new byte[257];
		byte [] inbyte = {0};
		int i=0,igot=0;
		while(i<256) {
			try {igot=inout.read(inbyte);}
			catch(IOException iex) {
				talk("IOException was: "+iex);
				return null;
				}
			if(igot == 1 && inbyte[0] != 10 && inbyte[0] != '\n' && inbyte[0]  != 13)
				buf[i++]=inbyte[0];
			else {
				if(i>0) {
					//talk("readOneLine got a line with "+i+" bytes");
					return (new String(buf,(byte)0,0,i));
					}
				else return null;
				}
			}
		talk("readOneLine hit line with more than 256 characters, namely this line:");
		talk(new String(buf,(byte)0,0,i));
		return ("overflow line");
		}
		
	private void closeOutputFile(RandomAccessFile inoutf) {
		try{inoutf.close();} catch(IOException e) {
			talk("closeOutputFile("+inoutf+") failed ???");};
		}		

	private void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
}

class OnePoint {
	int intens=-5,intensMax=-6,intensMin=256,nmeasurements=0,myTube=-7;
	String fileNames=null;
	GainCurve myCurve=null;
	double myConc;
	
	public OnePoint(int tube, GainCurve parentCurve) {
		myTube=tube;
		myCurve = parentCurve;
		}
		
	public void addMeasurement(int bright, double conc, String fileName) {
		if(intens<0) { // this is the first measurement made for this gain/slit/tube concentration.
			intens = bright;
			intensMax = bright;
			intensMin = bright;
			myConc=conc;
			nmeasurements=1;
			this.fileNames=fileName;
			}
		else {
			intens += bright;
			if(bright > intensMax)
				intensMax = bright;
			if(bright < intensMin)
				intensMin = bright;
			++nmeasurements;
			fileNames= fileNames+"\n"+fileName;
			}
		}
		
	public String contents() {
		if(!isValid()) return ("No valid measurement");
		return ("gain="+myCurve.myGain+"  slit="+myCurve.mySlit+" tube="+myTube+"  conc="+myConc+
				" Imin="+getMin()+"  I="+ getReading()+" Imax="+getMax());
		}
	
	public boolean isValid() {
		if(intens<0) return false;
		else return true;
		}
	
	public double getConcentration() {
		return myConc;
		}
		
	public int getReading() {
		return ((int)(.499 + (double)intens/((double)nmeasurements)));
		}
	public int getMax() {
		return intensMax;
		}
	public int getMin() {
		return intensMin;
		}
	}

class GainCurve {
	OnePoint [] point = new	OnePoint[26];
	int dimmestTube = -1;
	int brightestTube = 7777;
	int myGain=-2;
	int mySlit = -3;
	int myCurveSet=0;
	int tubeCounter=-4;
	String myName="?";
	boolean ifRed=false;
	Color myColor,myBrightColor;
	int myRed=255, myGreen=255, myBlue=255;
	boolean ifValid=false;
	int clickJMenuId=0;
	static final int [] reds =   {255,255,0  ,0  ,255,60 ,255,255,128,255};
	static final int [] greens = {0  ,255,255,255,0  ,80 ,255,128,255,128};
	static final int [] blues =  {0  ,0  ,0  ,255,255,255,255,0  ,128,128};
	
	public GainCurve(int slit, int gainIndex, int curveSet, String prefix, boolean ifRed) {
		talk("GainCurveConstructor("+slit+", "+gainIndex+", "+curveSet+", "+prefix+", "+ifRed+" )");
		int tube;
		myGain=gainIndex+60;
		myCurveSet = curveSet;
		mySlit=slit;
		//              red yel gre cyn mag blu whi org lime cornflowerBlue
		double gainMult = .01*(double)(gainIndex+60);
		myRed = (int)(gainMult*reds[myCurveSet]);
		myGreen = (int)(gainMult*greens[myCurveSet]);
		myBlue = (int)(gainMult*blues[myCurveSet]);
		myColor = new Color(myRed,myGreen,myBlue);
		myBrightColor = new Color(reds[myCurveSet],greens[myCurveSet],blues[myCurveSet]);
		for(tube=0;tube<26;tube++)
			point[tube] = new OnePoint(tube, this);
		this.ifRed=ifRed;	
		myName = prefix+"G"+(60+gainIndex)+"S"+slit;
		talk("GainCurveConstructor, myname = "+myName);
		}
	
	public Color myBrightestColor() {
		return myBrightColor;
		}
			
	private void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}

	public void addMeasurement(int tube, double conc, int intens, String fileName) {
		if(intens<=0) talk("GainCurve.addMeasurement got intens="+intens+" for tube "+tube+" and conc="+conc);
		point[tube].addMeasurement(intens, conc, fileName);
		if(tube>dimmestTube)
			dimmestTube=tube;
		if(tube<brightestTube)
			brightestTube=tube;
		ifValid=true;
		}
	
	public boolean isValid() {
		return ifValid;
		}
	
	public boolean setupFirstConcentration() {
		if(!isValid()) return false;
		tubeCounter = dimmestTube;
		return true;
		}
	
	public OnePoint nextConcentration() {
		OnePoint retn;
		if(!isValid() || tubeCounter<=0)
			 return null;
		if(point[tubeCounter].isValid())
			retn = point[tubeCounter]; 
		else
			retn =  null;
		--tubeCounter;
		while(tubeCounter >= 0 && !(point[tubeCounter].isValid()) )
			--tubeCounter;
		return retn;
		}
	
	}
		
