package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.lang.*;
import java.util.*;
import java.awt.*;  
import javax.swing.*;

class DeconvolverPSF {
	double [][] radProfile; // this is the data from which the delx, dely, delsect are built
	int nradpts=10,numProfiles=2,beadSect=0;
	double dxyorig, dzorig; // the voxel size of the stack from which the PSF was measured

	int numXYZpoints=0;
	
	int nSectsUp=0,nSectsDown=0;
	int [][] delxZup,delyZup; // the first index corresponds to the z-section displacement
								// at or above the point  
	double [][] glowZup; // the second index gives the glow-point number
	int [] numPtsInUpSects;

	int [][] delxZdown,delyZdown; // the first index corresponds to the z-section displacement
								// at or below the point  
	double [][] glowZdown; // the second index gives the glow-point number
	int [] numPtsInDownSects;
	
	double dxy,dz; // the voxel size of the stack to be deconvolved
	
	int minx=0,maxx=0,miny=0,maxy=0;
	int sectionPicDimension=1;
	
	private boolean ifConesUpsideDown =false;

	public boolean ifUsingConesUpsideDown() {
		return ifConesUpsideDown;
		}
	
	public PSFdata copyData() {
		int k;
		PSFdata thedata = new PSFdata();
		thedata.nradpts = nradpts;
		thedata.numProfiles = numProfiles;
		thedata.beadSect = beadSect;
		thedata.dxyorig = dxyorig;
		thedata.dzorig = dzorig;
		thedata.numXYZpoints = numXYZpoints;
		thedata.nSectsUp = nSectsUp;
		thedata.nSectsDown = nSectsDown;
		thedata.dxy = dxy;
		thedata.dz = dz;
		thedata.minx = minx;
		thedata.maxx = maxx;
		thedata.miny = miny;
		thedata.maxy = maxy;
		thedata.sectionPicDimension = sectionPicDimension;
		
		thedata.delxZup = new int[nSectsUp][];
		thedata.delyZup = new int[nSectsUp][];
		thedata.glowZup = new double[nSectsUp][];
		thedata.numPtsInUpSects = cloneArray(numPtsInUpSects);
		for(k=0;k<nSectsUp;k++) {
			thedata.delxZup[k] =  cloneArray(delxZup[k]);
			thedata.delyZup[k] =  cloneArray(delyZup[k]);
			thedata.glowZup[k] =  cloneArray(glowZup[k]);
			}

		thedata.delxZdown = new int[nSectsDown][];
		thedata.delyZdown = new int[nSectsDown][];
		thedata.glowZdown = new double[nSectsDown][];
		thedata.numPtsInDownSects = cloneArray(numPtsInDownSects);
		for(k=0;k<nSectsDown;k++) {
			thedata.delxZdown[k] =  cloneArray(delxZdown[k]);
			thedata.delyZdown[k] =  cloneArray(delyZdown[k]);
			thedata.glowZdown[k] =  cloneArray(glowZdown[k]);
			}
		
		return thedata;
		}	
	
	private int [] cloneArray(int [] aray) {
		int len = aray.length;
		int [] retn = new int[len];
		for(int i=0;i<len;i++)
			retn[i] = aray[i];
		return aray;
		}
		
	private double [] cloneArray(double [] aray) {
		int len = aray.length;
		double [] retn = new double[len];
		for(int i=0;i<len;i++)
			retn[i] = aray[i];
		return aray;
		}
		
	public synchronized boolean flipPSFCones() { // calling this exchanges the upSects & downSects
	// the first time you call it, it sets ifConesUpsideDown==true, the second time it sets ifConesUpsideDown==false,
	// etc.
	 
		int tmpint = nSectsUp;  nSectsUp = nSectsDown;  nSectsDown = tmpint;

		int [][] tempintaa = delxZup;   delxZup=delxZdown;  delxZdown = tempintaa;
		tempintaa = delyZup;   delyZup=delyZdown;  delyZdown = tempintaa;
		
		double [][] tempdbl = glowZup;  glowZup = glowZdown;  glowZdown=tempdbl;
		
		int [] tempinta = numPtsInUpSects;  numPtsInUpSects=numPtsInDownSects;  numPtsInDownSects=tempinta;
		ifConesUpsideDown = !ifConesUpsideDown;
		talk("ifConesUpsideDown = "+ifConesUpsideDown+", nsectsUp = "+nSectsUp+", nsectsDown = "+nSectsDown);
		tmpint=0;
		for(int i=0;i<nSectsUp;i++)
			tmpint += numPtsInUpSects[i];
		talk("there are "+tmpint+" glowing points above the bead");
		tmpint=0;
		for(int i=0;i<nSectsDown;i++)
			tmpint += numPtsInDownSects[i];
		talk("and there are "+tmpint+" glowing points below the bead");
		return ifConesUpsideDown;
		}
				
	public byte[] getBlursRaster(int ncols,int nrows) {
		int stampWidth = ncols/10;
		int stampHeight = nrows/10;
		int npix = ncols*nrows, row,col,xorg,yorg,i,j,x,y;
		byte [] blurras = new byte[npix];
		byte [] onestamp;
		int iz,stamprow,stampcol;
		for(row=0; row< nrows; row++) // blank the array
			for(col=0; col < ncols; col++)
				blurras[row*ncols + col]=0;
				
		iz = nSectsUp;
		stamprow=0;
		stampcol=0;
		while(iz>0) {
			--iz;
			onestamp = stampAtSomeZ( iz, stampWidth, stampHeight);
			xorg = stampcol*stampWidth;
			yorg = stamprow*stampHeight;
			for(i=0;i<stampHeight;i++)		// copy the stamp to where it goes	
				for(j=0;j<stampWidth;j++) {
					x = xorg+j;
					y = yorg+i;
					blurras[y*ncols + x] = onestamp[i*stampWidth+j];
					}
			if((++stamprow)>9) {
				stamprow=0;
				++stampcol;
				}
			}			

		iz = nSectsDown;
		stamprow=9;
		stampcol=9;
		while(iz>0) {
			--iz;
			onestamp = stampAtSomeZ( iz, stampWidth, stampHeight);
			xorg = stampcol*stampWidth;
			yorg = stamprow*stampHeight;
			for(i=0;i<stampHeight;i++)		// copy the stamp to where it goes	
				for(j=0;j<stampWidth;j++) {
					x = xorg+j;
					y = yorg+i;
					blurras[y*ncols + x] = onestamp[i*stampWidth+j];
					}
			if((--stamprow)<0) {
				stamprow=9;
				--stampcol;
				}
			}			
		// count the number of points that are 'on'
		x = 0;
		for(i=0;i<npix;i++)
		 	if(blurras[i] !=0)
		 		++x;
		talk("This DPSF has "+x+" glowing points in it.");	
		return blurras;
		}
					
	public DeconvolverPSF(double [][] radProf,int nradpts, int numProfiles, int beadSect,
				double dxyorig, double dzorig, double dxy, double dz) {
		this.nradpts = nradpts;
		this.numProfiles = numProfiles;
		this.beadSect = beadSect;
		this.dzorig = dzorig;
		this.dxyorig=dxyorig;
		radProfile = new double[numProfiles][nradpts];
		for(int i=0;i<numProfiles;i++) // make a local copy
			for(int j=0;j<nradpts;j++)
				radProfile[i][j] = radProf[i][j];
		fillRasters(dxy,dz);
		}
		
	public byte[] getGridRaster(int ncols,int nrows) {
		int stampWidth = ncols/10;
		int stampHeight = nrows/10;
		byte gridshade = (byte)(80);
		int npix = ncols*nrows, row,col;
		byte [] gridras = new byte[npix];
		for(row=0; row< nrows; row++) // blank the array
			for(col=0; col < ncols; col++)
				gridras[row*ncols + col]=0;
		for(row=0;row<nrows;row += stampHeight) // make the horizontal grid lines
			for(col=0;col<ncols;col++)
				gridras[row*ncols + col] = gridshade;				

		for(col=0;col<ncols;col += stampWidth)
			for(row=0;row<nrows;row++)
				gridras[row*ncols + col] = gridshade;
		return gridras;
		}
					
	public byte [] stampAtSomeZ(int iz, int nc, int nr) {
		/* returns an array whose length is the square of sectionPicDimension, so that's how
		you find out its dimensions. This array stores the brightness values in section 
		number iz. When iz>=0, this comes from the up sections, and when iz<0 his data 
		comes from the down sections */
		int i,row,col,npts,npix = nc*nr, x0=nc/2, y0=nr/2;
		byte [] retn = new byte[npix];
		int [] delx,dely;
		double [] glow;
		
		for(row=0; row< nr; row++) // blank the stamp
			for(col=0; col < nc; col++)
				retn[row*nc + col]=0;
		// paint in the pixels;
		if(iz>=0) {
			glow = glowZup[iz];
			npts = numPtsInUpSects[iz]; 
			delx = delxZup[iz];
			dely = delyZup[iz];
			}
		else {
			iz = -iz;
			glow = glowZdown[iz];
			npts = numPtsInDownSects[iz]; 
			delx = delxZdown[iz];
			dely = delyZdown[iz];
			}
		
		for(i=0;i<npts;i++) {
			row = dely[i]+y0;
			col = delx[i]+x0;
			if(row>=0 && row <nr && col>=0 && col <nc)
				retn[row*nc + col] = (byte)(glow[i] +.49);
			}
		return retn;
		}
			
	public int fillRasters(double dxy, double dz) { // can call this for different dxy and dz
		int npt,ntot;
		this.dz = dz;
		this.dxy = dxy;
		int maxirad=0;
		int ir,iz,ix,iy,i;
		
		int ntempts=100*100;
		minx=0;  maxx=0;  miny=0;  maxy=0;
		numXYZpoints=0;
		nSectsUp= 1+(int)((numProfiles-beadSect)*dzorig/dz);
		delxZup = new int[nSectsUp][];
		delyZup = new int[nSectsUp][];
		glowZup = new double[nSectsUp][];
		numPtsInUpSects = new int[nSectsUp];

		nSectsDown=1+(int)(beadSect*dzorig/dz);
		delxZdown = new int[nSectsDown][];
		delyZdown = new int[nSectsDown][];
		glowZdown = new double[nSectsUp][];
		numPtsInDownSects = new int[nSectsDown];
		
		talk("Reload rasters in PointSpreadFunction object. nSectsUp="+nSectsUp);
		talk("nSectsDown = "+nSectsDown+", voxel width, dxy = "+dxy+" um, voxel depth, dz = "+dz+" um");
		
		for(i=0; i<numProfiles; i++) // find the farthest xy displacement of the glowing pts
			for(ir=0; ir<nradpts; ir++)
				if(radProfile[i][ir]>0. && maxirad<ir)
					maxirad = ir;
		maxirad = (int)((maxirad*dxyorig)/dxy);
		talk("    this means the glow spreads "+maxirad+" voxels from the source for this dxy = "+dxy+" stack");
		
		
		ntempts = 10+4*(1+maxirad)*(1+maxirad);
		int [] tempdelx = new int[ntempts]; 
		int [] tempdely = new int[ntempts]; 
		int [] tempdelsect = new int[ntempts]; 
		double [] tempGlow = new double[ntempts]; 
		double glow;
		
		// load the up section data
		ntot=0;
		for(iz = 0; iz<nSectsUp; iz++) {
			npt=0;
			for(ix= (-maxirad); ix <= maxirad; ix++)
				for(iy= (-maxirad); iy <= maxirad; iy++) {
					if((glow=theGlowAt(ix,iy,iz)) > 0.) {//  make another point
						tempdelx[npt]=ix;
						tempdely[npt]=iy;
						tempGlow[npt]=glow;
						if(ix<minx) minx=ix;
						if(ix>maxx) maxx=ix;
						if(iy<miny) miny=iy;
						if(iy>maxy) maxy=iy;
						++npt;
						}}
			delxZup[iz] = new int[npt];
			delyZup[iz] = new int[npt];
			glowZup[iz] = new double[npt];
			numPtsInUpSects[iz]=npt;
			numXYZpoints += npt;
			talk("There are "+npt+" points in up section "+iz);
			ntot += npt;
			for(i=0;i<npt;i++) {
				delxZup[iz][i] = tempdelx[i];
				delyZup[iz][i] = tempdely[i];
				glowZup[iz][i] = tempGlow[i];
				}
			}
talk("There are "+ntot+" glowing points in the UP sections. Now load the down sections...");
		// load the down section data
		ntot=0;
		for(iz = 0; iz<nSectsDown; iz++) {
			npt=0;
			for(ix= (-maxirad); ix <= maxirad; ix++)
				for(iy= (-maxirad); iy <= maxirad; iy++)
					if((glow=theGlowAt(ix,iy,(-iz))) > 0.) {//  make another point
						tempdelx[npt]=ix;
						tempdely[npt]=iy;
						tempGlow[npt]=glow;
						if(ix<minx) minx=ix;
						if(ix>maxx) maxx=ix;
						if(iy<miny) miny=iy;
						if(iy>maxy) maxy=iy;
						++npt;
						}
		
			// make the real array.
			delxZdown[iz] = new int[npt];
			delyZdown[iz] = new int[npt];
			glowZdown[iz] = new double[npt];
			numPtsInDownSects[iz]=npt;
			numXYZpoints += npt;
			talk("There are "+npt+" points in down section "+iz);
			ntot += npt;
			for(i=0;i<npt;i++) {
				delxZdown[iz][i] = tempdelx[i];
				delyZdown[iz][i] = tempdely[i];
				glowZdown[iz][i] = tempGlow[i];
				}
			}
		if((sectionPicDimension = 1+ maxx-minx) < (1 +maxy-miny))
			sectionPicDimension = (1 +maxy-miny);
		talk("There are "+ntot+" glowing points in the DOWN sections. Now load the down sections...");
		talk("There are "+numXYZpoints+" points that glow in the DeconvolverPSF");
		talk("A sectionPic's height and width is "+sectionPicDimension); 
		return numXYZpoints;
		} 


	double theGlowAt(int ix, int iy,int iz) {
		double xinc = ix*dxy,  yinc = iy*dxy;
		double xyrad = Math.sqrt(xinc*xinc + yinc*yinc);
		double zinc = iz*dz;
		double irwt,irpwt,izswt,izspwt,izradavg,izpradavg;
		int izs,izsp,ir,irp,i;
		
		// find the two bounding sections in the original data
		if(zinc>=0) { // one of the zup sections
			izs = (int)(beadSect+zinc/dzorig);
			izspwt = (zinc/dzorig) - (int)(zinc/dzorig); // cannot be larger than 1.0
			izswt =  1. - izspwt;
			}
		else  {// one of the down sections
			izs = (int)(beadSect+zinc/dzorig)-1;
			izspwt =  (int)(zinc/dzorig) - (zinc/dzorig); // cannot be larger than 1.0
			izswt =  1. - izspwt;
			}
		izsp = izs+1;	

		ir  = (int)(xyrad/dxyorig);
		irp = ir+1;
		irpwt = xyrad/dxyorig - ir;
		irwt = 1.-irpwt;
		
		if(irp>=nradpts || izs <0 || izsp >= numProfiles ||irp >=  nradpts || ir<0) 
			return -1.;
		 // we have data to return a glow by interpolation
		 izradavg = irwt*radProfile[izs][ir] + irpwt*radProfile[izs][irp];
		 izpradavg = irwt*radProfile[izsp][ir] + irpwt*radProfile[izsp][irp];
		 return (izswt*izradavg + izspwt*izpradavg);
		 }

	static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
}