package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.lang.*;
import java.util.*;
import java.awt.*;  
import javax.swing.*;

class FlattenFltr {
	SegmenterNode nd=null;
	short [] work=null;
	int workrows=-1, workcols=-1;
	int [][] offs = null;
	boolean [][] ifInCirc = null;
	double [][] avgPins=null;
	int avgPinsCols=0,avgPinsRows=0;
	int npix=-1;
	int twoRadius;
	int Radius=-1;
	int bigMesh = 32;  
	
	public FlattenFltr(SegmenterNode nd) {
		this.nd=nd;
		}
	
	private void getAvgPins(double avgBright) {
		int x,y,i,j;
		double av;
		for(i=0;i<avgPinsRows;i++) {
			if((y = i*bigMesh) >= workrows)
				y=workrows-1;
			for(j=0;j<avgPinsCols;j++) {
				if((x = j*bigMesh) >= workcols)
					x=workcols-1;
				av = averageInCircle(x, y);
				if(av == 0.0)
					avgPins[i][j] = 1.0;
				else 
					avgPins[i][j] = avgBright/av;
				}
			}
		// now low pass filter this array twice
		double [][] newPin= new double[avgPinsRows][avgPinsCols];
		double sum;
		int k,l,pass,insum;
		for(pass=0;pass<2;pass++) {
		  for(i=0;i<avgPinsRows;i++)
			for(j=0;j<avgPinsCols;j++) {
				sum=0.;  insum=0;
				for(k=-1;k<2;k++) 
					for(l=-1;l<2;l++) 
						if(i+k>=0 && i+k<avgPinsRows && j+l>=0 && j+l < avgPinsCols) {
							sum += avgPins[i+k][j+l];
							++insum;
							}
				newPin[i][j] = sum/insum;
				}
		  for(i=0;i<avgPinsRows;i++)
			for(j=0;j<avgPinsCols;j++) 
				avgPins[i][j] = newPin[i][j];
		  }
		}
	
	private double getSplinedBoost(int col,int row) { // extrapolate the boost factor from nearby elements in avgPins
		int i = row/bigMesh;
		int j = col/bigMesh;
		int xleft = col - bigMesh*j; 
		int xright = bigMesh - xleft;
		int ytop = row - bigMesh*i; 
		int ybot = bigMesh - ytop;
		double leftAvg,rightAvg;
		if(i+1 >= avgPinsRows)
			leftAvg = avgPins[i][j];
		else
			leftAvg = (avgPins[i][j]*ybot + avgPins[i+1][j]*ytop)/bigMesh;
		
		if(j+1 >= avgPinsCols) 
			return leftAvg;
			
		if(i+1 >= avgPinsRows)
			rightAvg = avgPins[i][j+1];
		else	
			rightAvg = (avgPins[i][j+1]*ybot + avgPins[i+1][j+1]*ytop)/bigMesh;
			
		return ((leftAvg*xright + rightAvg*xleft)/bigMesh);
		}
			
	private void makeArrays(int incols, int inrows, int radius) {	
		int i,j;
		if(npix < 0 || radius != Radius || work==null || incols != workcols || inrows != workrows || ifInCirc == null) {
			workrows=inrows;
			workcols=incols;
			npix = incols*inrows;
			work = new short[npix];
			Radius = radius;
			twoRadius = 2*radius;
			bigMesh = (2*radius)/4; // set this larger to speed up calculation but coarsen the boost grid.
			avgPinsCols = 1+incols/bigMesh;
			avgPinsRows = 1+inrows/bigMesh;
			avgPins = new double[avgPinsRows][avgPinsCols];
			ifInCirc = new boolean[twoRadius+1][twoRadius+1];
			offs = new int[twoRadius+1][twoRadius+1];
			twoRadius = 2*radius;
			int radsq = radius*radius;
			for(i=0;i<=twoRadius;i++) 
				for(j=0;j<=twoRadius;j++) {
					if( (i-radius)*(i-radius) + (j-radius)*(j-radius) <= radsq) 
						ifInCirc[i][j]=true;
					else 
						ifInCirc[i][j]=false;
					offs[i][j] = (i-radius)*incols + j-radius;
					}
			}
		}
					
	public void tellWhatYouDo() {
		nd.talk("");
		nd.talk("The FlattenFltr filter has a single input raster and three parameters, 'tooDim', 'radius', and 'AvgBrt'.");
		nd.talk("It equilibrates the average brightness (to the value AvgBrt) across the whole raster. It sets");
		nd.talk("pixels dimmer than tooDim to zero. It computes the original brightness in a circular region surrounding");
		nd.talk("each pixel and multiplies the brightness of that pixel by the factor needed to bring the neighborhood");
		nd.talk("average brightness to AvgBrt.");
		nd.talk("Use a value of 'radius' that spans several features in the micrograph. Use this filter only for");
		nd.talk("grayscale images.");
		}
		
	private double averageInCircle(int x, int y) {
		// returns the average brightness in a circle surrounding row y and column x in the work[] array
		int pix,yir, xjr, i,j,offxy = y*workcols + x;
		int sum = 0, pixin = 0;
		for(i=0;i <= twoRadius; i++)
			if((yir = y + i - Radius) >= 0 && yir < workrows) 
				for(j=0; j <= twoRadius; j++)
					if((xjr = x + j -Radius) >= 0 && xjr < workcols && ifInCirc[i][j] && (pix=work[offxy+offs[i][j]]) > 0) {
						++pixin;
						sum += pix;
						}
		if(pixin>0) return ((double )sum/(double )pixin);
		else return 0.0;
		}
		
  public boolean doit(short [] inras, int incols, int inrows, int thresh, int radius, int avgBrt, short [] outras) {
	int i,j,row,col,offset;
	short pix;
	double boostFac;
	makeArrays(incols,inrows,radius); 

	for(i=0;i<npix;i++) // ignore pixels dimmer than thresh
  		work[i] = inras[i] >= thresh? inras[i] : 0;
  	
  	getAvgPins((double )avgBrt);
  	
  	for(row=0;row<inrows;row++)
  		for(col=0;col<incols;col++) {
  			offset = row*incols+col;
  			pix =(short )(.49 + getSplinedBoost(col,row)*work[offset]);
  			// pix = (short )(48*getSplinedBoost(col,row)); /* to show the boost field  */
			outras[offset] = pix > 255? 255 : pix;
			}
			
	return(true);
	}

}