package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.lang.*;
import java.util.*;
import java.awt.*;  import javax.swing.*;

class Histogram extends JFrame implements java.awt.event.ActionListener,java.awt.event.WindowListener,java.awt.event.ItemListener {
	int width=256,height=256;
	HistogramCanvas canvas;
	Stats statsJPanel = null;
	OneHistogram redHist=null, greenHist=null, blueHist=null;  // the three histograms we display.	
	OneProfile redProf=null, greenProf=null, blueProf=null;  // the three profiles we display.	
	JTextField horizLoc=null,vertLoc=null;
	JPanel locpan = null;
	int maxVertVal=256;
	CallibrationStats CallibStats = null;  // used for the special histogramType==7 case 
	
	HistogramBox myBox=null;
	boolean ifShowRed=false, ifShowGreen=false, ifShowBlue=false;
	int ifJustOneSection=-1;
	boolean ifMonochrome=false;
	int colorChoice=1;
	Font myFont=null;
	JComboBox whatYouCanDo=null;
	int histogramType=1; // 1 is histogram in boxed region of this section
						// 2 means stacked total accumulation histograms in 3-D volume
						// 3 means conventional histogram for all voxels in a 3-D volume
						// 4 means profile along a linear transect in this single section
						// 6 means z-core profile (Rdim,Gdim,Bdim as abs dark thresholds)
						// 8 z-core profile (Rdim,Gdim,Bdim as % of darkest pixels to discard)
						// 7 (special) do full stats on a directory full of callibration standards pics;
						//          type 7 done in Histogram because various methods here are useful for
						//          that task.	

	int zmin,zmax;
	short [][]drawramps=null; 
	float [] coreAvg=null;
	double plotscale=1;
	int scale= 255;
	int nPtsInBox=1;
	int  Rthresh=0, Gthresh=0, Bthresh=0;
	boolean ifStackDone=false;
	int howTall=256;
	int howWide=256;
	double grandmax=0;
	double pixelsPerMicrometer = 22.;
	static final int pixPerCoreTic=3;

	long redPixAtPeak=1;
	double [] redNormHist; double [] redTotHist; double redTotal;
	int redMean=1,redPeak=1,redStdDev=1,redNbright=1,redNblazing=1;
	
	long greenPixAtPeak=1;
	double [] greenNormHist; double [] greenTotHist; double greenTotal;
	int greenMean=1,greenPeak=1,greenStdDev=1,greenNbright=1,greenNblazing=1;
	
	long bluePixAtPeak=1;
	double [] blueNormHist; double [] blueTotHist; double blueTotal;
	int blueMean=1,bluePeak=1,blueStdDev=1,blueNbright=1,blueNblazing=1;

	boolean ifDiskReadWorked = false;
	/******************data to restore to statsJPanel*************/
	boolean ST_ifBoxIsVisible=true,ST_ifInnerBoxValid=true,ST_ifTransectLineVisible=true;
	double ST_enlargeRatio=1.,ST_shrinkRatio=1.;
	int ST_curJFrame = 1,ST_width=3,ST_blur=30;
	/******************data to restore to statsJPanel*************/
	
	double howBright=10;
	Color redd=new Color(255,50,50);
	Color greenn = new Color(100,255,100);
	Color bluee = new Color(100,140,255);
	String redFilename="nameless", greenFilename="nameless", blueFilename="nameless",myDiskName=null; 

	public void tellLocation(int xx, int yy) {
		horizLoc.setText("x="+xx+"/"+howWide);
		yy = (int)(.49 + (double)((howTall-yy)*scale)/(double) howTall);
		vertLoc.setText("val="+yy+"/"+scale);
		}
			
	public void finalize() {
	talk("                                                                 Histogram snuffs"); 
	}
	
// catch WindowListener events  
	public void windowActivated(java.awt.event.WindowEvent event) {
		// talk("window activated");
		}
	public void windowDeactivated(java.awt.event.WindowEvent event) {
		//  talk("window Deactivated");
		}
	public void windowOpened(java.awt.event.WindowEvent event) {}
	public void windowIconified(java.awt.event.WindowEvent event) {}
	public void windowDeiconified(java.awt.event.WindowEvent event) {}
	public void windowClosing(java.awt.event.WindowEvent event) {
	    setVisible(false);
		statsJPanel=null;
		if(myBox  != null)   myBox.sepuku(); 			myBox=null;  
		if(CallibStats!=null) CallibStats.sepuku();  	CallibStats = null;  
		if(canvas  != null)   canvas.sepuku();   		canvas=null; 
		if(redHist  != null)   redHist.sepuku();     	redHist=null; 
		if(greenHist != null)   greenHist.sepuku();   greenHist=null; 
		if(blueHist  != null)   blueHist.sepuku();    blueHist=null;
		if(redProf  != null)   redProf.sepuku();     	redProf=null; 
		if(greenProf  != null)   greenProf.sepuku();  greenProf=null; 
		if(blueProf != null)   blueProf.sepuku();    	blueProf=null; 
		drawramps=null; 	coreAvg=null;  redd=null; greenn=null; bluee=null;
		redNormHist=null; redTotHist=null; 
		greenNormHist=null; greenTotHist=null; 
		blueNormHist=null; blueTotHist=null; 
		talk("Histogram panel frees memory assets");
		dispose();
		}
	public void windowClosed(java.awt.event.WindowEvent event) {
		}
	 

 	public void writeToDisk(ArchiveMarks am) {
    	am.writeNum(width,"    width");
   		am.writeNum(height,"    height");
     	am.writeNum(ifShowRed,"    ifShowRed");
   		am.writeNum(ifShowGreen,"    ifShowGreen");
     	am.writeNum(ifShowBlue,"    ifShowBlue");
    	am.writeNum(ifJustOneSection,"    ifJustOneSection");
     	am.writeNum(colorChoice,"    colorChoice");
    	am.writeNum(zmin,"    zmin");
     	am.writeNum(zmax,"    zmax");
    	am.writeNum(scale,"    scale");
     	am.writeNum(Rthresh,"    Rthresh");
     	am.writeNum(Bthresh,"    Bthresh");
    	am.writeNum(Gthresh,"    Gthresh");
     	am.writeNum(histogramType,"    histogramType");
     	am.writeNum(pixelsPerMicrometer,"    pixelsPerMicrometer");
     	am.writeNum(ST_ifBoxIsVisible,"    ST_ifBoxIsVisible");
     	am.writeNum(ST_ifInnerBoxValid,"    ST_ifInnerBoxValid");
     	am.writeNum(ST_ifTransectLineVisible,"    ST_ifTransectLineVisible");
     	am.writeNum(ST_enlargeRatio,"    ST_enlargeRatio");
     	am.writeNum(ST_shrinkRatio,"    ST_shrinkRatio");
     	am.writeNum(ST_curJFrame,"    ST_curJFrame");
     	am.writeNum(ST_curJFrame,"    ST_width");
     	am.writeNum(ST_blur,"    ST_blur");
     	myBox.writeToDisk(am);
     	am.writeAString(redFilename);
     	am.writeAString(greenFilename);
     	am.writeAString(blueFilename);
    	myDiskName = noPath(am.filename);
    	repaint();
 		}
 
 	public Histogram(Stats stp) {
		statsJPanel=stp;
		myBox=new HistogramBox(stp.picJPanel,stp);
		stp.myBox.transferDataTo(myBox);		
	    ArchiveMarks reader=new ArchiveMarks();
	    reader.readHistogram(this);
	    addWindowListener(this);
		}

 	public void readFromDisk(ArchiveMarks am) {
	  	double badval = -7777.777;
	  	int ibool;
		width = (int)(am.getNextFileDouble(badval));
		height = (int)(am.getNextFileDouble(badval));
 		ibool = (int)(am.getNextFileDouble(badval)); ifShowRed = ibool!=0? true : false;
 		ibool = (int)(am.getNextFileDouble(badval)); ifShowGreen = ibool!=0? true : false;
 		ibool = (int)(am.getNextFileDouble(badval)); ifShowBlue = ibool!=0? true : false;
 		
		ifJustOneSection = (int)(am.getNextFileDouble(badval));
		colorChoice = (int)(am.getNextFileDouble(badval));
		zmin = (int)(am.getNextFileDouble(badval));
 		zmax = (int)(am.getNextFileDouble(badval));
		scale = (int)(am.getNextFileDouble(badval));
		Rthresh = (int)(am.getNextFileDouble(badval));
		Bthresh = (int)(am.getNextFileDouble(badval));
		Gthresh = (int)(am.getNextFileDouble(badval));
		histogramType = (int)(am.getNextFileDouble(badval));
		pixelsPerMicrometer = am.getNextFileDouble(badval);
		
		
 		ibool = (int)(am.getNextFileDouble(badval)); ST_ifBoxIsVisible = ibool!=0? true : false;
 		ibool = (int)(am.getNextFileDouble(badval)); ST_ifInnerBoxValid = ibool!=0? true : false;
 		ibool = (int)(am.getNextFileDouble(badval)); ST_ifTransectLineVisible = ibool!=0? true : false;
		
		ST_enlargeRatio = am.getNextFileDouble(badval);
		ST_shrinkRatio = am.getNextFileDouble(badval);
		ST_curJFrame = (int)(am.getNextFileDouble(badval));
		ST_width = (int)(am.getNextFileDouble(badval));
		ST_blur = (int)(am.getNextFileDouble(badval));
     	myBox.readFromDisk(am);
    	redFilename = am.readOneLine();
    	greenFilename = am.readOneLine();
    	blueFilename = am.readOneLine();
    	myDiskName = noPath(am.filename);
 		}
		
	private String noPath(String pathname) {
		int lastSlashIndex = pathname.lastIndexOf('/');	
		return pathname.substring(lastSlashIndex+1);
		}
			
	public void actOnMouseClick(int x,int y)  {
		if(histogramType==7 && CallibStats != null)
			CallibStats.actOnMouseClick(x,y);
		}
		
	private void makeLocationTextFields() {
		WidgetMaker wmk = new WidgetMaker(this);
		horizLoc = wmk.makeAJTextField("x=0",14);
		vertLoc = wmk.makeAJTextField("val=0",14);
		locpan = new JPanel();
		locpan.setLayout(new GridLayout(1,3));
		locpan.add(horizLoc);
		locpan.add(vertLoc);
		}
			
	public Histogram(Stats stp,  double pixelsPerMicrometer,  String hstName) {
		super("Statistical analysis of callibration standard pics");		
		statsJPanel=stp;
		this.colorChoice=1; // use the red channel for monochrome analysis.
		this.zmin=0;
		this.zmax=0;
		WidgetMaker wmk = new WidgetMaker(this);
		myFont = wmk.alp;
		setFont(myFont);
		this.histogramType = 7;  // do full stats on a directory full of callibration standards pics
		// this constructor is a special one for setting up statistical analysis of a directory
		// tree full of callibration standards pics.
		this.scale=511;
		this.pixelsPerMicrometer = pixelsPerMicrometer;
		this.Rthresh=0;
		this.Gthresh=0;
		this.Bthresh=0;
		
		myBox=new HistogramBox(stp.picJPanel,stp);
		stp.myBox.transferDataTo(myBox);
		
 		ST_curJFrame = myBox.picJPanel.curStack.curJFrame;
 		this.redFilename = null;
 		this.greenFilename = null;
 		this.blueFilename = null;
 		if(hstName != null)
 			myDiskName = hstName;

		ifJustOneSection=-1;
		ifStackDone=false;
		whatYouCanDo = wmk.makeAJComboBox(
			new String [] {"histogram options",
			"print black on white",
			"print in colors",
			"overlay another set of gain curves",
			"transfer quadratic gain curve fit to red channel LUT",
			"transfer quadratic gain curve fit to green channel LUT",
			"transfer quadratic gain curve fit to blue channel LUT",
			"purge empty directories from gain curve directory"});	
		whatYouCanDo.setLightWeightPopupEnabled(false);	
		ifShowRed=true;
		ifShowGreen=false;
		ifShowBlue=false;
		nPtsInBox = myBox.myTransect.npixInBlur; 
		redProf = new OneProfile(myBox, Rthresh, redd, 2);
		CallibStats = new CallibrationStats(this,myBox,redProf);
		
		ifJustOneSection=(zmin+zmax)/2;
		howTall=510;
		howWide=1000;
		plotscale= (double )howTall/((double)scale);
		CallibStats.overlayAnotherCurveSet(); // compute the statistics, scanning a directory tree the user picks.
		setBackground(Color.black);
		getContentPane().setLayout(new BorderLayout());
		canvas = new HistogramCanvas(this);
		canvas.rescale(howTall,howWide);
		makeLocationTextFields();
		locpan.add(whatYouCanDo);
		getContentPane().add("North",locpan);
		getContentPane().add("Center",canvas);
		show();
		setSize(howWide+3,howTall+160);
		ifStackDone=true;
		canvas.forceRepaint=true;
		repaint();
		addWindowListener(this);
		whatYouCanDo.setVisible(true);
		}			
		
		
	public Histogram(int histogramType, int zmin, int zmax, int colorChoice,
				int Rthresh,int Gthresh,int Bthresh, int scale, 
						double pixelsPerMicrometer, Stats stp,
							String redfn, String greenfn, String bluefn, String hstName) {
		super("section stack("+zmin+" thru "+zmax+")");
		statsJPanel=stp;
		this.colorChoice=colorChoice;
		this.zmin=zmin;
		this.zmax=zmax;
		this.histogramType = histogramType;
		this.scale=scale;
		this.pixelsPerMicrometer = pixelsPerMicrometer;
		this.Rthresh=Rthresh;
		this.Gthresh=Gthresh;
		this.Bthresh=Bthresh;
		WidgetMaker wmk = new WidgetMaker(this);
		myFont = wmk.alp;
		setFont(myFont);
		myBox=new HistogramBox(stp.picJPanel,stp);
		stp.myBox.transferDataTo(myBox);		

 		ST_curJFrame = myBox.picJPanel.curStack.curJFrame;
 		if(redfn==null)  {
 			//talk("get new redfilename = "+myBox.picJPanel.redFilename);
 			this.redFilename = noPath(myBox.picJPanel.redFilename);
 			}
 		else this.redFilename = redfn;

 		if(greenfn==null) {
 			//talk("get new redfilename = "+myBox.picJPanel.greenFilename);
 			this.greenFilename = noPath(myBox.picJPanel.greenFilename);
 			}
 		else this.greenFilename = greenfn;

 		if(bluefn==null) {
 			//  talk("get new redfilename = "+myBox.picJPanel.blueFilename);
 			this.blueFilename = noPath(myBox.picJPanel.blueFilename);
 			}
 		else this.blueFilename = bluefn;

 		if(hstName != null)
 			myDiskName = hstName;

		ifJustOneSection=-1;
		ifStackDone=false;
	// set up the menu...
	// histogramType    // 1 is histogram in region of this section
						// 2 means stacked total accumulation histograms in 3-D volume
						// 3 means conventional histogram for all voxels in a 3-D volume
						// 4 means profile along a linear transect in this single section
						// 5 means profile along a linear transect in a 3-D stack.
						// 6 means z-core profile (Rdim,Gdim,Bdim as abs dark thresholds)
						// 7 means do full stats on a directory full of callibration standards pics.
						// 8 z-core profile (Rdim,Gdim,Bdim as % of darkest pixels to discard)
				 		// 9 bleach correction drop all pixels dimmer than RGBdims 
						// 10 bleach correction RGBdims are percentage pf darkest pixels to discard

		switch(histogramType) { // add special  options 
			case 1:
			case 3:
				whatYouCanDo = wmk.makeAJComboBox(new String [] {
					"histogram options",
					"write as stats file to disk",
					"become current stats object",
					"print self",
					"print black on white",
					"print in colors",
					"unfiltered histograms",
					"1  Low Pass Filter iteration",
					"2  Low Pass Filter iterations",
					"3  Low Pass Filter iterations",
					"4  Low Pass Filter iterations"});
				break;
			case 2:
			case 4:
			case 5:	
			default:		
				whatYouCanDo = wmk.makeAJComboBox(new String [] {
					"histogram options",
					"write as stats file to disk",
					"become current stats object",
					"print self",
					"print black on white",
					"print in colors"});
				break;
			case 6:			
			case 8:			
			case 9:			
			case 10:			
				whatYouCanDo = wmk.makeAJComboBox(new String [] {
					"histogram options",
					"write as stats file to disk",
					"become current stats object",
					"print self",
					"print black on white",
					"print in colors",
					"Write z-coreProfile data to disk file"});
				break;
			}
		whatYouCanDo.setLightWeightPopupEnabled(false);	
			
		switch (colorChoice) {
			case 1 :
				ifShowRed=true;
				break;
			case 2 :
				ifShowGreen=true;
				break;
			case 3 :
				ifShowBlue=true;
				break;
			case 4 :
				ifShowRed=true;
				ifShowGreen=true;
				ifShowBlue=true;
				break;
			}
	// histogramType=      1 is histogram in boxed region of this section
						// 2 means stacked total accumulation histograms in 3-D volume
						// 3 means conventional histogram for all voxels in a 3-D volume
						// 4 means profile along a linear transect in this single section
						// 5 means profile along a linear transect in a 3-D stack.
						// 6 means z-core profile (Rdim,Gdim,Bdim as abs dark thresholds)
						// 7 means do full stats on a directory full of callibration standards pics.
						// 8 z-core profile (Rdim,Gdim,Bdim as % of darkest pixels to discard)
				 		// 9 bleach correction drop all pixels dimmer than RGBdims 
						// 10 bleach correction RGBdims are percentage pf darkest pixels to discard
		if(histogramType==4 || histogramType==5) { // profiles along a transect
		  nPtsInBox = myBox.myTransect.npixInBlur; 
		  if(ifShowRed) 
			redProf = new OneProfile(myBox, Rthresh, redd, 2);
		  if(ifShowGreen) 
			greenProf = new OneProfile(myBox, Gthresh, greenn, 85);
		  if(ifShowBlue) 
			blueProf = new OneProfile(myBox, Bthresh, bluee, 170);
			}
		else {
		  nPtsInBox = myBox.resetVoxPointers(); // computes the number and 
		  if(ifShowRed) 
			redHist = new OneHistogram(myBox, Rthresh, redd, 2);
		  if(ifShowGreen) 
			greenHist = new OneHistogram(myBox, Gthresh, greenn, 85);
		  if(ifShowBlue) 
			blueHist = new OneHistogram(myBox, Bthresh, bluee, 170);
			}

		switch (histogramType) {
			case 1: 
				ifJustOneSection=(zmin+zmax)/2;
				howTall = 256;
				howWide=256;
				plotscale= (double )howTall/((double)nPtsInBox*scale);
				makeSingleSectionHistogram(ifJustOneSection);
				this.setTitle("section "+ifJustOneSection);
				break;
			case 2:   
				howTall=320;
				howWide=256;
				plotscale= (double )howTall/((double)nPtsInBox*scale);
				makeStackedTotalsHistogram();
				break;
			case 3:
				howTall = 256;
				howWide=256;
				plotscale= (double )howTall/((double)nPtsInBox*scale);
				makeVolumeHistogram();
				break;
			case 4: // profile along a transect in single section.
				ifJustOneSection=(zmin+zmax)/2;
				this.setTitle("section "+ifJustOneSection);
				howTall=255;
				howWide=(int )(myBox.myTransect.transectLength);
				plotscale= (double )howTall/((double)scale);
				makeSingleProfile(ifJustOneSection);
				break;
			case 5: // profile along a transect in single section.
				howTall=255;
				howWide=(int)(myBox.myTransect.transectLength);
				plotscale= (double )howTall/(scale);
				makeStackedProfile();
				break;						
			case 6: // 6 means z-core profile thru a 3-D stack.
			case 8: // 6 means z-core profile thru a 3-D stack.
				howTall=320;
				howWide=(int)(pixPerCoreTic*(zmax-zmin));
				plotscale= (double )howTall/(255.0);
				makeZCoreProfile();
				break;
			case 9: // 6 means z-core profile thru a 3-D stack.
			case 10: // 6 means z-core profile thru a 3-D stack.
				howTall=320;
				howWide=(int)(pixPerCoreTic*(zmax-zmin));
				plotscale= (double )howTall/(255.0);
				makeBleachCorrectionProfile();
				break;
			}

		setBackground(Color.black);
		getContentPane().setLayout(new BorderLayout());
		canvas = new HistogramCanvas(this);
		canvas.rescale(howTall,howWide);
		makeLocationTextFields();
		locpan.add(whatYouCanDo);
		getContentPane().add("North",locpan);
		getContentPane().add("Center",canvas);
		
		show();
		if(histogramType==1 || histogramType==3 || histogramType==4) 
				setSize(howWide+3,howTall+182);
		else setSize(howWide+3,howTall+100);
		repaint();
		addWindowListener(this);
		whatYouCanDo.setVisible(true);
		}

//catch Choice events...	
	public void itemStateChanged(java.awt.event.ItemEvent event)
		{
		String arg = (String)(((JComboBox)(event.getSource())).getSelectedItem());
	    	if(arg.equals("write as stats file to disk")) {
	    		talk("write as stats file to disk");
	    		ArchiveMarks writer=new ArchiveMarks();
	    		writer.writeHistogram(this);
	    		}
	    		
	    	if(arg.equals("Write z-coreProfile data to disk file")) {
	    		if(histogramType==6 || histogramType==8 || histogramType==9 || histogramType==10)
	    			write_zCoreDataToDisk();
	    		}
	    		
	    	if(arg.equals("become current stats object")) {
	    		talk("become current stats object");
	    		statsJPanel.adoptNewHistogram(this,true);
	    		}
	    		
	    	if(arg.equals("unfiltered histograms")) {
				myBox.numLowPassIterations=0;
				talk("Histogram suffers no low pass filtering iterations");
				canvas.forceRepaint=true;
				canvas.repaint();
	    		}
	    		
	    	if(arg.equals("1  Low Pass Filter iteration")) {
				myBox.numLowPassIterations=1;
				talk("Putting histogram through "+myBox.numLowPassIterations+" low pass filtering iterations");
				canvas.forceRepaint=true;
				canvas.repaint();
	    		}
	    		
	    	if(arg.equals("2  Low Pass Filter iterations")) {
				myBox.numLowPassIterations=2;
				talk("Putting histogram through "+myBox.numLowPassIterations+" low pass filtering iterations");
				canvas.forceRepaint=true;
				canvas.repaint();
	    		}
	    	if(arg.equals("3  Low Pass Filter iterations")) {
				myBox.numLowPassIterations=3;
				talk("Putting histogram through "+myBox.numLowPassIterations+" low pass filtering iterations");
				canvas.forceRepaint=true;
				canvas.repaint();
	    		}
	    	if(arg.equals("4  Low Pass Filter iterations")) {
				myBox.numLowPassIterations=4;
				talk("Putting histogram through "+myBox.numLowPassIterations+" low pass filtering iterations");
				canvas.forceRepaint=true;
				canvas.repaint();
	    		}

	    	if(arg.equals("print self")) {
	    		talk("print self");
	    		}
	    	if(arg.equals("print black on white")) {
	    		ifMonochrome=true;
				setBackground(Color.white);
				canvas.setBackground(Color.white);
				myBox.setColors(ifMonochrome);
	    		}
	    	if(arg.equals("print in colors")) {
	    		ifMonochrome=false;
				setBackground(Color.black);
				canvas.setBackground(Color.black);
				myBox.setColors(ifMonochrome);
	    		}
	    		
	    	if(arg.equals("transfer quadratic gain curve fit to red channel LUT")) {
	    		if(CallibStats != null && CallibStats.gainCurveFitCoeffs != null && statsJPanel != null) { 
	    			statsJPanel.picJPanel.setNewGainCurveFit(0,CallibStats.concentrationMappedTo255,
	    				CallibStats.gainCurveFitCoeffs,true);
					}
	    		}
	    		
	    	if(arg.equals("transfer quadratic gain curve fit to green channel LUT")) {
	    		if(CallibStats != null && CallibStats.gainCurveFitCoeffs != null && statsJPanel != null) { 
	    			statsJPanel.picJPanel.setNewGainCurveFit(1,CallibStats.concentrationMappedTo255,
	    				CallibStats.gainCurveFitCoeffs,true);
					}
	    		}
	    		
	    	if(arg.equals("transfer quadratic gain curve fit to blue channel LUT")) {
	    		if(CallibStats != null && CallibStats.gainCurveFitCoeffs != null && statsJPanel != null) { 
	    			statsJPanel.picJPanel.setNewGainCurveFit(2,CallibStats.concentrationMappedTo255,
	    				CallibStats.gainCurveFitCoeffs,true);
					}
	    		}
	    		
	    		
	    	if(arg.equals("overlay another set of gain curves")) {
	    		if(CallibStats != null)
	    			CallibStats.overlayAnotherCurveSet();
				canvas.forceRepaint=true;
				canvas.repaint();
	    		}
	    		
	    	if(arg.equals("purge empty directories from gain curve directory")) {
	    		if(CallibStats != null)
	    			CallibStats.purgeEmptyDirectories();
	    		}
		}

     public void copyDataFrom(Stats astat) {
		ST_ifBoxIsVisible = astat.ifBoxIsVisible;
		ST_ifInnerBoxValid = astat.ifInnerBoxValid;
		ST_ifTransectLineVisible=astat.ifTransectLineVisible;		
		ST_width=astat.tranwidth;
		ST_blur=astat.tranblur;
		ST_enlargeRatio =  astat.enlargeRatio;
		ST_shrinkRatio =  astat.shrinkRatio;
    	}
    	
	
// catch button and menu events...
	public void actionPerformed(java.awt.event.ActionEvent event) {
		talk("actionPerformed...");
		if(event != null) {
			String arg = event.getActionCommand();
			talk("Should not happen. Got arg="+arg+" in actionPerformed of class Histogram????");
			}
	}
	
	public void drawHistograms(Graphics g) {
		if(!ifStackDone) {
			talk("stack not finished");
			return;
			}
		switch (histogramType) {
			case 1:
				drawHorizGridLines(g);
				if(blueHist != null && ifShowBlue) 
					blueHist.draw(g,blueFilename,myDiskName);
				if(greenHist != null && ifShowGreen) 
					greenHist.draw(g,greenFilename,myDiskName);
				if(redHist != null && ifShowRed) 
					redHist.draw(g,redFilename,myDiskName);
			break;
		
			case 2:
				drawStackedtotals(g);
				break;
			case 3:
				drawVolumeHistogram(g);
				break;
			case 4:
				drawHorizGridLines(g);
				if(blueProf != null && ifShowBlue) 
					blueProf.draw(g,howTall,scale,blueFilename,myDiskName);
				if(greenProf != null && ifShowGreen) 
					greenProf.draw(g,howTall,scale,greenFilename,myDiskName);
				if(redProf != null && ifShowRed) 
					redProf.draw(g,howTall,scale,redFilename,myDiskName);
				break;
			case 5:
				drawStackedProfs(g);
				break;
			case 6:
			case 8:
				drawZCoreProf(g);
				break;
			case 9:
			case 10:
				drawZCoreProf(g);
				break;
			case 7:
				drawHorizGridLines(g);
				drawCallibStats(g);
				break;
			}
		}
			
		
	private boolean makeAHistogram(BioradFile rawdat,OneHistogram theHistogram, int sectnum,
				String colorID) {
		byte []  raster;
		boolean retval=false;
		try {
			raster = rawdat.readShiftedBioradRaster(sectnum);
			if(raster!=null) { // make the red histogram
			       theHistogram.makeNewHistogram(raster);
			       retval=true;
				 }}
	 		catch(RuntimeException e) {		
			talk("Couldn't read "+colorID+" channel raw data for section "+sectnum);
			talk("RuntimeException=="+e);
			}
		return retval;
		}

	private boolean makeAProfile(BioradFile rawdat,OneProfile theProfile, int sectnum,
				String colorID) {
		byte []  raster;
		int ier=0;
		boolean retval=false;
		statsJPanel.webCacheLevel(-1);
		try {
			raster = rawdat.readShiftedBioradRaster(sectnum);
			if(raster!=null) { // make the red histogram
			       theProfile.makeNewProfile(raster);
			       retval=true;
				 }}
	 		catch(RuntimeException e) {		
			talk("Couldn't read "+colorID+" channel raw data for section "+sectnum+", ier="+ier);
			talk("RuntimeException=="+e);
			}
		statsJPanel.webCacheLevel(0);
		return retval;
		}
			
	private void makeSingleSectionHistogram(int nsect) {
		BioradFile rawdat;
		if(ifShowRed && (rawdat = myBox.picJPanel.curStack.redfile) != null) 
			makeAHistogram(rawdat,redHist,ifJustOneSection,"red");
		if(ifShowGreen && (rawdat = myBox.picJPanel.curStack.greenfile) != null) 
			makeAHistogram(rawdat,greenHist,ifJustOneSection,"green");
		if(ifShowBlue && (rawdat = myBox.picJPanel.curStack.bluefile) != null) 
			makeAHistogram(rawdat,blueHist,ifJustOneSection,"blue");
		ifStackDone=true;
		}

	private void makeSingleProfile(int nsect) {
		BioradFile rawdat;
		statsJPanel.webCacheLevel(-1);
		if(ifShowRed && (rawdat = myBox.picJPanel.curStack.redfile) != null) {
			makeAProfile(rawdat,redProf, ifJustOneSection,"red"); 
			redProf.captionX=1;
			}
		if(ifShowGreen && (rawdat = myBox.picJPanel.curStack.greenfile) != null) {
			makeAProfile(rawdat,greenProf, ifJustOneSection,"green");
			greenProf.captionX=89;
			}
		if(ifShowBlue && (rawdat = myBox.picJPanel.curStack.bluefile) != null)  {
			makeAProfile(rawdat,blueProf, ifJustOneSection,"blue");
			blueProf.captionX=177;
			}
		ifStackDone=true;
		statsJPanel.webCacheLevel(0);
		}

	private void cannotDo(String stg) {
		talk("can do only one color channel at a time for 3-D histogram stacks");
		talk("supressing "+stg);
		}
	
	private int purgeDarkestPixels(int [] hsto,double discardDarkFraction) {
		int junkPix,npix=0,i;
		for(i=0;i<256;i++)
			npix += hsto[i]; // count all the pixels.
		junkPix = (int)(discardDarkFraction*npix + .499);
		while(junkPix>=npix) --junkPix;
		npix -= junkPix;
		i=0;
		while(i<256 && junkPix>0) {
			if(junkPix>=hsto[i]) {
				junkPix -= hsto[i];
				hsto[i]=0;
				}
			else { 
				hsto[i] -= junkPix;
				junkPix=0;
				}
			++i;
			}
		return npix;
		}
	

	public int upDateRGBdims(int irgb)  {
		int Thresh;
		myBox.upDateRGBdims();
		Rthresh = statsJPanel.Rthresh;
		Gthresh = statsJPanel.Gthresh;
		Bthresh = statsJPanel.Bthresh;
		if(irgb==0)  return Rthresh;
		if(irgb==1)  return Gthresh;
		return Bthresh;			
		}		
		
	private void makeBleachCorrectionProfile() {
		BioradFile rawdat=null;
		BioradStackConfig myStackConfig = null;
		int nPreScansects=0;
		int [] rawProf;
		double [] grandTotal;
		double total,prevFac,maxCorrFac=-1, maxVal;
		int nsects=0;		
		int k,i,j,pix,npix,dimmestPix,mapsto,Thresh=0,irgb=0;
		String cword="red";
		int [] xy={0,0};
		int [] hsto;
		long kk,sum;
		byte [] ras;
		boolean ifWanderingCoreProfile = myBox.ifCoreBoxesWander;
		setTitle("gain/bleach correction factor");
		if(!statsJPanel.ifReallyMakeHistogram) {
			talk("Push the 'Make the Histogram' button to actually make the histogram");
			return;
			}
		talk("Doing auto bleach correction");
		// do only one color... 
		redPixAtPeak=0;
		grandmax=-1;	 
		if(ifShowRed) {
			rawdat = myBox.picJPanel.curStack.redfile;
			irgb=0;  cword="red";
			ifShowGreen=false;
			ifShowBlue=false;
			}
		if(ifShowGreen) {
			rawdat = myBox.picJPanel.curStack.greenfile;
			irgb=1;  cword="green";
			ifShowRed=false;
			ifShowBlue=false;
			}
		if(ifShowBlue) {
			rawdat = myBox.picJPanel.curStack.bluefile;
			irgb=2;  cword="blue";
			ifShowRed=false;
			ifShowGreen=false;
			}
		Thresh=upDateRGBdims(irgb);
		if(rawdat==null) {
			talk("OOPs. cannot find the raw data...");
			return;
			}
		myStackConfig = rawdat.myConfig;
		if(!myStackConfig.ifGainCurveCorrected_bleachCorrected_stack) {
			talk("You do not have the kind of stack to which I can apply automatic bleach correction.");
			return;
			}
		if(!statsJPanel.picJPanel.assertRawDataTransformationMode(1)) {
			talk("You must first define a bleach correction curve...");
			return;
			}
		myStackConfig.bleachCorrectPreScanSections(); // uses lambda1, lambda2, fastrat...
		nPreScansects = myStackConfig.firstRealSection;	
		zmin=0;
		zmax = rawdat.numJFrames-1;
		nsects=zmax-zmin+1;		
		drawramps = new short[2][nsects];
		coreAvg = new float[nsects+1];  // will hold the floating point average value
										// of each section. coreAvg[nsects] is set to the avg brightness
		for(i=0;i<nsects;i++) {
			coreAvg[i]=-1;  // flag that it's not done yet.
			myStackConfig.dimThresh[i]=-1;
			}
			
		for(i=0;i<nPreScansects;i++) {
			coreAvg[i] = -2;
			coreAvg[myStackConfig.takeWanderingCoreBoundaryFromSection(i)] = -2;
			} // now compute the ave brightness in each section whose coreAvg == -2
		hsto = new int[256];// this will hold the histogram
		
		int fullWidth=statsJPanel.picJPanel.fullWidth;
		if(rawdat != null) {
			for(i=0;i<nsects;i++) {
				if(coreAvg[i] <= -1.5) 
				  try {
					ras = rawdat.readShiftedBioradRaster(i+zmin);
					if(ifWanderingCoreProfile) {
					    myBox.setSection(i+zmin);
						j=myBox.resetVoxPointers();
						talk("box has "+j+" pixels in section "+i);
						}
					// boundaries appropriate for this frame. 
					myBox.resetVoxCycle();
					for(j=0;j<256;j++) hsto[j]=0; // clear the histogram array 
					while(myBox.voxCycle(xy)) {
						if((pix = ras[xy[1]*fullWidth+xy[0]]) <0) 
						pix += 256;
						++hsto[pix];
						}
						
					Thresh=upDateRGBdims(irgb);
					myStackConfig.dimThresh[i]=Thresh;
					if(histogramType==10) // Rdim is the percentage of darkest pixels to discard
						purgeDarkestPixels(hsto,(.01*(double)Thresh));
					else // ignore pixels dimmer than the threshold Rdim 
						for(j=0;j<Thresh;j++)	
							hsto[j]=0;
					sum = 0;
					npix=0;
					dimmestPix=-1;
					for(j=0;j<256;j++) {
						if(dimmestPix<0 && hsto[j]>0)
							dimmestPix=j;
						sum += (j*hsto[j]);
						npix += hsto[j];
						}					
					coreAvg[i] = (float)(((double)sum)/(.01+npix));
					myStackConfig.controlBrightness[i]=coreAvg[i];
					if(histogramType==10)
					  talk(cword+" sect "+(i+zmin)+" has avg intensity = "+coreAvg[i]+
						", among "+npix+" pixels ("+Rthresh+" % dimmest pixels < "+dimmestPix+" discarded)");
					else 
					  talk(cword+" sect "+(i+zmin)+" has avg intensity = "+coreAvg[i]+" among "+npix+
						" (pixels brighter than "+Rthresh);
					}
	 		catch(RuntimeException e) {		
			talk("make z-core profile couldn't read "+cword+" channel raw data for section "+i);
			talk("RuntimeException=="+e);
			}
		  } 
		}

		  // on fall through we have avg brightness data needed to compute the 
		  // bleach correction ratios for each pre-scan section and each of the finely space
		  // sections whose heights correspond to some pre-scan section.
		  for(i=0;i<nPreScansects;i++) {
				myStackConfig.boostFacToUnBleach[i] = 1./myStackConfig.preBleach(i);
				mapsto = myStackConfig.takeWanderingCoreBoundaryFromSection(i);
				myStackConfig.boostFacToUnBleach[mapsto] = 
					myStackConfig.boostFacToUnBleach[i]*coreAvg[i]/coreAvg[mapsto];
				talk("boost factor for sect "+mapsto+" = "+myStackConfig.boostFacToUnBleach[mapsto]+
				" = "+myStackConfig.boostFacToUnBleach[i]+" * "+coreAvg[i]+ " / "+coreAvg[mapsto]);
				}
		// special case... Use the same correction factor for the first real section as for the
		// last pre-scan section. 
		myStackConfig.boostFacToUnBleach[nPreScansects] = 1./myStackConfig.preBleach(nPreScansects);
		// now interpolate the correction factors for the finely-spaced sections that do not
		// correspond to any pre-scan section.
		myStackConfig.lowPassBoostFacs(4);
		// find the max value...
		  for(i=0;i<nsects;i++) {
		  	coreAvg[i] = (float)myStackConfig.boostFacToUnBleach[i];
		  	if(maxCorrFac < coreAvg[i])
		  		maxCorrFac = coreAvg[i];
		  	}
		  	
		maxVal = maxCorrFac -1.0 + .25;		 
		plotscale= (double )howTall/(maxVal);
		for(i=0;i<nsects;i++) {
			coreAvg[i] = (float)myStackConfig.boostFacToUnBleach[i];
			drawramps[0][i] = (short )(pixPerCoreTic*i); // the horizontal plot coordinate
			drawramps[1][i] = (short)(plotscale*(maxVal - (coreAvg[i]-1.0) ));
			}
		ifStackDone=true;
	}

	private void makeZCoreProfile() {
		BioradFile rawdat=null;
		int [] rawProf;
		double [] grandTotal;
		double total,avgbright;
		int nsects=zmax-zmin+1;		
		int k,i,j,pix,npix,dimmestPix,Thresh=0;
		String cword="purple";
		int [] xy={0,0};
		int [] hsto;
		long kk,sum;
		byte [] ras;
		boolean ifWanderingCoreProfile = myBox.ifCoreBoxesWander;
		if(!statsJPanel.ifReallyMakeHistogram) {
			talk("Push the 'Make the Histogram' button to actually make the histogram");
			return;
			}
		talk("Coring "+nsects+" sections: "+zmin+" to "+zmax);
		// do only one color... 
		redPixAtPeak=0;
		grandmax=-1;	 
		if(ifShowRed) {
			rawdat = myBox.picJPanel.curStack.redfile;
			Thresh=Rthresh;  cword="red";
			ifShowGreen=false;
			ifShowBlue=false;
			}
		if(ifShowGreen) {
			rawdat = myBox.picJPanel.curStack.greenfile;
			Thresh=Gthresh;  cword="green";
			ifShowRed=false;
			ifShowBlue=false;
			}
		if(ifShowBlue) {
			rawdat = myBox.picJPanel.curStack.bluefile;
			Thresh=Bthresh;  cword="blue";
			ifShowRed=false;
			ifShowGreen=false;
			}
		if(rawdat==null) {
			talk("OOPs. cannot find the raw data...");
			return;
			}
		avgbright=0.;
		drawramps = new short[2][nsects];
		coreAvg = new float[nsects+1];  // will hold the floating point average value
										// of each section. coreAvg[nsects] is set to the avg brightness
		hsto = new int[256];// this will hold the histogram
		int fullWidth=statsJPanel.picJPanel.fullWidth;
			for(i=0;i<nsects;i++) {
				  try {
					ras = rawdat.readShiftedBioradRaster(i+zmin);
					if(ifWanderingCoreProfile) {
					    myBox.setSection(i+zmin);
						myBox.resetVoxPointers();
						}
					// boundaries appropriate for this frame. 
					myBox.resetVoxCycle();
					for(j=0;j<256;j++) hsto[j]=0; // clear the histogram array 
					while(myBox.voxCycle(xy)) {
						if((pix = ras[xy[1]*fullWidth+xy[0]]) <0) 
						pix += 256;
						++hsto[pix];
						}
					if(histogramType==8) // Rdim is the percentage of darkest pixels to discard
						purgeDarkestPixels(hsto,(.01*(double)Thresh));
					else // ignore pixels dimmer than the threshold Rdim 
						for(j=0;j<Thresh;j++)	
							hsto[j]=0;
					sum = 0;
					npix=0;
					dimmestPix=-1;
					for(j=0;j<256;j++) {
						if(dimmestPix<0 && hsto[j]>0)
							dimmestPix=j;
						sum += (j*hsto[j]);
						npix += hsto[j];
						}					
					drawramps[0][i] = (short )(pixPerCoreTic*i); // the horizontal plot coordinate
					coreAvg[i] = (float)(((double)sum)/(.01+npix));
					avgbright += coreAvg[i];
					if(coreAvg[i]>grandmax) { // keep track where the maximum occurs
						grandmax = coreAvg[i];
						redPixAtPeak=i+zmin;
						}
					coreAvg[i] = (float)(.01*Math.round(100.*coreAvg[i]));
					drawramps[1][i] = (short)(plotscale*(255-.499-coreAvg[i]));
					if(histogramType==8)
					  talk(cword+" sect "+(i+zmin)+" has avg intensity = "+coreAvg[i]+
						", among "+npix+" pixels ("+Rthresh+" % dimmest pixels < "+dimmestPix+" discarded)");
					else 
					  talk(cword+" sect "+(i+zmin)+" has avg intensity = "+coreAvg[i]+" among "+npix+
						" (pixels brighter than "+Rthresh);
					}
	 		catch(RuntimeException e) {		
			talk("make z-core profile couldn't read "+cword+" channel raw data for section "+i);
			talk("RuntimeException=="+e);
			}
		  }
		
		avgbright = avgbright/(nsects+.000001);
		talk("The average brightness of the stack is "+avgbright);
		coreAvg[nsects] = (float)avgbright;
		ifStackDone=true;
		}
		
	private void makeStackedProfile() {
		BioradFile rawdat;
		int [] rawProf;
		double [] grandTotal;
		double total;
		int nsects=zmax-zmin+1;		
		int i,k;
		long kk;
		if(!statsJPanel.ifReallyMakeHistogram) {
			talk("Push the 'Make the Histogram' button to actually make the histogram");
			return;
			}
		talk("Doing sections "+zmin+" to "+zmax);
		statsJPanel.webCacheLevel(-1);
		// do only one color... 
		if(ifShowRed) {
			if(ifShowGreen) {
				if(ifShowGreen) cannotDo("green and blue");
				ifShowGreen=false;
				ifShowBlue=false;}
				}
		if(ifShowGreen) {
			if(ifShowBlue) cannotDo("blue");
			ifShowBlue=false;
			}
		grandTotal = new double[howWide];
		drawramps = new short[nsects][howWide];
		for(i=0;i<howWide;i++) grandTotal[i]=0;
		if(ifShowRed) {
			for(i=0;i<nsects;i++) {
				if((rawdat = myBox.picJPanel.curStack.redfile) != null) 
					makeAProfile(rawdat,redProf, i+zmin,"red");
				rawProf=redProf.getRawProf();
				talk("red section "+(i+zmin));
				for(k=0;k<howWide;k++) {					
					grandTotal[k] += rawProf[k];
					drawramps[i][k] = (short)(howTall-(plotscale*grandTotal[k]));
					}
				}
			}	
		if(ifShowGreen) {
			for(i=0;i<nsects;i++) {
				if((rawdat = myBox.picJPanel.curStack.greenfile) != null) 
					makeAProfile(rawdat,greenProf, i+zmin,"green");
				rawProf=greenProf.getRawProf();
				talk("green section "+(i+zmin));
				for(k=0;k<howWide;k++) {					
					grandTotal[k] += rawProf[k];
					drawramps[i][k] = (short)(howTall-(plotscale*grandTotal[k]));
					}
				}
			}	
		if(ifShowBlue) {
			for(i=0;i<nsects;i++) {
				if((rawdat = myBox.picJPanel.curStack.bluefile) != null) 
					makeAProfile(rawdat,blueProf, i+zmin,"blue");
				rawProf=blueProf.getRawProf();
				talk("blue section "+(i+zmin));
				for(k=0;k<howWide;k++) {					
					grandTotal[k] += rawProf[k];
					drawramps[i][k] = (short)(howTall-(plotscale*grandTotal[k]));
					}
				}
			}	
		grandTotal=null;
		ifStackDone=true;
		statsJPanel.webCacheLevel(-1);
		}
				
	private void makeStackedTotalsHistogram() {
		BioradFile rawdat;
		double [] normTotHist;
		double [] grandTotal;
		double total;
		int nsects=zmax-zmin+1;		
		int i,k;
		long kk;
		if(!statsJPanel.ifReallyMakeHistogram) {
			talk("Push the 'Make the Histogram' button to actually make the histogram");
			return;
			}
		talk("Doing sections "+zmin+" to "+zmax);
		statsJPanel.webCacheLevel(-1);
		// do only one color... 
		if(ifShowRed) {
			if(ifShowGreen) {
				if(ifShowGreen) cannotDo("green and blue");
				ifShowGreen=false;
				ifShowBlue=false;}
				}
		if(ifShowGreen) {
			if(ifShowBlue) cannotDo("blue");
			ifShowBlue=false;
			}
		grandTotal = new double[256];
		drawramps = new short[nsects][256];
		for(i=0;i<256;i++) grandTotal[i]=0;
		talk("make stacked histograms, scale="+scale);
		if(ifShowRed) {
				for(i=0;i<nsects;i++) {
				if((rawdat = myBox.picJPanel.curStack.redfile) != null) 
					makeAHistogram(rawdat,redHist, i+zmin,"red");
				normTotHist=redHist.getnormalTotalHist();
				talk("red section "+(i+zmin)+" has total = "+normTotHist[256]);
				total = normTotHist[256]/255.;
				for(k=0;k<256;k++) {					
					grandTotal[k] += total*normTotHist[k];
					drawramps[i][k] = (short)(howTall-(plotscale*grandTotal[k]));
					}
				}
			}	
		if(ifShowGreen) {
				for(i=0;i<nsects;i++) {
				if((rawdat = myBox.picJPanel.curStack.greenfile) != null) 
					makeAHistogram(rawdat,greenHist, i+zmin,"green");
				normTotHist=greenHist.getnormalTotalHist();
				talk("green section "+(i+zmin)+" has total = "+normTotHist[256]);
				total = normTotHist[256]/255.;
				for(k=0;k<256;k++) {					
					grandTotal[k] += total*normTotHist[k];
					drawramps[i][k] = (short)(howTall-(plotscale*grandTotal[k]));
					}
				}
			}	
		if(ifShowBlue) {
				for(i=0;i<nsects;i++) {
				if((rawdat = myBox.picJPanel.curStack.bluefile) != null) 
					makeAHistogram(rawdat,blueHist, i+zmin,"blue");
				normTotHist=blueHist.getnormalTotalHist();
				talk("blue section "+(i+zmin)+" has total = "+normTotHist[256]);
				total = normTotHist[256]/255.;
				for(k=0;k<256;k++) {					
					grandTotal[k] += total*normTotHist[k];
					drawramps[i][k] = (short)(howTall-(plotscale*grandTotal[k]));
					}
				}
			}	
		grandmax = grandTotal[255];
		howBright = (grandmax/((double)nPtsInBox));
		grandTotal=null;
		ifStackDone=true;
		statsJPanel.webCacheLevel(0);
		}
		
	private void makeVolumeHistogram() {
		BioradFile rawdat;
		double [] normTotHist;
		double [] normVolHist;
		int [] aHist;
		double total,dev;
		int nsects=zmax-zmin+1;		
		int i,k;
		if(!statsJPanel.ifReallyMakeHistogram) {
			talk("Push the 'Make the Histogram' button to actually make the histogram");
			return;
			}
		talk("Doing sections "+zmin+" to "+zmax);
		// do only one color... 
		statsJPanel.webCacheLevel(-1);
		if(ifShowRed) {
			redNormHist = new double[256];
			redTotHist = new double[256];
			for(i=0;i<256;i++) redNormHist[i]=0;
			for(i=nsects-1;i>=0;i--) {
				grandmax=0;
				if((rawdat = myBox.picJPanel.curStack.redfile) != null) 
					makeAHistogram(rawdat,redHist, i+zmin,"red");
				aHist = redHist.getRawHST();
				for(k=0;k<256;k++)  {
					if(k>=Rthresh) {				
						redNormHist[k] += aHist[k];
						grandmax += k*aHist[k];
						}
					}
				talk("red section "+(i+zmin)+" has total = "+grandmax);
				}
			redPixAtPeak=(long)redNormHist[0];
			redPeak=0;
			redTotHist[0] = 0;
			redNbright = 0;
			redNblazing=0;
			for(k=1;k<256;k++) {
				redNbright += redNormHist[k]; // the first few are zero...
				if(k>250) redNblazing += redNormHist[k];
				redTotHist[k] = redTotHist[k-1] + k*redNormHist[k];
				if(redNormHist[k] > redPixAtPeak) {				
					redPixAtPeak=(long)redNormHist[k];
					redPeak=k;
					}
				}
			// now normalize 
			redTotal = redTotHist[255];
			redMean = (int)(.499+redTotal/(1+redNbright));
			redStdDev=0;
			for(k=1;k<256;k++) {
				dev = k - redMean;
				redStdDev += redNormHist[k]*dev*dev;				 
				redNormHist[k] = (1+255*redNormHist[k])/redPixAtPeak;
				redTotHist[k] = (1+255*redTotHist[k])/redTotal;
				}
			redStdDev = (int)(.499+Math.sqrt(redStdDev/redNbright));
			}	
		if(ifShowGreen) {
			greenNormHist = new double[256];
			greenTotHist = new double[256];
			for(i=0;i<256;i++) greenNormHist[i]=0;
			for(i=nsects-1;i>=0;i--) {
				grandmax=0;
				if((rawdat = myBox.picJPanel.curStack.greenfile) != null) 
					makeAHistogram(rawdat,greenHist, i+zmin,"green");
				aHist= greenHist.getRawHST();
				for(k=0;k<256;k++)  {
					if(k>=Gthresh) {				
						greenNormHist[k] += aHist[k];
						grandmax += k*aHist[k];
						}
					}
				talk("green section "+(i+zmin)+" has total = "+grandmax);
				}
			greenPixAtPeak=(long)greenNormHist[0];
			greenPeak=0;
			greenTotHist[0] = 0;
			greenNbright = 0;
			greenNblazing=0;
			for(k=1;k<256;k++) {
				greenNbright += greenNormHist[k]; // the first few are zero...
				if(k>250) greenNblazing += greenNormHist[k];
				greenTotHist[k] = greenTotHist[k-1] + k*greenNormHist[k];
				if(greenNormHist[k] > greenPixAtPeak) {				
					greenPixAtPeak=(long)greenNormHist[k];
					greenPeak=k;
					}
				}
			// now normalize 
			greenTotal = greenTotHist[255];
			greenMean = (int)(.499+greenTotal/(1+greenNbright));
			greenStdDev=0;
			for(k=1;k<256;k++) {
				dev = k - greenMean;
				greenStdDev += greenNormHist[k]*dev*dev;				 
				greenNormHist[k] = (1+255*greenNormHist[k])/greenPixAtPeak;
				greenTotHist[k] = (1+255*greenTotHist[k])/greenTotal;
				}
			greenStdDev = (int)(.499+Math.sqrt(greenStdDev/greenNbright));
			}	
		if(ifShowBlue) {
			blueNormHist = new double[256];
			blueTotHist = new double[256];
			for(i=0;i<256;i++) blueNormHist[i]=0;
			for(i=nsects-1;i>=0;i--) {
				grandmax=0;
				if((rawdat = myBox.picJPanel.curStack.bluefile) != null) 
					makeAHistogram(rawdat,blueHist, i+zmin,"blue");
				aHist= blueHist.getRawHST();
				for(k=0;k<256;k++)  {
					if(k>=Bthresh) {				
						blueNormHist[k] += aHist[k];
						grandmax += k*aHist[k];
						}
					}
				talk("blue section "+(i+zmin)+" has total = "+grandmax);
				}
			bluePixAtPeak=(long)blueNormHist[0];
			bluePeak=0;
			blueTotHist[0] = 0;
			blueNbright = 0;
			blueNblazing=0;
			for(k=1;k<256;k++) {
				blueNbright += blueNormHist[k]; // the first few are zero...
				if(k>250) blueNblazing += blueNormHist[k];
				blueTotHist[k] = blueTotHist[k-1] + k*blueNormHist[k];
				if(blueNormHist[k] > bluePixAtPeak) {				
					bluePixAtPeak=(long)blueNormHist[k];
					bluePeak=k;
					}
				}
			// now normalize 
			blueTotal = blueTotHist[255];
			blueMean = (int)(.499+blueTotal/(1+blueNbright));
			blueStdDev=0;
			for(k=1;k<256;k++) {
				dev = k - blueMean;
				blueStdDev += blueNormHist[k]*dev*dev;				 
				blueNormHist[k] = (1+255*blueNormHist[k])/bluePixAtPeak;
				blueTotHist[k] = (1+255*blueTotHist[k])/blueTotal;
				}
			blueStdDev = (int)(.499+Math.sqrt(blueStdDev/blueNbright));
			}	
		ifStackDone=true;					
		statsJPanel.webCacheLevel(0);
		}
		
	private double truncate(double xx) {
		int i = (int)(10.0*xx);
		return (i/10.0);
		}		
			
	private void drawVolumeHistogram(Graphics g) {
		int i,k,xto,yto,xfrom,yfrom;
		int nsects=zmax-zmin+1,volume;
		int captionX = 10, capdy=15,capy;
		if(ifMonochrome) g.setColor(Color.black);
		capy=howTall+capdy;
			if(ifShowBlue) {
				if(!ifMonochrome) g.setColor(bluee);
				captionX=175;
				xto=0; yto=howTall - (int)(blueNormHist[0]);
				for(k=1;k<256;k++) {
						xfrom=xto;	yfrom=yto;
						xto=k;	yto=howTall - (int)(blueNormHist[k]);
						g.drawLine(xfrom,yfrom,xto,yto);
						}
				xto=0; yto=howTall - (int)(blueTotHist[0]);
				for(k=1;k<256;k++) {
						xfrom=xto;	yfrom=yto;
						xto=k;	yto=howTall - (int)(blueTotHist[k]);
						g.drawLine(xfrom,yfrom,xto,yto);
						}
				capy=howTall+capdy;
				capy=256+capdy;
				g.drawString("sumBP="+(int)blueTotal,captionX,capy);
				capy += capdy;
				volume  = (nsects*nPtsInBox);
				g.drawString("volume="+volume,captionX,capy);
				capy += capdy;
				g.drawString("BP/brightVox="+(.01*Math.round(100.0*(blueTotal/blueNbright))),captionX,capy);
				capy += capdy;
				g.drawString("BP/area="+(.01*Math.round(100.0*(blueTotal/nPtsInBox))),captionX,capy);
				capy += capdy;
				g.drawString("peak at "+bluePeak,captionX,capy);	
				capy += capdy;
				g.drawString(bluePixAtPeak+" pix@peak",captionX,capy);	
				capy += capdy;
				g.drawString("std dev="+blueStdDev,captionX,capy);
				capy += capdy;
				g.drawString(truncate(100.0*(double)blueNbright/(double)volume)+"% pix >",captionX,capy);
				capy += capdy;
				g.drawString("   dim thresh="+Bthresh,captionX,capy);
				capy += capdy;
				g.drawString(blueNblazing+" pix > 250",captionX,capy);	
				capy += capdy;
				g.drawString(("file="+blueFilename),captionX,capy);
				}
			if( ifShowGreen) {
				if(!ifMonochrome) g.setColor(greenn);
				captionX=88;
				xto=0; yto=howTall - (int)(greenNormHist[0]);
				for(k=1;k<256;k++) {
						xfrom=xto;	yfrom=yto;
						xto=k;	yto=howTall - (int)(greenNormHist[k]);
						g.drawLine(xfrom,yfrom,xto,yto);
						}
				xto=0; yto=howTall - (int)(greenTotHist[0]);
				for(k=1;k<256;k++) {
						xfrom=xto;	yfrom=yto;
						xto=k;	yto=howTall - (int)(greenTotHist[k]);
						g.drawLine(xfrom,yfrom,xto,yto);
						}
				capy=howTall+capdy;
				capy=256+capdy;
				g.drawString("sumBP="+(int)greenTotal,captionX,capy);
				capy += capdy;
				volume  = (nsects*nPtsInBox);
				g.drawString("volume="+volume,captionX,capy);
				capy += capdy;
				g.drawString("BP/brightVox="+(.01*Math.round(100.0*(greenTotal/greenNbright))),captionX,capy);
				capy += capdy;
				g.drawString("BP/area="+(.01*Math.round(100.0*(greenTotal/nPtsInBox))),
						captionX,capy);
				capy += capdy;
				g.drawString("peak at "+greenPeak,captionX,capy);	
				capy += capdy;
				g.drawString(greenPixAtPeak+" pix@peak",captionX,capy);	
				capy += capdy;
				g.drawString("std dev="+greenStdDev,captionX,capy);
				capy += capdy;
				g.drawString((truncate(100.0*(double)greenNbright/(double)volume))+"% pix >",captionX,capy);
				capy += capdy;
				g.drawString("   dim thresh="+Gthresh,captionX,capy);
				capy += capdy;
				g.drawString(greenNblazing+" pix > 250",captionX,capy);	
				capy += capdy;
				g.drawString(("file="+greenFilename),captionX,capy);
				}
			if(ifShowRed) {
				if(!ifMonochrome) g.setColor(redd);
				captionX=2;
				xto=0; yto=howTall - (int)(redNormHist[0]);
				for(k=1;k<256;k++) {
						xfrom=xto;	yfrom=yto;
						xto=k;	yto=howTall - (int)(redNormHist[k]);
						g.drawLine(xfrom,yfrom,xto,yto);
						}
				xto=0; yto=howTall - (int)(redTotHist[0]);
				for(k=1;k<256;k++) {
						xfrom=xto;	yfrom=yto;
						xto=k;	yto=howTall - (int)(redTotHist[k]);
						g.drawLine(xfrom,yfrom,xto,yto);
						}
				capy=howTall+capdy;
				capy=256+capdy;
				g.drawString("sumBP="+(int)redTotal,captionX,capy);
				capy += capdy;
				volume  = (nsects*nPtsInBox);
				g.drawString("volume="+volume,captionX,capy);
				capy += capdy;
				g.drawString("BP/brightVox="+(.01*Math.round(100.0*(redTotal/redNbright))),captionX,capy);
				capy += capdy;
				g.drawString("BP/area="+(.01*Math.round(100.0*(redTotal/nPtsInBox))),captionX,capy);
				capy += capdy;
				g.drawString("peak at "+redPeak,captionX,capy);	
				capy += capdy;
				g.drawString(redPixAtPeak+" pix@peak",captionX,capy);	
				capy += capdy;
				g.drawString("std dev="+redStdDev,captionX,capy);
				capy += capdy;
				g.drawString((truncate(100.0*(double)redNbright/(double)volume))+"% pix >",captionX,capy);
				capy += capdy;
				g.drawString("   dim thresh="+Rthresh,captionX,capy);
				capy += capdy;
				g.drawString(redNblazing+" pix > 250",captionX,capy);	
				capy += capdy;
				g.drawString(("file="+redFilename),captionX,capy);
				}
			if(myDiskName != null) {
				capy += capdy;
				g.drawString(("hst = "+myDiskName),2,capy);
				}
		}

	private void drawStackedtotals(Graphics g) {
		int i,k,xto,yto,xfrom,yfrom;
		int nsects=zmax-zmin+1;
		int thresh=0;
		int captionX = 2, capdy=15,capy;
		String Filename=null;
		if(ifMonochrome) g.setColor(Color.black);
			if(ifShowRed) {
				if(!ifMonochrome) g.setColor(redd);
				captionX=2;
				for(i=0;i<nsects;i++) {
					xto=0; yto=drawramps[i][0];
					for(k=1;k<256;k++) {
						xfrom=xto;	yfrom=yto;
						xto=k;	yto=drawramps[i][k];
						g.drawLine(xfrom,yfrom,xto,yto);
						}
					}
				thresh=Rthresh;
				Filename=redFilename;
				}
			if(ifShowGreen) {
				if(!ifMonochrome) g.setColor(greenn);
				captionX=50;
				for(i=0;i<nsects;i++) {
					xto=0; yto=drawramps[i][0];
					for(k=1;k<256;k++) {
						xfrom=xto;	yfrom=yto;
						xto=k;	yto=drawramps[i][k];
						g.drawLine(xfrom,yfrom,xto,yto);
						}
					}
				thresh=Gthresh;
				Filename=greenFilename;
				}
			if(ifShowBlue) {
				if(!ifMonochrome) g.setColor(bluee);
				captionX=100;
				for(i=0;i<nsects;i++) {
					xto=0; yto=drawramps[i][0];
					for(k=1;k<256;k++) {
						xfrom=xto;	yfrom=yto;
						xto=k;	yto=drawramps[i][k];
						g.drawLine(xfrom,yfrom,xto,yto);
						}
					}
				thresh=Bthresh;
				Filename=blueFilename;
				}
		if(ifMonochrome) g.setColor(Color.black);
			capy=howTall+capdy;
			g.drawString("TBP = "+grandmax,captionX,capy);
			capy += capdy;
			g.drawString(("BP/area="+(int)(.499+howBright)+", max="+scale),captionX,capy);
			capy += capdy;
			g.drawString(("dim threshold = "+thresh),captionX,capy);
			if(!ifMonochrome) g.setColor(Color.yellow);
			capy += capdy;
			g.drawString(nsects+"-section ("+zmin+" thru "+zmax+
				") histogram with volume="+(nsects*nPtsInBox)+" voxels" ,0,capy);
			capy += capdy;
			g.drawString("Histogram area = "+nPtsInBox+" pixels = "+
					truncate(nPtsInBox/(pixelsPerMicrometer*pixelsPerMicrometer))+" square um",0,capy);
			capy += capdy;
			g.drawString(("file="+Filename),captionX,capy);
			if(myDiskName != null) {
				capy += capdy;
				g.drawString(("hst = "+myDiskName),2,capy);
				}
		}
	
	public void writeZCoreToDisk(ArchiveMarks am) {
		int nlow,thresh=0,j,k,pix,minsect=0,maxsect=0,nsects=zmax-zmin+1;
		double min=255,max=0,avg=0;
		double [] lowpass = new double[nsects];
		double lowp,perCentChange=0,avgPerCentChange=0.;
		BioradStackConfig myConfig=null;
		String stg="";
		String filename="";
		talk("nsects="+nsects);
		if(ifShowBlue) {
				thresh=Bthresh;
				filename=blueFilename;
				myConfig = myBox.picJPanel.curStack.bluefile.myConfig;
				}
		if(ifShowGreen) {
				thresh=Gthresh;
				filename=greenFilename;
				myConfig = myBox.picJPanel.curStack.greenfile.myConfig;
				}
		if(ifShowRed) {
				thresh=Rthresh;
				filename=redFilename;
				myConfig = myBox.picJPanel.curStack.redfile.myConfig;
				}
		avg=0.;
		for(j=0;j<nsects;j++) {
				avg += coreAvg[j];
				if(coreAvg[j]>max) {
					max=coreAvg[j];
					maxsect=j;
					}
				if(coreAvg[j]<min) {
					min=coreAvg[j];
					minsect=j;
					}
			}
		avg = avg/(nsects+.000000001);
		for(j=0;j<nsects;j++) {
			stg = (""+(j+zmin)+"    "+coreAvg[j]+"    "+myConfig.getZHeightFromSectNum(j+zmin));
			am.writeAString(stg);
			talk(stg); 
	      }
		stg="sect  rawAvg  depthIn uM  ";  am.writeAString(stg); talk(stg);
		stg=" = number of sections of z-core data"; am.writeNum(nsects,stg); talk(nsects+stg);
		myBox.writeCoreBoxToDisk(am);
		stg = " Z-core profile through "+nsects+"-sections ("+zmin+" to "+zmax+")";
		am.writeAString(stg);	talk(stg);
		stg = (" average brightness = "+avg);
		am.writeAString(stg);	talk(stg);
		stg = (" Max value of "+max+" occurs in section "+maxsect);
		am.writeAString(stg);	talk(stg);
		stg=(" Min value of "+min+" occurs in section "+minsect);
		am.writeAString(stg);	talk(stg);
		stg = (" dim threshold = "+thresh);
		am.writeAString(stg);	talk(stg);
		stg=(" file="+filename);
		am.writeAString(stg);	talk(stg);
		if(myDiskName != null) {
			am.writeAString((" hst = "+myDiskName));
			}
		}
	
	private void write_zCoreDataToDisk() {
	    ArchiveMarks writer=new ArchiveMarks();
	    writer.writeZCore(this);
		}
	
	private void drawHorizGridLines(Graphics g) {
		int jump=10, majorJump=100;	
		Color dim=new Color(70,70,70), emp = new Color(110,140,120);
		if(ifMonochrome)  {
			dim=new Color(180,180,180);
			emp = new Color(110,110,110);
			}
		switch (howTall) {
			case 256:
			case 255:
			case 320:
				jump = 20;
				majorJump=100;
				break;							
			case 510:
				jump = 50;
				majorJump=100;
				break;			
			default :
				jump = (1+howTall)/10;
				majorJump = 100;
				break;
			}
					
		g.setColor(dim);//draw the frequent, dim, horizontal grid lines
		for(int i=howTall-jump;i>=0;i -= jump)  
			g.drawLine(0,i,howWide,i);	
			
		g.setColor(emp);//draw the major, emphasized, horizontal grid lines
		for(int i=howTall-majorJump;i>=0;i -= majorJump)  
			g.drawLine(0,i,howWide,i);	
		}
			
	private void drawZCoreProf(Graphics g) {
		int nsects=zmax-zmin+1;
		int grid,xfrom,xto,yfrom,yto;
		int captionX = 10, capdy=15,capy,thresh=0,i,j;
		double avgbright=0.;
		String filename="";
		if(ifMonochrome) g.setColor(Color.lightGray);
		else g.setColor(Color.darkGray); // draw the tic marks
		grid=0;
		while(grid<zmin) grid +=5; // draw grid lines every five sections.
		for(i=0;i<nsects;i++) { // vertical grid lines
			if((i + zmin) == grid) {
				g.drawLine(drawramps[0][i],0,drawramps[0][i],howTall);
				grid += 5;
				}
			}
			
		if(histogramType<=8) 
			drawHorizGridLines(g);
			
		if(ifMonochrome) g.setColor(Color.black);
		captionX=2;
		if(ifShowRed) {
				if(!ifMonochrome) g.setColor(redd);
				thresh=Rthresh;
				filename=redFilename;
				}
		if(ifShowGreen) {
				if(!ifMonochrome) g.setColor(greenn);
				thresh=Gthresh;
				filename=greenFilename;
				}
		if(ifShowBlue) {
				if(!ifMonochrome) g.setColor(bluee);
				thresh=Bthresh;
				filename=blueFilename;
				}
		xto = drawramps[0][0];
		yto = drawramps[1][0];
		for(j=0;j<nsects-1;j++) {
			xfrom = xto;
			yfrom = yto;
			xto = drawramps[0][j+1];
			yto = drawramps[1][j+1];
			g.drawLine(xfrom,yfrom,xto,yto);
			// talk("to ("+xto+", "+yto+")");
			}			
		capy=howTall+capdy;
		g.drawString(("dim threshold = "+thresh+". Full scale==255"),captionX,capy);				
		capy += capdy;
		g.drawString(("Avg brightness thru all sections = "+truncate(coreAvg[nsects])),
							captionX,capy);
		capy += capdy;
		g.drawString(("Max value of "+(int)(grandmax)+" occurs in section "+redPixAtPeak),captionX,capy);
		capy += capdy;
		g.drawString("Z-core profile through "+nsects+"-sections ("+zmin+" to "+zmax+")",captionX,capy);
		capy += capdy;
		g.drawString(("file="+filename+". Grid lines @5-sect marks"),captionX,capy);
		
		if(myDiskName != null) {
			capy += capdy;
			g.drawString(("hst = "+myDiskName),2,capy);
			}
		}
	
	private void drawCallibStats(Graphics g) {
		drawHorizGridLines(g);			
		CallibStats.draw(g,howWide,howWide);
		}

	private void drawStackedProfs(Graphics g) {
		int i,k,xto,yto,xfrom,yfrom;
		int nsects=zmax-zmin+1;
		int captionX = 10, capdy=15,capy;
		double tic;
		if(!ifMonochrome) g.setColor(Color.yellow); // draw the tic marks 5 microns apart
		else g.setColor(Color.black);
		tic = -5*pixelsPerMicrometer;
		for(i=0;i<howWide;i++)  // draw the tic marks along horizontal axis
			if(i-tic >= 5*pixelsPerMicrometer) {
				g.drawLine(i,howTall,i,howTall+6);
				tic += 5*pixelsPerMicrometer;
				}
		drawHorizGridLines(g);				
		capy=howTall+capdy;
		g.drawString("stacked profiles for "+nsects+"-sections ("+zmin+" thru "+zmax+")",captionX,capy);
		capy += capdy;
		g.drawString("Transect is " +(int)myBox.myTransect.transectLength+" pix = "+
			truncate(myBox.myTransect.transectLength/myBox.myTransect.pixelsPerMicrometer)+" um long.",captionX,capy);
		capy += capdy;
		g.drawString("avgeraging box is "+myBox.myTransect.width+" pix wide", captionX,capy);
		capy += capdy;		
		g.drawString("  by "+myBox.myTransect.blur+ " pix blur along transect", captionX,capy);
		capy += capdy;
		g.drawString("maxVerticalScale = "+scale,captionX,capy);;				
		capy += capdy;
			if(ifShowRed) {
				if(!ifMonochrome) g.setColor(redd);
				captionX=2;
				for(i=0;i<nsects;i++) {
					xto=0; yto=drawramps[i][0];
					for(k=1;k<howWide;k++) {
						xfrom=xto;	yfrom=yto;
						xto=k;	yto=drawramps[i][k];
						g.drawLine(xfrom,yfrom,xto,yto);
						}
					}
				g.drawString("dim threshold = "+Rthresh,captionX,capy);				
				capy += capdy;
				g.drawString(("file="+redFilename),captionX,capy);
				}
			if(ifShowGreen) {
				if(!ifMonochrome) g.setColor(greenn);
				captionX=30;
				for(i=0;i<nsects;i++) {
					xto=0; yto=drawramps[i][0];
					for(k=1;k<howWide;k++) {
						xfrom=xto;	yfrom=yto;
						xto=k;	yto=drawramps[i][k];
						g.drawLine(xfrom,yfrom,xto,yto);
						}
					}
				g.drawString("dim threshold = "+Gthresh,captionX,capy);				
				capy += capdy;
				g.drawString(("file="+greenFilename),captionX,capy);
				}
			if(ifShowBlue) {
				if(!ifMonochrome) g.setColor(bluee);
				captionX=60;
				for(i=0;i<nsects;i++) {
					xto=0; yto=drawramps[i][0];
					for(k=1;k<howWide;k++) {
						xfrom=xto;	yfrom=yto;
						xto=k;	yto=drawramps[i][k];
						g.drawLine(xfrom,yfrom,xto,yto);
						}
					}
				g.drawString("dim threshold = "+Bthresh,captionX,capy);				
				capy += capdy;
				g.drawString(("file="+blueFilename),captionX,capy);
				}
			if(myDiskName != null) {
				capy += capdy;
				g.drawString(("hst = "+myDiskName),2,capy);
				}
		}

	static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
		
	public void paint(Graphics g) {
		//talk("This should never happen. Histogram.paint() called");
		paint2D((Graphics2D)g);
		}
		
	public void update(Graphics g) {
		// talk("Histogram.update() called. Should never happen");
		paint2D((Graphics2D)g);
		}

	public void update2D(Graphics2D g) {
		paint2D(g);
		}

	public void paint2D(Graphics2D g) {
		width = getSize().width;		
		height=getSize().height;
		if(ifMonochrome) g.setColor(Color.white);
		else g.setColor(Color.black);
		g.fillRect(0, 0, width, height);
		canvas.paint(g);
		whatYouCanDo.setVisible(true);
		whatYouCanDo.paint(g);
		}
}
	
class HistogramCanvas extends Canvas implements  java.awt.event.MouseListener,java.awt.event.MouseMotionListener {
	Histogram parent=null;
	Color newColor=Color.magenta;
	int oldParentWidth=0,  oldParentHeight=0;
	int howTall=256;
	int howWide=256;
	int mouseStartX = 0;
	int mouseStartY = 0;
	boolean forceRepaint=false;
	
	public HistogramCanvas(Histogram parent) {
		int i;
		this.parent=parent;
		setBackground(Color.black);
		setSize(howWide,howTall);
		setFont(parent.myFont);
		addMouseListener(this);
	    addMouseMotionListener(this);
   		}
   	
   	public void sepuku() {
   		parent=null;
   		}
   	   		
   	public void finalize() { 
   		talk("                                                                           HistogramCanvas snuffs"); 
   		}
   			
   	public void rescale(int howT, int howW) {
   		howTall=howT;
   		howWide=howW;
   		setSize(howWide,howTall);
   		}

	public void paint(Graphics g) {
		//talk("This should never happen. HistogramCanvas.paint() called");
		paint2D((Graphics2D)g);
		}
		
	public void update(Graphics g) {
		talk("HistogramCanvas.update() called. Should never happen");
		paint2D((Graphics2D)g);
		}

	public void update2D(Graphics2D g) {
		paint2D(g);
		}

	public void paint2D(Graphics2D g) {
	    if(parent.width != oldParentWidth || parent.height != oldParentHeight || forceRepaint) {
	    	forceRepaint=false;
			oldParentWidth = parent.width; 
			oldParentHeight = parent.height;
			if(!parent.ifMonochrome) g.setColor(Color.black);
			else g.setColor(Color.white);
			g.fillRect(0,0,parent.width,parent.height);
			}
		if(!parent.ifMonochrome) g.setColor(Color.yellow);
		else g.setColor(Color.black);
		g.drawLine(0,0,howWide-1,0);
		g.drawLine(howWide-1,0,howWide-1,howTall);
		g.drawLine(howWide-1,howTall,0,howTall);
		g.drawLine(0,howTall,0,0);
		parent.drawHistograms(g);
		}

	static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
		
	public void mouseClicked(java.awt.event.MouseEvent event) {
		//talk("Mouse Clicked at ("+event.getX()+", "+event.getY()+")");
		}
		
	public void mousePressed(java.awt.event.MouseEvent event) {
		mouseStartX = event.getX();
		mouseStartY = event.getY();
		}
		
	public void mouseReleased(java.awt.event.MouseEvent event) {	
		//talk("Mouse Released at ("+event.getX()+", "+event.getY()+")");
		parent.actOnMouseClick(mouseStartX,mouseStartY);
		}  
	
	public void mouseEntered(java.awt.event.MouseEvent event) {
		//talk("Mouse entered at ("+event.getX()+", "+event.getY()+")");
		}
		
	public void mouseExited(java.awt.event.MouseEvent event) {
		//talk("Mouse exited at ("+event.getX()+", "+event.getY()+")");
		}
		
	public void mouseMoved(java.awt.event.MouseEvent event){
		parent.tellLocation(event.getX(),event.getY());
		}
			
	public void mouseDragged(java.awt.event.MouseEvent event){
		int x=event.getX(),y=event.getY();
		//talk("Mouse Draged to ("+x+", "+y+")");
		}
}  
