package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.lang.*;
import java.util.*;
import java.awt.*;  import javax.swing.*;
import java.io.*;
import java.awt.image.*;

class LightSuckerClient implements Runnable {
	double noglow = 1.0; // the pixel brightness so low that it's not worth subtracting the glow from it.
		// make noglow larger to speed up the calculation.
	LeakedLight leakingSection=null;
	LeakedLight [] deconSection=null;
	int numSectionsToDeconvolve=-1;
	ConvolverInterface ConFace;
	PSFdata PSFD;
	int nicesleep;
	double mainParameter;
	
	// This class  never knows what the value of DPSF.ifConesUpsideDown is.
	 
	int leaksectnum,nSectsUp,nSectsDown;
	String taskList = "[]";

	int [] deconSectNum;
	int rgbchan=0;
	int npix,numDeconvolutionIterations;
	int raswidth,rasheight;

	boolean [] sectionNeeded;
	int sectsInStack;
	
	Thread unleakThread=null;
	String uniqueID="123";
	
	public LightSuckerClient(LightSuckerClientInterface LSCI, String uniqueID) {
		this.uniqueID = uniqueID;
		ConFace = new ConvolverInterface(LSCI,uniqueID);
		PSFD = ConFace.getPSFData();
		start();				 
		}
		
	public void sepuku() {
		stop(); // stop the run thread
		if(leakingSection !=null)
			leakingSection.sepuku();
		leakingSection=null;
		
		if(deconSection !=null)
			for(int i=0;i<numSectionsToDeconvolve;i++) {
				deconSection[i].sepuku();
				deconSection[i]=null;
				}
		deconSection=null;
		ConFace.sepuku();
		ConFace=null;
		} 
		
		
	 public void start() {
		if(unleakThread==null) {
			unleakThread = new Thread(this);
			}
		unleakThread.start();
		}
		
	public void stop() {
		if(unleakThread!=null) 
			unleakThread.stop();
		unleakThread = null;
		}

	private boolean getNextAssignment() {
		//  this is a place holder for a future version that asks 
		//  for the the Next Assignment from Deconvolver() 
		int leakSectLow,leakSectHi;
		int [] task;
		int i,j;
		if(ConFace.isGameOver())
			sepuku();
		task = ConFace.getAssignedSectionsToDeconvolve();
		if(task==null) {
			return false;
			}
		else {
			// we now have an assignment
			numSectionsToDeconvolve = task[0];
			deconSectNum = new int[numSectionsToDeconvolve];
			deconSection = new LeakedLight[numSectionsToDeconvolve];
			nSectsUp = PSFD.nSectsUp;
			nSectsDown = PSFD.nSectsDown;
			String stg = " gets assignment to deconvolve "+numSectionsToDeconvolve+" sections ";
			taskList = "[";
			for(i=0;i<numSectionsToDeconvolve;i++) {
				deconSectNum[i] = task[i+1];
				taskList = taskList +" "+task[i+1];
				deconSection[i] = fetchASection(deconSectNum[i],true,false);
				deconSection[i].zeroLeaksum();
				}
			taskList = taskList +"]";
			ConFace.report1(uniqueID+"-> "+stg);
			ConFace.DSDreport(stg+taskList);
			// establish sections that leak light into the deconSectNum[] sections to be deconvolved
			sectsInStack=ConFace.getSectsInStack();
			sectionNeeded = new boolean[sectsInStack];
			for(i = 0;i < sectsInStack;i++)
				sectionNeeded[i]=false;
			for(i=0;i<numSectionsToDeconvolve;i++) {
				if((leakSectLow =1 + deconSectNum[i] - nSectsUp) <0)
					leakSectLow=0;
			//deconSectNum - leakSectNum >0 uses nSectsUp and == nSectsUp when leaksect==leakSectLow
		
				if((leakSectHi = deconSectNum[i] + nSectsDown-1) >= sectsInStack)
					leakSectHi = sectsInStack-1;
				for(j=leakSectLow;j<=leakSectHi;j++)
					sectionNeeded[j]=true;
				}
			return true;
			}
		}
		
	private int nextLeakingSection() {
		for(int i=0;i<sectsInStack;i++)
			if(sectionNeeded[i]) {
				sectionNeeded[i]=false;
				return i;
				}
		return -1;
		}
	
	
		public void gatherLightOriginal() {
		int i, x,y,k,iz,numpoints,row,col;
		int [] dlx=null,dly=null;
		double theglow,lightRemoved=0;
		double [] gl=null;
		float [] sumlight;
		//gather light leaked to all the deconSect's by section number leaksectnum
			leakingSection = fetchASection(leaksectnum,true,false);
			for(k=0;k<numSectionsToDeconvolve;k++) {
				iz = leaksectnum - deconSectNum[k];
				sumlight = deconSection[k].leaksum;
			// leaksectnum starts at deconSectNum - DPSF.nSectsUp, so iz starts at - DPSF.nSectsUp
			//deconSectNum - leaksectnum >0 uses nSectsUp and == nSectsUp when leaksect==leakSectLow
				if(iz<0) {
					iz = -iz;
					if(iz < nSectsUp) {
						numpoints = PSFD.numPtsInUpSects[iz];
						dlx =  PSFD.delxZup[iz]; 
						dly =  PSFD.delyZup[iz]; 
						gl =  PSFD.glowZup[iz];
						}
					else numpoints=0;
					}
				else {
					if(iz < nSectsDown) {
						numpoints =  PSFD.numPtsInDownSects[iz];
						dlx =  PSFD.delxZdown[iz]; 
						dly =  PSFD.delyZdown[iz]; 
						gl =  PSFD.glowZdown[iz];
						}
					else numpoints=0;
					}
					
				if(numpoints>0) {
				  ConFace.DSDreport("computing light that leaked from sect "+leaksectnum+" into sect "+
				  			deconSectNum[k]+", taskList="+taskList);
				  noglow = ConFace.getNoGlow();
				  if(iz != 0)
		 			for(row=0; row<rasheight; row++) {
		 				for(col = 0; col<raswidth; ++col) { // row,col identifies the voxel in sectionThatLeaked
		 					if((theglow = leakingSection.origSect[row*raswidth+col]) > noglow) {
		 						theglow = theglow/(100000.*(double)(PSFD.numXYZpoints));
		 						for(i=0;i<numpoints;i++)
		 							if((x = col+dlx[i]) >=0 && x<raswidth && (y=row+dly[i]) >= 0 && y<rasheight) 
		 								sumlight[y*raswidth + x] += gl[i]*theglow;
		 						}
		 					}
						}
			 	  else // don't subtract a voxel from itself
		 			for(row=0; row<rasheight; row++) {
		 				for(col = 0; col<raswidth; ++col) { // row,col identifies the voxel in sectionThatLeaked
		 					if((theglow = leakingSection.origSect[row*raswidth+col]) > noglow) {
		 						theglow = theglow/(100000.*(double)(PSFD.numXYZpoints));
		 						for(i=0;i<numpoints;i++)
		 							if((x = col+dlx[i]) >=0 && x<raswidth && (y=row+dly[i]) >= 0 && y<rasheight &&
		 									x != col && y != row)
		 							sumlight[y*raswidth + x] += gl[i]*theglow;
		 						}
		 					}
						}
					try{Thread.sleep(nicesleep);}  catch(InterruptedException e) {;} 
					}
				  }	// end of for(k=0;k<numSectionsToDeconvolve;k++)  loop that gathered all light leaking from one
				  // section called leakingSection	 					
		 		leakingSection.sepuku();
		 		leakingSection=null;
		}
					


	public void gatherLightImproved() {  // this cryptic code speeds up the calculation by wasting some memory
		// the main improvement is bypassing the slow conditionals
		//  if((x = col+dlx[i]) >=0 && x<raswidth && (y=row+dly[i]) >= 0 && y<rasheight)  ...
		int i, x,y,k,iz,numpoints,row,col,rowoff;
		int [] dlx=null,dly=null;
		double theglow,lightRemoved=0,eentz = 1.0/(100000.*(double)(PSFD.numXYZpoints)) ;
		double [] gl=null;

		int overrows = rasheight+PSFD.sectionPicDimension;
		int overcols = raswidth+PSFD.sectionPicDimension;
		int overnpix = overrows*overcols;
		int overoff;
		float [] sumlight = new float[overnpix];
		float [] collector;
		int overoffset = -overcols*PSFD.miny - PSFD.minx;
		boolean ifFirstPass=true;
		// make the sumlight[] array large enough to accommodate border overruns.

		//gather light leaked to all the deconSect's by section number leaksectnum
			leakingSection = fetchASection(leaksectnum,true,false);
			for(k=0;k<numSectionsToDeconvolve;k++) { // this is the main loop which counts through all sections to be deconvolved
				iz = leaksectnum - deconSectNum[k];
				//old sumlight = deconSection[k].leaksum;
				for(i=0;i<overnpix;i++) // clear the sumlight array
					sumlight[i]=0;
			nicesleep = ConFace.getNiceSleep();
			// leaksectnum starts at deconSectNum - DPSF.nSectsUp, so iz starts at - DPSF.nSectsUp
			//deconSectNum - leaksectnum >0 uses nSectsUp and == nSectsUp when leaksect==leakSectLow
				if(iz<0) {
					iz = -iz;
					if(iz < nSectsUp) {
						numpoints = PSFD.numPtsInUpSects[iz];
						dlx =  PSFD.delxZup[iz]; 
						dly =  PSFD.delyZup[iz]; 
						gl =  PSFD.glowZup[iz];
						}
					else numpoints=0;
					}
				else { // this is also the iz==0 case
					if(iz < nSectsDown) {
						numpoints =  PSFD.numPtsInDownSects[iz];
						dlx =  PSFD.delxZdown[iz]; 
						dly =  PSFD.delyZdown[iz]; 
						gl =  PSFD.glowZdown[iz];
						}
					else numpoints=0;
					}
					
				if(numpoints>0) {
				  ConFace.DSDreport("computing light that leaked from sect "+leaksectnum+" into sect "+
				  			deconSectNum[k]+", taskList="+taskList);
				  noglow = ConFace.getNoGlow();
				  if(iz != 0) { // the leaking section is not the current section
		 			  for(row=0; row<rasheight; row++) {
		 				rowoff = row*raswidth;
		 				for(col = 0; col<raswidth; ++col) { // row,col identifies the voxel in sectionThatLeaked
		 					if((theglow = leakingSection.origSect[rowoff + col]) > noglow) {
		 						theglow = eentz*theglow;
		 						for(i=0;i<numpoints;i++)  
		 							sumlight[overoffset + (row+dly[i])*overcols + col+dlx[i]] += gl[i]*theglow;
		 								//old sumlight[y*raswidth + x] += gl[i]*theglow;
		 						}
		 					}
						   }
						}						
						
						
			 	  else {// The current section IS the leaking section; don't subtract a voxel from itself
		 			  for(row=0; row<rasheight; row++) {
		 				rowoff = row*raswidth;
		 				for(col = 0; col<raswidth; ++col) { // row,col identifies the voxel in sectionThatLeaked
		 					if((theglow = leakingSection.origSect[rowoff+col]) > noglow) {
		 						theglow = eentz*theglow;
		 						for(i=0;i<numpoints;i++) {
		 							if((x=col+dlx[i]) != col && (y=row+dly[i]) != row) // this line is only difference from iz != 0 loop
		 								sumlight[overoffset + y*overcols + x] += gl[i]*theglow;
		 						}
		 					}
						}
						
					  }
					 }
					  
			/* now add what's in the sumlight array, namely the light that leaked from section number leaksectnum into
			section number deconSectNum[k], to the deconSection[k].leaksum array */	
					collector = deconSection[k].leaksum;
		 			for(row=0; row<rasheight; row++) {
		 				rowoff = row*raswidth;
		 				overoff = overoffset + row*overcols;
		 				for(col = 0; col<raswidth; ++col)  // row,col identifies the voxel in sectionThatLeaked
  							collector[rowoff+col] += sumlight[overoff + col];
  						}
  						
					try{Thread.sleep(nicesleep);}  catch(InterruptedException e) {;} 
					}
				  }	// end of for(k=0;k<numSectionsToDeconvolve;k++)  loop that gathered all light leaking from one
				  // section called leakingSection	 					
		 		leakingSection.sepuku();
		 		leakingSection=null;
		}
					

	public void gatherLight() {  // this cryptic code speeds up the calculation by wasting some memory
		// the main improvement is bypassing the slow conditionals
		//  if((x = col+dlx[i]) >=0 && x<raswidth && (y=row+dly[i]) >= 0 && y<rasheight)  ...
		int i, x,y,k,iz,numpoints,row,col,rowoff;
		int [] dlx=null,dly=null;
		double theglow,lightRemoved=0,eentz = 1.0/(100000.*(double)(PSFD.numXYZpoints)) ;
		double [] gl=null;

		int overrows = rasheight+PSFD.sectionPicDimension;
		int overcols = raswidth+PSFD.sectionPicDimension;
		int overnpix = overrows*overcols;
		int overoff;
		float [] sumlight = new float[overnpix];
		float [] collector;
		int overoffset = -overcols*PSFD.miny - PSFD.minx;
		boolean ifFirstPass=true;
		// make the sumlight[] array large enough to accommodate border overruns.

		//gather light leaked to all the deconSect's by section number leaksectnum
			leakingSection = fetchASection(leaksectnum,true,false);
			for(k=0;k<numSectionsToDeconvolve;k++) { // this is the main loop which counts through all sections to be deconvolved
				iz = leaksectnum - deconSectNum[k];
				//old sumlight = deconSection[k].leaksum;
				for(i=0;i<overnpix;i++) // clear the sumlight array
					sumlight[i]=0;
			nicesleep = ConFace.getNiceSleep();
			// leaksectnum starts at deconSectNum - DPSF.nSectsUp, so iz starts at - DPSF.nSectsUp
			//deconSectNum - leaksectnum >0 uses nSectsUp and == nSectsUp when leaksect==leakSectLow
				if(iz<0) {
					iz = -iz;
					if(iz < nSectsUp) {
						numpoints = PSFD.numPtsInUpSects[iz];
						dlx =  PSFD.delxZup[iz]; 
						dly =  PSFD.delyZup[iz]; 
						gl =  PSFD.glowZup[iz];
						}
					else numpoints=0;
					}
				else { // this is also the iz==0 case
					if(iz < nSectsDown) {
						numpoints =  PSFD.numPtsInDownSects[iz];
						dlx =  PSFD.delxZdown[iz]; 
						dly =  PSFD.delyZdown[iz]; 
						gl =  PSFD.glowZdown[iz];
						}
					else numpoints=0;
					}
					
				if(numpoints>0) {
				  ConFace.DSDreport("computing light that leaked from sect "+leaksectnum+" into sect "+
				  			deconSectNum[k]+", taskList="+taskList);
				  noglow = ConFace.getNoGlow();
		 			  for(row=0; row<rasheight; row++) {
		 				rowoff = row*raswidth;
		 				for(col = 0; col<raswidth; ++col) { // row,col identifies the voxel in sectionThatLeaked
		 					if((theglow = leakingSection.origSect[rowoff + col]) > noglow) {
		 						theglow = eentz*theglow;
		 						for(i=0;i<numpoints;i++)  
		 							sumlight[overoffset + (row+dly[i])*overcols + col+dlx[i]] += gl[i]*theglow;
		 								//old sumlight[y*raswidth + x] += gl[i]*theglow;
		 						}
		 					}
						   }
									  
			/* now add what's in the sumlight array, namely the light that leaked from section number leaksectnum into
			section number deconSectNum[k], to the deconSection[k].leaksum array */	
					collector = deconSection[k].leaksum;
		 			for(row=0; row<rasheight; row++) {
		 				rowoff = row*raswidth;
		 				overoff = overoffset + row*overcols;
		 				for(col = 0; col<raswidth; ++col)  // row,col identifies the voxel in sectionThatLeaked
  							collector[rowoff+col] += sumlight[overoff + col];
  						}
  						
					try{Thread.sleep(nicesleep);}  catch(InterruptedException e) {;} 
					}
				  }	// end of for(k=0;k<numSectionsToDeconvolve;k++)  loop that gathered all light leaking from one
				  // section called leakingSection	 					
		 		leakingSection.sepuku();
		 		leakingSection=null;
		}
			
	
	public void run() {
		int i,x,y,k;
		
		while (true) { // endless loop
			while(!getNextAssignment()) {
				ConFace.DSDreport("awaiting next task");
				try{Thread.sleep(2000);}  catch(InterruptedException e) {;}
				talk(uniqueID+"-->  awaiting next task");
				}
				
			ConFace.DSDreport("got assignment to operate on "+numSectionsToDeconvolve+" sections");

			// as soon as getNextAssignment() returns true, there will be a deconvolution
			// task set up and ready to do.
			while((leaksectnum=nextLeakingSection()) >= 0) {
				while(ConFace.ifPaused()) {
					ConFace.DSDreport(uniqueID+" paused tasklist="+taskList);
					talk(uniqueID+"paused tasklist="+taskList);	
					try{Thread.sleep(2000);}  catch(InterruptedException e) {;} 
					}
					
				//talk(uniqueID+"call gatherLight()");	
				gatherLight();
				//talk(uniqueID+"returned from gatherLight()");	
				
				try{Thread.sleep(nicesleep);}  catch(InterruptedException e) {;} 
		 		}  // end of while((leaksectnum=nextLeakingSection()) >= 0) loop
		
			while(ConFace.ifPaused()) {
				talk(uniqueID+"sleeping through pause mode");	
				try{Thread.sleep(2000);}  catch(InterruptedException e) {;} 
				}
			// now return all the leaksum arrays to Deconvolver central
			ConFace.DSDreport("returning leaksum arrays for sections "+taskList); 
			talk(uniqueID+"returning leaksum arrays for sections "+taskList);	
			
			for(k=0;k<numSectionsToDeconvolve;k++) {
				deconSection[k].status = 2;
				talk(uniqueID+"--->returnLeaksumToDeconvolver("+k+")");	
				returnLeaksumToDeconvolver(k); // return the leaksum array to the Deconvolver()
				deconSection[k].sepuku(); // free storage
				deconSection[k]=null;}
			deconSection=null;
			numSectionsToDeconvolve=-1;
			try{Thread.sleep(nicesleep);}  catch(InterruptedException e) {;} 
			talk(uniqueID+"--->termination of one pass through endless loop");	
			
			}// end of endless loop
		}
	
	
		
	private LeakedLight fetchASection(int whichSection, boolean ifGetOrigSection,
					boolean ifGetLeaksum) {
			/* this is a place holder for what will eventually get data via a TCP/IP socket
				from a server that is running the only deconvolver(), which maintains the
				archival copies of all sections, and coordinates everything.  */
			int [] parms = ConFace.getSectionParams(whichSection);
			raswidth = parms[0];
			rasheight = parms[1];
			rgbchan =  parms[2];
			
			if(whichSection != parms[4]) {
				talk("************* terrible mistake occured. whichSection="+whichSection+
					" != (parms[4] = "+ parms[4]+")");
				}
			npix = rasheight*raswidth;
			mainParameter = ConFace.getMainParameter();

			LeakedLight retn = new LeakedLight(raswidth, rasheight, whichSection, rgbchan, mainParameter);
			retn.numDeconvolutionIterations = parms[3];
			if(ifGetOrigSection) 
				retn.origSect = ConFace.copyOrigSect(whichSection,npix);
				
			if(ifGetLeaksum) {
				retn.leaksum = ConFace.copyLeaksum(whichSection,npix);
				retn.status = parms[5];
				}
			return retn;
			}	
		
	private void returnLeaksumToDeconvolver(int k) {
		// this is a placeholder for a future version that will use a TCP/IP socket to return answers	
		ConFace.recordLightLeakedFromSection(deconSectNum[k],deconSection[k].leaksum,deconSection[k].status);
		}
					
	static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
}
