package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.io.*;
import java.lang.*;
import java.util.*;
import java.awt.*;  import javax.swing.*;
/* all requests from a LightSuckerClient object to the hub Deconvolver object come through this interface
which will eventually implement all transactions using TCPIP sockets */

class LightSuckerClientInterface {
	Deconvolver parent;
	boolean ifdead = false;
	int beat = 2000;
	
	public LightSuckerClientInterface(Deconvolver parent) {
		this.parent = parent;
		ifdead = false;
		talk("LightSuckerClientInterface Constructor makes another one...");		
		}
	
	public void sepuku() {
		parent=null;
		ifdead=true;
		}
	
	private long timeNow() {
		long milliSecondsElapsedSince = (new Date()).getTime();
		return milliSecondsElapsedSince/1000; // returns seconds
		}			
	
	private void report2(String stg) {
		parent.statusReport(stg);
		parent.report2(stg);
		}
			
	public synchronized void DSDreport(String stg, String uniqueID) {
		parent.DSDreport(stg,uniqueID); // no return
		}
	
	public synchronized void report1(String stg,String uniqueID) {
		report2(uniqueID+"->");
		parent.report1(stg);  // no return
		}
		
	public synchronized void recordLightLeakedFromSection(int deconSectNum, float [] leaksum, 
				int deconSectStatus, String uniqueID) {
		report2(uniqueID+"->recordLightLeakedFromSection("+deconSectNum+")");
		parent.logtalk(uniqueID+"->recordLightLeakedFromSection("+deconSectNum+")");
		parent.recordLightLeakedFromSection(deconSectNum,leaksum,deconSectStatus); 
		} 
		
	public synchronized  int [] getAssignedSectionsToDeconvolve(String uniqueID) {
		report2(uniqueID+"->getListOfSectionsToDeconvolve()");
		int [] answer = parent.getListOfSectionsToDeconvolve(); // returns an int[] array of unpredictavble size
		if(answer == null) 
			parent.logtalk("LightSuckerClientInterface.getAssignedSectionsToDeconvolve() from "+uniqueID+" returns null");
		else
			parent.logtalk("LightSuckerClientInterface.getAssignedSectionsToDeconvolve() from "+uniqueID+" returns list of "+answer[0]+" sections.");
		return answer;
		}
	
	public synchronized PSFdata getPSFData(String uniqueID) {
		report2(uniqueID+"->");
		return parent.DPSF.copyData(); // returns an entire data-filled object
		}

	public synchronized boolean isGameOver(String uniqueID) {
		return parent.isGameOver();
		}


	public synchronized int getNiceSleep(String uniqueID) {
		// report2(uniqueID+"->getNiceSleep()");
		return (1+parent.nicesleep); // returns one int
		}
		
	public synchronized int getSectsInStack(String uniqueID) {
		report2(uniqueID+"->getSectsInStack()");
		return parent.sectsInStack; // returns one int
		}
	
	public synchronized double getNoGlow(String uniqueID) {
		// report2(uniqueID+"->getNoGlow()");
		return parent.getNoGlow(); //  returns one double
		}

	public synchronized int ifPaused(String uniqueID) {
		// report2(uniqueID+"->ifPaused()");
		if(parent.ifPaused)
			return 1; 
		else return 0; // returns one int
		}
	
	public synchronized int [] getSectionParams(int whichSection,String uniqueID) {
		report2(uniqueID+"->getSectionParams("+whichSection+")");
		return parent.getSectionParams(whichSection); // an int[6] array
		}

	public double getMainParameter(String uniqueID) {
		report2(uniqueID+"->getMainParameter()");
		return parent.mainParameter; // returns one double
		}
	
	public synchronized void copyOrigSect(int whichSection,float [] aray,String uniqueID) {
		report2(uniqueID+"->copyOrigSect("+whichSection+")");
		parent.copyOrigSect(whichSection,aray); // will return a huge float[] array
		}

	public synchronized void copyLeaksum(int whichSection,float [] aray,String uniqueID) {
		report2(uniqueID+"->copyLeaksum("+whichSection+")");
		parent.copyLeaksum(whichSection,aray);  // will return a huge float[] array
		}
		
		
/* below are methods useful for packing anything into an int array 
   (whose first entry gives the length of the original array) and then recovering what
   was packed at the other end of a TCP/IP socket call.
   
	public static int floatToIntBits(floatvalue) returns in the int a bit representation of the float value
	and 
	public static float intBitsToFloat(intbits) rebuilds the float from the int.
	Thus the next method packs a float array into an int [] array whose first entry is the number of entries in
	the original array
	
 */
   
	private synchronized int [] floatArrayToIntArray(float [] flo) {
		int len = flo.length;
		int [] retn = new int[len+1];
		retn[0]=len;
		while(len > 0) {
			retn[len] = Float.floatToIntBits(flo[len-1]);
			len--;
			}
		return retn;
		}
	
	private synchronized float [] floatArrayFromIntArray(int [] inaray) {
		// this works only to unpack what floatArrayToIntArray() packs
		int len = inaray[0];
		float [] retn = new float[len];
		for(int i=0;i<len;i++)
			retn[i]= Float.intBitsToFloat(inaray[i+1]);
		return retn;
		}
	
	private synchronized int [] oneDoubleToIntBits(double dbl) {
		int [] retn = new int[2];
		retn[0]=1;
		retn[1] = Float.floatToIntBits((float)(dbl));
		return retn;
		}
		
	private synchronized double  oneDoubleFromIntArray(int [] inaray) {
		// this works only to unpack what oneDoubleToIntBits() packs
		return (double)(Float.intBitsToFloat(inaray[1]));
		}
	
	
	private synchronized int [] intArrayToIntArray(int [] nums) {
		int len = nums.length;
		int [] retn = new int[len+1];
		retn[0]=len;
		while(len > 0) {
			retn[len] = nums[len-1];
			len--;
			}
		return retn;
		}
	
	private synchronized int [] intArrayFromIntArray(int [] inaray) {
		// this works only to unpack what intArrayToIntArray() packs
		int len = inaray[0];
		int [] retn = new int[len];
		for(int i=0;i<len;i++)
			retn[i]= inaray[i+1];
		return retn;
		}
		
		
	static  synchronized void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
}