package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.lang.*;
import java.util.*;
import java.awt.*;  import javax.swing.*;

class LowPassFltr {
	SegmenterNode nd=null;

	public LowPassFltr(SegmenterNode nd) {
		this.nd=nd;
		}
		
	public void tellWhatYouDo() {
		nd.talk("                   ");
		nd.talk("The LowPassFltr has a single input raster and two parameters");
		nd.talk("called 'iter8' and 'weight'. It performs iter8 successive iterations of");
		nd.talk("a standard LaPlace-operator spatial averaging filter.");
		nd.talk("In each of the iter8 passes, each pixel is replaced by ");
		nd.talk("((255*itsVal + weight*(sum of pixel values in 8 nearest neighbors))/denom");
		nd.talk("  where denom = 255 + 8*weight");
		nd.talk("To make the neighbors have smaller influence, reduce weight toward 0");
		nd.talk("To make the sum of all neighbors have the same weight as the central");
		nd.talk("                 pixel, set weight=255/8 = 32");
		nd.talk("To make EACH neighbor count as much as the central pixel, set weight=255.");
		}
				
public boolean doit(short [] inras, int incols, int inrows, int weight, int npass, short [] outras) {
		int npix = incols*inrows;
		int row,col,sidemass,rowcoloff=0,k,pass;
		int [] offsets = {-1-incols, -incols, -incols+1,
		                     -1,                 1,
		                    incols-1,  incols,  incols+1};
		int denom = 255 + 8*weight; 
		int rowlim = inrows-1;
		int collim=incols-1; 
		short [] workin = new short[npix]; // temporary work space
		short [] temp;
		// if(nd.diags) nd.talk("LowPassFltr with weight="+weight+" and npass="+npass);
		for(row=0;row<npix;row++) outras[row]=inras[row]; // copy inras to outras
			// so after first swap we ger the right input
		for(pass=0;pass<npass;pass++) {
		  temp = outras;
		  outras = workin;
		  workin = temp;  
		  for(row=1;row<rowlim;row++) {
			for(col=1;col<collim;col++) {
				rowcoloff=row*incols + col;
				sidemass=0;
				for(k=0;k<8;k++)
					sidemass += workin[rowcoloff + offsets[k]];
				outras[rowcoloff] = (short)((255*workin[rowcoloff] + weight*sidemass)/denom);
			}}}	
		return true;	
		}
		
}
