package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.lang.*;
import java.util.*;
import java.awt.*;  import javax.swing.*;

class OneHistogram extends JFrame{ //This class computes and holds single histograms
	int width=256,height=256;
	int thresh  = 0; // all intensities less than this are ignored.
	int [] hist = null;
	int [] normalHist = null;
	double [] normalTotalsHist=null;
	double maxTotal=0;
	int Nbright = 1; // number of pixels having intensity greater than thresh.
	double mean=128;
	int median=128;
	int Nblazing=0;
	int captionX = 2;
	int maxbright=10;
	int ihistmax=0;
	double stdDev=1.;
	boolean hasHistogramChanged=true;
	int currentNumLowPasses=0;
	Color mycolor = Color.red;
	HistogramBox myBox=null;
	int [] temp = new int[256];
	
	public boolean hasAnythingChanged() {
		if(hasHistogramChanged || 
			currentNumLowPasses != myBox.numLowPassIterations) {
				return true;
				}
		return false;
		}
		
	public void sepuku() {
		hist = null;
		normalHist = null;
		normalTotalsHist=null;
		myBox=null;
		}
	
	public void finalize() {
		talk("                                                                           OneHistogram snuffs");
		}
		
	public OneHistogram(HistogramBox myBox,int thresh, Color crayon, int captionX) {
		this.thresh = thresh;
		this.captionX = captionX;
		this.myBox=myBox;
		mycolor = crayon;
		hist = new int[256];
		normalHist = new int[256];
		normalTotalsHist = new double[257]; // last entry holds maxTotal
		for(int i=0;i<256;i++) hist[i]=1;
		hasHistogramChanged=true;
		}
	
	public void makeNewHistogram(byte [] ras) {
		int i,pix,npix;
		int [] xy={0,0};
		int fullWidth=myBox.picJPanel.fullWidth;
	    for(i=0;i<256;i++)
		  	hist[i]=0;
		npix=0;
		myBox.resetVoxCycle();
		while(myBox.voxCycle(xy)) {
			++npix;
			if((pix = ras[xy[1]*fullWidth+xy[0]]) <0) 
				pix += 256;
			++hist[pix];
			}	 
		hasHistogramChanged=true;
	    }
	
	public double getMean() {
		if(hasAnythingChanged() ) 
			getNormalizedHistogram();
		return mean;
		}
				
	public double getStdDev() {
		if(hasAnythingChanged()) 
			getNormalizedHistogram();
		return stdDev;
		}
				
	public int [] getRawHST() {
		return hist;
		}
	
	private double truncate(double xx) {
		int i = (int)(10.0*xx);
		return (i/10.0);
		}		
			
	public void draw(Graphics g, String fileName, String hstName) { // draw the histogram
		int i,j,intavg;
		int x,y,xn=0, yn, capdy=15,capy;
		if(hasAnythingChanged()) 
			getNormalizedHistogram();
		if(myBox.ifMonochrome)
			g.setColor(Color.black);
		else
			g.setColor(mycolor);
		yn=255-normalHist[0];
		for(i=1;i<256;i++) {
			x=xn;	y=yn; 
			xn=i;	yn=255-normalHist[i];
			g.drawLine(x,y,xn,yn);
			}
			
		yn=(int)(255-normalTotalsHist[0]);  xn=0;
		for(i=1;i<256;i++) {
			x=xn;	y=yn; 
			xn=i;	yn=(int)(255-normalTotalsHist[i]);
			g.drawLine(x,y,xn,yn);
			}
		x=(int)(mean+.499);
		g.drawLine(x,260,x,200); // draw vertical line at mean
		// print some statistics
		capy=256+capdy;
		g.drawString("TBP = "+(int)maxTotal,captionX,capy);
		capy += capdy;
		g.drawString("area="+myBox.NPixInHistobox,captionX,capy);
		capy += capdy;
		g.drawString("avg Probe/Bright",captionX,capy);
		capy += capdy;
		intavg = (int)(100.0*(maxTotal/Nbright));
		g.drawString("   pixel="+(.01*(double)intavg),captionX,capy);
		capy += capdy;
		g.drawString("peak at "+ihistmax,captionX,capy);	
		capy += capdy;
		g.drawString(hist[ihistmax]+" pix@peak",captionX,capy);	
		capy += capdy;
		g.drawString("std dev="+stdDev,captionX,capy);
		capy += capdy;
		g.drawString(truncate(100.0*(double)Nbright/(double)myBox.NPixInHistobox)+"% pix >",captionX,capy);
		capy += capdy;
		g.drawString("   dim thresh="+thresh,captionX,capy);
		capy += capdy;
		g.drawString(Nblazing+" pix > 250",captionX,capy);	
		capy += capdy;
		g.drawString(("file="+fileName),captionX,capy);
		if(hstName != null) {
				capy += capdy;
				g.drawString(("hst = "+hstName),2,capy);
				}
		capy += capdy;
		g.drawString(("lowPassIter8= "+currentNumLowPasses),captionX,capy);
		}

	static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
		
	public double [] getnormalTotalHist() {
		getNormalizedHistogram();
		return 	normalTotalsHist;
		}
		
	public int [] lowPassFilter(int [] hst) {
		currentNumLowPasses=myBox.numLowPassIterations;
		int [] retn = new int[256];
		switch (currentNumLowPasses) {
			case 0:
			default:
				for(int i=0;i<256;i++)
					retn[i] = hst[i];					
				break;
			case 1: // one low pass
				oneLowPass(hst,retn);  
				break;				
			case 2: // two low passes
				oneLowPass(hst,temp);
				oneLowPass(temp,retn);
				break;				
			case 3:  // three low passes
				oneLowPass(hst,retn);    
				oneLowPass(retn,temp);  
				oneLowPass(temp,retn);  // 3 passes
				break;
			case 4: // four low passes
				oneLowPass(hst,temp);   
				oneLowPass(temp,retn);
				oneLowPass(retn,temp);  
				oneLowPass(temp,retn);
				break;				
			}
		return retn;				
		}

		private void oneLowPass(int [] hin,int [] hout) {
		int i;
		hout[0]=(int)(.499 + (2.0*hin[0]+hin[1])/3.0);
		hout[255] = (int)(.499 + (2.0*hin[255]+hin[254])/3.0);
		for(i=1;i<255;i++)
			hout[i] = (int)(.499 + (hin[i-1]+hin[i]+hin[i+1])/3.00);
		}	
			
	public int [] getNormalizedHistogram() {
		int i,sum,histi,j;
		double dev;
		if(!hasAnythingChanged()) return normalHist;
		Nbright=0;  sum=0;  maxbright=0; Nblazing=0; ihistmax=0;
		int [] lphist = lowPassFilter(hist);
		for(i=thresh;i<256;i++) { // find the max value greater than thresh
			histi=lphist[i];
			if(i>thresh) { 
				sum += histi*i;
				Nbright += histi;			
				if(histi>maxbright)  { 
					maxbright=histi;
					ihistmax=i;
					}
				}
			}
		if(Nbright<=0) Nbright=1;
		mean = (double)sum/(double )Nbright;
		stdDev = 0;
		Nblazing=0;
		normalTotalsHist[0] = 0.;
		maxTotal=0.;
		for(i=0;i<256;i++) {
			if(i<thresh) {
				normalHist[i]=0;
				normalTotalsHist[i] = 0.;
				}
			else {
				normalHist[i] = 
					(int )(.499 + (255.0*lphist[i])/(double )maxbright);
				normalTotalsHist[i] = maxTotal + lphist[i]*i;
				maxTotal = normalTotalsHist[i];
				if(i>250) Nblazing += lphist[i];
				dev = i - mean;
				stdDev += lphist[i]*dev*dev;
				}
			}
		for(i=0;i<256;i++) 
			normalTotalsHist[i] = (255.*normalTotalsHist[i])/(maxTotal);
			
		normalTotalsHist[256] = maxTotal;
		
		j=(int)(10.*mean);
		mean = j/10.;
		stdDev = Math.sqrt(stdDev/(Nbright-1.)); 
		j=(int)(10.*stdDev);
		stdDev = j/10.;
		hasHistogramChanged=false;
		return normalHist;	
		}	
}  
