package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.io.*;
import java.lang.*;
import java.util.*;
import java.awt.*;  import javax.swing.*;

class PickColors extends JFrame implements java.awt.event.ActionListener,java.awt.event.ItemListener,java.awt.event.WindowListener {
	ShowPicJPanel picJPanel;
	int nomwidth=250, nomheight=168;
	JPanel redPan,greenPan, bluePan, rednessPan,greennessPan,bluenessPan,
			colorJPanel, artTogglePan,togglePan,gainCurveToggleJPanel;
	int [] redColors = {255,0,0};
	int [] greenColors = {0,255,0};
	int [] blueColors = {0,0,255};
	Color myBlue = 	new Color(100,100,255);

	boolean ifReturnStdPrimaries=true;
	JComboBox rawDataXformChoice;
	Color bblue = new Color(90,  90 , 255);
	JButton setRed, setGreen,  setBlue,rlut,glut,blut,toggleStd,rampBoostToggle;
	CustomRGBLUT custLUT;
	StackGradient stackramp=null;
		
	JTextField rredText, rgreenText,  rblueText;
	JTextField gredText, ggreenText,  gblueText;
	JTextField bredText, bgreenText,  bblueText;

	boolean stackGradientOn = false;
	
	public double [] getBoosts(int sectionNum) {
		if(!stackGradientOn || stackramp==null) {
			double [] boosts = new double[3];
			boosts[0]=1.0; boosts[1]=1.0; boosts[2]=1.0;
			return boosts; // this flags that there is no boosting.
			}
		else 
			return (stackramp.getBoosts(sectionNum));
		}
		
    public PickColors(ShowPicJPanel picJPanel) {   
        super("Pick Colors");
		WidgetMaker wmk = new WidgetMaker(this);
		this.picJPanel = picJPanel;
        getContentPane().setLayout(new BorderLayout());    // NEED to specify a layout manager!!
    	setSize(nomwidth, nomheight);
		setBackground(Color.black);
		colorJPanel = new JPanel();
		colorJPanel.setLayout(new GridLayout(0, 1));
	
		gainCurveToggleJPanel=new JPanel();
		gainCurveToggleJPanel.setLayout(new GridLayout(1,2));
		gainCurveToggleJPanel.setBackground(Color.lightGray);
				
		rawDataXformChoice = wmk.makeAJComboBox(new String [] {"Use raw data",
					"Transform raw data by gain curve",
					"Ungain/Unbleach raw data"});
		gainCurveToggleJPanel.add(rawDataXformChoice);
		rampBoostToggle = wmk.makeAJButton("activateStackGradient");
		gainCurveToggleJPanel.add(rampBoostToggle);
		
		artTogglePan=new JPanel();
		artTogglePan.setLayout(new BorderLayout());
		artTogglePan.setBackground(Color.lightGray);
		toggleStd = wmk.makeAJButton("Using RGB standard primaries (push to toggle)");
		artTogglePan.add("Center", toggleStd);
		
		togglePan=new JPanel();
		togglePan.setLayout(new GridLayout(1,3));
		togglePan.setBackground(Color.lightGray);
		togglePan.add(wmk.makeAJButton("gray toggle"));
		togglePan.add(wmk.makeAJButton("all 3 colors"));		
		togglePan.add(wmk.makeAJButton("redraw"));
	
		redPan = new JPanel();
		redPan.setLayout(new GridLayout(1,5));
		rredText = wmk.makeAJTextField("255", 4,wmk.lightred);
		rgreenText = wmk.makeAJTextField("0", 4,wmk.lightgreen);
		rblueText = wmk.makeAJTextField("0", 4,wmk.lightblue);
	setRed = wmk.makeAJButton("set red",Color.red);
	rlut = wmk.makeAJButton("r LUT",Color.red);
	redPan.add(rredText); redPan.add(rgreenText);  redPan.add(rblueText); 
	redPan.add(setRed);  redPan.add(rlut);
	rednessPan = new JPanel();
	rednessPan.setLayout(new GridLayout(1,4));
	rednessPan.add(wmk.makeAJButton("red only",Color.red));
	rednessPan.add( wmk.makeAJButton("red off",Color.red));
	rednessPan.add(wmk.makeAJButton("rLUT=gLUT",Color.red));
	rednessPan.add(wmk.makeAJButton("rLUT=bLUT",Color.red));


	greenPan = new JPanel();
	greenPan.setLayout(new GridLayout(1,5));
	greenPan.setBackground(Color.green);
	gredText = wmk.makeAJTextField("0", 4,wmk.lightred);
	ggreenText = wmk.makeAJTextField("255", 4,wmk.lightgreen);
	gblueText = wmk.makeAJTextField("0", 4,wmk.lightblue);
	setGreen = wmk.makeAJButton("set green",Color.green);
	glut=wmk.makeAJButton("g LUT",Color.green);
	greenPan.add(gredText); greenPan.add(ggreenText);  greenPan.add(gblueText); 
	greenPan.add(setGreen);  greenPan.add(glut); 

	greennessPan = new JPanel();
	greennessPan.setLayout(new GridLayout(1,4));
	greennessPan.setBackground(Color.green);
	greennessPan.add(wmk.makeAJButton("green only",Color.green));
	greennessPan.add( wmk.makeAJButton("green off",Color.green));
	greennessPan.add(wmk.makeAJButton("gLUT=rLUT",Color.green));
	greennessPan.add(wmk.makeAJButton("gLUT=bLUT",Color.green));


	bluePan = new JPanel();
	bluePan.setLayout(new GridLayout(1,5));
	bluePan.setBackground(bblue);
	bredText = wmk.makeAJTextField("0", 4,wmk.lightred);
	bgreenText = wmk.makeAJTextField("0", 4,wmk.lightgreen);
	bblueText = wmk.makeAJTextField("255", 4,wmk.lightblue);
	setBlue = wmk.makeAJButton("set blue",myBlue);
	blut=wmk.makeAJButton("b LUT",myBlue);
	bluePan.add(bredText); bluePan.add(bgreenText);  bluePan.add(bblueText); 
	bluePan.add(setBlue);  bluePan.add(blut);

	bluenessPan = new JPanel();
	bluenessPan.setLayout(new GridLayout(1,4));
	bluenessPan.setBackground(bblue);
	bluenessPan.add(wmk.makeAJButton("blue only",myBlue));
	bluenessPan.add( wmk.makeAJButton("blue off",myBlue));
	bluenessPan.add(wmk.makeAJButton("bLUT=rLUT",myBlue));
	bluenessPan.add(wmk.makeAJButton("bLUT=gLUT",myBlue));


	colorJPanel.add(redPan);	
	colorJPanel.add(greenPan);	
	colorJPanel.add(bluePan);
	colorJPanel.add(artTogglePan);
	colorJPanel.add(togglePan);
	colorJPanel.add(rednessPan);	
	colorJPanel.add(greennessPan);	
	colorJPanel.add(bluenessPan);	
	colorJPanel.add(gainCurveToggleJPanel);
	getContentPane().add("Center", colorJPanel);
    JMenu overlayJMenu = new JMenu("colors to/from files");
	overlayJMenu.add(wmk.makeAJMenuItem("write art colors to disk file"));
	overlayJMenu.add(wmk.makeAJMenuItem("read art colors from disk file."));	
	JMenuBar mb = new JMenuBar();
	mb.add(overlayJMenu);
	setJMenuBar(mb);
	addWindowListener(this);
	// initialize the custom LUT editor
	custLUT = new CustomRGBLUT();
  }  

		 
// catch WindowListener events  
	public void windowActivated(java.awt.event.WindowEvent event) {
		// talk("window activated");
		}
	public void windowDeactivated(java.awt.event.WindowEvent event) {
		//  talk("window Deactivated");
		}
	public void windowOpened(java.awt.event.WindowEvent event) {}
	public void windowIconified(java.awt.event.WindowEvent event) {}
	public void windowDeiconified(java.awt.event.WindowEvent event) {}
	public void windowClosing(java.awt.event.WindowEvent event) {
	    setVisible(false);
		}
	public void windowClosed(java.awt.event.WindowEvent event) {
		}
	 
	public void archiveColors(ArchiveMarks disker,boolean ifWrite) {
		int i;
		double badval = 7777.777;
		int [] priLUT;
		if(ifWrite) { // write art color info to disk
			for(i=0;i<3;i++) disker.writeNum(redColors[i],(" = redColors["+i+"]"));
			for(i=0;i<3;i++) disker.writeNum(greenColors[i],(" = greenColors["+i+"]"));
			for(i=0;i<3;i++) disker.writeNum(blueColors[i],(" = blueColors["+i+"]"));
			priLUT = getRedLUT();
			for(i=0;i<255;i+=16)
			    disker.write16Ints(priLUT, i,  ("priRedLUT["+i+"]"));
			priLUT = getGreenLUT();
			for(i=0;i<255;i+=16)
			    disker.write16Ints(priLUT, i,  ("priGreenLUT["+i+"]"));
			priLUT = getBlueLUT();
			for(i=0;i<255;i+=16)
			    disker.write16Ints(priLUT, i,  ("priBlueLUT["+i+"]"));
			return;
			}
		else { // recover art color info from disk and set text fields 
			for(i=0;i<3;i++) redColors[i] = (int)disker.getNextFileDouble(badval);
			for(i=0;i<3;i++) greenColors[i] = (int)disker.getNextFileDouble(badval);
			for(i=0;i<3;i++) blueColors[i] = (int)disker.getNextFileDouble(badval);
			priLUT=new int[256];
			for(i=0;i<255;i+=16)
				disker.read16Ints(priLUT, i);
			custLUT.acceptNewRedLUT(priLUT);
			for(i=0;i<255;i+=16)
			    disker.read16Ints(priLUT, i);
			custLUT.acceptNewGreenLUT(priLUT);
			for(i=0;i<255;i+=16)
			    disker.read16Ints(priLUT, i);
			custLUT.acceptNewBlueLUT(priLUT);

			setColors();
			getAllColors();
			}
		}
		
    private void setColors() {
	for(int i=0;i<3;i++) {
	    if(redColors[i]<0) redColors[i]=0;
	    if(redColors[i]>255) redColors[i]=255;
	    if(greenColors[i]<0) greenColors[i]=0;
	    if(greenColors[i]>255) greenColors[i]=255;
	    if(blueColors[i]<0) blueColors[i]=0;
	    if(blueColors[i]>255) blueColors[i]=255;
	    }
	rredText.setText((""+redColors[0]));
	rgreenText.setText((""+redColors[1]));
	rblueText.setText((""+redColors[2]));
			
	gredText.setText((""+greenColors[0]));
	ggreenText.setText((""+greenColors[1]));
	gblueText.setText((""+greenColors[2]));
			
	bredText.setText((""+blueColors[0]));
	bgreenText.setText((""+blueColors[1]));
	bblueText.setText((""+blueColors[2]));
	
	Color curcol=new Color(redColors[0],redColors[1],redColors[2]);
	Color compcol=new Color(255-redColors[0],255-redColors[1],255-redColors[2]);
	setRed.setBackground(curcol); setRed.setForeground(compcol); setRed.setLabel("set red");
	rlut.setBackground(curcol); rlut.setForeground(compcol); rlut.setText("r LUT");
	
	curcol=new Color(greenColors[0],greenColors[1],greenColors[2]);
	compcol=new Color(255-greenColors[0],255-greenColors[1],255-greenColors[2]);
	setGreen.setBackground(curcol); setGreen.setForeground(compcol); setGreen.setText("set green");
	glut.setBackground(curcol); glut.setForeground(compcol); glut.setText("g LUT");
	
	curcol=new Color(blueColors[0],blueColors[1],blueColors[2]);
	compcol=new Color(255-blueColors[0],255-blueColors[1],255-blueColors[2]);
	setBlue.setBackground(curcol); setBlue.setForeground(compcol); setBlue.setText("set blue");
	blut.setBackground(curcol); blut.setForeground(compcol); blut.setText("b LUT");
	}
	
    public int [] getRedLUT() {
	return custLUT.getRedLUT();
	}

    public int [] getGreenLUT() {
	return custLUT.getGreenLUT();
	}

    public int [] getBlueLUT() {
	return custLUT.getBlueLUT();
	}


  public int [] getRedsColors() {
	if(ifReturnStdPrimaries) {
	   int [] truered = {255, 0, 0};
	   return truered;
	    }
	else return (redColors);
	}
	
    public int [] getGreensColors() {
	if(ifReturnStdPrimaries) {
	   int [] truegreen = {0, 255, 0};
	   return truegreen;
	    }
	else return(greenColors);
	}
    public int [] getBluesColors() {
	if(ifReturnStdPrimaries) {
	   int [] trueblue = {0, 0, 255};
	   return trueblue;
	    }
	else return(blueColors);
	}
	
    private void getTheRedColors() {
	redColors[0] = (int)(Double.valueOf(rredText.getText()).doubleValue());	
	redColors[1] = (int)(Double.valueOf(rgreenText.getText()).doubleValue());	
	redColors[2] = (int)(Double.valueOf(rblueText.getText()).doubleValue());
	}
	
    private void getTheGreenColors() {
	greenColors[0] = (int)(Double.valueOf(gredText.getText()).doubleValue());	
	greenColors[1] = (int)(Double.valueOf(ggreenText.getText()).doubleValue());	
	greenColors[2] = (int)(Double.valueOf(gblueText.getText()).doubleValue());
	}

    private void getTheBlueColors() {
	blueColors[0] = (int)(Double.valueOf(bredText.getText()).doubleValue());	
	blueColors[1] = (int)(Double.valueOf(bgreenText.getText()).doubleValue());	
	blueColors[2] = (int)(Double.valueOf(bblueText.getText()).doubleValue());
	}

// catch button events...
	public void actionPerformed(java.awt.event.ActionEvent event) {
		String arg = event.getActionCommand();
 		picJPanel.showYourself();       
		if(arg.equals("write art colors to disk file"))  {	
	    	ArchiveMarks disker = new ArchiveMarks();
			disker.readWriteColors(true,this);
			disker=null; 
		  	return;
		  	}
		  
		if(arg.equals("read art colors from disk file."))  {
	    	ArchiveMarks disker = new ArchiveMarks();
			disker.readWriteColors(false,this);
			disker=null; 
		  	return;
		  	}


	    if(arg.equals("activateStackGradient")) {
	    	stackGradientOn = true;
	    	rampBoostToggle.setText("TurnOffStackGradient");	 
			if(stackramp==null)
				stackramp = new StackGradient();
			stackramp.setVisible(true);
	    	return;
	    	}
	    
	    if(arg.equals("TurnOffStackGradient")) {
	    	stackGradientOn = false;
	    	rampBoostToggle.setText("activateStackGradient");
			if(stackramp != null)
				stackramp.setVisible(false);
	    	return;
	    	}
	    
	    if(arg.equals("set red")) {
	    	getTheRedColors();
	    	new PainterColorPicker(350,200,"redChanColors",this,
	    		redColors[0],redColors[1],redColors[2],rredText,rgreenText,rblueText);
	    	}

	    if(arg.equals("set green")) {
	    	getTheGreenColors();
	    	new PainterColorPicker(350,320,"greenChanColors",this,
	    		greenColors[0],greenColors[1],greenColors[2],gredText,ggreenText,gblueText);
	    	}
		
	    if(arg.equals("set blue")) {
			getTheBlueColors();
	    	new PainterColorPicker(350,440,"blueChanColors",this,
	    		blueColors[0],blueColors[1],blueColors[2],bredText,bgreenText,bblueText);
	    	}
    		
	    if(arg.equals("r LUT")) {
		custLUT.setVisible(true);
		custLUT.getNewLUT(0);
 		picJPanel.forceNewDrawing();
    		}

		
	    if(arg.equals("g LUT")) {
		custLUT.setVisible(true);
		custLUT.getNewLUT(1);
 		picJPanel.forceNewDrawing();
    		}
		
	    if(arg.equals("b LUT")) {
		custLUT.setVisible(true);
		custLUT.getNewLUT(2);
 		picJPanel.forceNewDrawing();
    		}
    		
	    if(arg.equals("gray toggle")) {
 			picJPanel.forceNewDrawing();
    		picJPanel.grayToggle();
    		}

	    if(arg.equals("all 3 colors")) {
    		picJPanel.allColors();
    		forcedRedraw();
    		}

	    if(arg.equals("redraw")) 
 			forcedRedraw();
    		
    	if(arg.equals("red only")) {picJPanel.redOnly();}
    	if(arg.equals("red off")) {picJPanel.redOff();}
    	if(arg.equals("green only")) {picJPanel.greenOnly();}
    	if(arg.equals("green off")) {picJPanel.greenOff();}
   		if(arg.equals("blue only")) {picJPanel.blueOnly();}
    	if(arg.equals("blue off")) {picJPanel.blueOff();}
    			
   		if(arg.equals("rLUT=gLUT")) {custLUT.transferLUT(1,0);forcedRedraw();}
   		if(arg.equals("rLUT=bLUT")) {custLUT.transferLUT(2,0);forcedRedraw();}   			
   		if(arg.equals("gLUT=rLUT")) {custLUT.transferLUT(0,1);forcedRedraw();}   			
   		if(arg.equals("gLUT=bLUT")) {custLUT.transferLUT(2,1);forcedRedraw();}
		if(arg.equals("bLUT=rLUT")) {custLUT.transferLUT(0,2);forcedRedraw();}   			
   		if(arg.equals("bLUT=gLUT")) {custLUT.transferLUT(1,2);forcedRedraw();} 
   			
 	    if(arg.equals("Using RGB standard primaries (push to toggle)")) {
			ifReturnStdPrimaries = false;
			toggleStd.setText("Using 3 art colors above (push to toggle)");
			forcedRedraw();
    		}
	    
	    if(arg.equals("Using 3 art colors above (push to toggle)")) {
			ifReturnStdPrimaries = true;
			toggleStd.setText("Using RGB standard primaries (push to toggle)");;
			forcedRedraw();
    		}
	}
	
	public void getAllColors() {
		//talk("PickColors.getAllColors()");
		getTheRedColors();
		getTheGreenColors();
		getTheBlueColors();
		forcedRedraw();
		setColors();
		}
		
		
	public void forceRGBPrimaryMode() {
			ifReturnStdPrimaries = true;
			toggleStd.setText("Using RGB standard primaries (push to toggle)");
    		}

//catch JComboBox events...	
	public void itemStateChanged(java.awt.event.ItemEvent event)
		{
		//String arg = (String)(event.getItem());
		String arg = (String)(((JComboBox)(event.getSource())).getSelectedItem());

      	if(arg.equals("Use raw data")) {
			picJPanel.assertRawDataTransformationMode(0);
			forcedRedraw();
	    	}
      	if(arg.equals("Transform raw data by gain curve")) {
			picJPanel.assertRawDataTransformationMode(1);
			forcedRedraw();
	    	}
      	if(arg.equals("Ungain/Unbleach raw data")) {
			picJPanel.assertRawDataTransformationMode(2);
			forcedRedraw();
	    	}	    
		}

    
    private void forcedRedraw() {
 		picJPanel.forceNewDrawing();
 		picJPanel.reDraw();
   		}

	public void changeJPanelSize(double ratio) {
		int pwidth = getSize().width;
		int pheight = getSize().height;
		pwidth = (int)(ratio*pwidth +.4999);
		pheight = (int)(ratio*pheight +.4999);
		setSize(pwidth,pheight);
		show();
		}

static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
}

