package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.awt.*;  import javax.swing.*;
import java.applet.*;
import java.util.*; 
import java.io.*;

class Query extends JFrame{
	static QueryJPanel mQP=null;  
	boolean answer=false;	
	
	Query(String question, String trueAns, String falseAns,
			AnsBuf theans,Color backcol) {
		// this one sets up a yes-or-no query
		int width = 15*(5+question.length());
		int i = 7*(trueAns.length() + falseAns.length());
		if(i>width) width=i;
		width += 30;
		getContentPane().setLayout(new FlowLayout());
		setBackground(backcol);
		setTitle(question);
		mQP = new QueryJPanel(width,60);
		mQP.initialize(trueAns,falseAns,theans,backcol);
		getContentPane().add(mQP);
		resize(width,60);
		move(100,120);
		show();
		mQP.repaint();
		}

	Query(String question, AnsBuf theans,Color backcol) {
		// this one sets up for a query demanding a double or a string
		int width = 15*(5+question.length());
		int i = 7*(25);
		if(i>width) width=i;
		i += 30;
		getContentPane().setLayout(new FlowLayout());
		setBackground(backcol);
		setTitle(question);
		mQP = new QueryJPanel(width,80);
		mQP.initialize(theans,backcol);
		getContentPane().add(mQP);
		resize(width,80);
		move(100,120);
		show();
		mQP.repaint();
		}
		
	public boolean handleEvent(Event evtObj) {
		if(evtObj.id == Event.WINDOW_DESTROY) {
			dispose();
			return true;
			}
		return super.handleEvent(evtObj);
		} 
	
	public void off() {
		mQP=null;
		dispose();
		}	
}

class QueryJPanel extends JPanel{
	JButton yesJButton,noJButton,okJButton;
	AnsBuf theans=null;
    JTextField textInput = null;
	int iwh=0; /*   	1 means get true/false when the yesJButton/noJButton is pushed
					2 means get a double value from the text field when okJButton is pushed
					3 means get a String value from the text fiels when the okJButton is pushes 
					4 means get a long value from the text field when the okJButton is pushed  */
					
	
	QueryJPanel(int width,int height) {
		resize(width,height);
		requestFocus();
		}

	public void initialize(String trueAns, String falseAns,
		AnsBuf thns,Color backcol) { // this is for a yes/no query
		theans=thns;
		iwh = theans.iwh;
		setLayout(new FlowLayout());
		setBackground(backcol);
		add((noJButton = new JButton(falseAns)));
		add((yesJButton = new JButton(trueAns)));
		}
		
	public void initialize(AnsBuf thns, Color backcol) {
		theans=thns; // this is for a query demanding a String or a double or a long answer
		iwh = theans.iwh;
		setLayout(new FlowLayout());
		setBackground(backcol);
		if(iwh==2) {
			textInput = new JTextField(""+theans.doubleAns,16);			
			okJButton = new JButton("click to enter");
			}
		if(iwh==3)  {
			textInput = new JTextField("",36);			
			okJButton = new JButton("click to enter");
			}
		if(iwh==4)  {
			textInput = new JTextField(""+theans.longAns,16);			
			okJButton = new JButton("click to enter");
			}
			
		add(textInput);
		add(okJButton);
		}
		
	public boolean action(Event e,Object arg) {
        if(e.target instanceof JButton) { 
        
        if(e.target == yesJButton) {
        	theans.gotIt(true);
        	} 

        if(e.target == noJButton) {
        	theans.gotIt(false);
        	}
        	
        if(e.target == okJButton) {
        	switch (iwh) {
        		case 2 : // fetch a double from the text field.
					theans.gotIt(Double.valueOf(textInput.getText()).doubleValue());
        			break;
        		
        		case 3 : // fetch a string from the text field.
					theans.gotIt(textInput.getText());
        			break;
        			
        		case 4 : // fetch a long from the text field.
					theans.gotIt(Long.valueOf(textInput.getText()).longValue());
        			break;
        		
        			}
        		}
        		
        return true;
		}
		return false;
	}
}	

class AnsBuf {
	boolean ifAnswered=false;
	boolean booleanAns=false;
	String stringAns = "";
	double doubleAns=0;
	long longAns=0;
	int iwh=0; /*   	1 means get true/false when the yesJButton/noJButton is pushed
					2 means get a double value from the text field when okJButton is pushed
					3 means get a String value from the text fiels when the okJButton is pushed 
					4 means get a long value from the text field when the okJButton is pushed  */
	
	synchronized public boolean gotAnsQ() {
		if(!ifAnswered) {
			try {wait();}
		catch(InterruptedException e) {
			 }
		}
		ifAnswered=false;
		notify();
		return true;
		}

	synchronized public void gotIt(boolean yorn) {
		if(ifAnswered) {
			try {wait();}
			catch(InterruptedException e) {
				}
			}
		booleanAns=yorn;
		ifAnswered=true;
		notify();
		}
	
	synchronized public void gotIt(double dbl) {
		if(ifAnswered) {
			try {wait();}
			catch(InterruptedException e) {
				}
			}
		doubleAns=dbl;
		ifAnswered=true;
		notify();
		}
	
	synchronized public void gotIt(long lgl) {
		if(ifAnswered) {
			try {wait();}
			catch(InterruptedException e) {
				}
			}
		longAns=lgl;
		ifAnswered=true;
		notify();
		}
	
	
	synchronized public void gotIt(String stg) {
		if(ifAnswered) {
			try {wait();}
			catch(InterruptedException e) {
				}
			}
		stringAns=stg;
		ifAnswered=true;
		notify();
		}
	
}

class queryThread implements Runnable {
	Thread t;
	AnsBuf theans;
	
	queryThread(AnsBuf theans) {
		this.theans = theans;
		t = new Thread(this,"query thread");
		t.start();
		}
	
	public void run() {
		theans.gotAnsQ();
		}
}
	
