package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.io.*;
import java.lang.*;
import java.util.*;
import java.awt.*;  import javax.swing.*;


class SectionNavigator extends JFrame implements Runnable, java.awt.event.ActionListener,java.awt.event.WindowListener {
	ShowPicJPanel picJPanel;
	private JTextField bigJumpText,scaleFacText,fpsField;
	private int movieDirection = 0; // -1 means going backwards at j fps, +1 means going forward at j fps
									// 0 means movie is not playing
	long lastMovieFrameTime=-1;
	int fps = 8;
	
	private int bigjump=10;
	int nomwidth=115, nomheight=320;
 	JTextField Xtext,Ytext,Ztext,rednessText,greennessText,bluenessText,coordinateLabel;
 	JButton movieLoopButton;
 	String whatXis="x";
 	String whatYis="y";
 	String whatZis="z";
	double windowScaleFac=1.0;
	Thread movieThread=null;
	boolean ifLoopMovie=true;
		
	private long milliTimeNow() {
		return (new Date()).getTime(); // returns milliseconds seconds
		}			
	
	public void run() {
		int frameinterval = 100;
		long elapsedMs;
		while(true) {
			if(movieDirection != 0) {
				if(movieDirection != 0 && (elapsedMs=(milliTimeNow() - lastMovieFrameTime)) > 1000./(double)fps) {
				    try{fps = (int)(Double.valueOf(fpsField.getText()).doubleValue());}
				    catch(NumberFormatException e) {
				    	talk("Something is wrong with the text you typed in the 'fps =' text field");
				    	fps = 10;
				    	}	
					if(fps<1) fps=1;
					frameinterval = (int)(1.+ 1000./(double) fps); // sleep time = one tenth of movie frame time.
					if((frameinterval -= 10) < 10) frameinterval=10;
					// talk("elapsedMs = "+elapsedMs);
   					if(movieDirection<0) 
   						picJPanel.curStack.showPreviousJFrameUrgently(frameinterval);
   					else 
   						picJPanel.curStack.showNextJFrameUrgently(frameinterval);
					
					// see if to reverse movie direction
					if(movieDirection>0 && picJPanel.curStack.curJFrame >= picJPanel.curStack.NJFrames-1) {
						// at end of movie
						if(ifLoopMovie) {
							picJPanel.curStack.showFirstJFrame();
							}
						else 
							movieDirection = -1;
						}
					else if(picJPanel.curStack.curJFrame <= 0)  {
						// at beginning of movie
						if(ifLoopMovie) {
							picJPanel.curStack.showLastJFrame();
							}
						else 
							movieDirection = 1;
						}
					// picJPanel.parent.updateStackSliderPosition();
					lastMovieFrameTime = milliTimeNow();
					}
				}
			else frameinterval = 5000;
			try{Thread.sleep(1 + frameinterval/10);}
			catch(InterruptedException e) {;}
			}
		}
				
    public SectionNavigator() {   
    super("navigator");
    JButton bt;
	WidgetMaker wmk = new WidgetMaker(this);
    getContentPane().setLayout(new GridLayout(0,1)); 
    setSize(nomwidth, nomheight);
    /* layout
    jump back
    --section
    ++section
    jump ahead
    
    <-movie movie->
    stop the movie
    movie fps [    ]
    
    jump=10
    scaleFac=1.0
    redraw
    prev pin
    next pin
    Z nums on/off
    maps to
    "x   y    z"
    Xtext Ytext Ztext
    "red green blue"
    rednessText greennessText bluenessText
    keyHelp
    fileNames
    		*/
    
    getContentPane().add(wmk.makeAJButton("jump back"));
    getContentPane().add(wmk.makeAJButton("--section"));
    getContentPane().add(wmk.makeAJButton("++section"));
    getContentPane().add(wmk.makeAJButton("jump ahead"));
    
    JPanel jumpSizePanel=new JPanel();
	jumpSizePanel.setLayout(new GridLayout(1,2));
	jumpSizePanel.add(wmk.makeAJLabel("jump--->"));
	bigJumpText = wmk.makeAJTextField((""+bigjump), 4);
	jumpSizePanel.add(bigJumpText);
    getContentPane().add(jumpSizePanel);

 // new movie stuff
    
    JPanel movieDirectionPanel=new JPanel();
	movieDirectionPanel.setLayout(new GridLayout(1,2));
	movieDirectionPanel.add(wmk.makeAJButton("<-mvy"));
	movieDirectionPanel.add(wmk.makeAJButton("mvy->"));
    getContentPane().add(movieDirectionPanel);
    
    JPanel movieControlPanel=new JPanel();
	movieControlPanel.setLayout(new GridLayout(1,2));
	movieControlPanel.add(wmk.makeAJButton("stopMvy"));
	movieLoopButton = wmk.makeAJButton("loop");
	movieControlPanel.add(movieLoopButton);
    getContentPane().add(movieControlPanel);
	
	
    JPanel movieFPSPanel=new JPanel();
	movieFPSPanel.setLayout(new GridLayout(1,2));
	movieFPSPanel.add(wmk.makeAJLabel("fps = "));
	fpsField = wmk.makeAJTextField(""+fps,4);
	movieFPSPanel.add(fpsField);
	getContentPane().add(movieFPSPanel);
    
    // Panels changed to JPanels November 2002
    JPanel scalePanel=new JPanel();
	scalePanel.setLayout(new GridLayout(1,2));
	scalePanel.add(wmk.makeAJLabel("scale->"));
	scaleFacText = wmk.makeAJTextField((""+windowScaleFac), 6);
	scalePanel.add(scaleFacText);
    getContentPane().add(scalePanel);

    getContentPane().add(wmk.makeAJButton("redraw"));
    getContentPane().add(wmk.makeAJButton("prev pin"));
    getContentPane().add(wmk.makeAJButton("next pin"));
    getContentPane().add(wmk.makeAJButton("Znums toggle"));
    getContentPane().add(wmk.makeAJButton("maps to"));

	coordinateLabel=wmk.makeAJTextField("   x       y       z",18);
	getContentPane().add(coordinateLabel);
	JPanel locationPanel=new JPanel();
	locationPanel.setLayout(new GridLayout(1,3));
	Xtext = wmk.makeAJTextField((""+0),4);
	Ytext = wmk.makeAJTextField((""+0),4);
	Ztext = wmk.makeAJTextField((""+0),4);
	locationPanel.add(Xtext);
	locationPanel.add(Ytext);
	locationPanel.add(Ztext);
    getContentPane().add(locationPanel);

	getContentPane().add(wmk.makeAJLabel("red    green    blue"));
	JPanel colorsPanel=new JPanel();
	colorsPanel.setLayout(new GridLayout(1,3));
	rednessText = wmk.makeAJTextField((""+0),3);
	greennessText = wmk.makeAJTextField((""+0),3);
	bluenessText = wmk.makeAJTextField((""+0),3);
    colorsPanel.add(rednessText);
    colorsPanel.add(greennessText);
    colorsPanel.add(bluenessText);
    getContentPane().add(colorsPanel);
    
    getContentPane().add(wmk.makeAJButton("keyHelp"));
    getContentPane().add(wmk.makeAJButton("fileNames"));

	addWindowListener(this);
	movieThread = new Thread(this);
	movieThread.start();
	}
	
	public void zSubjunctive(int newZ) {
		Ztext.setText(""+newZ);
		}
		
	public void labelSectionTypes(int sectionType) {
		switch (sectionType) {
			case 0 : // XY sections
				whatXis="x";
				whatYis="y";
				whatZis="z";
				break;
			case 1 :  // XZ sections
				whatXis="x";
				whatYis="z";
				whatZis="y";
				break;
			case 2 :  // ZY sections
				whatXis="y";
				whatYis="z";
				whatZis="x";
				break;
			}
	}	
	
	public void pixelReport(int x, int y, int sect, int redbri,int greenbri,int bluebri) {
		coordinateLabel.setText("   "+whatXis+"       "+whatYis+"       "+whatZis);		
		Xtext.setText(""+x);
		Ytext.setText(""+y);
		Ztext.setText(""+sect);
		rednessText.setText(""+redbri);
		greennessText.setText(""+greenbri);
		bluenessText.setText(""+bluebri);
		}

	public void remora(int xmainWindowRight, int ymainWindow) {
		setLocation(xmainWindowRight + 9,ymainWindow);
		setVisible(true);
		}
		
// catch WindowListener events  
	public void windowActivated(java.awt.event.WindowEvent event) {
		// talk("window activated");
		}
	public void windowDeactivated(java.awt.event.WindowEvent event) {
		//  talk("window Deactivated");
		}
	public void windowOpened(java.awt.event.WindowEvent event) {}
	public void windowIconified(java.awt.event.WindowEvent event) {}
	public void windowDeiconified(java.awt.event.WindowEvent event) {}
	public void windowClosing(java.awt.event.WindowEvent event) {
	    setVisible(false);
		}
	public void windowClosed(java.awt.event.WindowEvent event) {
		}
	
	public void makeAJump(int iwh) {
		switch (iwh) {
			case (-2) : // big jump backwards
   				adjustBigJump(0);
   				picJPanel.curStack.showJumpJFrame(-bigjump);
				break;
			case (-1) : // move one section backwards
   				picJPanel.curStack.showPreviousJFrame();
				break;
			case (1) : // move one section forward
   				picJPanel.curStack.showNextJFrame();
				break;
			case (2) : // big jump forward
   				adjustBigJump(0);
   				picJPanel.curStack.showJumpJFrame(bigjump);
				break;
				}
			picJPanel.parent.updateStackSliderPosition();
		}
	 
// catch button events...
	public void actionPerformed(java.awt.event.ActionEvent event) {
		String arg = event.getActionCommand();
		
 		picJPanel.showYourself();       
	    if(arg.equals("fileNames")) {
			picJPanel.tellFileNames();
			return;
	    	}
		
	    if(arg.equals("mvy->")) {
    		movieDirection = 1;
    		picJPanel.movieStatus(true);
			return;
	    	}

	    if(arg.equals("<-mvy")) {
    		picJPanel.movieStatus(true);
    		picJPanel.ifTellTiming=true;
    		movieDirection = -1;
			return;
	    	}

	    if(arg.equals("stopMvy")) {
    		movieDirection = 0;
    		picJPanel.ifTellTiming=false;
    		picJPanel.movieStatus(false);
			return;
	    	}

	    if(arg.equals("loop")) {
    		movieLoopButton.setText("<---->");
			ifLoopMovie=false;
			return;
	    	}

	    if(arg.equals("<---->")) {
    		movieLoopButton.setText("loop");
			ifLoopMovie=true;
			return;
	    	}



	    if(arg.equals("keyHelp")) {
			picJPanel.instructions();
			return;
	    	}

    	if(arg.equals("redraw")) {
    		picJPanel.reDraw();
    		}

   		if(arg.startsWith("++section")) {
			makeAJump(1);
   			}
   		if(arg.startsWith("--section")) {
			makeAJump(-1);
   			}
   		if(arg.startsWith("jump ahead")) {
			makeAJump(2);
   			}
   		if(arg.startsWith("jump back")) {
			makeAJump(-2);
   			}
   			
   		if(arg.equals("prev pin")) {
   			jumpToPinnedSection(-1);
			picJPanel.parent.updateStackSliderPosition();
   			}
   			
   		if(arg.equals("maps to")) {
   			jumpToMappedToSection();
			picJPanel.parent.updateStackSliderPosition();
   			}
   			
   		if(arg.equals("next pin")) {
   			jumpToPinnedSection(1);
			picJPanel.parent.updateStackSliderPosition();
   			}
   			
   		if(arg.equals("Znums toggle")) {
   			picJPanel.ifDrawZNumbers=!(picJPanel.ifDrawZNumbers); 
   			picJPanel.reDraw(); 
   			}


		}
   
   private void jumpToPinnedSection(int iwh) {
		int ito = picJPanel.statsJPanel.myBox.identifyPinnedSectionNumber(iwh);
		if(ito>=0)
			picJPanel.curStack.showAJFrame(ito);
		else 
			talk("The pin buttons work only when a wandering core stack is defined.");
		}
   
   private void jumpToMappedToSection() {
   	int curJFrame = picJPanel.curStack.curJFrame;
   		BioradFile chanFile=picJPanel.redfile;
   		if(chanFile==null) 
   			chanFile=picJPanel.greenfile;
   		if(chanFile==null) 
   			chanFile=picJPanel.bluefile;
   		if(chanFile != null && chanFile.myConfig.ifGainCurveCorrected_bleachCorrected_stack) {
   			if(curJFrame < chanFile.myConfig.firstRealSection)
				picJPanel.curStack.showAJFrame(chanFile.myConfig.takeWanderingCoreBoundaryFromSection(curJFrame));
			else  // jump to the closest section to one that gets mapped to curJFrame
				picJPanel.curStack.showAJFrame(chanFile.myConfig.closestSectMappedToMe(curJFrame));
			}
		else 
			talk("This option works only for autoBleachCorrectionSection");
   		}
   			   
   	public int bigJump() { return bigjump; }
    
	public void adjustBigJump(int iwh) {
		if(iwh>0)	
			++bigjump;
		if(iwh<0) {
			if(--bigjump < 2) bigjump=2;
			}
		if(iwh==0) { // getvalue from text Field
			bigjump = (int)(Double.valueOf(bigJumpText.getText()).doubleValue());	
			}	
		bigJumpText.setText(""+bigjump);
		}

	public double getScaleFac() {
		windowScaleFac = Double.valueOf(scaleFacText.getText()).doubleValue();	
		double oneTeenth = 1./16.;
		if(windowScaleFac < oneTeenth) windowScaleFac=oneTeenth;
		if(windowScaleFac > 1.0) windowScaleFac=1.0;
		int quanta = (int)(.01+ windowScaleFac/oneTeenth);
		windowScaleFac = quanta*oneTeenth;
		scaleFacText.setText(""+windowScaleFac);
		return windowScaleFac;
		}

	public void changeJPanelSize(double ratio) {
		int pwidth = getSize().width;
		int pheight = getSize().height;
		pwidth = (int)(ratio*pwidth +.4999);
		pheight = (int)(ratio*pheight +.4999);
		setSize(pwidth,pheight);
		show();
		}

	static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
}

