package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.awt.*;  import javax.swing.*;
import java.awt.image.*;
import java.util.*;
import java.lang.*;
import java.io.*;

/* implements a drawing canvas that will hold a vertical column of three postage stamp images
   having text-marked spots in each of the three pictures which, if clicked, tell whether to 
   dismiss, promote to "contender" status, or pick as "current best".
   */
class SegTowerCanvas extends Canvas implements java.awt.event.MouseListener{
	SegTower parent=null;
	int towerWidth=100,towerHeight=133;
	int border=3;
	int nhigh = 3;
	int ID=-1;
	int borderRed=0, borderGreen=180, borderBlue=120;
	int [] towerRas=null;
	SegPic [] thePics = new SegPic[nhigh];
	int stampXorg=0, stampYorg=0, stampHeight=100;
	int stampWidth=(4*stampHeight)/3;
	static int verTextOffset=8,charWidth=6;
	int moustStopX=0;
	int moustStopY=0;
	int mouseStartX = 0;
	int mouseStartY = 0;
	Image towerImage=null,offScrImage=null;

	public SegTowerCanvas(int ID, int stampWidth, int stampHeight, SegTower parent) {
		this.parent = parent;
		this.stampXorg=stampXorg;     // this will change
		this.stampYorg=stampYorg;     // this will change
		this.stampWidth=stampWidth;   // this dimensions will never change
		this.stampHeight=stampHeight; // this dimensions will never change
	
		int i;
		newTowerSize();
		this.ID = ID;
		for(i=0;i<nhigh;i++) { // ppaint the rectangle black
			thePics[i] = new SegPic(this);
			}
		updateTowerRas();
		setLocation(ID*(4+towerWidth),0);
		setVisible(true);
		addMouseListener(this);
		repaint();
		}
	
	public void dismissAllBut(int survivingStatus) {
		for(int i=0;i<3;i++)
			if(thePics[i].status >=0 && thePics[i].status<survivingStatus  && thePics[i].status <3)
				thePics[i].empty();
		repaint();
		}
		
	private void newTowerSize() {
		towerWidth=2*border + stampWidth;         // 2*3 is to make 3-pixel borders on the left and right
		towerHeight = (nhigh+1)*border + (3*stampWidth*nhigh)/4;// 4*3 is to make 4 horizontal separation borders.
		setSize(towerWidth,towerHeight);
		towerRas = new int[towerWidth*towerHeight];
		}
		
	public void changeStampSizes(int newStampWidth,int newStampHeight) {
		this.stampHeight=newStampHeight;
		this.stampWidth=newStampWidth;
		newTowerSize();
		for(int i=0;i<nhigh;i++) {
			thePics[i].changeStampSizes(newStampWidth,newStampWidth);
			thePics[i].isStampCurrent = false;
			}
		}
		
	public void sepuku() {
		for(int i=0;i<nhigh;i++) {
			thePics[i].sepuku();
			thePics[i]=null;
			}
		parent=null;
		thePics=null;
		setVisible(false);
		}
		
	public void newStampLocation(int stampXorig,int stampYorig) {
		stampXorg=stampXorig;
		stampYorg=stampYorig;
		}
		
	public void noOneIsBest() {
		for(int i=0;i<nhigh;i++)
			thePics[i].revokeBestStatus();
		repaint();
		}
	
	public int firstNonContender()	{
		for(int i=0;i<nhigh;i++)
			if(thePics[i].status < 2)
				return i;
		return -1;
		}
		
	public int firstEmptyPic()	{
		for(int i=0;i<nhigh;i++)
			if(thePics[i].status < 0)
				return i;
		return -1;
		}
		
	public void emptyPics() {
		for(int i=0;i<nhigh;i++)
			thePics[i].empty();
		repaint();
		}

	public boolean updateTowerRas() {
		boolean somethingChanged=false;
		for(int i=0;i<nhigh;i++) { // refill each SegPic's rectangle 
		// talk("i="+i+" border="+border+", and towerRow="+(border+i*(stampHeight+border)));
			 if(thePics[i].copyStampToTowerRas(stampXorg, stampYorg, stampWidth, stampHeight,
					towerWidth, border+i*(stampHeight+border), border, towerRas))
						somethingChanged = true;
			}
		return somethingChanged;
		}
		
	public void update(Graphics2D g) {
		paint(g);
		}
	
	static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
		
	public void mouseClicked(java.awt.event.MouseEvent event) {
		// talk("Mouse Clicked at ("+event.getX()+", "+event.getY()+")");
		}
		
	public void mousePressed(java.awt.event.MouseEvent event) {
		 mouseStartX = event.getX();
		 mouseStartY = event.getY();
		// talk("Mouse Pressed at "+mouseStartX+", "+mouseStartX);		    
		}
		
		
	public void mouseEntered(java.awt.event.MouseEvent event) {
		// talk("Mouse entered at ("+event.getX()+", "+event.getY()+")");
		}
		
	public void mouseExited(java.awt.event.MouseEvent event) {
		// talk("Mouse exited at ("+event.getX()+", "+event.getY()+")");
		}
		
	public void promoteBestYet(int lastPicClicked) {
		if(lastPicClicked>=0 && thePics[lastPicClicked].status>=0) {
			parent.parent.noOneIsBest();
			thePics[lastPicClicked].promoteToBestYet();
			parent.parent.setBestYet(thePics[lastPicClicked]);
			repaint();
			}
	    return;
	    }

	public void promoteContender(int lastPicClicked) {
		if(lastPicClicked>=0  && thePics[lastPicClicked].status>=0) {
			thePics[lastPicClicked].promoteToContender();
			repaint();
			}
	    return;
	    }


	public void demoteContender(int lastPicClicked) {
		if(lastPicClicked>=0  && thePics[lastPicClicked].status>=0) {
			thePics[lastPicClicked].demoteContender();
			repaint();
			}
	    return;
	    }

	public void dismiss(int lastPicClicked) {
		if(lastPicClicked>=0) {
			if(thePics[lastPicClicked].status ==3) 
				talk("You cannot dismiss the bestYet picture");
			else	
				thePics[lastPicClicked].dismissPic();
			repaint();
			}
	    return;
	    }

	public void mouseReleased(java.awt.event.MouseEvent event) {	
		moustStopX = event.getX();
		moustStopY = event.getY();
		int id,xc,yc;
		// see which pic the mouse clicked in...
		for(id=0;id<nhigh;id++) {
			xc = towerWidth/2;
			yc = border + id*(stampHeight+border) +stampHeight/2;
			if(Math.abs(moustStopY - yc) <= stampHeight/2 && 
				Math.abs(moustStopX - xc) <= stampWidth/2) {					
					if(parent.parent.lastClickedPicClickedAgainQ(ID, id)) {
						// second click on this pic. If in upper right corner, promote to contender and unregister
						if(Math.abs(moustStopY - (yc - stampHeight/2)) <= 33 && 
								Math.abs(moustStopX - (xc + stampWidth/2)) <= 33) {
									if(thePics[id].status >=0 && thePics[id].status < 2) {
										promoteContender(id);
										talk("Promote to contender");
										parent.parent.unRegisterLastClickedPic(122);
										}
										return;	
									}					
						//If in upper left corner, promote to  best yet 
						if(Math.abs(moustStopY - (yc - stampHeight/2)) <= 33 && 
								Math.abs(moustStopX - (xc - stampWidth/2)) <= 33) {
									if(thePics[id].status >=0 && thePics[id].status < 3) {
										promoteBestYet(id);
										parent.parent.unRegisterLastClickedPic(120);
										}
									return;	
									}					
						// if in lower right, empty this pic..
						if(Math.abs(moustStopY - (yc + stampHeight/2)) <= 33 && 
								Math.abs(moustStopX - (xc + stampWidth/2)) <= 33) {
									dismiss(id);
									parent.parent.unRegisterLastClickedPic(121);
									return;	
									}
						// if in lower left, demote from contender this pic..
						if(Math.abs(moustStopY - (yc + stampHeight/2)) <= 33 && 
								Math.abs(moustStopX - (xc - stampWidth/2)) <= 33) {
									if(thePics[id].status ==2)
									demoteContender(id);
									parent.parent.unRegisterLastClickedPic(121);
									return;	
									}
						}					
					parent.parent.registerLastClickedPic(ID, id);
					if(thePics[id].attentionPaid())
						repaint();
					return;
			}}
		}

	private void drawStatusJLabels(GraphicsPs gg) {
		int i;
		for(i=0;i<nhigh;i++) {
				gg.newColor(0,255,0);
				if(thePics[i].status >=0) {
				   gg.drawString((thePics[i].statusString()+"  "+thePics[i].nPixOn),
							4,border + i*(stampHeight+border) + verTextOffset);	
				   gg.drawString(("sect "+thePics[i].sectionNumber),
							stampWidth-46,border + i*(stampHeight+border) + verTextOffset);	
					}	
				else		
				   gg.drawString(thePics[i].statusString(),
							4,border + i*(stampHeight+border) + verTextOffset);				
				}
		}
	
	public void highlightBestYetStamp(GraphicsPs gg) {
		int x,xb,yb,ytop;
		for(int i=0; i<nhigh;i++)
			if(thePics[i].status==3) {// this is the unique bestYet pic
				gg.newColor(255, 0, 0);
				ytop = i*(stampHeight+border);
				for(x=0; x<towerWidth; x+=(stampWidth+border))
					for(xb=0;xb<border;xb++)
						gg.drawLine(x+xb,ytop,x+xb,ytop+stampHeight+border); // draw the vertical borders
			
				for(yb=0;yb<border;yb++) {
					gg.drawLine(0, yb+ytop, towerWidth,yb+ytop); // draw the horizontal borders
					gg.drawLine(0, yb+ytop+border+stampHeight, towerWidth,yb+ytop+border+stampHeight); // draw the horizontal borders
					}
			return;
			}
		return;
		}
		

	public void highlightContenders(GraphicsPs gg) {
		int x,xb,yb,ytop;
		for(int i=0; i<nhigh;i++)
			if(thePics[i].status==2) {// this is a contender
				gg.newColor(0, 255, 0);
				ytop = i*(stampHeight+border);
				for(x=0; x<towerWidth; x+=(stampWidth+border))
					for(xb=0;xb<border;xb++)
						gg.drawLine(x+xb,ytop,x+xb,ytop+stampHeight+border); // draw the vertical borders
			
				for(yb=0;yb<border;yb++) {
					gg.drawLine(0, yb+ytop, towerWidth,yb+ytop); // draw the horizontal borders
					gg.drawLine(0, yb+ytop+border+stampHeight, towerWidth,yb+ytop+border+stampHeight); // draw the horizontal borders
					}
			}
		return;
		}
		
	public void highlightNewPics(GraphicsPs gg) {
		int x,xb,yb,ytop;
		for(int i=0; i<nhigh;i++)
			if(thePics[i].status==0) {// this is a new pic
				gg.newColor(100, 100, 255);
				ytop = i*(stampHeight+border);
				for(x=0; x<towerWidth; x+=(stampWidth+border))
					for(xb=0;xb<border;xb++)
						gg.drawLine(x+xb,ytop,x+xb,ytop+stampHeight+border); // draw the vertical borders
			
				for(yb=0;yb<border;yb++) {
					gg.drawLine(0, yb+ytop, towerWidth,yb+ytop); // draw the horizontal borders
					gg.drawLine(0, yb+ytop+border+stampHeight, towerWidth,yb+ytop+border+stampHeight); // draw the horizontal borders
					}
			}
		return;
		}
		
	public void paint(Graphics g) {
		//talk("This should never happen. segTowerCanvas.paint() called");
		paint2D((Graphics2D)g);
		}
		
	public void update(Graphics g) {
		// talk("segTowerCanvas.update() called. Should never happen");
		paint2D((Graphics2D)g);
		}

	public void update2D(Graphics2D g) {
		paint2D(g);
		}


	public void paint2D(Graphics2D g) {
		int	x,xb,yb, y;
		GraphicsPs gg;		
		gg = new GraphicsPs(towerWidth,towerHeight,g,false);
		setSize(towerWidth,towerHeight);
		if(updateTowerRas()) { // update the towerRas;	
	      	towerImage = createImage(new MemoryImageSource(towerWidth, towerHeight,
			 	ColorModel.getRGBdefault(), towerRas, 0, towerWidth));
			 }
		try {
		    if(towerImage != null) {
			    gg.drawImage(towerImage, 0, 0, towerWidth,towerHeight, this);
			    }
		    }
		  catch (Exception e) {
			System.out.println("Error painting tower panel: " + e.toString());
			}
		gg.newColor(borderRed, borderGreen, borderBlue);
		for(x=0; x<towerWidth; x+=(stampWidth+border))
			for(xb=0;xb<border;xb++)
				gg.drawLine(x+xb,0,x+xb,towerHeight); // draw the vertical borders
			
		for(y=0; y<towerHeight; y+=(stampHeight+border))
			for(yb=0;yb<border;yb++)
				gg.drawLine(0, y+yb, towerWidth,y+yb); // draw the horizontal borders
		drawStatusJLabels(gg);
		highlightNewPics(gg);
		highlightContenders(gg);			
		highlightBestYetStamp(gg);
		setVisible(true);
		}
			
}