package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.lang.*;
import java.util.*;
import java.awt.*;  import javax.swing.*;

class SegmenterNode extends JFrame implements java.awt.event.ActionListener,java.awt.event.ItemListener, java.awt.event.WindowListener {
boolean diags = true; // set true to make the various filters announce themselves
int whatAmI = 0;
int myXcen = 100;
int width=100,height=40;
int parJPanelHeight=100;
int myYcen=100;
int nparams=0;  
SegPar [] thePars = new SegPar[4]; // 4 is the most parameters we'll ever have
int myPath=1;
int myStep=1;
int inStep1=-1;
int inPath1=-1;
int inPath2=-2;
int inStep2=-2;
WidgetMaker wmk;
int myred=128,mygreen=128,myblue=128;
int myTextred=255,myTextgreen=255,myTextblue=255;
long gen = 0;
long in1Gen=-1,in2Gen=-1;
boolean ifShowSelf = false;
Segmenter parent;
JComboBox whatIBecome;
boolean ifMakeChoiceBar=true;
int MaskColorsData=0;
boolean ifActive=false;
boolean anyChangeSinceLastRun=true;
boolean justRun=false;
int imageInputNode=-1; // if positive, then this is an image input node...
int sectionNumber=-1;
int nullID = 18;
//                      0       1        2         3        4          5          6           7           8           9       10      11       12          13       14       15       16         17       18      19      20      21      22      23      24      25
String [] whats =    {"wire","flatten","erode","thicken","ridges",  "BlkWht",  "lowpass",  "NiceBalls","NiceTubes","invert","skel","undust","ThreshAnd","Focus","DimEdges","Zgrad","gapFill","AndOrAvg","null","chan1","chan2","chan3","mask1","mask2","mask3","mask4"};
boolean [] if2Inputs={false,  false,   false,   false,    false,     false,      false,      true,       false,     false,   false,  false,    true,      false,    false,   false,   false,   true,    false,  false,  false,  false,  true,  true,   true,   true};   
int possibilities=26;

short [] shortRas=null;
int rasRows=-1,rasCols=-1;

InvertFltr inversion=null;
BlackWhiteFltr BlkWht=null;
ThreshAndFltr ThreshAnd=null;
LowPassFltr lowpass = null;
ErodeFltr erode = null;
DimEdgesFltr dimedges = null;
ThickenFltr thicken = null;
ZGradFltr zgrad=null;
RidgesFltr ridges=null;
UndustFltr unduster = null; 
FillThread gapplugger = null;
NiceTubes tuber=null;
NiceBalls makeBalls = null;
SkelFltr skeletor = null;
FlattenFltr flatten = null;
AndOrAvgFltr andoravg = null;
FocusFltr focsharp = null;

TwoColorPicker displayColorPicker=null;

	public void sepuku() {
		parent=null;
		whatIBecome=null;
		dispose();
		}

	public void finalize() {
		//talk("                                                                 SegmenterNode snuffs");
		}
		
private int setHeight(int maxHeight) {
	int ht=10;
	nparams=0;
	switch (whatAmI) {
		case 0:  // wire; no parameters
			ht=80;
			break;
		case 1: // FlattenFltr; 3 pars=cutoff
			ht=120;
			nparams=3;
			break;
		case 2: //  erode; 2 pars
			ht=110;
			nparams=2;
			break;
		case 3:  // thicken; 3 pars
			ht=120;
			nparams=3;
			break;
		case 4: //  ridges; 3 pars (2 are doublePrecision)
			ht=130;
			nparams=3;
			break;
		case 5: //   BlkWht; 1 par
			ht=100;
			nparams=1;
			break;
		case 6:  // lowpass; 2 pars
			ht=110;
			nparams=2;
			break;
		case 7: //   makeBalls; 2 pars -- one input raster
			ht=110;
			nparams=2;
			break;
		case 8: //  NiceTubes; 3 pars (2 are doublePrecision)
			ht=135;
			nparams=3;
			break;
		case 9: // invert; no parameters
			ht=80;
			break;
		case 10: // skel; 2 pars
			ht=110;
			nparams=2;
			break;
		case 11: // undust; 1 pars
			ht=105;
			nparams=1;
			break;
		case 12: // ThreshAnd; 2 pars -- two inputs
			ht=110;
			nparams=2;
			break;
		case 13: // FocusFltr; 1 pars
			ht=105;
			nparams=1;
			break;
		case 14: // DimEdgesFltr; 2 par
			ht=110;
			nparams=2;
			break;
		case 15: // Zgrad; 4 par
			ht=170;
			nparams=4;
			break;
		case 16: //gapFill: 1 par
			ht=105;
			nparams=1;
			break;
		case 17: // AndOrAvg: 3 pars
			ht=125;
			nparams=3;
			break;			
		case 18: // null; no pars
			ht=67;
			break;
		case 19: //  chan1	
		case 20: //  chan2	
		case 21: //  chan3	
			ht=20;
			break;
		case 22: // mask1
		case 23: // mask2
		case 24: // mask3
		case 25: // mask4
			ht=115;
			break;
		default : // who knows?
			ht=60;
			break;
		}
	if(ht>maxHeight) return maxHeight;
	else return ht;			
	}

public void decodeParamValues(int p0,int p1,int p2,int p3)	{
	switch (whatAmI) {
		case 0:  // wire; no parameters
			break;
		case 1: // threshold; 3 pars tooDim, radius, avgBrt
			thePars[0].setVal(p0); // tooDim
			thePars[1].setVal(p1); // radius
			thePars[2].setVal(p2); // avgBrt
			break;
		case 2: //  erode; 2 pars
			thePars[0].setVal(p0);
			thePars[1].setVal(p1);
			break;
		case 3:  // thicken; 3 pars
			thePars[0].setVal(p0);
			thePars[1].setVal(p1);
			thePars[2].setVal(p2);
			break;
		case 4: //  ridges; 3 pars (2 are doublePrecision)
			thePars[0].setVal(p0); //pixthresh
			thePars[1].setVal(.0001*p1); //peakthresh
			thePars[2].setVal(.0001*p2); //peaktol
			break;
		case 5: //   BlkWht; 1 par
			thePars[0].setVal(p0);
			break;
		case 6:  // lowpass; 2 pars
			thePars[0].setVal(p0);
			thePars[1].setVal(p1);
			break;
		case 7: //   makeBalls; 2 pars -- two inputs
			talk("SegmenterNode.decodeParamValues, NiceBalls: po="+p0+", p1="+p1+", p2="+p2+", p3="+p3);
			thePars[0].setVal(p0);
			thePars[1].setVal(.0001*p1); // one double precision parameter
			break;
		case 8: //  NiceTubes; 3 pars (2 are doublePrecision)
			thePars[0].setVal(p0);
			thePars[1].setVal(.0001*p1);
			thePars[2].setVal(.0001*p2);
			break;
		case 9: // invert; no parameters
			break;
		case 10: // skel; 2 pars
			thePars[0].setVal(p0);
			thePars[1].setVal(p1);
			break;
		case 11: // undust; 1 pars
			thePars[0].setVal(p0);
			break;
		case 12: // ThreshAnd; 2 pars
			thePars[0].setVal(p0);
			thePars[1].setVal(p1);
			break;
		case 13: // FocusFltr; 1 pars
			thePars[0].setVal(p0);
			break;
		case 14 : // DimEdgesFltr, 2 par
			thePars[0].setVal(p0);
			thePars[1].setVal(p1);
			break;			
		case 15 : // Zgrad, 2 par
			thePars[0].setVal(.0001*p0);
			thePars[1].setVal(.0001*p1);
			thePars[2].setVal(p2);
			thePars[3].setVal(.0001*p3);
			break;			
		case 16 : // FillThread, 0 par
			thePars[0].setVal(p0);
			break;			
		case 17 : // AndOrAvg, 3 par
			thePars[0].setVal(p0);
			thePars[1].setVal(p1);
			thePars[2].setVal(p2);
			break;			
		case 18 : // null
		case 19: // chan1
		case 20: // chan2
		case 21: // chan3
		case 22: // mask1
		case 23: // mask2
		case 24: // mask3
		case 25: // mask4
		default :
			break;
		}
	anyChangeSinceLastRun=true; 
	}

	public int howManyParametersToVary() {
		int i,nvary=0;
		for(i=0;i<nparams;i++)
			if(thePars[i] != null && thePars[i].ifVary)
				++nvary;
		return nvary;
		}
				
private int [] encodeParsToFourInts() {
	int [] fourints = new int[4];
	for(int i=0;i<4;i++) 
		fourints[i]= i>=nparams? 0 : thePars[i].intVal;
	return fourints;
	}	
	
public String encodeInto12Ints(int [] aray) {
/* the order is 
  0    1     2       3       4       5       6       7       8       9       10     11
path step whatAmI inPath1 inStep1 inPath2 inStep2   par0    par1    par2    par3  extra */
	aray[0] = myPath;
	aray[1] = myStep;
	aray[2] = whatAmI;
	aray[3] = inPath1;
	aray[4] = inStep1;
	aray[5] = inPath2;
	aray[6] = inStep2;
	int [] fourInts = encodeParsToFourInts();
	aray[7] = fourInts[0];
	aray[8] = fourInts[1];
	aray[9] = fourInts[2];
	aray[10] =fourInts[3];
	aray[11] = 0; // the extra
	if(whatAmI>=0) return (whats[whatAmI]);
	else return "empty";
	}
	
public void setTextColor(GraphicsPs gg) {
	gg.newColor(myTextred, myTextgreen, myTextblue);
	}

public void setTextColorDim(GraphicsPs gg) {
	gg.newColor((2*myTextred)/4, (2*myTextgreen)/4, (2*myTextblue)/4);
	}
	
private void setSpecialCoords() {
	parent.setFillHeight();
	int topht = 30;
	int heightOffset = (parent.yinc)/2;
	myred=0;mygreen=0;myblue=0;
	int [] theReds =   {170,255,0,  0,  255,255,0,  100};
	int [] theGreens = {170,0,  255,0,  255,0,  255,255};
	int [] theBlues =  {170,0,  0  ,255,0,  255,255,100};
	int colornum = myStep;
	if(colornum<0) colornum=0;
	while(colornum>7) colornum -= 8; // set the default colors for text and links
	myTextred=theReds[colornum];
	myTextgreen=theGreens[colornum];
	myTextblue=theBlues[colornum];
	if(whatAmI==possibilities-4) { // mask1
		myYcen=topht;
		myXcen=parent.width - width/5 - 6;
		myred=0;mygreen=255;myblue=180;
		myTextred=255; myTextgreen=180; myTextblue=180;
		ifMakeChoiceBar=false;
		}
	if(whatAmI==possibilities-3) { // mask2
		myYcen=topht+heightOffset;
		myXcen=parent.width - width/5 - 6;
		myTextred=180; myTextgreen=255; myTextblue=180;
		myred=170; mygreen=0; myblue=170;
		ifMakeChoiceBar=false;
		}
	if(whatAmI==possibilities-2) { // mask3
		myYcen=topht + 2*heightOffset;
		myXcen=parent.width - width/5 - 6;
		myTextred=180; myTextgreen=180; myTextblue=255;
		myred=150; mygreen=120; myblue=0;
		ifMakeChoiceBar=false;
		}
	if(whatAmI==possibilities-1) { // mask4
		myYcen=topht + 3*heightOffset ;
		myXcen=parent.width - width/5 - 6;
		myTextred=255; myTextgreen=255; myTextblue=180;
		myred=100; mygreen=120; myblue=240;
		ifMakeChoiceBar=false;
		}
	else if (whatAmI==possibilities-7) { // chan1
		myYcen=topht;
		myXcen=width/5;
		myred=255;mygreen=0;myblue=0;
		myTextred=0;myTextgreen=255;myTextblue=255;
		imageInputNode=1;
		ifMakeChoiceBar=false;
		}
	else if (whatAmI==possibilities-6) { // chan2
		myYcen=topht+heightOffset;
		myXcen = width/5;
		myred=0;mygreen=255;myblue=0;
		myTextred=255;myTextgreen=0;myTextblue=255;
		ifMakeChoiceBar=false;
		imageInputNode=2;
		}
	else if (whatAmI==possibilities-5) {// chan3
		myYcen=topht + 2*heightOffset;
		myXcen=width/5;
		myred=0;mygreen=0;myblue=255;
		myTextred=255;myTextgreen=255;myTextblue=0;
		ifMakeChoiceBar=false;
		imageInputNode=3;
		}
	else if (whatAmI==nullID) {//  null node
		ifShowSelf = false;
		ifMakeChoiceBar=true;
		}
	}	

public SegmenterNode(Segmenter parent, int path, int step, int whatAmI, int xcen, int ycen,
	int width, int maxHeight, int inPath1, int inStep1, int inPath2, int inStep2, 
		boolean ifShowSelf) {
	this.parent=parent;
	this.width = width;
	myPath=path;
	myStep=step;
	this.whatAmI=whatAmI;
	wmk = new WidgetMaker(this);
	this.height=maxHeight;
	parJPanelHeight=setHeight(maxHeight);
	this.inPath1=inPath1;
	this.inStep1=inStep1;
	this.inPath2=inPath2;
	this.inStep2=inStep2;
	if(whatAmI>=0 && !if2Inputs[whatAmI]) {
		this.inPath2=-2;
		this.inStep2=-2;
		}
	this.ifShowSelf=ifShowSelf;
	myXcen=xcen;
	myYcen=ycen;
	setSpecialCoords();
	if(whatAmI<0) {
		ifMakeChoiceBar=false;
		}
	if(ifMakeChoiceBar) {
		getContentPane().setLayout(new GridLayout(0,1));
		whatIBecome = wmk.makeAJComboBox(nullID+1,whats);
		whatIBecome.setSelectedIndex(whatAmI);
		getContentPane().add(whatIBecome);
		}
		
	if(whatAmI >= 0)
		makeParRows();
	
	setSize((width),parJPanelHeight); // if the user changes this size, let the new size persist.
	addWindowListener(this);
	if(whatAmI>=0) setTitle(whats[whatAmI]);
	makeTheFilter();
	wmk=null;
	}

public void flagShowNode(GraphicsPs gg) {
	gg.newColor(255,255,255);
	gg.fillCenteredOval(myXcen,myYcen-12,20,8); // the sclera
	gg.newColor(128,80,40);
	gg.fillCenteredOval(myXcen,myYcen-12,5,5); // the pupil
	}
	
public boolean backPropActiveStatus() {
	if(!ifActive) return false;
	boolean ifActivatedAnything=false;
	if(inPath1>=0 && inStep1>=0 && !parent.nodes[inPath1][inStep1].ifActive) {
		parent.nodes[inPath1][inStep1].ifActive = true;
		// mark that node active
		ifActivatedAnything=true;
		}
	if(inPath2>=0 && inStep2>=0 && !parent.nodes[inPath2][inStep2].ifActive) {
		parent.nodes[inPath2][inStep2].ifActive = true;
		// mark that node active
		ifActivatedAnything=true;
		}
	return ifActivatedAnything;
	}
	
private void makeParRows() {
	int i;
	nparams=0;
	switch (whatAmI) {
		case -1:  // empty
		case 0:  // wire; no parameters
			addShowJButton();
			break;
		case 1: // FlattenFltr; 3 pars=cutoff
			thePars[0] = new SegPar(16,0,255,"tooDim",this); 
			thePars[1] = new SegPar(32,32,96,"radius",this); 
			thePars[2] = new SegPar(100,64,180,"avgBrt",this); 
			nparams=3;
			for(i=0;i<nparams;i++) getContentPane().add(thePars[i]);
			addShowJButton();
			break;
		case 2: //  erode; 2 pars
			thePars[0] = new SegPar(130,0,255,"cutoff",this); 
			thePars[1] = new SegPar(3,1,255,"iter8",this); 
			nparams=2;
			for(i=0;i<nparams;i++) getContentPane().add(thePars[i]);
			addShowJButton();
			break;
		case 3:  // thicken; 3 pars
			thePars[0] = new SegPar(255,0,255,"fill",this); 
			thePars[1] = new SegPar(5,1,7,"Nnghb",this); 
			thePars[2] = new SegPar(2,1,255,"iter8",this); 
			nparams=3;
			for(i=0;i<nparams;i++) getContentPane().add(thePars[i]);
			addShowJButton();
			break;
		case 4: //  ridges; 3 pars (2 are doublePrecision)
			thePars[0] = new SegPar(70,0,255,"pixthresh",this); 
			thePars[1] = new SegPar(1.40,.8,3.,"peakthresh",this); 
			thePars[2] = new SegPar(1.08,.8,1.4,"peaktol",this); 
			nparams=3;
			for(i=0;i<nparams;i++) getContentPane().add(thePars[i]);
			addShowJButton();
			break;
		case 5: //   BlkWht; 1 par
			thePars[0] = new SegPar(70,0,255,"thresh",this); 
			nparams=1;
			for(i=0;i<nparams;i++) getContentPane().add(thePars[i]);
			addShowJButton();
			break;
		case 6:  // lowpass; 2 pars
			thePars[0] = new SegPar(200,0,255,"weight",this); 
			thePars[1] = new SegPar(3,1,255,"iter8",this); 
			nparams=2;
			for(i=0;i<nparams;i++) getContentPane().add(thePars[i]);
			addShowJButton();
			break;
		case 7: //   NiceBalls; 2 pars -- one inputs
			thePars[0] = new SegPar(20,0,255,"maxrad",this); 
			thePars[1] = new SegPar(.901,0,.9999,"passingScore",this); 
			nparams=2;
			for(i=0;i<nparams;i++) 
				getContentPane().add(thePars[i]);
			addShowJButton();
			break;
		case 8: //  NiceTubes; 3 pars int spotthresh, double linetol, double maxtubewidth
			thePars[0] = new SegPar(12,0,255,"spotThresh",this); 
			thePars[1] = new SegPar(1.40,.8,3.0,"lineTol",this); 
			thePars[2] = new SegPar(2.5,.8,3.5,"maxTubeDia",this); 
			nparams=3;
			for(i=0;i<nparams;i++) getContentPane().add(thePars[i]);
			addShowJButton();
			break;
		case 9: // invert; no parameters
			addShowJButton();
			break;
		case 10: // skel; 2 pars
			thePars[0] = new SegPar(100,0,255,"minPix",this); 
			thePars[1] = new SegPar(1,1,255,"iter8",this); 
			nparams=2;
			for(i=0;i<nparams;i++) getContentPane().add(thePars[i]);
			addShowJButton();
			break;
		case 11: // undust; 1 pars
			thePars[0] = new SegPar(1,0,50,"grain",this); 
			nparams=1;
			for(i=0;i<nparams;i++) getContentPane().add(thePars[i]);
			addShowJButton();
			break;
		case 12: // ThreshAnd; 2 pars
			thePars[0] = new SegPar(128,0,255,"Athresh",this); 
			thePars[1] = new SegPar(128,0,255,"Bthresh",this); 
			nparams=2;
			for(i=0;i<nparams;i++) getContentPane().add(thePars[i]);
			addShowJButton();
			break;
		case 13: // FocusFltr; 1 pars
			thePars[0] = new SegPar(8,0,30,"rad",this); 
			nparams=1;
			for(i=0;i<nparams;i++) getContentPane().add(thePars[i]);
			addShowJButton();
			break;
		case 14: //   DimEdgesFltr; 2 par
			thePars[0] = new SegPar(70,0,255,"thresh",this); 
			thePars[1] = new SegPar(1,1,12,"iter8",this); 
			nparams=2;
			for(i=0;i<nparams;i++) getContentPane().add(thePars[i]);
			addShowJButton();
			break;
		case 15: //   Zgrad; 4 par
			thePars[0] = new SegPar(1.0,0.0,10.,"botBoost",this); 
			thePars[1] = new SegPar(1.0,0.0,10.,"topBoost",this); 
			thePars[2] = new SegPar(1,1,2048,"midDepth",this); 
			thePars[3] = new SegPar(1.0,0.0,10.,"midBoost",this); 
			nparams=4;
			for(i=0;i<nparams;i++) getContentPane().add(thePars[i]);
			addShowJButton();
			break;
		case 16: //   FillThread; 0 par
			nparams=1;
			thePars[0] = new SegPar(1,-7,256,"iter8",this); 
			for(i=0;i<nparams;i++) getContentPane().add(thePars[i]);
			addShowJButton();
			break;
		case 17: //   AndOrAvg filter, 3 par
			nparams=3;
			thePars[0] = new SegPar(1,1,4,"select",this); 
			thePars[1] = new SegPar(128,0,255,"Athresh",this); 
			thePars[2] = new SegPar(128,0,255,"Bthresh",this); 
			for(i=0;i<nparams;i++) getContentPane().add(thePars[i]);
			addShowJButton();
			break;
		case 18 : // null
			break;
		case 19: // chan1
		case 20: // chan2
		case 21: // chan3
			addShowJButton();
			break;
		case 22: // mask1
		case 23: // mask2
		case 24: // mask3
		case 25: // mask4
			addMaskJButtons();
			break;
		}
	for(i=0;i<nparams;i++)
		getContentPane().add(thePars[i]);
	}
	
private void addShowJButton() {
	JPanel showHelp = new JPanel();
	showHelp.setLayout(new GridLayout(1,2));
	showHelp.add(wmk.makeAJButton("show"));
	showHelp.add(wmk.makeAJButton("help"));
	getContentPane().add(showHelp);
	}
		
private void addMaskJButtons() {
	getContentPane().setLayout(new GridLayout(0,1));
	getContentPane().add(wmk.makeAJButton("showMask"));
	getContentPane().add(wmk.makeAJButton("Data&Mask"));
	getContentPane().add(wmk.makeAJButton("DataThruMask"));
	getContentPane().add(wmk.makeAJButton("JustData"));
	getContentPane().add(wmk.makeAJButton("PickColors"));
	}
		
public void actionPerformed(java.awt.event.ActionEvent event) {
		String arg = event.getActionCommand();
	    if(arg.equals("show")){
	    	parent.setShowNode(myPath,myStep);
	    	return;
    		}
	    if(arg.equals("help")){
	    	helpMessage();
	    	return;
    		}
	    if(arg.equals("avail")) {
	    	talk("available button...");
	    	return;
    		}
	    if(arg.equals("showMask")) {
	    	parent.setShowNode(myPath,myStep);
	    	talk("show the raw mask at row="+myPath+", col="+myStep+" in black.white");
	    	MaskColorsData=0;
	    	return;
    		}
    		
	   if(arg.equals("PickColors")) { 
	    	// select colors for displaying mask and data
	    	parent.tellAllColors();
	    	if(displayColorPicker==null) 
	    		switch (whatAmI)  {
	    			default:
	    			case 22 : // red mask
	    				displayColorPicker = new TwoColorPicker("chan1 colors",
	    					"chan1, selected bits", "maskSelects",
	    					"chan1, discarded bits", "maskRejects", parent.maskrgbs[0]);
	    				break;
	    			case 23 : // green mask
	    				displayColorPicker = new TwoColorPicker("chan2 colors",
	    					"chan2, selected bits", "maskSelects",
	    					"chan2, discarded bits", "maskRejects", parent.maskrgbs[1]);
	    				break;
	    			case 24 : // blue mask
	    				displayColorPicker = new TwoColorPicker("chan3 colors",
	    					"chan3, selected bits", "maskSelects",
	    					"chan3, discarded bits", "maskRejects", parent.maskrgbs[2]);
	    				break;
	    			}
	    	else displayColorPicker.showYourself();
	    	return;
    		}
    		
	    if(arg.equals("Data&Mask")) {
	    	parent.setShowNode(myPath,myStep);
	    	talk("show input data colored through mask at row="+myPath+", col="+myStep+" (but which input channel?)");
	    	MaskColorsData=1;
	    	return;
    		}
    		
	    if(arg.equals("DataThruMask")) {
	    	parent.setShowNode(myPath,myStep);
	    	talk("show input data colored through mask at row="+myPath+", col="+myStep+" (but which input channel?)");
	    	MaskColorsData=2;
	    	return;
    		}
    		
	    if(arg.equals("JustData")) {
	    	parent.setShowNode(myPath,myStep);
	    	talk("show input data colored through mask at row="+myPath+", col="+myStep+" (but which input channel?)");
	    	MaskColorsData=3;
	    	return;
    		}
    	return;
    	}

	public void newGeometry(int path, int step,int xcen,int ycen,int width,int maxHeight){
		if(whatAmI<0) return;
		this.width=width;
		parJPanelHeight=setHeight(maxHeight);
		this.height=maxHeight;
		if(whatAmI<=nullID) {
			myStep=step;
			myPath=path;
			myXcen=xcen;
			myYcen=ycen;
			}
		else {
			setSpecialCoords();
			}
		}
		
	public void display(int xparent,int yparent) {
		int rectHeight=parJPanelHeight;
		if(whatAmI >= 22 && whatAmI <= 25) // it's a mask output...
			rectHeight = 30;
		setLocation(xparent+myXcen-width/2,yparent+myYcen + 30 - rectHeight/4);
		if(ifShowSelf) {
			//talk("display: row "+myPath+", col "+myStep+" should be visible");
			setVisible(true);
			}
		else {
			//talk("row "+myPath+", col "+myStep+" is completely invisible");
			setVisible(false);
			}		
		}
		
	public void appear() {
		//talk("appearance of row "+myPath+", col "+myStep);
		ifShowSelf=true;
		}
	
	public void disappear() {
		//talk("row "+myPath+", col "+myStep+" disappears");
		ifShowSelf=false;
		}
	
	public void eraseInputLinks() {
		eraseInput1Links();
		eraseInput2Links();
		}
			
	public void eraseInput1Links() {
		inStep1=-1; inPath1=-1;
		anyChangeSinceLastRun=true;
		}
		
	public void eraseInput2Links() {
		if(whatAmI <0) return;
		if(if2Inputs[whatAmI]) {
			inPath2=-1; inStep2=-1;}
		else {
			inPath2=-2; inStep2=-2;}
		anyChangeSinceLastRun=true;
		}
			
	public void drawNode(GraphicsPs gg) {
		int rady = (3*height)/4;
		int radx = (3*width)/4;
		int rectHeight=parJPanelHeight;
		if(whatAmI >= 22 && whatAmI <= 25) // it's a mask output...
			rectHeight = 30;
		if(whatAmI<0) return;
		if(whatAmI<nullID) {
			if(justRun) {
				gg.newColor(255,255,255);
				gg.drawLine(myXcen+10, myYcen-radx/2-8,myXcen+10, myYcen+radx/2+8);
				}
			gg.newColor(myred,mygreen,myblue);
			gg.fillCenteredOval(myXcen, myYcen, radx, rady);
			if(ifActive) {
				gg.newColor(255,255,0);
				gg.drawCenteredOval(myXcen, myYcen, radx, rady);
				gg.drawCenteredOval(myXcen, myYcen, 2+radx, 2+rady);
				}
			gg.newColor(myTextred,myTextgreen,myTextblue);
			gg.drawString(whats[whatAmI],myXcen-(2*radx)/5, 5+myYcen);
			}
		else if(whatAmI>nullID){ // draw a rectangle... 
			if(justRun) {
				gg.newColor(255,255,255);
				gg.drawLine(myXcen+10, myYcen-rectHeight/2-8,myXcen+10, myYcen+rectHeight/2+8);
				}
			gg.newColor(myred,mygreen,myblue);
			gg.fillCenteredRect(myXcen,myYcen,width/2,rectHeight);
			if(ifActive) {
				gg.newColor(255,255,0); // yellow bounds mark active nodes
				gg.drawCenteredRect(myXcen,myYcen,width/2,rectHeight);
				gg.drawCenteredRect(myXcen,myYcen,2+width/2,2+rectHeight);
				}
			gg.newColor(myTextred,myTextgreen,myTextblue);
			gg.drawString(whats[whatAmI],myXcen-width/5, 5+myYcen);
			}
		}
	
	public boolean mouseHitMeQ(int x, int y) {
		if(whatAmI<0) return false;
		int rady = (3*height)/4;
		int radx = (3*width)/4;
		if(whatAmI>nullID) rady=parJPanelHeight;
		if(x<myXcen-radx/2 || x>myXcen+radx/2 || y<myYcen-rady/2 ||y>myYcen+rady/2)
			return false;
		else return true;
		}
	
	public int availableInput() {
		if(inStep1<0 && inPath1<0) return 1;
		else if(inStep2==(-1) && inPath2==(-1)) return 2;
		else return (-1);
		}
		
static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}

	// catch WindowListener events  
	public void windowActivated(java.awt.event.WindowEvent event) {
		// talk("window activated");
		}
	public void windowDeactivated(java.awt.event.WindowEvent event) {
		//  talk("window Deactivated");
		}
	public void windowOpened(java.awt.event.WindowEvent event) {}
	public void windowIconified(java.awt.event.WindowEvent event) {}
	public void windowDeiconified(java.awt.event.WindowEvent event) {}
	public void windowClosing(java.awt.event.WindowEvent event) {
		parent.hideAllParamJPanels(); // hide all the parameter panels...
		}
		
	public void windowClosed(java.awt.event.WindowEvent event) {
		}


//catch Choice events...	
	public void itemStateChanged(java.awt.event.ItemEvent event) {
		String arg = (String)(((JComboBox)(event.getSource())).getSelectedItem());
		if(whatAmI < 0) return; // should be 0
	    	for(int i=0;i<possibilities;i++)
	    	  if(arg.equals(whats[i])) {
	    		if(i==whatAmI) {
	    			return; // because nothing changed...
	    			}
	    	    //talk("SegmenterNode become "+ whats[i]);
	    		if(if2Inputs[i]) { // two inputs are possible
	    			//talk("SegmenterNode  two inputs are possible ");
	    			if(inPath2<-1) inPath2=-1;
	    			if(inStep2<-1) inStep2=-1;
	    			}
	    		else { // only one input is possible
	    			inPath2=-2;
	    			inStep2=-2;
	    			}
	    		if(i==nullID) { // null object
	    			if(parent!=null) {
						parent.cutLinksTo(myPath,myStep);
						parent.repaintAll();
						}
	    			inPath2=-2;
	    			inStep2=-2;
	    			inPath1=-2;
	    			inStep1=-2;
	    			}
				if(parent!=null) 
	    			parent.replaceNode(this,i); // this will lead to a call of this object's sepuku method...
    			}
	}
	
public boolean 	readParameterValues() {
	// for any textField that is not null, read in the parameter value and compare it to previous value...
	boolean someParameterChanged=false;
	int i;
		for(i=0;i<nparams;i++) 
			if(thePars[i] == null) talk("OOPs, SegPar has null thePars");
			else thePars[i].getVal(); // this will set someParameterChanged==true if any parameter changed
	return someParameterChanged; 
	}

public void someParChanged() {
	anyChangeSinceLastRun=true;
	}
	
public void writeNodeData(ArchiveMarks disker) {
	}

	private void makeShortRas() {
		if(rasRows != parent.parent.fullHeight || 
				rasCols != parent.parent.fullWidth) {
					rasRows = parent.parent.fullHeight;
					rasCols = parent.parent.fullWidth;
					shortRas = null;
					}
		if(shortRas==null)
			shortRas = new short[rasRows*rasCols];
		}
		
public void runTheFilter(long calcGen) { 
	int i,j,npix;
	if(!ifActive || whatAmI==nullID || whatAmI<0) return; // only do it if this node is currently active
	if(imageInputNode>=0) { // sectioNumber is the number of the section to use in the main image panel
		if(sectionNumber != parent.parent.curStack.curJFrame) {
			sectionNumber = parent.parent.curStack.curJFrame;
			anyChangeSinceLastRun=true;
			talk("Now using frame nmber "+sectionNumber+" as input to chan1, chan2, chan3");
			}
		}
	
	else {
		// see if inputs are newer than or as new as the current calcGen
		if(inPath1>=0 && inStep1>=0) {
			if(parent.nodes[inPath1][inStep1].gen >= calcGen)
				anyChangeSinceLastRun=true; // because there is new input 1 for this node
			if(parent.nodes[inPath1][inStep1].gen != in1Gen)
				anyChangeSinceLastRun=true; // because there is new input 1 for this node
			}
		if(inPath2>=0 && inStep2>=0) {
			if(parent.nodes[inPath2][inStep2].gen >= calcGen)
				anyChangeSinceLastRun=true; // because there is new input 2 for this node
			if(parent.nodes[inPath2][inStep2].gen != in2Gen)
				anyChangeSinceLastRun=true; // because there is new input 1 for this node
			}
		
		if(readParameterValues() || howManyParametersToVary()>0) 
		// readParameterValues() returns true if any parameter value changed...
		// if howManyParametersToVary()>0 then some parameter will change when getMutantVal is called
			anyChangeSinceLastRun=true;
		}
		
	  if(!anyChangeSinceLastRun) return;  // leaving gen at older value.
	
	// here make a big switch that calls the code to do whatever the filter does.
	if(imageInputNode<0) { // not a raw channel image input node
		doTheFilterOperation();
		}
		
	else { // get input image from the appropriate input channel
		BioradFile chanfile;
		switch (whatAmI) {
			case 19 : // red channel
				chanfile=parent.parent.curStack.redfile;
				if(chanfile != null)
					break;
			case 20 : // green channel
				chanfile=parent.parent.curStack.greenfile;
				if(chanfile != null)
					break;
			case 21 : // blue channel
				chanfile=parent.parent.curStack.bluefile;
				break;
			default :
				chanfile=null;
			}
		if(chanfile != null) {
			short apix;
			byte [] byteras = chanfile.readShiftedBioradRaster(sectionNumber);
			// this bypasses caching
			makeShortRas();
			if(shortRas !=null && byteras != null) {
				parent.curJFrameDepth = parent.parent.curStack.NJFrames;
				parent.curSect = parent.parent.curStack.lastJFrameShown;
				npix = rasRows*rasCols;
				for(i=0;i<npix;i++)  {
					if((apix=byteras[i]) < 0) 
						apix += 256;
					shortRas[i] = apix;
					}
				}
			}
		}
	gen=calcGen;
	justRun=true;
	anyChangeSinceLastRun=false;  // any parameter change, or change of input links will change this to true
	if(imageInputNode<0) { // not a raw channel image input node
		if(inPath1>=0 && inStep1>=0) // update reckoning of input calc gen on input1
			in1Gen=parent.nodes[inPath1][inStep1].gen;
		if(inPath1>=0 && inStep2>=0) // update reckoning of input calc gen on input2
			in2Gen=parent.nodes[inPath2][inStep2].gen;
		}
	}

	private void helpMessage() {
	  switch (whatAmI) {
		case 0:  // wire; no parameters
			talk("The wire transfers its single input raster to its output raster, unchanged");
			break;
		case 1: // FlattenFltr;
			if(flatten!=null) flatten.tellWhatYouDo();
			break;
		case 2: //  erode; 2 pars
			if(erode!=null) erode.tellWhatYouDo();
			break;
		case 3:  // thicken; 3 pars
			if(thicken!=null) thicken.tellWhatYouDo();
			break;

		case 4: //  ridges; 3 pars (2 are doublePrecision)
			if(ridges != null) ridges.tellWhatYouDo();
			break;
			
		case 5: //   BlkWht; 1 par
			if(BlkWht!=null) BlkWht.tellWhatYouDo();
			break;
		case 6:  // lowpass; 2 pars
			if(lowpass!=null) lowpass.tellWhatYouDo();
			break;
		case 9: // invert; no parameters
			if(inversion!=null)  inversion.tellWhatYouDo();
			break;
		case 11: // undust; 1 pars
			if(unduster!=null)  unduster.tellWhatYouDo();
			break;
			
		case 8: //  NiceTubes; 3 pars: int spotthresh,double linetol, double maxtubewidth
			if(tuber!=null) tuber.tellWhatYouDo();
			break;

		case 7: //   makeBalls; 2 pars -- one input
			if(makeBalls!=null)  makeBalls.tellWhatYouDo();
			break;

		case 10: // skel; 2 pars
			if(skeletor!=null)  skeletor.tellWhatYouDo();
			break;

		case 12: // ThreshAnd; 2 pars -- two inputs
			if(ThreshAnd!=null)  ThreshAnd.tellWhatYouDo();
			break;
		case 13: // FocusFltr; 1 pars
			if(focsharp!=null)  focsharp.tellWhatYouDo();
			break;
		case 14: //   DimEdgesFltr; 1 par
			if(dimedges!=null)  dimedges.tellWhatYouDo();
			break;
		case 15: //   ZgradFltr; 1 par
			if(zgrad!=null)  zgrad.tellWhatYouDo();
			break;
		case 16: //   FillThread; 0 par
			if(gapplugger!=null)  gapplugger.tellWhatYouDo();
			break;
		default : // who knows?
			// for trials, just take image from first input
			break;
		case 17: // AndOrAvg; 3 pars
			if(andoravg != null)  andoravg.tellWhatYouDo();
			break;
		case 18: // null; no pars
			talk("null element does nothing");
			break;
		case 22: // mask1
		case 23: // mask2
		case 24: // mask3
		case 25: // mask4
			break;			
	  	}
		talk("----------------------------------------------------------------------------");
		talk("");
	}
	
	private void makeTheFilter() {
	  switch (whatAmI) {
		case 1: // FlattenFltr; 
			if(flatten==null) flatten=new FlattenFltr(this);
			break;
		case 2: //  erode; 2 pars
			if(erode==null) erode=new ErodeFltr(this);
			break;
		case 3:  // thicken; 3 pars
			if(thicken==null) thicken=new ThickenFltr(this);
			break;

		case 4: //  ridges; 3 pars (2 are doublePrecision)
			if(ridges == null) ridges=new RidgesFltr(this);
			break;
			
		case 5: //   BlkWht; 1 par
			if(BlkWht==null) BlkWht=new BlackWhiteFltr(this);
			break;
		case 6:  // lowpass; 2 pars
			if(lowpass==null) lowpass=new LowPassFltr(this);
			break;
		case 9: // invert; no parameters
			if(inversion==null) inversion=new InvertFltr(this);
			break;
		case 11: // undust; 1 pars
			if(unduster==null) unduster=new UndustFltr(this);
			break;
			
		case 8: //  NiceTubes; 3 pars: int spotthresh,double linetol, double maxtubewidth
			if(tuber==null) 
				tuber = new NiceTubes(this);
			break;

		case 7: //   NiceBalls; 2 pars -- one input
			if(makeBalls==null) 
				makeBalls=new NiceBalls(this);
			break;

		case 10: // skel; 2 pars
			if(skeletor==null) skeletor=new SkelFltr(this);
			break;

		case 12: //  ThreshAnd; 2 pars -- two inputs
			if(ThreshAnd==null) ThreshAnd=new ThreshAndFltr(this);
			break;
		case 13: // FocusFltr; 3 pars
			if(focsharp == null) focsharp = new FocusFltr(this);
			break;
		case 14: //   DimEdgesFltr; 1 par
			if(dimedges==null) dimedges=new DimEdgesFltr(this);
			break;
		case 15: //   ZgradFltr; 1 par
			if(zgrad==null) zgrad=new ZGradFltr(this);
			break;
		case 16: //   FillThread; 0 par
			if(gapplugger==null) gapplugger=new FillThread(this);
			break;
		case 17: //   AndOrAvg; 3 par
			if(andoravg == null) andoravg=new AndOrAvgFltr(this);
			break;
		case 0:  // wire; no parameters
		default : // who knows?
			// for trials, just take image from first input
			break;
		case 18: // null; no pars
			break;
		case 22: // mask1
		case 23: // mask2
		case 24: // mask3
		case 25: // mask4
			break;			
	  	}
	}
		
	public void doTheFilterOperation() {
		int offset=0;
	  if(inPath1<0 || inStep1<00) { 
	  		talk("inPath1 <0 or inStep1<0 at path="+myPath+", and step="+myStep); 
			shortRas=null; rasRows = -1;  rasCols = -1;
			return;
			}
	  int inRasCols = (parent.nodes[inPath1][inStep1]).rasCols;
	  int inRasRows = (parent.nodes[inPath1][inStep1]).rasRows;
	  short [] inRas = (parent.nodes[inPath1][inStep1]).shortRas;
	  
	  int inRasCols2=2,inRasRows2=2;
	  short [] inRas2=null;
	  for(int i=0;i<nparams;i++) {
	  		if(parent.ifVaryingParameters && thePars[i].ifVary) 
	  			thePars[i].getMutantVal();
	  		else 
	  			thePars[i].getVal();	  		 
	  		} 
	  		
	  // talk("whatAmI="+whatAmI+", which is "+whats[whatAmI]);
	  	
	  switch (whatAmI) {
		case 1: // FlattenFltr;
			makeShortRas();
			if(flatten==null) flatten=new FlattenFltr(this);
			if(!flatten.doit(inRas,inRasCols,inRasRows,thePars[0].intVal,thePars[1].intVal,thePars[2].intVal,shortRas)) 
				talk("FlattenFltr failed at path="+myPath+", and step="+myStep);
			break;
		case 2: //  erode; 2 pars
			makeShortRas();
			if(erode==null) erode=new ErodeFltr(this);
			if(!erode.doit(inRas,inRasCols,inRasRows,thePars[0].intVal,thePars[1].intVal,shortRas)) 
				talk("ErodeFltr failed at path="+myPath+", and step="+myStep);
			break;
		case 3:  // thicken; 3 pars
			makeShortRas();
			if(thicken==null) thicken=new ThickenFltr(this);
			if(!thicken.doit(inRas,inRasCols,inRasRows,thePars[0].intVal,thePars[1].intVal,thePars[2].intVal,shortRas)) 
				talk("ThickenFltr failed at path="+myPath+", and step="+myStep);
			break;

		case 4: //  ridges; 3 pars (2 are doublePrecision)
			makeShortRas();
			if(ridges == null) ridges=new RidgesFltr(this);
			if(!ridges.doit(inRas, inRasCols, inRasRows, thePars[0].intVal, 
				 .0001*(thePars[1].intVal), .0001*(thePars[2].intVal), shortRas))
					talk("RidgesFltr failed at path="+myPath+", and step="+myStep);
			break;
			
		case 5: //   BlkWht; 1 par
			makeShortRas();
			if(BlkWht==null) BlkWht=new BlackWhiteFltr(this);
			if(!BlkWht.doit(inRas,inRasCols,inRasRows,thePars[0].intVal,shortRas)) 
				talk("BlkWhtFltr failed at path="+myPath+", and step="+myStep);
			break;
		case 6:  // lowpass; 2 pars
			makeShortRas();
			if(lowpass==null) lowpass=new LowPassFltr(this);
//              doit(short [] inras, int incols, int inrows, int weight, int npass, short [] outras)
			if(!lowpass.doit(inRas,inRasCols,inRasRows,thePars[0].intVal,thePars[1].intVal,shortRas)) 
				talk("LowPassFltr failed at path="+myPath+", and step="+myStep);
			break;
		case 9: // invert; no parameters
			makeShortRas();
			if(inversion==null) inversion=new InvertFltr(this);
			if(!inversion.doit(inRas,inRasCols,inRasRows,shortRas))
				talk("InvertFltr failed at path="+myPath+", and step="+myStep);
			break;
		case 10: // skel; 2 pars
			makeShortRas();
			if(skeletor==null) skeletor=new SkelFltr(this);
//              doit(short [] inras, int incols, int inrows, int minPix, int iter8, short [] outras)
			talk("iter8 in skeletor is "+thePars[1].intVal);
			if(!skeletor.doit(inRas,inRasCols,inRasRows,thePars[0].intVal,thePars[1].intVal,shortRas)) 
				talk("SkelFltr failed at path="+myPath+", and step="+myStep);
			break;
		case 11: // undust; 1 pars
			makeShortRas();
			if(unduster==null) unduster=new UndustFltr(this);
			if(!unduster.doit(inRas,inRasCols,inRasRows,thePars[0].intVal,shortRas)) 
				talk("UndustFltr failed at path="+myPath+", and step="+myStep);
			break;
			
		case 7: //   makeBalls; 2 pars -- one inputs
			if(makeBalls==null) 
				makeBalls=new NiceBalls(this);
			makeShortRas();
			if(!makeBalls.doit(inRas, inRasCols,inRasRows,  thePars[0].intVal,  .0001*(thePars[1].intVal),
				                  inRas2,inRasCols2,inRasRows2,shortRas))  
					talk("NiceBalls failed at path="+myPath+", and step="+myStep);
			break;
			
		case 8: //  NiceTubes; 3 pars: int spotthresh,double linetol, double maxtubewidth
			if(tuber==null) 
				tuber = new NiceTubes(this);
			makeShortRas();
			if(!tuber.doit(inRas,inRasCols,inRasRows,thePars[0].intVal,.0001*(thePars[1].intVal),
					.0001*(thePars[2].intVal),shortRas)) 
						talk("NiceTubes failed at path="+myPath+", and step="+myStep);
			break;

		case 12: // ThreshAnd; 2 pars -- two inputs
			if(inPath2<0 || inStep2<0) {
	  			talk("inPath2 <0 or inStep2<0 at path="+myPath+", and step="+myStep); 
				shortRas=null; rasRows = -1;  rasCols = -1;
				return;
				}
	  		inRasCols2 = (parent.nodes[inPath2][inStep2]).rasCols;
	  		inRasRows2 = (parent.nodes[inPath2][inStep2]).rasRows;
	  		inRas2 = (parent.nodes[inPath2][inStep2]).shortRas;
			if(inRas2==null || inPath2<0 || inStep2<0 || inRasCols2 != inRasCols || inRasRows2 != inRasRows) {
	  			talk("screwup at path="+myPath+", and step="+myStep); 
	  			talk("inRas2="+inRas2);
	  			talk("inPath2="+inPath2);
	  			talk("inStep2="+inStep2);
	  			talk("inRasCols2="+inRasCols2+" =?= ");
	  			talk("                      inRasCols="+inRasCols);
	  			talk("inRasRows2="+inRasRows2+" =?= ");
	  			talk("                      inRasRows="+inRasRows);
				shortRas=null; rasRows = -1;  rasCols = -1;
				return;
				}
			makeShortRas();
				talk("Some node becomes an ThreshAndFltr");
				if(ThreshAnd==null) ThreshAnd=new ThreshAndFltr(this);
				if(!ThreshAnd.doit(inRas, inRasCols,inRasRows,  thePars[0].intVal,
				                  inRas2,inRasCols2,inRasRows2,thePars[1].intVal,shortRas)) 
					talk("ThreshAndFltr failed at path="+myPath+", and step="+myStep);
			break;
		case 13: // FocusFltr; 1 pars
			if(focsharp==null) focsharp=new FocusFltr(this);
			makeShortRas();
			if(!focsharp.doit(inRas,inRasCols,inRasRows, thePars[0].intVal, shortRas)) 
				talk("FocusFltr failed at path="+myPath+", and step="+myStep);
			break;
		case 14: //   DimEdgesFltr; 1 par
			makeShortRas();
			if(dimedges==null) dimedges=new DimEdgesFltr(this);
			if(!dimedges.doit(inRas,inRasCols,inRasRows, thePars[0].intVal, thePars[1].intVal, shortRas)) 
				talk("DimEdgesFltr failed at path="+myPath+", and step="+myStep);
			break;
		case 15: //   ZgradFltr; 1 par
			makeShortRas();
			if(zgrad==null) zgrad=new ZGradFltr(this);
			int junk=parent.curSect/2;
			if(!zgrad.doit(inRas,inRasCols,inRasRows, .0001*(thePars[0].intVal), .0001*(thePars[1].intVal),
				thePars[2].intVal, .0001*(thePars[3].intVal), parent.curJFrameDepth, parent.curSect,parent.ifXYsect, shortRas)) 
					talk("ZGradFltr failed at path="+myPath+", and step="+myStep);
			break;
		case 16: //   FillThread; 0 par
			makeShortRas();
			if(gapplugger==null) gapplugger=new FillThread(this);
			if(!gapplugger.doit(inRas,inRasCols,inRasRows,thePars[0].intVal,shortRas)) 
				talk("FillThread failed at path="+myPath+", and step="+myStep);
			break;
		case 17: //   AndOrAvg, 3 par
			if(inPath2<0 || inStep2<0) {
	  			talk("inPath2 < 0 or inStep2 < 0 at path="+myPath+", and step="+myStep); 
				shortRas=null; rasRows = -1;  rasCols = -1;
				return;
				}
	  		inRasCols2 = (parent.nodes[inPath2][inStep2]).rasCols;
	  		inRasRows2 = (parent.nodes[inPath2][inStep2]).rasRows;
	  		inRas2 = (parent.nodes[inPath2][inStep2]).shortRas;
			if(inRas2==null || inPath2<0 || inStep2<0 || inRasCols2 != inRasCols || inRasRows2 != inRasRows) {
	  			talk("screwup at path="+myPath+", and step="+myStep); 
	  			talk("inRas2="+inRas2);
	  			talk("inPath2="+inPath2);
	  			talk("inStep2="+inStep2);
	  			talk("inRasCols2="+inRasCols2+" =?= ");
	  			talk("                      inRasCols="+inRasCols);
	  			talk("inRasRows2="+inRasRows2+" =?= ");
	  			talk("                      inRasRows="+inRasRows);
				shortRas=null; rasRows = -1;  rasCols = -1;
				return;
				}
			makeShortRas();
			if(andoravg==null) andoravg=new AndOrAvgFltr(this);			
			if(!andoravg.doit(inRas, inRasCols,inRasRows,  thePars[0].intVal,  thePars[1].intVal,
				                  inRas2,inRasCols2,inRasRows2,thePars[2].intVal,shortRas)) 
					talk("AndOrAvgFltr failed at path="+myPath+", and step="+myStep);
			break;
			
		case 0:  // wire; no parameters
		default : // who knows?
			// for trials, just take image from first input
			shortRas=(parent.nodes[inPath1][inStep1]).shortRas;
			rasRows=(parent.nodes[inPath1][inStep1]).rasRows;
			rasCols=(parent.nodes[inPath1][inStep1]).rasCols;
			break;
		case 18: // null; no pars
			shortRas=null; rasRows = -1;  rasCols = -1;
			break;
		case 22: // mask1
		case 23: // mask2
		case 24: // mask3
		case 25: // mask4
			rasRows=(parent.nodes[inPath1][inStep1]).rasRows;
			rasCols=(parent.nodes[inPath1][inStep1]).rasCols;
			shortRas=(parent.nodes[inPath1][inStep1]).shortRas;
			break;
			}
		}
		
				
}
