package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.lang.*;
import java.util.*;
import java.awt.*;  import javax.swing.*;



class ShowLUT extends JFrame  implements java.awt.event.ActionListener,java.awt.event.WindowListener {
	int width=256,height=300;
	LUTCanvas canvas;
	JTextField lam1=null,lam2=null,fastr=null,conc255=null;
	double lambda1=0., lambda2=0., fastrat=.5, concentrationMappedTo255=.1;
	int ifRGB=0;
	String colorWord="red";
	double [] gainCurveFitCoeffs={0.,0.,0.,0.}; // can be set by a CallibrationStats window.
	byte [] gainCurveLUT=null;
	int modeLUT = 0;
	
	public void setNewGainCurveFit(double 	concentrationMappedTo255, double [] newcoeffs,
				boolean ifPurgeCache) { 
		this.concentrationMappedTo255 = concentrationMappedTo255;
		conc255.setText(""+concentrationMappedTo255);
		if(newcoeffs == null) {
			assertRawDataTransformationMode(0);
			talk("OOps setNewGainCurveFit called with null gainCurveFitCoeffs[] array");
			return;
			}
		for(int i=0;i<3;i++) {
	    	gainCurveFitCoeffs[i]=newcoeffs[i];
	    	canvas.myMaster.gainCurveFitCoeffs[i]=newcoeffs[i];
	    	}
		/* talk("gainCurveFitCoeffs transfered to "+colorWord+" LUT =  ("+
			gainCurveFitCoeffs[0]+", "+gainCurveFitCoeffs[1]+", "+
				gainCurveFitCoeffs[2]+")"); */
		makeGainCurveCorrectionLUT();
		//talk(colorWord+" concentration mapped to intensity 255 = "+concentrationMappedTo255);
		if(ifPurgeCache) 
			canvas.myMaster.myBioradFile.cachePurge();
		}
	
	public void makeGainCurveCorrectionLUT() {
		gainCurveLUT = new byte[384];
		int i,k;
		int [] theLUT = new int[256];
		String stg;
		for(i=0;i<384;i++) gainCurveLUT[i]=0;
		getLambdas(); // in case they changed.
		for(i=0;i<256;i++) {
			if(i<128)
				k = 256+i;
			else 
				k = i;
			theLUT[i] = gainCorrectedLookupTable(i);
			gainCurveLUT[k] = (byte)theLUT[i];
			}
		}
		
   public int gainCorrectedLookupTable(int pix) {
 	double dpix = pix;
 	double linConc = gainCurveFitCoeffs[0] + dpix*(gainCurveFitCoeffs[1] + dpix*gainCurveFitCoeffs[2]);
	pix = (int)(.499 + 255.0*linConc/concentrationMappedTo255);
   	if(pix>255) return 255;
   	if(pix<0) return 0;
   	return pix;
    }

	public int assertRawDataTransformationMode(int modeLUT) { // try to change mode
		// return the mode actually set.
		// modeLUT==0 means no transformation.
		// modeLUT==1 means use the gain curve that came from the callibrationStats panel
		// modeLUT==2 means use the Ungain/Unbleach  corrections specified by the
		//          special header info encoded in the first section of the stack, if it's there.
		int was = this.modeLUT;
		this.modeLUT = modeLUT;
		if(modeLUT==1) {
			// turn it on if there are actually some gainCurveFitCoeffs defined
			if(gainCurveFitCoeffs!=null) 
				setNewGainCurveFit(concentrationMappedTo255, gainCurveFitCoeffs,false);							
			else this.modeLUT = was;
			}
		if(was != this.modeLUT)
			canvas.myMaster.myBioradFile.cachePurge();
		return this.modeLUT;
		}

	public void getLambdas() {
	   lambda1 = (Double.valueOf(lam1.getText()).doubleValue());
	   lambda2 = (Double.valueOf(lam2.getText()).doubleValue());
	   fastrat = (Double.valueOf(fastr.getText()).doubleValue());
	   concentrationMappedTo255 = (Double.valueOf(conc255.getText()).doubleValue());
	   canvas.myMaster.lambda1=lambda1;
	   canvas.myMaster.lambda2=lambda2;
	   canvas.myMaster.fastrat=fastrat;
	   canvas.myMaster.concentrationMappedTo255=concentrationMappedTo255;
	   for(int i=0;i<3;i++)
	   		canvas.myMaster.gainCurveFitCoeffs[i] = gainCurveFitCoeffs[i];	   	
	   // talk("lambda1="+lambda1+", lambda2="+lambda2+", fastrat="+fastrat+", concentrationMappedTo255="+
		}

	public void setLambdas(double lambda1,double lambda2,double fastrat,double concentrationMappedTo255) {
		this.lambda1=lambda1;
		this.lambda2=lambda2;
		this.fastrat=fastrat;
		this.concentrationMappedTo255 = concentrationMappedTo255;
		lam1.setText(""+lambda1);
		lam2.setText(""+lambda2);
		fastr.setText(""+fastrat);
		conc255.setText(""+concentrationMappedTo255);
		}
			
	public ShowLUT(int [] theLUT, String stg, int ifRGB,Color col1, Color col2, Color col3, Color col4) {
		super("Identity Map");
		this.ifRGB=ifRGB;
		setBackground(new Color(64, 0, 64));
		getContentPane().setLayout(new BorderLayout());
		setSize(width,height);
		canvas = new LUTCanvas(this,theLUT,stg,col1, col2, col3, col4);
		lam1 = new JTextField("0.0");
		lam2 = new JTextField("0.0");
		fastr = new JTextField("0.5");
		conc255 = new JTextField("0.1");
		JPanel gampars = new JPanel();
		gampars.setBackground(col3);
		gampars.setLayout(new GridLayout(2,4));
		
		Color texthue=Color.cyan;
		colorWord="red";
		if(ifRGB==1) {colorWord="green"; texthue=Color.white;}
		if(ifRGB==2) {colorWord="blue"; texthue=Color.yellow;}

		JLabel alab = new JLabel("lambda1");
		alab.setForeground(texthue);
		gampars.add(alab);
		
		alab = new JLabel("lambda2");
		alab.setForeground(texthue);
		gampars.add(alab);

		alab = new JLabel("fastrat");
		alab.setForeground(texthue);
		gampars.add(alab);

		alab = new JLabel("conc->255");
		alab.setForeground(texthue);
		gampars.add(alab);

		lam1.setForeground(texthue);
		lam2.setForeground(texthue);
		fastr.setForeground(texthue);
		conc255.setForeground(texthue);
		gampars.add(lam1);
		gampars.add(lam2);
		gampars.add(fastr);		
		gampars.add(conc255);		
		getContentPane().add("Center",canvas);
		getContentPane().add("South",gampars);
		JMenu showLUTJMenu = new JMenu("print LUT");
		showLUTJMenu.add(new JMenuItem("print the LUT values"));
		showLUTJMenu.add(new JMenuItem("print Gain/Bleach correction data"));
		showLUTJMenu.add(new JMenuItem("write Gain/Bleach correction data to disk file"));
		showLUTJMenu.add(new JMenuItem("Initiate Auto Bleach Correction; write BioradFileHeader"));
		JMenuBar mb = new JMenuBar();
		mb.add(showLUTJMenu);
		setJMenuBar(mb);
		showLUTJMenu.addActionListener(this);
		addWindowListener(this);
		}
		
// catch button and menu events...
	public void actionPerformed(java.awt.event.ActionEvent event) {
		String arg = event.getActionCommand();
	    	if(arg.equals("print the LUT values")) {
	    		canvas.printCurrentLUT();
	    		}
	    	if(arg.equals("print Gain/Bleach correction data")) {
	    		if(canvas.myMaster != null)
	    			canvas.myMaster.printHeaderData();
	    		}

	    	if(arg.equals("write Gain/Bleach correction data to disk file")) {
	    		if(canvas.myMaster != null)
	    			canvas.myMaster.writeHeaderDataToDiskFile();
	    		}

	    	if(arg.equals("Initiate Auto Bleach Correction; write BioradFileHeader")) {
	    		if(canvas.myMaster != null) {
	    			talk("try to initiate auto bleach correction");
	    			canvas.myMaster.initBleachCorrection();
	    			}
	    		}
	    }

		
private static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}

	
	public void paint(Graphics g) {
		//talk("This should never happen. showLUT.paint() called");
		paint2D((Graphics2D)g);
		}
		
	public void update(Graphics g) {
		//talk("showLUT.update() called. Should never happen");
		paint2D((Graphics2D)g);
		}

	public void update2D(Graphics2D g) {
		paint2D(g);
		}

	public void paint2D(Graphics2D g) {
		width = getSize().width;		
		height=getSize().height;
		g.setColor(new Color(64, 0, 64));
		g.fillRect(0, 0, width, height);
		canvas.paint2D(g, true);
		}
		
   	public void acceptNextLUT(int [] theLUT, String caption, String theTitle) {
		setTitle(theTitle);
 		canvas.acceptNextLUT(theLUT,caption);
   		}
   	   		
	public void sepuku() {
			setVisible(false);
			dispose();
			}
		
// catch WindowListener events  
	public void windowActivated(java.awt.event.WindowEvent event) {
		// talk("window activated");
		}
	public void windowDeactivated(java.awt.event.WindowEvent event) {
		//  talk("window Deactivated");
		}
	public void windowOpened(java.awt.event.WindowEvent event) {}
	public void windowIconified(java.awt.event.WindowEvent event) {}
	public void windowDeiconified(java.awt.event.WindowEvent event) {}
	public void windowClosing(java.awt.event.WindowEvent event) {
	    setVisible(false);
		}
	public void windowClosed(java.awt.event.WindowEvent event) {
		}

}

class LUTCanvas extends Canvas {
	int [] currentLUT, oldLUT1, oldLUT2, oldLUT3;
	ShowLUT parent=null;
	BioradStackConfig myMaster=null;
	boolean justFinished=true;
	int oldParentWidth=0,  oldParentHeight=0;
	Color currentcolor=	Color.white;
	Color priorcolor = Color.cyan;
	Color previouspriorcolor= Color.magenta;
	Color oldestcolor = Color.blue;
	String message=null;

	public void printCurrentLUT() {
		talk("in ---> out (change)");
		for(int i=0;i<256;i++)
			talk(""+i+"  ---> "+currentLUT[i]+"  ("+(currentLUT[i]-i)+")");
		}
	
	public LUTCanvas(ShowLUT parent, int [] theLUT, String stg , Color col1, Color col2, Color col3, Color col4) {
		int i;
		this.parent=parent;
		currentLUT = new int[256];
		for(i=0;i<256;i++) {
			currentLUT[i]=theLUT[i];
			}
		currentcolor = col1;
		priorcolor = col2;
		previouspriorcolor = col3;
		oldestcolor = col4;
		message=stg;
		setBackground(new Color(64, 0, 64));
		setSize(256,256);
		justFinished=true;
   		}
   		
   	public void acceptNextLUT(int [] theLUT,String stg) {
		message=stg;
   		oldLUT3=oldLUT2;
   		oldLUT2=oldLUT1;
   		oldLUT1 = currentLUT;
   		currentLUT = new int[256];
   		for(int i=0;i<256;i++) 
   			currentLUT[i]=theLUT[i];
   		}
   		
	public void paint(Graphics g) {
		//talk("This should never happen. LUTCanvas.paint() called");
		paint2D((Graphics2D)g);
		}
		
	public void update(Graphics g) {
		talk("LUTCanvas.update() called. Should never happen");
		paint2D((Graphics2D)g);
		}

	public void update2D(Graphics2D g) {
		paint2D(g);
		}

	public void paint2D(Graphics2D g,  boolean ifVirgin) {
	    justFinished=ifVirgin;
	    repaint();
	    }
	    
	public void paint2D(Graphics2D g) {
	    if(justFinished || parent.width != oldParentWidth || 
		parent.height != oldParentHeight) {
		int grid;
			oldParentWidth = parent.width; 
			oldParentHeight = parent.height;
			g.setColor(Color.darkGray);
			g.fillRect(0,0,255,255);
			g.setColor(Color.gray);
			g.drawLine(0,255,255,0); // the 45o line
		// draw the grid lines...
		for(grid=255; grid >= 0 ;grid -= 32) { 
		    g.drawLine(grid, 0, grid, 255);
		    g.drawLine(0, grid, 255, grid);
		    }
		if(oldLUT3 != null) { 
			g.setColor(oldestcolor); 
			drawLUT(g,oldLUT3,255);
			g.drawString("3-rd previous LUT",10,75);
			}
		if(oldLUT2 != null) { 
			g.setColor(previouspriorcolor); 
			drawLUT(g,oldLUT2,255); 
			g.drawString("2-nd previous LUT",10,60);
			}
		if(oldLUT1 != null) { 
			g.setColor(priorcolor); 
			drawLUT(g,oldLUT1,255); 
			g.drawString("previous LUT",10,45);
			}
		g.setColor(currentcolor); 	
		drawLUT(g,currentLUT,255);
		g.drawString("current LUT transforms raw data",10,30);
		g.drawString(message,10,15);
		}
	}


	private void drawLUT(Graphics g,int [] LUT,int kk) {
		int i;
		int x,y,xn=0, yn=255-LUT[0];
		for(i=0;i<=kk;i++) {
			x=xn;	y=yn; 
			xn=i;	yn=255-LUT[i];
			g.drawLine(x,y,xn,yn);
			}
		}		

	static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
}  
