package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.io.*;
import java.lang.*;
import java.util.*;
import java.awt.*;  import javax.swing.*;


class TransectLine {	
	int XMIN=0,XMAX=10,YMIN=0,YMAX=10; // the max raster size
			//  the bounds of the current histogram.
	Color mycolor = Color.white;
	ShowPicJPanel picJPanel=null;
	Stats statsJPanel=null;
	HistogramBox parent;
	int draggingVertexNumber = -1;
	int [] polyxcoords={(int)(.2*XMAX), (int)(.8*XMAX)};
	int [] polyycoords={(YMAX+YMIN)/2, (YMAX+YMIN)/2};
	double [] unitVec = {1.,0.};
	double [] crossVec = {0,1.};
	double transectLength = 1.;
	boolean ifMonochrome=false;
	int nverts=2;
	double cenx,ceny;
	int width=30;
	int blur = 4;
	int npixInBlur = blur*width;
	double pixelsPerMicrometer=22.;
	double hwidth=.5*width;
	double hblur= .5*blur;
	final static int MOUSE_MOVES_OUTERBOX_CENTER = 23;
	final static int MOUSE_MOVES_INNERBOX_CENTER = 26;
	final static int MOUSE_CLICKS_DEFINE_POLYGON = 24;	
	final static int MOUSE_DRAGS_POLYGON_VERTEX = 25;	
	final static int MOUSE_CLICKS_DEFINE_TRANSECT = 28;	
	final static int MOUSE_DRAGS_TRANSECT_VERTEX = 29;	
	final static int MOUSE_MOVES_TRANSECT_CENTER = 27;

	public void sepuku() {
		picJPanel=null;
		statsJPanel=null;
		polyxcoords=null;
		polyycoords=null;
		unitVec =null;
		crossVec =null;
		}
	
	public void finalize() {
		talk("                                                                           TransectLine snuffs");
		}
			
	public int readFromDisk(ArchiveMarks am) {
	  	double badval = -7777.777;
	  	int ibool,i;
		i = (int)(am.getNextFileDouble(badval,true)); // saves whole input line
		if(am.mostRecentLineRead.indexOf("= numCoreSects") >0) 
			return (i); // flag the we encountered the first line of a wandering core stack.
		XMIN=i;
		XMAX = (int)(am.getNextFileDouble(badval));
		YMIN = (int)(am.getNextFileDouble(badval));
		YMAX = (int)(am.getNextFileDouble(badval));
		if(XMAX>picJPanel.fullWidth)
			XMAX=picJPanel.fullWidth;
		if(YMAX>picJPanel.fullHeight)
			YMAX=picJPanel.fullHeight;

		Color mycolor = Color.white;

		cenx = (int)(am.getNextFileDouble(badval));
		ceny = (int)(am.getNextFileDouble(badval));
		transectLength = am.getNextFileDouble(badval);
		width = (int)(am.getNextFileDouble(badval));
		blur = (int)(am.getNextFileDouble(badval));
		npixInBlur = blur*width;
		hwidth=.5*width;
		hblur= .5*blur;
		
		pixelsPerMicrometer = am.getNextFileDouble(badval);
		polyxcoords = new int[2];
		polyycoords = new int[2];
		unitVec = new double[2];
		crossVec = new double[2];
 		for(i=0;i<nverts;i++)
   			polyxcoords[i] = (int)(am.getNextFileDouble(badval));
  		for(i=0;i<nverts;i++)
   			polyycoords[i] = (int)(am.getNextFileDouble(badval));
  		for(i=0;i<2;i++)
   			unitVec[i] = (int)(am.getNextFileDouble(badval));
  		for(i=0;i<2;i++)
   			crossVec[i] = (int)(am.getNextFileDouble(badval));
   		return (-1);
     	}
     	
 	public void writeToDisk(ArchiveMarks am, String stg) {
 		int i;
		if(XMAX>picJPanel.fullWidth)
			XMAX=picJPanel.fullWidth;
		if(YMAX>picJPanel.fullHeight)
			YMAX=picJPanel.fullHeight;
   		am.writeNum(XMIN,("    XMIN  <---"+stg));
   		am.writeNum(XMAX,("    XMIN"));
   		am.writeNum(YMIN,("    XMIN"));
   		am.writeNum(YMAX,("    XMIN"));
   		am.writeNum(cenx,"    cenx");
   		am.writeNum(ceny,"    ceny");
   		am.writeNum(transectLength,"    transectLength");
   		am.writeNum(width,"    width");
   		am.writeNum(blur,"    blur");
   		am.writeNum(pixelsPerMicrometer,"    pixelsPerMicrometer");
   		for(i=0;i<nverts;i++)
   			am.writeNum(polyxcoords[i],("   polyxcoords["+i+"]"));
  		for(i=0;i<nverts;i++)
   			am.writeNum(polyycoords[i],("   polyycoords["+i+"]"));
   		for(i=0;i<2;i++)
   			am.writeNum(unitVec[i],("   unitVec["+i+"]  <---"+stg));
   		for(i=0;i<2;i++)
   			am.writeNum(crossVec[i],("   crossVec["+i+"]  <---"+stg));
 		}
	
	public void transferDataTo(TransectLine atran) {
		int j;
		if(XMAX>picJPanel.fullWidth)
			XMAX=picJPanel.fullWidth;
		if(YMAX>picJPanel.fullHeight)
			YMAX=picJPanel.fullHeight;
		atran.XMIN=XMIN; atran.XMAX=XMAX; atran.YMIN=YMIN; atran.YMAX=YMAX;
		atran.mycolor = mycolor;
		atran.picJPanel=picJPanel;
		atran.statsJPanel=statsJPanel;
		atran.polyxcoords = new int[2];
		atran.polyycoords = new int[2];
		atran.nverts=2;
		for(j=0;j<2;j++) {
			atran.polyxcoords[j] = polyxcoords[j];
			atran.polyycoords[j] = polyycoords[j];
			}
		atran.unitVec[0] = unitVec[0];
		atran.unitVec[1] = unitVec[1];
		atran.crossVec[0] = crossVec[0];
		atran.crossVec[1] = crossVec[1];
		
		atran.transectLength = transectLength;
		atran.nverts=2;
		atran.cenx=cenx;	atran.ceny=ceny;
		atran.width=width;
		atran.blur = blur;
		atran.npixInBlur = npixInBlur;
		atran.pixelsPerMicrometer=pixelsPerMicrometer;
		atran.hwidth=hwidth;
		atran.hblur= hblur;
		}

	
	public TransectLine(ShowPicJPanel picJPanel, Stats statsJPanel, HistogramBox parent) {
		this.parent = parent;
		this.picJPanel=picJPanel;
		this.statsJPanel=statsJPanel;
		XMAX=picJPanel.fullWidth;
		YMAX=picJPanel.fullHeight;
		getUnitVec();
		}
	
	public void setParams(int width, int blur, double pixelsPerMicrometer) {
		this.width=width;
		this.blur=blur;
		this.pixelsPerMicrometer = pixelsPerMicrometer;
		if(pixelsPerMicrometer<4) {
			talk("*********TransectLine.setParams  warning!!! pixelsPerMicrometer= "+pixelsPerMicrometer);
			pixelsPerMicrometer=22.5;
			}
		npixInBlur = blur*width;
		hwidth=.5*width;
		hblur= .5*blur;
		}
	
	public void resetProfile() {
		}
		
	public double [] getUnitVec() {
		while(polyxcoords[0]==polyxcoords[1] && polyycoords[0]==polyycoords[1])
			++polyxcoords[0];
		unitVec[0] = polyxcoords[1]-polyxcoords[0];	
		unitVec[1] = polyycoords[1]-polyycoords[0];
		transectLength = unitVec[0]*unitVec[0] + unitVec[1]*unitVec[1];
		transectLength = Math.sqrt(transectLength);
		unitVec[0]= unitVec[0]/transectLength;
		unitVec[1] = unitVec[1]/transectLength;
		crossVec[0] = -unitVec[1];
		crossVec[1] =  unitVec[0];
		return unitVec;
		}
		
	public void rescale(double enlargeRatio) { // boost or shrink size of the box...
		int nx0,nx1,ny0,ny1;
		getUnitVec();
		cenx = polyxcoords[0] + .5*transectLength*unitVec[0];
		ceny = polyycoords[0] + .5*transectLength*unitVec[1];
		transectLength *= enlargeRatio;
		// don't let transect line go outside the big box
		while(outOfXBoundsQ(nx0=(int)(cenx-.5*transectLength*unitVec[0])) ||
				outOfXBoundsQ(nx1=(int)(cenx+.5*transectLength*unitVec[0])) ||
				outOfYBoundsQ(ny0=(int)(ceny-.5*transectLength*unitVec[1])) ||
				outOfYBoundsQ(ny1=(int)(ceny+.5*transectLength*unitVec[1])))
							transectLength -= .5;
		// make sure first endpoint is to the left...
		if(unitVec[0]<0) { // flip the line orientation
			polyxcoords[1]=nx0;
			polyxcoords[0]=nx1;
			polyycoords[1]=ny0;
			polyycoords[0]=ny1;
			getUnitVec();
			if(draggingVertexNumber>=0) 
				draggingVertexNumber = 1-draggingVertexNumber;
			}
		else {
			polyxcoords[0]=nx0;
			polyxcoords[1]=nx1;
			polyycoords[0]=ny0;
			polyycoords[1]=ny1;
			}
		statsJPanel.transectPixels.setText(""+transectLength);
		}
		
	private boolean outOfXBoundsQ(int x) {
		if(x<XMIN) return true;
		if(x>=XMAX) return true;
		else return false;
		}		
	private boolean outOfYBoundsQ(int y) {
		if(y<YMIN) return true;
		if(y>=YMAX) return true;
		else return false;
		}		
			
	public void moveCenterTo(int x, int y) {
		cenx=x;	ceny=y;
		polyxcoords[0]=(int)(cenx-.5*transectLength*unitVec[0]);
		polyxcoords[1]=(int)(cenx+.5*transectLength*unitVec[0]);
		polyycoords[0]=(int)(ceny-.5*transectLength*unitVec[1]);
		polyycoords[1]=(int)(ceny+.5*transectLength*unitVec[1]);
		rescale(1.0);
		}
	
	public void draw(Graphics g, int xorg, int yorg) { // draw the x-y outline of the boundary
		double xfrom,xto,yfrom,yto, xc, yc; 
		double hwidth = 0.5*width;
		double hblur = 0.5*blur;
		double tic;
		int i;
		if(ifMonochrome) g.setColor(Color.black);
		else g.setColor(mycolor);
		g.drawLine(xorg+polyxcoords[0],yorg+polyycoords[0],
				xorg+polyxcoords[1],yorg+polyycoords[1]);
		getUnitVec();
		tic = 0;
		for(i=0;i<=transectLength;i++) 
		  if(i >= tic) {
		  	tic += 5*pixelsPerMicrometer;
			xc = xorg+polyxcoords[0] + i*unitVec[0];
			yc = yorg+polyycoords[0] + i*unitVec[1];
			xfrom = xc + hwidth*crossVec[0] - hblur*unitVec[0];
			yfrom = yc + hwidth*crossVec[1] - hblur*unitVec[1];
		
			xto = xfrom + 2.*hblur*unitVec[0];
			yto = yfrom + 2.*hblur*unitVec[1];
			g.drawLine((int)(xfrom),(int)(yfrom),(int)(xto),(int)(yto));
		
			xfrom = xto - 2.*hwidth*crossVec[0];
			yfrom = yto - 2.*hwidth*crossVec[1];
			g.drawLine((int)(xfrom),(int)(yfrom),(int)(xto),(int)(yto));

			xto = xfrom -  2.*hblur*unitVec[0];
			yto = yfrom -  2.*hblur*unitVec[1];
			g.drawLine((int)(xfrom),(int)(yfrom),(int)(xto),(int)(yto));
		
			xfrom = xto + 2.*hwidth*crossVec[0];
			yfrom = yto + 2.*hwidth*crossVec[1];
			g.drawLine((int)(xfrom),(int)(yfrom),(int)(xto),(int)(yto));
			}
		}
		
	public void setupToDefineNewTransect() {
		nverts=0;
		}
		
	public boolean nextVertexOfTransect(int x,int y) {
			polyxcoords[nverts]=x;
			polyycoords[nverts]=y;
			++nverts;
			if(nverts==1) { // it's the first vertex...
	    		talk("first transect endpoint defined");
				return true;
				}
			if(nverts>1) { // end of the line
				rescale(1.0);
				}
			return true;
	    }
			
	static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
		
	
	public void moveVertex(int vertnumber, int x, int y) {
			polyxcoords[vertnumber]=x;
			polyycoords[vertnumber]=y;
			rescale(1.0);
			}
		
	public void primeToDragAVertex() {
		draggingVertexNumber=-1;
		}	
			
	public void moveVertex(int x, int y) { 
			if(draggingVertexNumber<0) 
				draggingVertexNumber=closestVertex(x,y);
			moveVertex(draggingVertexNumber,x,y);
			}
		
	private int distSq(int x,int y,int xx, int yy) {
		int dx=x-xx;
		int dy=y-yy;
		return(dx*dx + dy*dy);
		}
	
	public boolean blurCoords(int i,int b, int w, int [] xy) {
		/* i counts along the transect. 
		b counts blur position (parallel to transect
		w counts position off transect perpendicular to it
		xy array returns the x,y coords of the pixel */
		double along = i - hblur + b;
		double off = w -hwidth;
 		xy[0] = (int)(.499 + polyxcoords[0] + along*unitVec[0] +off*crossVec[0]);
		xy[1] = (int)(.499 + polyycoords[0] + along*unitVec[1] +off*crossVec[1]);
		/* if(i==10 || i==20 ||i==30) 
			talk("i="+i+",b="+b+", w="+w+"x="+xy[0]+", y="+xy[1]);  */
		return ( !outOfXBoundsQ(xy[0]) && !outOfYBoundsQ(xy[1]));
		}
	
	private int closestVertex(int x, int y) {
		int dsq,ibest=0,i;
		int distsq=distSq(polyxcoords[ibest],polyycoords[ibest],x,y);
		for(i=1;i<nverts;i++) {
			dsq=distSq(polyxcoords[i],polyycoords[i],x,y);
			if(dsq<distsq) {
				ibest=i;
				distsq=dsq;	
				}
			}
		return ibest;
		}
}