package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.lang.*;
import java.util.*;
import java.awt.*;  import javax.swing.*;

class UndustFltr {
	SegmenterNode nd=null;

	public UndustFltr(SegmenterNode nd) {
		this.nd=nd;
		} 
		
	public boolean eraseSingleDots(short [] inras,int NCOLS,int NROWS, short[] outras) {
    	int i,j,loc;
        for(i=0;i<NROWS;i++)
    	  for(j=0;j<NCOLS-1;j++) {
    		loc=i*(NCOLS)+j;
    		outras[loc]=inras[loc]; // copy each pixel
    		if(inras[loc]>126 && i>0 && j>0 && i<NROWS-1 && j<NCOLS-1 &&
    			inras[loc-1]<127 && inras[loc+1]<127 && inras[loc-NCOLS]<127 &&
    			inras[loc+NCOLS]<127 && inras[loc-NCOLS-1]<127 && inras[loc-NCOLS+1]<127 &&
    			inras[loc+NCOLS-1]<127 && inras[loc+NCOLS+1]<127) 
    				outras[loc]=0; //erase this on pixel because it has no neighbors.
    		}
    return true;
	}

	public void tellWhatYouDo() {
		nd.talk("");
		nd.talk("The undust filter removes isolated small bits. It has one parameter,");
		nd.talk("grain.  An isolated patch of pixels which are on and which fits inside a");
		nd.talk("circle whose diameter = grain is erased (set to zero values).");
		}

public boolean doit(short [] redin,int NCOLS,int NROWS, int dustdia, short[] trueoutras) {
	nd.talk("UndustFltr with dustdia="+dustdia);
	if(dustdia<=1) return eraseSingleDots(redin,NCOLS,NROWS,trueoutras);
	int atx=0,aty=0,i;
	SwarmOfPoints firstSwarm=null,lastSwarm=null,aswarm=null,bswarm=null;
	int npix = NCOLS*NROWS;
	short [] red = new short[npix];
	short [] outras = new short[npix];
	int numSinglets=0;
	for(i=0;i<npix;i++) 
		red[i] = redin[i];
		
	while(aty < NROWS) {
		while(atx <NCOLS) {
			if(israspixelon(aty, atx, red, NCOLS, NROWS)) {
				aswarm = new SwarmOfPoints(red,NCOLS,NROWS,atx,aty,outras);
				aswarm.collectConnectedSwarm();
				if(firstSwarm==null) { // this is the first swarm found
					firstSwarm=aswarm;
					lastSwarm=aswarm;
					}
				else {
					lastSwarm.nextSwarm=aswarm;
					lastSwarm=aswarm;
					}
				}
			++atx;
			}
		atx=0;
		++aty;
		} // when fall through, we have found all connected pieces.
	// outras has all points in it, red[] has none. Restore red[] and 
	// Remove the small bits from outras
	for(i=0;i<npix;i++)
		red[i]=outras[i];
	aswarm = firstSwarm;
	while(aswarm != null) {
		if(aswarm.getLocations() <= dustdia) {
			if(aswarm.numPoints==1) --numSinglets;
			aswarm = aswarm.vacateAndDissolve(outras);
			}
		else aswarm = aswarm.dissolve(); 
		}
	for(i=0;i<npix;i++)
		trueoutras[i]=outras[i];
	return true;
	}
	
private boolean israspixelon(int y, int x, short [] ras, int NCOLS, int NROWS) {
    if(x<0 || x>=NCOLS) return false;
    if(y<0 || y>=NROWS) return false;
    if(ras[y*NCOLS+x]>127) return true;
    else return false;    
	}

}

