package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.awt.*;  
import javax.swing.*;
import java.awt.image.*;
import java.io.*;
import java.net.*;
import java.lang.*;
import java.util.*;
import infoCCD.*;

class ZseriesMainWindow extends JFrame  implements Runnable, java.awt.event.MouseListener, java.awt.event.ActionListener,javax.swing.event.ChangeListener, java.awt.event.WindowListener{
// the run() thread is responsible for loading new triads of files when appropriate...

	static int fileGetXLoc=0,fileGetYLoc=0;
	ShowPicJPanel	itsJPanel;
	String aFileNameThatWorked=null;
	int sectionType = 0;
	boolean ifWebStacks=false;
	JMenu fileJMenu, showJMenu, viewJMenu;
	String red_filename = "";
	String green_filename = "";
	String blue_filename = "";
	int buttonJPanelHeight=20;
	int edgeLoss=1;
	JSlider stackSlider,xscrollSlider,yscrollSlider;
	int stackSliderMax = 10;
	int stackSliderValue = -1;
	int stackSliderSlidTo=-1;	
	boolean setSliderTrigger=false, setScrollTrigger=false;

	PickColors getArtColors;
	ColorSeparator separator;
	GoToMarks goJPanel;
	AlignMarks alignJPanel;
	SelectMarks selectJPanel;
	DockMarks dockJPanel;
	Stereo3DReconstruction reconstruction3DPanel;
	SlideWholeStack slideJPanel;
	GoToHotSpots hotJumpJPanel;
	Stats histogramJPanel;
	Centrosomes centsJPanel;
	BlackHoles holesJPanel;
	MarkerShape shapeJPanel;
	Segmenter segmentationJPanel;
	SectionNavigator navigatorJPanel;
	MoveMarksAndSections moveJPanel;
	ShowLUT exhibitRedChanLUT,exhibitGreenChanLUT,exhibitBlueChanLUT;
	writePointSpreadFile writePSF;
	DeconvolverPSF DPSF;
	Deconvolver decon;
	IfUseSqrtTransform iusqt;
	
	RandomAccessFile mouseCllickIntensityStats=null;
	RandomAccessFile triadNames;
	String triadRedName,triadGreenName,triadBlueName;
	Thread TriadNamesThread=null;
	boolean UseTriadFileNames = true; // set false to ignore the file called staxTriadNames in /vefFire0/

	int xDrawOrg=0, yDrawOrg=0, xscrolled=50, yscrolled=50,xsc=50,ysc=50,winwidth=512,winheight=512;	
		 	// (xDrawOrg,yDrawOrg) tells what location of the drawn image goes at the upper left corner
		 	// it's controlled by sliders whose current positions are at (xscrolled,yscrolled).
	double windowScaleFac=1.0;
	StereoPairFrame [] stereoHotPics = new StereoPairFrame [32];

	
	public void setHotPic(StereoPairFrame apic, boolean ifAuto) {
		int i, firstHole=-1, nhots = stereoHotPics.length;
		if(ifAuto) { // add to the list
			for(i=0;i<nhots;i++)
				if(stereoHotPics[i] == null) {
					if(firstHole<0)
						firstHole=i;
					}
				else if(stereoHotPics[i].equals(apic)) 
					return; // it's already there
			// on fall through, apic is not in the list.
			if(firstHole>=0)  // can add it
				stereoHotPics[firstHole] = apic;
			return;
			}
		else { // remove from the list 
			for(i=0;i<nhots;i++)
				if(stereoHotPics[i] != null && stereoHotPics[i].equals(apic)) 
					stereoHotPics[i] = null; // remove it
			}
		}


public void setDrawOrg(int curWidth, int curHeight) {
	/* When xscrolled==100, means picture slides all the way to the right which means xDrawOrg=0.
	   When xscrolled==0, means picture slides all the way to the left which means xDrawOrg=curWidth-winwidth.
	   So, */
	//winwidth=this.size().width;
	//winheight=this.size().height;
	double oldScaleFac;
	if(!itsJPanel.moviePlaying) {
		oldScaleFac = windowScaleFac;
		windowScaleFac = navigatorJPanel.getScaleFac();
		winwidth = (int)(windowScaleFac*itsJPanel.fullWidth);
		winheight = (int)(windowScaleFac*itsJPanel.fullHeight);
		if(oldScaleFac != windowScaleFac) 
    		setSize(winwidth+edgeLoss, winheight+2*buttonJPanelHeight);    	
		
		recruitRemora();
		}	

	if(curWidth <= winwidth) 
		xDrawOrg=0;		
	else 
		xDrawOrg=-(int )(.499 + .01*(xscrolled)*(curWidth-winwidth));
	
	if(curHeight <= winheight) 
		yDrawOrg=0;
	else  
		yDrawOrg=-(int )(.499 + .01*(100-yscrolled)*(curHeight-winheight));			
	}

    public void recruitRemora() {
		navigatorJPanel.remora(getLocation().x + this.size().width, getLocation().y);
		}
		
    public void exhibitFileName(String stg) {
		if(stg==null)
		stg = "null string";
		int i=stg.indexOf("vefFire");
		if(i<0) i=0;
    	navigatorJPanel.setTitle(stg.substring(i));
    	}	
    	
	public void run() {
		int i;
		int gotThreeGoodNames=0;
		boolean namesStillMatch=false;
		while(UseTriadFileNames) {
			try{Thread.sleep(2000);}
			catch(InterruptedException e) { talk("oops");}							
			try{triadNames = new RandomAccessFile((itsJPanel.whereIsVefFire0+"staxTriadNames"),"r");}
			catch(IOException e) {talk("cannot open "+itsJPanel.whereIsVefFire0+"staxTriadNames. Got exception="+e);}
			if(triadNames != null) {;
				gotThreeGoodNames=0;
				namesStillMatch=true;
				triadRedName=readOneLine();
				if(triadRedName != null) {
					++gotThreeGoodNames;
					if(!triadRedName.equals(itsJPanel.lastOpenedRedName) ) {
						namesStillMatch=false;
						talk("Open new red channel file called <"+triadRedName+">");
						}
					}
						
				triadGreenName=readOneLine();
				if(triadGreenName != null) {
					++gotThreeGoodNames;
					if(!triadGreenName.equals(itsJPanel.lastOpenedGreenName) ) {
						talk("Open new green channel file called <"+triadGreenName+">");
						namesStillMatch=false;
						}
					}
						
				triadBlueName=readOneLine();
				if(triadBlueName != null) {
					++gotThreeGoodNames;
					if(!triadBlueName.equals(itsJPanel.lastOpenedBlueName) ) {
						namesStillMatch=false;
						talk("Open new blue channel file called <"+triadBlueName+">");
						}
					}						
				try{triadNames.close();}
				catch(IOException e) {talk("double oops");}
				
				if(gotThreeGoodNames==3 && !namesStillMatch) {
					talk("Found new file names in /tmp/staxTriadNames and will now open them");
					itsJPanel.getNewStacks(triadRedName,triadGreenName,triadBlueName,null,0,
							aFileNameThatWorked);
					// sleep while the files get opened...
					try{Thread.sleep(2000);}
					catch(InterruptedException e) { talk("oops");}	
					updateStackSliderPosition();						
					}
					
				}
			}
			
		}
		

	private String readOneLine() {
		byte [] buf = new byte[129];
		byte [] inbyte = {0};
		int i=0,igot=0;
		while(i<128) {
			igot=0;
			if(triadNames != null) try {igot=triadNames.read(inbyte);}
			  catch(IOException iex) {
				talk("IOException was: "+iex);
				return null;
				}
			if(igot != 1) return null;
			if(inbyte[0] == 13 || inbyte[0] == 10 || inbyte[0] == (byte)'\n') {
				if(i>0) {
					return (new String(buf,(byte)0,0,i));
					}
				else  return "";
				}
			else
				buf[i++]=inbyte[0];
			}
		return ("overflow line");
		}
		
	public void blotOutHoles(BioradStackConfig BSC, byte [] ras, int ncols, int nrows, int nsect ) {
		if(holesJPanel != null) {
			holesJPanel.blotOutHoles(BSC, ras, ncols, nrows, nsect);
			}
		}
	   

	public void updateStackSliderPosition() {
		if(itsJPanel.moviePlaying) return; // don't slow down the movie
		stackSliderValue = itsJPanel.curStack.curJFrame;
		// talk("updateStackSliderPosition(): z= "+stackSliderValue);
		navigatorJPanel.zSubjunctive(stackSliderValue);
		if(stackSlider.getMaximum() != itsJPanel.curStack.howManySections()) 
			resetStackSlider();			
		stackSlider.setValue(stackSliderValue);
		}
		
	public void resetStackSlider() {
		stackSliderMax = itsJPanel.curStack.howManySections();
		stackSlider.setMaximum(stackSliderMax);
		stackSliderValue = stackSliderMax/2;
		}

	public void stateChanged(javax.swing.event.ChangeEvent e) {
		stackSliderSlidTo = stackSlider.getValue();
		xsc = xscrollSlider.getValue();
		ysc = yscrollSlider.getValue();
		// talk("xsc="+xsc+"("+xscrolled+") , ysc= "+ysc+"("+yscrolled+"), stackSliderSlidTo = "+stackSliderSlidTo);
		if(stackSliderSlidTo<0 | 
		        stackSlider.getMaximum() != itsJPanel.curStack.howManySections()) {
			resetStackSlider();
			talk("Reset the stackSlider to new Maximum ="+itsJPanel.curStack.howManySections());
			}
		else {
			setSliderTrigger=true;
			navigatorJPanel.zSubjunctive(stackSliderSlidTo);
			}
		if(xsc != xscrolled || ysc != yscrolled) 
			setScrollTrigger=true;			
		} 
		
	public void mouseReleased(java.awt.event.MouseEvent event) {
		if(stackSliderSlidTo<0) {
			talk("mouse released and must reseet stackSlider");
			resetStackSlider();
			stackSliderSlidTo = stackSliderValue;
			}
		// only when the mouse is released do we move to the section the slider locates....
		if(setSliderTrigger && stackSliderSlidTo != itsJPanel.curStack.curJFrame) {
			itsJPanel.curStack.showAJFrame(stackSliderSlidTo);
			}
		setSliderTrigger=false;
		if(setScrollTrigger) {// scrolled position changed, force redraw
			// talk("setScrollTrigger==true, redraw scrolled pic");
			xscrolled = xsc;
			yscrolled = ysc;
			itsJPanel.reallyReDraw();
			}
		setScrollTrigger=false;
		}			
			
	public void mouseClicked(java.awt.event.MouseEvent event) {
		//talk("Mouse Clicked at ("+event.getX()+", "+event.getY()+")");
		}
	public void mouseEntered(java.awt.event.MouseEvent event) {
		//talk("Mouse entered at ("+event.getX()+", "+event.getY()+")");
		}
	public void mouseExited(java.awt.event.MouseEvent event) {
		//talk("Mouse exited at ("+event.getX()+", "+event.getY()+")");
		}
	public void mousePressed(java.awt.event.MouseEvent event) {
		//talk("Mouse pressed at ("+event.getX()+", "+event.getY()+")");
		}
		
    public ZseriesMainWindow()  {   
        super("Sections from biorad Z-series");
        getContentPane().setLayout(new BorderLayout());    // NEED to specify a layout manager!!
	Color myBlue = 	new Color(100,100,255);
	
	setBackground(new Color(210,220,240));
	WidgetMaker wmk = new WidgetMaker(this);
	
	int [] temp = new int[256];
	for(int i=0;i<256;i++) temp[i]=i;

	// make and show menu ------------------
	JMenuBar mb = new JMenuBar();
	fileJMenu = new JMenu("File");
	fileJMenu.add(wmk.makeAJMenuItem("Interlace"));
	fileJMenu.add(wmk.makeAJMenuItem("load new triad of files..."));
	fileJMenu.add(wmk.makeAJMenuItem("load new merged single-section file"));
	fileJMenu.add(wmk.makeAJMenuItem("save color tiff image in a file..."));
	fileJMenu.add(wmk.makeAJMenuItem("save image in a Jpeg file..."));	
	fileJMenu.add(wmk.makeAJMenuItem("write images in stack as jpg movie files to local disk"));
	fileJMenu.add(wmk.makeAJMenuItem("write images in stack as movie file to web server"));			
	fileJMenu.add(wmk.makeAJMenuItem("Write marks/alignment info to disk"));
	fileJMenu.add(wmk.makeAJMenuItem("Recover marks/alignment info from a disk file"));
	fileJMenu.add(wmk.makeAJMenuItem("Write e3-format shift/centrosome list to a disk file"));
	fileJMenu.add(wmk.makeAJMenuItem("write, from a stack, a single section biorad.pic to disk"));	
	fileJMenu.add(wmk.makeAJMenuItem("start to accumulate mouseClickIntensity vs sect# into a file"));	
	fileJMenu.add(wmk.makeAJMenuItem("stop accumulating mouseClickIntensity vs sect# into file"));	
	fileJMenu.add(wmk.makeAJMenuItem("load XY sections from triad of net servers"));
	fileJMenu.add(wmk.makeAJMenuItem("load XZ sections from triad of net servers"));
	fileJMenu.add(wmk.makeAJMenuItem("load ZY sections from triad of net servers"));
	fileJMenu.add(wmk.makeAJMenuItem("Stop everything"));
	fileJMenu.add(wmk.makeAJMenuItem("Quit"));
	mb.add(fileJMenu);
	showJMenu = new JMenu("Show");
	showJMenu.add(wmk.makeAJMenuItem("Section navigator Panel"));
	showJMenu.add(wmk.makeAJMenuItem("Pick Colors Panel"));
	showJMenu.add(wmk.makeAJMenuItem("stereo 3D reconstruction panel"));
	showJMenu.add(wmk.makeAJMenuItem("Color Separator Panel"));
	showJMenu.add(wmk.makeAJMenuItem("histogram panel"));	
	showJMenu.add(wmk.makeAJMenuItem("Panel for moving markers and/or sections"));
	showJMenu.add(wmk.makeAJMenuItem("Select marker Panel"));
	showJMenu.add(wmk.makeAJMenuItem("GoTo marks Panel"));
	showJMenu.add(wmk.makeAJMenuItem("Align marks Panel"));
	showJMenu.add(wmk.makeAJMenuItem("Dock marker Panel"));
	showJMenu.add(wmk.makeAJMenuItem("global slider Panel"));
	showJMenu.add(wmk.makeAJMenuItem("hot spots Panel"));
	showJMenu.add(wmk.makeAJMenuItem("Centrosomes Panel"));
	showJMenu.add(wmk.makeAJMenuItem("Black Holes Panel"));
	showJMenu.add(wmk.makeAJMenuItem("Marker Shape Panel"));
	showJMenu.add(wmk.makeAJMenuItem("R/G/B LUT's transforming raw input data"));
	showJMenu.add(wmk.makeAJMenuItem("data segmentation panel"));
	showJMenu.add(wmk.makeAJMenuItem("measure/retrieve point spread functions"));
	showJMenu.add(wmk.makeAJMenuItem("deconvolver"));
	showJMenu.addActionListener(this);
	mb.add(showJMenu);

	viewJMenu = new JMenu("view");
	viewJMenu.add(wmk.makeAJMenuItem("show all auxiliary Panels"));
	viewJMenu.add(wmk.makeAJMenuItem("enlarge all auxiliary Panels"));
	viewJMenu.add(wmk.makeAJMenuItem("shrink all auxiliary Panels"));
	viewJMenu.add(wmk.makeAJMenuItem("Hide all auxiliary Panels"));
	viewJMenu.addActionListener(this);
	mb.add(viewJMenu);
	
	new Info(true,true,false);
	mb.add(Info.infoMenu);
	
	setJMenuBar(mb);
	// make and show menu ------------------

 	red_filename = getExistingFileName(this,"Pick the red file (Instructions panel explains)");
	green_filename = getExistingFileName(this,"Pick the green file (Instructions panel explains)");
	blue_filename = getExistingFileName(this,"Pick the blue file (Instructions panel explains)");

	exhibitBlueChanLUT = new ShowLUT(temp,"Blue channel's LUT",2,new Color(0,100,255),
							new Color(0,(5*100)/6,(5*255)/6), new Color(0,(4*100)/6,(4*255)/6), 
									new Color(0,100/2,255/2));
	exhibitGreenChanLUT = new ShowLUT(temp,"Green channel's LUT",1,new Color(0,255,0),
							new Color(0,(5*255)/6,0), new Color(0,(4*255)/6,0), 
									new Color(0,255/2,0));
	exhibitRedChanLUT = new ShowLUT(temp,"Red channel's LUT",0,new Color(255,0,0),
							new Color((5*255)/6,0,0), new Color((4*255)/6,0,0), 
									new Color(255/2,0,0));

	navigatorJPanel = new SectionNavigator(); 
	itsJPanel = new ShowPicJPanel(red_filename,green_filename,blue_filename, this);
	navigatorJPanel.picJPanel = itsJPanel;
	
	getContentPane().add("Center", itsJPanel);
	addWindowListener(this);

	reconstruction3DPanel = new Stereo3DReconstruction(itsJPanel);
	getArtColors = new PickColors(itsJPanel);
	goJPanel =      new GoToMarks(itsJPanel);
	alignJPanel =   new AlignMarks(itsJPanel);
	selectJPanel =  new SelectMarks(itsJPanel);
	dockJPanel =    new DockMarks(itsJPanel);
	slideJPanel =   new SlideWholeStack(itsJPanel);
	hotJumpJPanel = new GoToHotSpots(itsJPanel,itsJPanel.curStack.NJFrames);
	histogramJPanel = new Stats(itsJPanel);
	centsJPanel = new Centrosomes(itsJPanel);
	holesJPanel = new BlackHoles(itsJPanel);
	shapeJPanel = new MarkerShape(itsJPanel);
	segmentationJPanel = new Segmenter(itsJPanel);
	moveJPanel = new MoveMarksAndSections(itsJPanel);
	separator = new ColorSeparator(itsJPanel,getArtColors,itsJPanel.curStack);

	itsJPanel.setImageSize(1); // full size images
	itsJPanel.setVisible(true);
	
	itsJPanel.selectMarker(1);
	// show the essential panels...
	exhibitRedChanLUT.makeGainCurveCorrectionLUT();
	exhibitGreenChanLUT.makeGainCurveCorrectionLUT();
	exhibitBlueChanLUT.makeGainCurveCorrectionLUT();
	int ytopleft=10;
	getArtColors.setVisible(true);
	getArtColors.setLocation(10,ytopleft);
	ytopleft += (getArtColors.nomheight+15);
	
	reconstruction3DPanel.setVisible(true);
	reconstruction3DPanel.setLocation(10,ytopleft);
	ytopleft += (reconstruction3DPanel.nomheight+15);
	
	histogramJPanel.setVisible(true);
	histogramJPanel.setLocation(10,ytopleft);
	ytopleft += (reconstruction3DPanel.nomheight+5);
	
	navigatorJPanel.setVisible(true);
	stackSlider = new JSlider(JSlider.HORIZONTAL,0,stackSliderMax,
			0); // this is the slider that picks new sections.
	stackSlider.addChangeListener(this);
	stackSlider.setMajorTickSpacing(10);
	stackSlider.setPaintTicks(true);
	stackSlider.addMouseListener(this);			

	xscrollSlider = new JSlider(JSlider.HORIZONTAL,0,100,50);
	yscrollSlider = new JSlider(JSlider.VERTICAL,0,100,50);
	xscrollSlider.addChangeListener(this);
	yscrollSlider.addChangeListener(this);
	xscrollSlider.addMouseListener(this);
	yscrollSlider.addMouseListener(this);
	getContentPane().add("North",stackSlider);
	getContentPane().add("South",xscrollSlider);
	getContentPane().add("East",yscrollSlider);
   	resize();
	setVisible(true);
	setLocation(histogramJPanel.nomwidth+30,40);
	recruitRemora();	
	itsJPanel.holesJPanel=holesJPanel;
	itsJPanel.centsJPanel = centsJPanel;
	itsJPanel.statsJPanel = histogramJPanel;
	TriadNamesThread = new Thread(this);
	TriadNamesThread.start();
	updateStackSliderPosition();						
    }
    
    public void resize () {
    	if(itsJPanel != null)
    		setSize(itsJPanel.fullWidth+edgeLoss, itsJPanel.fullHeight+2*buttonJPanelHeight);
    	}
    	
	
	public boolean assertRawDataTransformationMode(int mode) {
	boolean ifAtLeastOneChanCanDoIt=false;
	if(exhibitRedChanLUT.assertRawDataTransformationMode(mode) == mode)
		ifAtLeastOneChanCanDoIt=true;
	if(exhibitGreenChanLUT.assertRawDataTransformationMode(mode) == mode)
		ifAtLeastOneChanCanDoIt=true;
	if(exhibitBlueChanLUT.assertRawDataTransformationMode(mode) == mode)
		ifAtLeastOneChanCanDoIt=true;
	return ifAtLeastOneChanCanDoIt;
	}
	
	public void repaintExhibitRGBChanLUTS() {
		exhibitRedChanLUT.repaint();
		exhibitGreenChanLUT.repaint();
		exhibitBlueChanLUT.repaint();
		}
		
	public void setNewGainCurveFit(int ifRGB, double concentrationMappedTo255,
				double [] gainCurveFitCoeffs, boolean ifPurgeCache) {
		switch (ifRGB) {
			case 0:
				exhibitRedChanLUT.setNewGainCurveFit(concentrationMappedTo255,gainCurveFitCoeffs,
					ifPurgeCache);
				break;
			case 1:
				exhibitGreenChanLUT.setNewGainCurveFit(concentrationMappedTo255,gainCurveFitCoeffs,
					ifPurgeCache);
				break;
			case 2:
				exhibitBlueChanLUT.setNewGainCurveFit(concentrationMappedTo255,gainCurveFitCoeffs,
					ifPurgeCache);
				break;
			}
		}
		
      	
	public void reDraw() {
    	itsJPanel.reDraw();
    	}

// catch WindowListener events  
	public void windowActivated(java.awt.event.WindowEvent event) {}
	public void windowDeactivated(java.awt.event.WindowEvent event) {}
	public void windowOpened(java.awt.event.WindowEvent event) {}
	public void windowIconified(java.awt.event.WindowEvent event) {}
	public void windowDeiconified(java.awt.event.WindowEvent event) {}
	public void windowClosing(java.awt.event.WindowEvent event) {
        talk("Exit via 'Quit' option in 'file' menu");
        setVisible(true);
		}
		
	public void windowClosed(java.awt.event.WindowEvent event) {
        talk("Exit via 'Quit' option in 'file' menu");
        setVisible(true);
		}
	
	public double [] getBoosts(int sectionNum) {
		if(getArtColors != null)
			return (getArtColors.getBoosts(sectionNum));
		else {
			double [] boosts = new double[3];
			boosts[0]=1.0; boosts[1]=1.0; boosts[2]=1.0;
			return boosts; // this flags that there is no boosting.
			}
		}
		
		 
    public int [] getRedsColors() {
		if(getArtColors != null)
	    	return (getArtColors.getRedsColors());
		else {
	    	int [] truered = {255, 0, 0};
	    	return truered;
	   		}
		}
		
    public int [] getRedsCustLUT() {
	if(getArtColors != null)
	    return (getArtColors.getRedLUT());
	else {
	    return ramp();
	    }
	}

    public int [] getGreensColors() {
		if(getArtColors != null)
	    	return (getArtColors.getGreensColors());
		else {
	    	int [] truegreen = {0, 255, 0};
	    	return truegreen;
	   		}
		}
		
    public int [] getGreensCustLUT() {
	if(getArtColors != null)
	    return (getArtColors.getGreenLUT());
	else {
	    return ramp();
	    }
	}
	
    public int [] getBluesColors() {
		if(getArtColors != null)
	    	return (getArtColors.getBluesColors());
		else {
	    	int [] trueblue = {0, 0, 255};
	    	return trueblue;
	   		}
		}
	
    public int [] getBluesCustLUT() {
	if(getArtColors != null)
	    return (getArtColors.getBlueLUT());
	else {
	    return ramp();
	    }
	}

    private int [] ramp() {
	    int [] ddd = new int[256];
	    for(int i=0;i<256;i++) ddd[i]=i;
	    return ddd;
	    }


// catch button and menu events...
	public void actionPerformed(java.awt.event.ActionEvent event) {
		String arg = event.getActionCommand();
	    if(arg.equals("Stop everything")) {
			itsJPanel.stopEverything();
	    	}
		
	    if(arg.equals("load XY sections from triad of net servers")) {
			itsJPanel.getNewStacks(navigatorJPanel,0,aFileNameThatWorked);
			updateStackSliderPosition();						
	    	}
	    	
	    if(arg.equals("load XZ sections from triad of net servers")) {
			itsJPanel.getNewStacks(navigatorJPanel,1,aFileNameThatWorked);
			updateStackSliderPosition();						
	    	}
	    	
	    if(arg.equals("load ZY sections from triad of net servers")) {
			itsJPanel.getNewStacks(navigatorJPanel,2,aFileNameThatWorked);
			updateStackSliderPosition();						
	    	}
	    	
	    if(arg.equals("load new merged single-section file")) {
			red_filename = getExistingFileName(this,"Pick the merged file");
			green_filename = new String(red_filename);
			blue_filename =  new String(red_filename);
			itsJPanel.getNewStacks(red_filename,green_filename,blue_filename,null,0,
					aFileNameThatWorked);
			updateStackSliderPosition();						
			itsJPanel.MergedPicShifts();						
	    	}
	    	
	    if(arg.equals("load new triad of files...")) {
			red_filename = getExistingFileName(this,"Pick the red file");
			green_filename = getExistingFileName(this,"Pick the green file");
			blue_filename = getExistingFileName(this,"Pick the blue file");
			itsJPanel.getNewStacks(red_filename,green_filename,blue_filename,null,0,
					aFileNameThatWorked);
			updateStackSliderPosition();
			itsJPanel.noShifts();						
	    	}
	    	
	    else if(arg.equals("start to accumulate mouseClickIntensity vs sect# into a file")) {
			if(mouseCllickIntensityStats!=null) {
				talk("Close previously activated mouseClickIntensity stats file before starting a new one.");
				}
			mouseCllickIntensityStats = openWritingFile(this,"Pick file to hold mouseClickIntensity stats");
			talk("Until you close it, each mouse click puts intensity data into "+mouseCllickIntensityStats);
			itsJPanel.initMouseClickIntensityStats();
	    	}
	    else if(arg.equals("stop accumulating mouseClickIntensity vs sect# into file")) {
			if(mouseCllickIntensityStats==null) {
				talk("You never activated the mouseClickIntensity stats file.");
				}
			itsJPanel.writeMouseClickIntensityStats();
			try {mouseCllickIntensityStats.close();}
			catch(IOException ecl) {
				talk("could not close file "+mouseCllickIntensityStats);
				}
			mouseCllickIntensityStats=null;
	    	}
	    
	    else if(arg.equals("write, from a stack, a single section biorad.pic to disk")) {
	    	String theFilename = getExistingFileName(this,"Pick the Biorad file");
	    	if(theFilename != null) {
	    		RasterWriter rawRasterToDisk = new RasterWriter(this,theFilename);
	    		rawRasterToDisk.setVisible(true);
	    		}
	    	}
	    	
	    else if(arg.equals("save color tiff image in a file...")) {
		    itsJPanel.writeTiff();
	    	}
	    	
	    else if(arg.equals("save image in a Jpeg file...")) {
		    itsJPanel.writeJpeg();
	    	}

	    else if(arg.equals("write images in stack as jpg movie files to local disk")) {
	    	talk("start writing movie frame files to local disk.");
	    	itsJPanel.ifWriteWebfiles=false;
		    itsJPanel.start();
	    	}

	    else if(arg.equals("write images in stack as movie file to web server")) {
	    	talk("start writing movie frame files to web server.");
	    	itsJPanel.ifWriteWebfiles=true;
		    itsJPanel.start();
	    	}

	    else if(arg.equals("Interlace")) {
		itsJPanel.changeInterlace();
	    	}
	    else if(arg.equals("Quit")) {
		talk("OK,  before quitting you get final chance to archive mark/shift data");
	    	ArchiveMarks disker = new ArchiveMarks();
		disker.readWriteMarks(true,itsJPanel.marks, getArtColors, 
			centsJPanel, holesJPanel, hotJumpJPanel,
			itsJPanel.curStack.Xshifts,itsJPanel.curStack.Yshifts, 
			    itsJPanel.curStack.Zshifts);
		disker=null; 
		setVisible(false);         // hide the JFrame
		dispose();      // tell windowing system to free resources
		System.exit(0); // exit
		}

	    else if(arg.equals("Pick Colors Panel")) {
			getArtColors.setVisible(true);
	    	}
	    else if(arg.equals("Color Separator Panel")) {
			separator.setVisible(true);
	    	}

	    else if(arg.equals("GoTo marks Panel")) {
			goJPanel.setVisible(true);
	    	}
	    	
	    else if(arg.equals("Align marks Panel")) {
			alignJPanel.setVisible(true);
	    	}
	    	
	    else if(arg.equals("Select marker Panel")) {
			selectJPanel.setVisible(true);
	    	}
	    	
	    else if(arg.equals("Panel for moving markers and/or sections")) {
			moveJPanel.setVisible(true);
	    	}
	    	

	    else if(arg.equals("stereo 3D reconstruction panel")) {
	    	reconstruction3DPanel.setVisible(true);
	    	}
	    	
	    else if(arg.equals("Dock marker Panel")) {
			dockJPanel.setVisible(true);
	    	}
	    else if(arg.equals("global slider Panel")) {
			slideJPanel.setVisible(true);
	    	}
	    	
	    else if(arg.equals("hot spots Panel")) {
			hotJumpJPanel.setVisible(true);
	    	}
	    	
	    else if(arg.equals("histogram panel")) {
			histogramJPanel.setVisible(true);	    	}
	    	
	    else if(arg.equals("Centrosomes Panel")) {
			centsJPanel.setVisible(true);
	    	}
	    	
	    else if(arg.equals("Black Holes Panel")) {
			holesJPanel.setVisible(true);
	    	}

	    else if(arg.equals("Marker Shape Panel")) {
			shapeJPanel.setVisible(true);
	    	}
	    else if(arg.equals("data segmentation panel")) {
			segmentationJPanel.setVisible(true);
	    	}
	    else if(arg.equals("measure/retrieve point spread functions")) {
	    	if(writePSF==null)
				writePSF = new writePointSpreadFile(itsJPanel);
	    	writePSF.setVisible(true);
	    	}	
	    else if(arg.equals("deconvolver")) {
	    	if(DPSF==null) {
	    		talk("You must first initialize the Deconvolver's point spread function object");
	    		talk("By clicking the 'load deconvolver PSF object' button...");
	    		return;
	    		}
	    	else {
	    		if(decon == null) // don't wipe out onhe that's already active
	    			decon = new Deconvolver(DPSF,itsJPanel);
	    		decon.setVisible(true);
	    		}
	    	}	

	    
	    else if(arg.equals("R/G/B LUT's transforming raw input data")) {
			exhibitBlueChanLUT.setVisible(true);
			exhibitGreenChanLUT.setVisible(true);	
			exhibitRedChanLUT.setVisible(true);
	    	}
	    	
	    else if(arg.equals("Section navigator Panel")) {
			navigatorJPanel.setVisible(true);
	    	}
	    	
	    else if(arg.equals("Move sections and markers Panel")) {
	    	}
	    	

	    else if(arg.equals("show all auxiliary Panels")) {
			exhibitGreenChanLUT.setVisible(true);
			exhibitBlueChanLUT.setVisible(true);
			exhibitRedChanLUT.setVisible(true);
			selectJPanel.setVisible(true);
			alignJPanel.setVisible(true);
			goJPanel.setVisible(true);
			getArtColors.setVisible(true);
			separator.setVisible(true);
			histogramJPanel.setVisible(true);
			dockJPanel.setVisible(true);
			reconstruction3DPanel.setVisible(true);
			slideJPanel.setVisible(true);
			hotJumpJPanel.setVisible(true);
			centsJPanel.setVisible(true);
			holesJPanel.setVisible(true);
			shapeJPanel.setVisible(true);
			segmentationJPanel.setVisible(true);
	    	if(writePSF!=null) writePSF.setVisible(true);
	    	if(decon!=null) decon.setVisible(true);
			navigatorJPanel.setVisible(true);
			moveJPanel.setVisible(true);
	    	}
	    	
	    else if(arg.equals("enlarge all auxiliary Panels")) {
	    	growJPanels(1.25);
	    	}
	    	
	    else if(arg.equals("shrink all auxiliary Panels")) {
	    	growJPanels(.85);
	    	}	    	

	    else if(arg.equals("Hide all auxiliary Panels")) {
			selectJPanel.setVisible(false);
			alignJPanel.setVisible(false);
			goJPanel.setVisible(false);
			getArtColors.setVisible(false);
			separator.setVisible(false);
			dockJPanel.setVisible(false);
			reconstruction3DPanel.setVisible(false);
			slideJPanel.setVisible(false);
			hotJumpJPanel.setVisible(false);
			histogramJPanel.setVisible(false);
			centsJPanel.setVisible(false);
			holesJPanel.setVisible(false);
			shapeJPanel.setVisible(false);
			navigatorJPanel.setVisible(false);
			moveJPanel.setVisible(false);
			segmentationJPanel.setVisible(false);
	    	if(writePSF!=null) writePSF.setVisible(false);
	    	if(decon!=null) decon.setVisible(false);
			exhibitGreenChanLUT.setVisible(true);
			exhibitBlueChanLUT.setVisible(true);
			exhibitRedChanLUT.setVisible(true);
	    	}
	    	
	    else if(arg.equals("Write marks/alignment info to disk")) {
	    	ArchiveMarks disker = new ArchiveMarks();
			disker.readWriteMarks(true,itsJPanel.marks, getArtColors, 
			centsJPanel, holesJPanel,hotJumpJPanel, 
			itsJPanel.curStack.Xshifts,itsJPanel.curStack.Yshifts, 
			itsJPanel.curStack.Zshifts);
			disker=null; 
	    	}
	    	
        else if(arg.equals("Recover marks/alignment info from a disk file")) {
	    	ArchiveMarks disker = new ArchiveMarks();
	    	int [] rlut=new int[256];
	    	int [] glut=new int[256];
	    	int [] blut=new int[256];
			disker.readWriteMarks(false,itsJPanel.marks, getArtColors, 
			centsJPanel, holesJPanel, hotJumpJPanel,
			itsJPanel.curStack.Xshifts,itsJPanel.curStack.Yshifts, 
			itsJPanel.curStack.Zshifts);
			disker=null; 
			itsJPanel.selectMarker(1);
			reDraw();
	    	}

        else if(arg.equals("Write e3-format shift/centrosome list to a disk file")) {
	    	ArchiveMarks disker = new ArchiveMarks();
		disker.writeCentrosomeList(centsJPanel,itsJPanel.curStack.Xshifts,
			itsJPanel.curStack.Yshifts, itsJPanel.curStack.Zshifts);
		disker=null; 
	    	}
    }
    
    public void growJPanels(double growRatio) {
		selectJPanel.changeJPanelSize(growRatio);
		alignJPanel.changeJPanelSize(growRatio);
		goJPanel.changeJPanelSize(growRatio);
		getArtColors.changeJPanelSize(growRatio);
		dockJPanel.changeJPanelSize(growRatio);
		reconstruction3DPanel.changeJPanelSize(growRatio);
		slideJPanel.changeJPanelSize(growRatio);
		hotJumpJPanel.changeJPanelSize(growRatio);
		histogramJPanel.changeJPanelSize(growRatio);
		centsJPanel.changeJPanelSize(growRatio);
		holesJPanel.changeJPanelSize(growRatio);
		shapeJPanel.changeJPanelSize(growRatio);
		segmentationJPanel.changeJPanelSize(growRatio);
		navigatorJPanel.changeJPanelSize(growRatio);
		moveJPanel.changeJPanelSize(growRatio);
		}
   
 	static  RandomAccessFile openWritingFile(JFrame theJFrame,String prompt) {
		RandomAccessFile outfile;
		FileDialog getFileToWrite = 
		   new FileDialog(theJFrame,prompt, FileDialog.SAVE);
		getFileToWrite.setVisible(true);
		String filename = getFileToWrite.getDirectory()+"/"+getFileToWrite.getFile();
		if(filename == null) return null;
		try{
			outfile = new RandomAccessFile(filename,"rw");
			}
		catch (IOException e) {
			talk("openWritingFile("+prompt+") failed");
			return null;
			}
		try{outfile.seek((long)0);}
		catch(IOException e) {};
		return outfile;
		}
		
	static String getExistingFileName(JFrame theJFrame,String prompt) {
		FileDialog fileGet = 
		   new FileDialog(theJFrame,prompt, FileDialog.LOAD);
		fileGet.setLocation(fileGetXLoc,fileGetYLoc);
		fileGet.setVisible(true);
		String theName = fileGet.getDirectory()+"/"+fileGet.getFile();
		Point pt = fileGet.getLocation();
		fileGetXLoc=pt.x;
		fileGetYLoc=pt.y;
		return (theName);
		}
		
	static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
}


