package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.awt.*;  
import javax.swing.*;
import java.util.*;
import java.io.*;
import java.net.*;

/* transaction by which server sends a byte array containing a section to client:
first 5 bytes in are the int SENDSECTION=2
next 5 bytes in are one of XYSECT=110, XZSECT=101, ZYSECT=011
next 5 bytes in define the int telling which section to send
Then the server returns:
first 5 bytes supply an int telling the number of columns in the raster.
next 5 bytes supply an int telling the number of rows in the raster.
next ncols*nrows bytes are byte array defining the image raster.*/

/* transaction by which server tells client the size of the stack (after 
		interpolated sections are added).   
first 5 bytes in are the int TELLSTACKSIZE=3
then the server returns:
first 5 bytes specify ncols = number of pixels in the x-direction.
next  5  bytes specify nrows = number of pixels in the y-direction.
next  5  bytes specify nsects = number of pixels in the s-direction.
*/		
		
/* transaction by which server reads a byte array from a client:
first 5 bytes in are the int READBYTEARRAY=1
second 5 bytes in are the int typeOfArray==?
third 5 bytes in supply an int giving the columns of the raster
fourth 5 bytes in supply an int giving the rows of the raster
the following rows*cols bytes are that many bytes constituting the byte array
 */
   

public class sectionConsumerClient implements Runnable {
	int ncols=1,nrows=1,nsects=1;
	int nxycols,nxyrows,nxypix,nxzcols,nxzrows,nxzpix,nzycols,
			nzyrows,nzypix;
	DataInputStream in,prein;
	DataOutputStream out,preout;
	final static int GEOFXCEPT=-2;
	final static int GIOEXCEPT=-1;
	final static int READBYTEARRAY=1;
	final static int SENDSECTION=2;
	final static int TELLSTACKSIZE=3;
	final static int STOPALLTRANSACTORS_RESTART=7;
	final static int XYSECT=110;
	final static int XZSECT=101;
	final static int ZYSECT=011;
	final static int GTERMINATE=10;
	Socket s=null,spre=null;
	 String goodIP=null;

	int portNum=-1;
	boolean ifOnLine=false;
	Thread lookAhead=null;
	int ifRGB=0;
	String chanName=" red ";
	
	// use the run thread to get what is likely to be "the next" section 
	int readyXYsectNum=-1,readyXZsectNum=-1,readyZYsectNum=-1;
	int jumpXYsectNum=-1,jumpXZsectNum=-1,jumpZYsectNum=-1;
	int justGotXYsect=-1,justGotXZsect=-1,justGotZYsect=-1;
	byte [] readyXYsect,readyXZsect,readyZYsect;
	byte [] jumpXYsect,jumpXZsect,jumpZYsect;
	
	SectionNavigator jumper;
	int cacheLevel=-777,previousCacheLevel=2;
			
	public sectionConsumerClient(String hostname,int portNum,SectionNavigator jumper,
				int ifRGB) throws Exception {
	this.portNum = portNum;
	this.jumper = jumper;
	this.ifRGB = ifRGB;
	if(ifRGB==0) chanName = " red ";
	if(ifRGB==1) chanName = " green ";
	if(ifRGB==2) chanName = " blue ";
	FindServerPortOnSubnet portCaller;
	InetAddress address = InetAddress.getLocalHost();
	talk("getLocalHost returns <"+address+">");
	if(hostname != null) { // only if a hostname was supplied do we try to
	// get its IP number
	  try {
		address = InetAddress.getByName(hostname);
		talk("getByName returns <"+address+">");
		s = new Socket(address,portNum);
		makeInAndOut(); // hoses other server sockets
		spre = new Socket(address,portNum);
		makePreinAndPreout();
		ifOnLine=true;
		}
	  catch (Exception e) {
		talk(" InetAddress.getByName("+hostname+") threw "+e);
		}
	
	  if(!ifOnLine) {	
		try {
			s = new Socket(hostname,portNum);
			makeInAndOut(); // hoses other server sockets               
			spre = new Socket(address,portNum);
			makePreinAndPreout();
			ifOnLine=true;
			}
		catch (Exception e) {
			talk(chanName+"client attempt to create socket threw "+e);
			}
		}
		}
		
	if(!ifOnLine) {	
		// try all IP's on this subnet
		String localHostName = (InetAddress.getLocalHost()).toString();
		talk("the local host name is <"+localHostName+">");
		// Now isolate the IP number prefix
		int lastSlash = localHostName.lastIndexOf('/');
		int lastDot = localHostName.lastIndexOf('.');
		String iprefix = localHostName.substring(lastSlash+1,lastDot+1);
		portCaller = new FindServerPortOnSubnet(portNum,iprefix,this);
		//talk("call snoop()");
		portCaller.snoop(); // hangs until a port is found or none can be found
		//talk("back from snoop() and with port on "+goodIP);
		//talk("GOT Socket = "+s);
 		if(s != null) { // we got one. 
			talk(chanName+" channel found server at port "+portNum+" uniquely at "+goodIP);
			makeInAndOut(); // hoses other server sockets

			try {
				spre = new Socket(goodIP,portNum);
				makePreinAndPreout();
				}
			catch(Exception e) {
				talk("failed to create sockets, s or spre, on "+ goodIP);
				}
			portCaller.sepuku();
			portCaller=null;
			ifOnLine=true;
			}
		else {
			s=null; spre=null;
			out=null; preout=null; in=null; prein=null;
			talk("No computer in the subnet "+iprefix+" would open a server port for"+
					chanName+"channel");
			return;
			}
		}
		
	if(ifOnLine) {
		talk("sectionConsumerClient constructor worked for "+chanName+"channel");
		cacheLevel=2;
		start();
		}
	}
	
public void sepuku() {
	if (s != null) {
		try{in.close();
		out.close();
		s.close();}
		catch(Exception e) {}
		}
	}
	
	public void start() {
		if(!ifOnLine) return;
		//talk("sectionConsumerClient.start()");
		if(lookAhead==null) {
			//talk("sectionConsumerClient.start() makes new thread");
			lookAhead = new Thread(this);
			}
		lookAhead.start();
		}
		
	public void stop() {
		//talk("sectionConsumerClient.stop()");
		cacheLevel=-1;
		if(lookAhead!=null) {
			lookAhead.stop();
			}
		}

	public void webCacheLevel(int lev) {
		if(!ifOnLine) return;
	/* lev = -1 means suspend all caching,
	   lev = 0 means restore to most recent positive level
	   lev = 1 means look ahead just one section,
	   lev = 2 means look ahead one section and also sectionNavigator.bigjump sections.
			*/
		/* talk("webCacheLevel called with lev = "+lev+", cacheLevel="+cacheLevel+
			", and previousCacheLevel="+previousCacheLevel); */
		if(lev<0) {
			stop();
			//talk("Turn off all web caching");
			return;
			}
		if(cacheLevel<0 && lev >= 0)  // caching thread now stopped
			start();
			
		if(previousCacheLevel<1) previousCacheLevel=1;
		if(lev==0) lev = previousCacheLevel;
		cacheLevel=lev;
		previousCacheLevel=cacheLevel;					
		}
		
	private void sleeper(int howLong) {
		try{Thread.sleep(howLong);}
		catch(InterruptedException e) {}
		}	
		
	public void run () {
		int bigjump; // is the number of sections the section navigator jumps ahead
					// or back...
		while(ifOnLine) {
			bigjump = jumper.bigJump();
			if(cacheLevel>0 && justGotXYsect >=0 && readyXYsectNum != justGotXYsect+1 &&
				justGotXYsect+1 < nsects) {
				readyXYsectNum=-2; // so no one uses readyXYsect while it's being replaced
				readyXYsect = getXYsect(justGotXYsect+1,true);
				readyXYsectNum = justGotXYsect+1;
				//talk(chanName+"channel preFetched XY section number "+readyXYsectNum);
				sleeper(100);
				bigjump = jumper.bigJump();
				}
				
			if(cacheLevel>1 && justGotXYsect >=0 && jumpXYsectNum != justGotXYsect+bigjump &&
				justGotXYsect+bigjump < nsects) {
				// fetch the next jump XY section
				jumpXYsectNum=-2; // so no one uses readyXYsect while it's being replaced
				jumpXYsect = getXYsect(justGotXYsect+bigjump,true);
				jumpXYsectNum = justGotXYsect+bigjump;
				//talk(chanName+"channel preFetched XY section number "+jumpXYsectNum);
				sleeper(100);
				bigjump = jumper.bigJump();
				}
				
			if(cacheLevel>0 && justGotXZsect >=0 && readyXZsectNum != justGotXZsect+1 &&
				justGotXZsect+1 < nrows) {
				// fetch the next XZ section
				readyXZsectNum=-2; // so no one uses readyXZsect while it's being replaced
				readyXZsect = getXZsect(justGotXZsect+1,true);
				readyXZsectNum = justGotXZsect+1;
				//talk(chanName+"channel preFetched XZ section number "+readyXZsectNum);
				sleeper(100);
				bigjump = jumper.bigJump();
				}
				
			if(cacheLevel>1 && justGotXZsect >=0 && jumpXZsectNum != justGotXZsect+bigjump &&
				justGotXZsect+bigjump < nrows) {
				// fetch the next XZ jump section
				jumpXZsectNum=-2; // so no one uses readyXZsect while it's being replaced
				jumpXZsect = getXZsect(justGotXZsect+bigjump,true);
				jumpXZsectNum = justGotXZsect+bigjump;
				//talk(chanName+"channel preFetched XZ section number "+jumpXZsectNum);
				sleeper(100);
				bigjump = jumper.bigJump();
				}
				
			if(cacheLevel>0 && justGotZYsect >=0 && readyZYsectNum != justGotZYsect+1 &&
				justGotZYsect+1 < ncols) {
				// fetch the next ZY section
				readyZYsectNum=-2; // so no one uses readyZYsect while it's being replaced
				readyZYsect = getZYsect(justGotZYsect+1,true);
				readyZYsectNum = justGotZYsect+1;
				//talk(chanName+"channel preFetched ZY section number "+readyZYsectNum);
				sleeper(100);
				bigjump = jumper.bigJump();
				}
				
			if(cacheLevel>1 && justGotZYsect >=0 && jumpZYsectNum != justGotZYsect+bigjump &&
				justGotZYsect+bigjump < ncols) {
				// fetch the next ZY jump section
				jumpZYsectNum=-2; // so no one uses readyZYsect while it's being replaced
				jumpZYsect = getZYsect(justGotZYsect+bigjump,true);
				jumpZYsectNum = justGotZYsect+bigjump;
				//talk(chanName+"channel preFetched ZY section number "+jumpZYsectNum);
				sleeper(100);
				bigjump = jumper.bigJump();
				}
				
			sleeper(750);
			}
		}

	private void makeInAndOut() {
		try {
			out = new DataOutputStream(s.getOutputStream());
			out.flush();
			}
		catch(Exception e) {
			talk("making out threw this exception "+e);
			}
		try{		
			in = new DataInputStream(s.getInputStream());
			}
		catch(Exception e) {
			talk("making in threw this exception "+e);
			}					
		//talk("try to terminateAllTransactorssButThis() ");
		try {out.write(encodeIntAs5Bytes(STOPALLTRANSACTORS_RESTART));}
		catch(IOException e) {
			talk(chanName+"channel oops: got ioException writing an int");
			}
		if(readAnInt(false) == STOPALLTRANSACTORS_RESTART)
			talk("terminateAllTransactorssButThis() worked");
		else talk("terminateAllTransactorssButThis()  failed");
		}
				
	private void makePreinAndPreout() {
		try {
			preout = new DataOutputStream(spre.getOutputStream());
			preout.flush();
			}
		catch(Exception e) {
			talk("making preout threw this exception "+e);
			}
		try{		
			prein = new DataInputStream(spre.getInputStream());
			}
		catch(Exception e) {
			talk("making prein threw this exception "+e);
			}					
		}
				
	private byte [] encodeIntAs5Bytes(int theInt) {
		byte [] byts = new byte[5];
		int rem,quot;
		if(theInt>=0) {
			byts[0] = (byte)(1); // first byte encodes the sign
			}
		else {
			byts[0]=(byte) (-1);
			theInt = -theInt;
			}
		for(int i=0;i<4;i++) {
			quot = theInt/256;
			rem = theInt - 256*quot;
			theInt=quot;
			if(rem>127) rem = rem-256;
			byts[4-i]=(byte)(rem);
			}
		return byts;	
		}
		
	private int decodeIntFrom5Bytes(byte [] the5Bytes) {
		int ibyte,theInt,i;
		theInt=0;
		for(i=0;i<4;i++) {
			if((ibyte=the5Bytes[i+1]) <0)
				ibyte+=256;
			theInt = 256*theInt + ibyte;
			}
		if(the5Bytes[0]<0)
			return (-theInt);
		else return (theInt);	
		}
		
	private void  writeAnInt(int theInt, boolean ifPre) {
		DataOutputStream outs = ifPre? preout : out;
		try {outs.write(encodeIntAs5Bytes(theInt));}
		catch(IOException e) {
			talk(chanName+"channel oops: got ioException writing an int");
			}
		}
		
	private int bytesAvailable(boolean ifPre) {
		int avail=0;
		DataInputStream ins = ifPre? prein : in;
		try {avail=ins.available();}
		catch(IOException e) {};
		return avail;
		}
				
	private void waitForEnoughBytes(int howManyBytes,boolean ifPre) {
		while(bytesAvailable(ifPre) < howManyBytes) {
			try {(Thread.currentThread()).sleep(20);}
			catch (InterruptedException e) {} 
			}			
		}
			
	private int readAnInt(boolean ifPre) {
		DataInputStream ins = ifPre? prein : in;
		byte [] in5 = new byte[5];
		int inread=0;
		waitForEnoughBytes(5,ifPre);
		try {inread=ins.read(in5);}
		catch (EOFException e) {return -2;}
		catch (IOException e) {return -1;}
		if(inread != 5) {
			talk(chanName+"channel oops: did not get 5 bytes when reading an int");
			return 0;
			}
		return (decodeIntFrom5Bytes(in5));
		}
		
	public int [] getStackSize() {
		if(!ifOnLine) return null;
		writeAnInt(TELLSTACKSIZE,false);
		int [] headerDat = new int[5];
		ncols = readAnInt(false); headerDat[0] =  ncols; 
		nrows = readAnInt(false); headerDat[1] = nrows;
		nsects = readAnInt(false); headerDat[2] = nsects;
		headerDat[3] = readAnInt(false); // merged
		headerDat[4] = readAnInt(false); // magic
		return headerDat;
		}
		
	public byte [] getXYsect  (int sectNum) {
		if(!ifOnLine) return null;
		return (getXYsect(sectNum,false));
		}
		
	private byte [] getXYsect(int sectNum, boolean ifPre) {
		if(!ifOnLine) return null;
		if(!ifPre && readyXYsectNum==sectNum) {
			//talk(chanName+"channel ready cache hit for XY section "+sectNum);
			justGotXYsect = sectNum;
			return readyXYsect;
			}
		if(!ifPre && jumpXYsectNum==sectNum) {
			//talk(chanName+"channel ready cache hit for jump XY section "+sectNum);
			justGotXYsect = sectNum;
			return jumpXYsect;
			}
		// otherwise, must get the section across the wire
		writeAnInt(SENDSECTION,ifPre);
		writeAnInt(XYSECT,ifPre);
		writeAnInt(sectNum,ifPre);
		nxycols = readAnInt(ifPre);
		nxyrows = readAnInt(ifPre);
		nxypix = nxycols*nxyrows;
		byte [] xysect = new byte[nxypix];
		if(ifPre) {
			readThePix(nxypix,xysect,true);
			return xysect;
			}
		else {
			readThePix(nxypix,xysect,false);
			justGotXYsect = sectNum;
			return xysect;
			}
		}
		
	 public byte [] getXYsect(int sectNum, byte [] xysect) {
		if(!ifOnLine) return null;
		if(readyXYsectNum==sectNum) {
			for(int i=0;i<nxypix;i++)
				xysect[i] = readyXYsect[i];
			//talk(chanName+"channel scored a cache hit for XY section "+sectNum);
			justGotXYsect = sectNum;
			return xysect;
			}
		if(jumpXYsectNum==sectNum) {
			for(int i=0;i<nxypix;i++)
				xysect[i] = jumpXYsect[i];
			//talk(chanName+"channel scored a cache hit for XY jump section "+sectNum);
			justGotXYsect = sectNum;
			return xysect;
			}
		// otherwise, must get the section across the wire
		writeAnInt(SENDSECTION,false);
		writeAnInt(XYSECT,false);
		writeAnInt(sectNum,false);
		nxycols = readAnInt(false);
		nxyrows = readAnInt(false);
		nxypix = nxycols*nxyrows;
		readThePix(nxypix,xysect,false);
		justGotXYsect = sectNum;
		return xysect;
		} 
		
	public byte [] getXZsect(int sectNum) {
		if(!ifOnLine) return null;
		return getXZsect(sectNum,false);
		}
		
	private byte [] getXZsect(int sectNum, boolean ifPre) {
		if(!ifOnLine) return null;
		if(!ifPre && readyXZsectNum==sectNum) {
			//talk(chanName+"channel ready cache hit for XZ section "+sectNum);
			justGotXZsect = sectNum;
			return readyXZsect;
			}
		if(!ifPre && jumpXZsectNum==sectNum) {
			//talk(chanName+"channel scored a cache hit for XZ jump section "+sectNum);
			justGotXZsect = sectNum;
			return jumpXZsect;
			}
		// otherwise, must get the section across the wire
		writeAnInt(SENDSECTION,ifPre);
		writeAnInt(XZSECT,ifPre);
		writeAnInt(sectNum,ifPre);
		nxzcols = readAnInt(ifPre);
		nxzrows = readAnInt(ifPre);
		nxzpix = nxzcols*nxzrows;		
		byte [] xzsect = new byte[nxzpix];
		if(ifPre) {
			readThePix(nxzpix,xzsect,true);
			return xzsect;
			}
		else {
			readThePix(nxzpix,xzsect,false);
			justGotXZsect = sectNum;
			return xzsect;
			}
		}
		
	public byte [] getXZsect(int sectNum, byte [] xzsect) {
		if(!ifOnLine) return null;
		if(readyXZsectNum==sectNum) {
			//talk(chanName+"channel ready cache hit for XZ section "+sectNum);
			for(int i=0;i<nxzpix;i++)
				xzsect[i] = readyXZsect[i];
			justGotXZsect = sectNum;
			return xzsect;
			}
		if(jumpXZsectNum==sectNum) {
			//talk(chanName+"channel scored a cache hit for XZ jump section "+sectNum);
			for(int i=0;i<nxzpix;i++)
				xzsect[i] = jumpXZsect[i];
			justGotXZsect = sectNum;
			return xzsect;
			}
		writeAnInt(SENDSECTION,false);
		writeAnInt(XZSECT,false);
		writeAnInt(sectNum,false);
		nxzcols = readAnInt(false);
		nxzrows = readAnInt(false);
		nxzpix = nxzcols*nxzrows;
		readThePix(nxzpix, xzsect,false);
		justGotXZsect = sectNum;
		return xzsect;
		} 
		
	public byte [] getZYsect(int sectNum) {
		if(!ifOnLine) return null;
		return getZYsect(sectNum,false);
		}
		
	private byte [] getZYsect(int sectNum, boolean ifPre) {
		if(!ifOnLine) return null;
		if(!ifPre && readyZYsectNum==sectNum) {
			//talk(chanName+"channel ready cache hit for ZY section "+sectNum);
			justGotZYsect = sectNum;
			return readyZYsect;
			}
		if(!ifPre && jumpZYsectNum==sectNum) {
			//talk(chanName+"channel scored a cache hit for ZY jump section "+sectNum);
			justGotZYsect = sectNum;
			return jumpZYsect;
			}
		// otherwise, must get the section across the wire
		writeAnInt(SENDSECTION,ifPre);
		writeAnInt(ZYSECT,ifPre);
		writeAnInt(sectNum,ifPre);
		nzycols = readAnInt(ifPre);
		nzyrows = readAnInt(ifPre);
		nzypix = nzycols*nzyrows;
		byte [] zysect = new byte[nzypix];
		if(ifPre) {
			readThePix(nzypix,zysect,true);
			return zysect;
			}
		else {
			readThePix(nzypix,zysect,false);
			justGotZYsect = sectNum;
			return zysect;
			}
		}
		
	public byte [] getZYsect(int sectNum, byte [] zysect) {
		if(!ifOnLine) return null;
		if(readyZYsectNum==sectNum) {
			//talk(chanName+"channel ready cache hit for ZY section "+sectNum);
			for(int i=0;i<nzypix;i++)
				zysect[i] = readyZYsect[i];
			justGotZYsect = sectNum;
			return zysect;
			}
		if(jumpZYsectNum==sectNum) {
			//talk(chanName+"channel scored a cache hit for ZY jump section "+sectNum);
			for(int i=0;i<nzypix;i++)
				zysect[i] = jumpZYsect[i];
			justGotZYsect = sectNum;
			return zysect;
			}
		writeAnInt(SENDSECTION,false);
		writeAnInt(ZYSECT,false);
		writeAnInt(sectNum,false);
		nzycols = readAnInt(false);
		nzyrows = readAnInt(false);
		nzypix = nzycols*nzyrows;
		readThePix(nzypix,zysect,false);
		justGotZYsect = sectNum;
		return zysect;
		} 
		
	private void readThePix(int npix, byte [] buf, boolean ifPre) {
		DataInputStream ins = ifPre? prein : in;
		int inread=0,ntogo;
		while(inread<npix) {
			ntogo = npix-inread;
			if(bytesAvailable(ifPre) <=0) {
				try { (Thread.currentThread()).sleep(30); }
				catch (InterruptedException e) {}
				} 		
			else {
				try {
					inread += ins.read(buf,inread,ntogo);
					}
				catch(IOException e) {
					talk(chanName+"channel oops, got IOException while reading zy section");
					}
				}
			}
		}
			
	public void sendRasterToServer(int 	arrayType, int rasterCols, int rasterRows,
		byte [] buf) {
/* transaction by which server reads a byte array from a client:
first 5 bytes in are the int READBYTEARRAY=1
second 5 bytes in are the int typeOfArray==?
third 5 bytes in supply an int giving the columns of the raster
fourth 5 bytes in supply an int giving the rows of the raster
the following rows*cols bytes are that many bytes constituting the byte array
 */
		writeAnInt(READBYTEARRAY,false);
		writeAnInt(arrayType,false);
		writeAnInt(rasterCols,false);
		writeAnInt(rasterRows,false);
		int npix = rasterCols*rasterRows;
		int nout = 0;
		int nchunk = 265*256;
		while(nout < npix) {
			if(nchunk > npix-nout)
				nchunk = npix-nout;
			try{ 
				out.write(buf,nout,nchunk);
				nout += nchunk;
 				}
 			catch(IOException e) {
 				talk(chanName+"channel OOPS; sendRasterToServer failed");
 				return;
 				}
 			}
		}
		
	static void talk(String p) {
		System.out.println(p);
	}	

}
